(*
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2015-2017)
License: LGPL (see file COPYING.LESSER)
*)

theory Parallel_Closed
imports
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  Critical_Pairs
  "../Rewriting/Multihole_Context"
  "../Rewriting/More_Abstract_Rewriting"
begin

subsection \<open>Measuring Overlap\<close>

fun overlap' :: "('f, 'v) mctxt \<times> ('f, 'v) term list \<Rightarrow> ('f, 'v) mctxt \<times> ('f, 'v) term list \<Rightarrow> ('f, 'v) term multiset"
where
  "overlap' (MHole, [l]) (C, ls) = mset ls"
| "overlap' (C, ls) (MHole, [l]) = mset ls"
| "overlap' (MFun f Cs, ls) (MFun g Ds, ls') =
   \<Union># {# overlap' (Cs ! i, partition_holes ls Cs ! i) (Ds ! i, partition_holes ls' Ds ! i). i \<in># mset [0..<length Cs] #}"
| "overlap' (MVar x, []) (C, ls) = {#}"
| "overlap' (C, ls) (MVar x, []) = {#}"
| "overlap' _ _ = undefined"

lemma overlap'_symmetric:
  assumes "s =\<^sub>f (C, ls)"  and "s =\<^sub>f (D, ls')"
  shows "overlap' (C, ls) (D, ls') = overlap' (D, ls') (C, ls)"
using assms
proof (induct C arbitrary: D s ls ls')
  case (MVar x) note C = this
  then have [simp]: "ls = []" by (auto dest!: eqfE)
  show ?case
  proof (cases D)
    case (MVar y) with C show ?thesis by (auto dest!: eqfE)
  next
    case MHole with C show ?thesis  by (auto dest!: eqf_MHoleE)
  qed auto
next
  case MHole note C = this
  then obtain l where [simp]: "ls = [l]" by (auto dest!: eqf_MHoleE)
  show ?case
  proof (cases D)
    case (MVar y) with C show ?thesis by (auto dest!: eqfE)
  next
    case MHole with C show ?thesis  by (auto dest!: eqf_MHoleE)
  qed auto
next
  case (MFun f cs) note C = this
  then obtain ss lss where s:"s = Fun f ss"
    and  lcs:"length ss = length cs" and llss:"length lss = length cs"
    and lss:"ls = concat lss" and lsics:"\<And> i. i < length cs \<Longrightarrow> ss ! i =\<^sub>f (cs ! i, lss ! i)"
    using C by (auto elim: eqf_MFunE)
  show ?case
  proof (cases D)
    case (MVar x) with C show ?thesis by (auto dest!: eqfE)
  next
    case MHole with C show ?thesis  by (auto dest!: eqf_MHoleE)
  next
    case (MFun g ds)
    with C obtain ts lss' where "s = Fun g ts"
      and  lds:"length ts = length ds" and llss':"length lss' = length ds"
      and lss':"ls' = concat lss'" and ls'ids:"\<And> i. i < length ds \<Longrightarrow> ts ! i =\<^sub>f (ds ! i, lss' ! i)"
      using C by (auto elim: eqf_MFunE)
    with s have [simp]: "g = f" "ts = ss" by auto
    have [simp]: "partition_holes ls cs = lss" by (metis eqfE(2) llss lsics lss partition_holes_concat_id)
    have [simp]: "partition_holes ls' ds = lss'" by (metis eqfE(2) llss' ls'ids lss' partition_holes_concat_id)
    { fix i
      assume i:"i < length cs"
      then have "cs ! i \<in> set cs" by auto
      from C(1)[OF this lsics[OF i], of "ds ! i" "lss' ! i"] ls'ids i
      have "overlap' (cs ! i, lss ! i) (ds ! i, lss' ! i) = overlap' (ds ! i, lss' ! i) (cs ! i, lss ! i)"
        using lds lcs by auto
    } note ieq = this
    then have "(map (\<lambda>i. overlap' (cs ! i, lss ! i)  (ds ! i, lss' ! i)) [0..<length cs]) =
      (map (\<lambda>i. overlap' (ds ! i, lss' ! i) (cs ! i, lss ! i)) [0..<length ds])"
      using nth_equalityI lcs lds by auto
    then have "{#overlap' (cs ! i, lss ! i)  (ds ! i, lss' ! i). i \<in># mset [0..<length cs]#} =
      {#overlap' (ds ! i, lss' ! i) (cs ! i, lss ! i). i \<in># mset [0..<length ds]#}"
      using mset_map by (metis (no_types))
    with MFun show ?thesis by auto
  qed
qed

lemma overlap'_bounded:
  assumes "s =\<^sub>f (C, ls)" and "s =\<^sub>f (D, ls')"
  shows "(overlap' (C, ls) (D, ls'), mset ls) \<in> (mult {\<lhd>})\<^sup>="
proof -
  have trans: "trans {\<lhd>}" unfolding trans_def using supt_trans by auto
  show ?thesis
  using assms
  proof (induct C arbitrary: D ls ls' s)
    case (MVar x) note C = this
    then have [simp]: "ls = []" by (auto dest!: eqfE)
    show ?case
    proof (cases D)
      case (MVar y) with C show ?thesis by (auto dest!: eqfE)
    next
      case MHole with C show ?thesis  by (auto dest!: eqf_MHoleE)
    qed auto
  next
    case MHole note C = this
    then obtain l where [simp]: "ls = [l]" "s = l" by (auto dest!: eqf_MHoleE)
    show ?case
    proof (cases D)
      case (MVar y) with C show ?thesis by (auto dest!: eqfE simp: mult_def)
    next
      case MHole with C show ?thesis  by (auto dest!: eqf_MHoleE)
    next
      case (MFun g ds)
      from eqf_MFun_imp_strict_subt[OF C(2)[unfolded MFun]] have "\<And>t. t \<in> set ls' \<Longrightarrow> l \<rhd> t"
        by auto
      then have "(mset ls', {#l#}) \<in> mult {\<lhd>}"
        using one_step_implies_mult [of "{#l#}" "mset ls'" "{\<lhd>}" "{#}"] by auto
      with MFun show ?thesis by simp
    qed
  next
    case (MFun f cs) note C = this
    then obtain ss lss where s:"s = Fun f ss"
      and  lcs:"length ss = length cs" and llss:"length lss = length cs"
      and lss:"ls = concat lss" and lsics:"\<And> i. i < length cs \<Longrightarrow> ss ! i =\<^sub>f (cs ! i, lss ! i)"
      using C by (auto elim: eqf_MFunE)
    show ?case
    proof (cases D)
      case (MVar y)
      with C show ?thesis using non_empty_empty_mult[of "mset ls"] by (auto dest!: eqfE)
    next
      case MHole
      with C obtain l' where [simp]: "ls' = [l']" "s = l'" by (auto dest!: eqf_MHoleE)
      from eqf_MFun_imp_strict_subt[OF C(2)] have "\<And>t. t \<in> set ls \<Longrightarrow> l' \<rhd> t"
        by auto
      then have "(mset ls, {#l'#}) \<in> mult {\<lhd>}"
        using one_step_implies_mult[of "{#l'#}" "mset ls" "{\<lhd>}" "{#}"]  by auto
      with MHole show ?thesis by simp
    next
      case (MFun g ds)
      with C obtain ts lss' where "s = Fun g ts"
        and  lds:"length ts = length ds" and llss':"length lss' = length ds"
        and lss':"ls' = concat lss'" and ls'ids:"\<And> i. i < length ds \<Longrightarrow> ts ! i =\<^sub>f (ds ! i, lss' ! i)"
        using C by (auto elim: eqf_MFunE)
      with s have [simp]: "g = f" "ts = ss" by auto
      have partlscs: "partition_holes ls cs = lss"
        by (metis eqfE(2) llss lsics lss partition_holes_concat_id)
      have partlsds:"partition_holes ls' ds = lss'"
        by (metis eqfE(2) llss' ls'ids lss' partition_holes_concat_id)
      let ?ol = "\<lambda>i. overlap' (cs ! i, lss ! i) (ds ! i, lss' ! i)"
      { fix i
        assume i:"i < length cs"
        then have "cs ! i \<in> set cs" by auto
        from C(1)[OF this lsics[OF i]] ls'ids i lds lcs
        have "(?ol i,  mset (lss ! i)) \<in> (mult {\<lhd>})\<^sup>=" by auto
      } note IH = this
      then have "\<forall>i < length cs. (map ?ol [0..<length cs] ! i , map mset lss ! i) \<in> (mult {\<lhd>})\<^sup>="
        using llss by auto
      with pointwise_mult_imp_mult[OF _ _ trans]
      have "(\<Union># mset (map ?ol [0..<length cs]), \<Union># mset (map mset lss)) \<in> (mult {\<lhd>})\<^sup>="
        using length_map llss map_upt_len_conv by (metis (lifting))
      then show ?thesis unfolding lss mset_concat_union MFun overlap'.simps
        using lss mset_map partlscs partlsds by metis
    qed
  qed
qed

subsection \<open>Parallel Rewriting using Multihole Contexts\<close>

lemma par_rstep_mctxtE:
  assumes "(s, t) \<in> par_rstep R"
  obtains C ss ts where "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "length ss = length ts"
    and "\<forall>i<length ss. (ss ! i, ts ! i) \<in> rrstep R"
proof -
  have "\<exists> C ss ts. s =\<^sub>f (C, ss) \<and> t =\<^sub>f (C, ts) \<and> length ss = length ts \<and>
    (\<forall>i<length ss. (ss ! i, ts ! i) \<in> rrstep R)" (is "\<exists>C ss ts. ?P s t C ss ts")
  using assms
  proof (induct)
    case (root_step s t \<sigma>)
    then have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rrstep R" by auto
    moreover have "s \<cdot> \<sigma> =\<^sub>f (MHole, [s \<cdot> \<sigma>])" and "t \<cdot> \<sigma> =\<^sub>f (MHole, [t \<cdot> \<sigma>])" by auto
    ultimately show ?case by force
  next
    case (par_step_var x)
    have "Var x =\<^sub>f (MVar x, [])" by auto
    then show ?case by force
  next
    case (par_step_fun ts ss f)
    then have "\<forall>i<length ts. \<exists>x. ?P (ss ! i) (ts ! i) (fst x) (fst (snd x)) (snd (snd x))" by force
    then obtain g where "\<forall>i<length ts. ?P (ss ! i) (ts ! i) (fst (g i)) (fst (snd (g i))) (snd (snd (g i)))"
      unfolding choice_iff' by blast
    moreover
    define Cs us vs where "Cs = map (\<lambda>i. fst (g i)) [0 ..< length ts]"
      and "us = map (\<lambda>i. fst (snd (g i))) [0 ..< length ts]"
      and "vs = map (\<lambda>i. snd (snd (g i))) [0 ..< length ts]"
    ultimately have [simp]: "length Cs = length ts"
        "length us = length ts" "length vs = length ts"
      and *: "\<forall>i<length us. ss ! i =\<^sub>f (Cs ! i, us ! i) \<and> ts ! i =\<^sub>f (Cs ! i, vs ! i) \<and>
        (\<forall>j<length (us ! i). (us ! i ! j, vs ! i ! j) \<in> rrstep R)"
      by simp_all
    define C where "C = MFun f Cs"
    have "Fun f ss =\<^sub>f (C, concat us)" and "Fun f ts =\<^sub>f (C, concat vs)"
      using * by (auto simp: C_def \<open>length ss = length ts\<close> intro: eqf_MFunI)
    moreover have "\<forall>i<length (concat us). (concat us ! i, concat vs ! i) \<in> rrstep R"
      using * by (intro concat_all_nth) (auto dest!: eqfE)
    moreover have "length (concat us) = length (concat vs)" by (metis calculation(1) calculation(2) eqfE(2))
    ultimately show ?case  by blast
  qed
  with that show ?thesis by blast
qed

lemma par_rstep_mctxt_rrstepI :
  assumes "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "\<forall>i<length ss. (ss ! i, ts ! i) \<in> rrstep R"
  shows "(s, t) \<in> par_rstep R"
by (meson assms contra_subsetD par_rstep_mctxt rrstep_imp_rstep rstep_par_rstep)

definition "par_rstep_mctxt R C ss ts =
  {(s, t). s =\<^sub>f (C, ss) \<and> t =\<^sub>f (C, ts) \<and> (\<forall>i<length ss. (ss ! i, ts ! i) \<in> rrstep R)}"

lemma par_rstep_par_rstep_mctxt_conv:
  "(s, t) \<in> par_rstep R \<longleftrightarrow> (\<exists>C ss ts. (s, t) \<in> par_rstep_mctxt R C ss ts)"
proof
  assume "(s, t) \<in> par_rstep R"
  from par_rstep_mctxtE[OF this] obtain C ss ts
    where " s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)" and "\<forall>i<length ss. (ss ! i, ts ! i) \<in> rrstep R"
    by metis
  then show "\<exists>C ss ts. (s, t) \<in> par_rstep_mctxt R C ss ts" by (auto simp: par_rstep_mctxt_def)
next
  assume "\<exists>C ss ts. (s, t) \<in> par_rstep_mctxt R C ss ts"
  then show "(s, t) \<in> par_rstep R"
    by (auto simp: par_rstep_mctxt_def intro: par_rstep_mctxt_rrstepI)
qed

text \<open>lemma which tells us, that a parallel rewrite step of s \<cdot> \<sigma> is either inside s i.e., we
  can split of a critical pair, or we can do the step completely inside \<sigma>\<close>
lemma par_rstep_mctxt_linear_subst:
  assumes "linear_term s"
  and "(s \<cdot> \<sigma>, t) \<in> par_rstep_mctxt R C ss ts"
  shows "(\<exists> \<tau>. t = s \<cdot> \<tau> \<and> (\<forall> x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> par_rstep R) \<or>
           (\<exists> D s' l r j C'. s = D\<langle>s'\<rangle> \<and> is_Fun s' \<and> (l, r) \<in> R \<and> (s' \<cdot> \<sigma> = l \<cdot> \<tau>)
             \<and> j < length ss \<and> hole_pos D \<in> hole_poss C
             \<and> (D \<cdot>\<^sub>c \<sigma>)\<langle>r \<cdot> \<tau>\<rangle> =\<^sub>f (C', remove_idx j ss) \<and> t =\<^sub>f (C', remove_idx j ts)))"
using assms
proof (induction s arbitrary: t C ss ts)
  case (Var x)
  then show ?case
    unfolding par_rstep_mctxt_def by (auto intro: exI[of _ "\<lambda>y. t"] par_rstep_mctxt_rrstepI)
next
  case (Fun f ss')
  then have s:"Fun f ss' \<cdot> \<sigma> =\<^sub>f (C, ss)" and t: "t =\<^sub>f (C, ts)"
    and steps : "\<forall>i<length ss. (ss ! i, ts ! i) \<in> rrstep R"
    and length: "length ss = length ts"
    unfolding par_rstep_mctxt_def by (auto dest!: eqfE)
  let ?ss = "map (\<lambda> s. s \<cdot> \<sigma>) ss'"
  show ?case
  proof (cases C)
    case (MHole)
    with t have t:"ts = [t]" by (auto elim: eqf_MHoleE)
    from s MHole have ss: "ss = [Fun f ?ss]" by (auto elim: eqf_MHoleE)
    with steps obtain \<tau> l r where l:"Fun f ?ss = l \<cdot> \<tau>" and r:"ts ! 0 = r \<cdot> \<tau>" and lr:"(l, r) \<in> R"
      unfolding rrstep_def' by auto
    show ?thesis
    proof (rule exI, rule disjI2, intro exI conjI)
      show "Fun f ss' = \<box>\<langle>Fun f ss'\<rangle>" by simp
      show "is_Fun (Fun f ss')" by auto
      from l show " Fun f ss' \<cdot> \<sigma> = l \<cdot> \<tau>" by auto
      from lr show "(l, r) \<in> R" .
      from ss show "0 < length ss" by auto
      show "(\<box> \<cdot>\<^sub>c \<sigma>)\<langle>r \<cdot> \<tau>\<rangle> =\<^sub>f (mctxt_of_term (r \<cdot> \<tau>), remove_idx 0 ss)"
        using ss unfolding remove_idx_def by auto
      show "t =\<^sub>f (mctxt_of_term (r \<cdot> \<tau>), remove_idx 0 ts)"
        using r t unfolding remove_idx_def by auto
      show "hole_pos \<box> \<in> hole_poss C" using MHole by auto
    qed
  next
    case (MFun g Cs)
    then have gf [simp]: "g = f" using s by (auto dest: eqfE)
    from s have "Fun f ?ss =\<^sub>f (C, ss)" by auto
    with MFun obtain sss where lCs:"length ?ss = length Cs" and lsss:"length sss = length Cs"
      and sss:"ss = concat sss" and ss'iC:"\<And> i. i < length Cs \<Longrightarrow> ?ss ! i =\<^sub>f (Cs ! i, sss ! i)"
      by (auto elim: eqf_MFunE)
    from s have "Fun f ?ss = fill_holes C ss" "num_holes C = length ss" by (auto dest: eqfE)
    from t MFun obtain ts' where t2 [simp]: "t = Fun f ts'" by (auto elim: eqf_MFunE)
    obtain tss where lCs2:"length ts' = length Cs" and ltss:"length tss = length Cs"
      and tss:"ts = concat tss" and ts'iC:"\<And> i. i < length Cs \<Longrightarrow> ts' ! i =\<^sub>f (Cs ! i, tss ! i)"
      using MFun t by (auto elim: eqf_MFunE)
    { fix i
      assume i:"i < length sss"
      then have "?ss ! i =\<^sub>f (Cs ! i, sss ! i)" using lsss ss'iC by auto
      moreover have "ts' ! i =\<^sub>f (Cs ! i, tss ! i)" using ltss ts'iC i lsss by auto
      ultimately have "length (sss ! i) = length (tss ! i)"  by (auto dest!: eqfE)
    } note lssstss = this
    let ?p1 = "\<lambda> \<tau> i. ts' ! i = ss' ! i \<cdot> \<tau> \<and> (\<forall> x \<in> vars_term (ss' ! i). (\<sigma> x, \<tau> x) \<in> par_rstep R)"
    let ?p2 = "\<lambda> \<tau> i. (\<exists> D s' l r j C'. ss' ! i = D\<langle>s'\<rangle> \<and> is_Fun s' \<and> (l, r) \<in> R \<and> s' \<cdot> \<sigma> = l \<cdot> \<tau> \<and>
               j < length (sss ! i) \<and> hole_pos D \<in> hole_poss (Cs ! i) \<and>
               (D \<cdot>\<^sub>c \<sigma>)\<langle>r \<cdot> \<tau>\<rangle> =\<^sub>f (C', remove_idx j (sss ! i)) \<and>
               ts' ! i  =\<^sub>f (C', remove_idx j (tss ! i)))"
    let ?p = "\<lambda> \<tau> i. ?p1 \<tau> i \<or> ?p2 \<tau> i"
     {
      fix i
      assume i: "i < length ss'"
      then have i2: "i < length Cs" using lCs by auto
      from ss'iC[OF i2] i have eqf:"ss' ! i \<cdot> \<sigma> =\<^sub>f (Cs ! i, sss ! i)" by simp
      from ts'iC[OF i2] i have eqf2:"ts' ! i =\<^sub>f (Cs ! i, tss ! i)" by simp
      from i have mem: "ss' ! i \<in> set ss'" by auto
      from Fun(2) i have lin:"linear_term (ss' ! i)" by auto
      from lssstss i lCs lsss ltss have len: "length (tss ! i) = length (sss ! i)" by auto
      { fix j
        assume j:"j < length (sss ! i)"
        let ?k = "sum_list (map length (take i sss)) + j"
        from lssstss have *:"map length (take i tss) = map length (take i sss)"
          by (simp add: lsss ltss nth_map_conv)
        have "?k < length ss" by (simp add: concat_nth_length i2 j lsss sss)
        moreover have "ss ! ?k = sss ! i ! j" using sss concat_nth[OF _ j] i lsss lCs by auto
        moreover have "ts ! ?k = tss ! i ! j" using tss concat_nth[of i tss j ?k] j i ltss lsss lCs * lssstss by simp
        ultimately have "(sss ! i ! j, tss ! i ! j) \<in> rrstep R" using steps by auto
      }
      then have rrstep: "\<forall>j<length (sss ! i). (sss ! i ! j, tss ! i ! j) \<in> rrstep R" by blast
      from eqf eqf2 rrstep have "(ss' ! i \<cdot> \<sigma>, ts' ! i) \<in> par_rstep_mctxt R (Cs ! i) (sss ! i) (tss ! i)"
        unfolding par_rstep_mctxt_def by auto
      from Fun.IH[OF mem lin this] have "\<exists> \<tau>. ?p \<tau> i" .
    }
    hence "\<forall>i. \<exists>\<tau>. i < length ss' \<longrightarrow> ?p \<tau> i" by blast
    from choice[OF this] obtain \<tau>s where \<tau>s: "\<And> i. i < length ss' \<Longrightarrow> ?p (\<tau>s i) i" by blast
    show ?thesis
    proof (cases "\<exists> i. i < length ss' \<and> ?p2 (\<tau>s i) i")
      case True
      then obtain i where iss': "i < length ss'" and p2: "?p2 (\<tau>s i) i" by blast+
      from iss' have its': "i < length ts'" using lCs lCs2 by auto
      from p2 obtain D s' l r j C' where ssi: "ss' ! i = D\<langle>s'\<rangle>" and "is_Fun s'" "(l, r) \<in> R"
        and s':"s' \<cdot> \<sigma> = l \<cdot> \<tau>s i" and j:"j < length (sss ! i)" and hpos: "hole_pos D \<in> hole_poss (Cs ! i)"
        and D\<sigma>r:"(D \<cdot>\<^sub>c \<sigma>)\<langle>r \<cdot> \<tau>s i\<rangle> =\<^sub>f (C', remove_idx j (sss ! i))"
        and ts'i:"ts' ! i  =\<^sub>f (C', remove_idx j (tss ! i))"
        by blast
      define bef where "bef = take i ss'"
      define aft where "aft = drop (Suc i) ss'"
      from id_take_nth_drop[OF iss', unfolded ssi] have ss':"ss' = bef @ D \<langle>s'\<rangle> # aft"
        using aft_def bef_def by blast
      let ?D = "More f bef D aft"
      let ?r = "(D \<cdot>\<^sub>c \<sigma>)\<langle>r \<cdot> \<tau>s i\<rangle>"
      let ?\<sigma> = "map (\<lambda> s. s \<cdot> \<sigma>)"
      let ?Cs = "take i Cs @ C' # drop (Suc i) Cs"
      have D: "(?D \<cdot>\<^sub>c \<sigma>)\<langle>r \<cdot> \<tau>s i\<rangle> = Fun f (?\<sigma> bef @ ?r # ?\<sigma> aft)" by simp
      have D2: "(?D \<cdot>\<^sub>c \<sigma>)\<langle>l \<cdot> \<tau>s i\<rangle> = Fun f ?ss" by (simp add: ss' s')
      show ?thesis unfolding ss'
      proof (rule exI[of _ "\<tau>s i"], rule disjI2, rule exI[of _ ?D], intro exI conjI)
        show "is_Fun s'" by fact
        show "(l, r) \<in> R" by fact
        show "s' \<cdot> \<sigma> = l \<cdot> \<tau>s i" by fact
        show "Fun f (bef @ D\<langle>s'\<rangle> # aft) = (More f bef D aft)\<langle>s'\<rangle>" by auto
        have "min (length Cs) i = i" using iss' lCs by auto
        then show "hole_pos (More f bef D aft) \<in> hole_poss C" using hpos MFun bef_def iss' lCs by auto
        from bef_def iss' have "length bef = i" by auto
        have iCs: "i < length Cs" using lCs iss' by auto
        let ?k = "sum_list (map length (take i sss)) + j"
        from lssstss have *:"map length (take i tss) = map length (take i sss)"
          by (simp add: lsss ltss nth_map_conv)
        show k:"?k < length ss" by (simp add: concat_nth_length iCs j lsss sss)
        have ssk:"ss ! ?k = sss ! i ! j" using sss concat_nth[OF _ j] iss' lsss lCs by auto
        have tsk:"ts ! ?k = tss ! i ! j" using tss concat_nth[of i tss j ?k] j iss' ltss lsss lCs * lssstss by simp
        from iss' lCs lsss have isss:"i < length sss" by auto
        let ?sssk = "take i sss @ remove_idx j (sss ! i) # drop (Suc i) sss"
        have sssk:"concat ?sssk = remove_idx ?k ss"  using concat_remove_idx[OF isss j] sss by auto
        have "length ?sssk = length ?Cs" using lsss by auto
        moreover have "length (?\<sigma> bef @ ?r # ?\<sigma> aft) = length ?Cs"
          using lCs ss' iss' by auto
        moreover {
          fix l
          assume "l < length ?Cs"
          then have l:"l < length Cs" using iCs by auto
          consider (le) "l < i" | (eq) "l = i" | (gt) "l > i" by fastforce
          then have "(?\<sigma> bef @ ?r # ?\<sigma> aft) ! l =\<^sub>f (?Cs ! l, ?sssk ! l)"
          proof (cases)
            case le
            have "ss' ! l \<cdot> \<sigma> =\<^sub>f (Cs ! l, sss ! l)" using ss'iC[OF l] nth_map lCs l by force
            with le show ?thesis using l lCs lsss by (auto simp: nth_append bef_def aft_def)
          next
            case eq
            then show ?thesis using l lCs lsss D\<sigma>r by (auto simp: nth_append bef_def aft_def)
          next
            case gt
            have "ss' ! l \<cdot> \<sigma> =\<^sub>f (Cs ! l, sss ! l)" using ss'iC[OF l] nth_map lCs l by force
            moreover have "min (length Cs) i = i" using iCs by auto
            ultimately show ?thesis using l lCs lsss gt by (auto simp: nth_append bef_def aft_def)
          qed
        }
        ultimately have "Fun f (?\<sigma> bef @ ?r # ?\<sigma> aft) =\<^sub>f (MFun f ?Cs, concat ?sssk)"
          using eqf_MFunI by fast
        then show "(?D \<cdot>\<^sub>c \<sigma>)\<langle>r \<cdot> \<tau>s i\<rangle> =\<^sub>f (MFun f ?Cs, remove_idx ?k ss)" unfolding D sssk .
        let ?tssk = "take i tss @ remove_idx j (tss ! i) # drop (Suc i) tss"
        have "?k = sum_list (map length (take i tss)) + j" using * by auto
        then have tssk:"concat ?tssk = remove_idx ?k ts" unfolding tss
          using concat_remove_idx[OF _ j[unfolded lssstss[OF isss]]] isss lsss ltss by auto
        have "length ?tssk = length ?Cs" using ltss by auto
        moreover have "length ts' = length ?Cs" using lCs2 lCs iss' by auto
        moreover {
          fix l
          assume "l < length ?Cs"
          then have l:"l < length Cs" using iCs by auto
          consider (le) "l < i" | (eq) "l = i" | (gt) "l > i" by fastforce
          then have "ts' ! l =\<^sub>f (?Cs ! l, ?tssk ! l)"
          proof (cases)
            case le
            with ts'iC[OF l] show ?thesis using l lCs2 ltss by (auto simp: nth_append)
          next
            case eq
            then show ?thesis using l lCs ltss ts'i by (auto simp: nth_append)
          next
            case gt
            moreover have "min (length Cs) i = i" using iCs by auto
            ultimately show ?thesis using l lCs ltss gt ts'iC[OF l] by (auto simp: nth_append)
          qed
        }
        ultimately have "Fun f ts' =\<^sub>f (MFun f ?Cs, concat ?tssk)"
          using eqf_MFunI by fast
        then show "t =\<^sub>f (MFun f ?Cs, remove_idx ?k ts)" unfolding tssk t2 .
      qed
    next
      case False
      with \<tau>s have \<tau>s: "\<And> i. i < length ss' \<Longrightarrow> ?p1 (\<tau>s i) i" by blast
      from Fun(2) have "is_partition (map vars_term ss')" by simp
      from subst_merge[OF this, of \<tau>s] obtain \<tau>
        where \<tau>: "\<And>i x. i < length ss' \<Longrightarrow> x \<in> vars_term (ss' ! i) \<Longrightarrow> \<tau> x = \<tau>s i x" by auto
      { fix i
        assume i: "i < length ss'"
        then have mem: "ss' ! i \<in> set ss'" by auto
        from \<tau>s[OF i] have p1: "?p1 (\<tau>s i) i" .
        have id: "ss' ! i \<cdot> (\<tau>s i) = ss' ! i \<cdot> \<tau>" by (rule term_subst_eq, rule \<tau>[OF i, symmetric])
        have "?p1 \<tau> i"
        proof (rule conjI[OF _ ballI])
          fix x
          assume x: "x \<in> vars_term (ss' ! i)"
          with p1 have step: "(\<sigma> x, \<tau>s i x) \<in> par_rstep R" by auto
          with \<tau>[OF i x] show "(\<sigma> x, \<tau> x) \<in> par_rstep R" by simp
        qed (insert p1[unfolded id], auto)
      } note p1 = this
      have p1: "\<And> i. i < length ss' \<Longrightarrow> ?p1 \<tau> i" by (rule p1)
      let ?ss = "map (\<lambda> s. s \<cdot> \<tau>) ss'"
      show ?thesis
      proof (rule exI[of _ \<tau>], rule disjI1, rule conjI[OF _ ballI])
        have "ts' = map (\<lambda> i. ts' ! i) [0 ..< (length ts')]" by (rule map_nth[symmetric])
        also have "...  = map (\<lambda> i. ?ss ! i) [0 ..< length ?ss]"  using p1 lCs2 lCs by auto
        also have "... = ?ss" by (rule map_nth)
        finally have ts: "ts' = ?ss" .
        then show "t = Fun f ss' \<cdot> \<tau>" by auto
      next
        fix x
        assume "x \<in> vars_term (Fun f ss')"
        then obtain s where s: "s \<in> set ss'" and x: "x \<in> vars_term s" by auto
        from s[unfolded set_conv_nth] obtain i where i: "i < length ss'" and s: "s = ss' ! i" by auto
        from p1[OF i] x[unfolded s]
        show "(\<sigma> x, \<tau> x) \<in> par_rstep R" by blast
      qed
    qed
  qed (insert s t, auto dest: eqfE)
qed

lemma step_in_subst_imp_par_diamond:
  assumes "t = r \<cdot> \<sigma>"
  and "u = l \<cdot> \<tau>"
  and "\<forall>x\<in>vars_term l. (\<sigma> x, \<tau> x) \<in> par_rstep R2"
  and "(l, r) \<in> R1"
  shows "\<exists>v. (t, v) \<in> par_rstep R2 \<and> (u, v) \<in> par_rstep R1"
proof -
  define \<delta> where "\<delta> = (\<lambda>x. if x \<in> vars_term l then \<tau> x else \<sigma> x)"
  have "\<forall>x\<in>vars_term l. \<tau> x = \<delta> x" by (auto simp: \<delta>_def)
  then have "l \<cdot> \<tau> = l \<cdot> \<delta>" using term_subst_eq_conv by auto
  moreover have "(l \<cdot> \<delta>, r \<cdot> \<delta>) \<in> par_rstep R1" using assms rstep_par_rstep by blast
  ultimately have "(l \<cdot> \<tau>, r \<cdot> \<delta>) \<in> par_rstep R1"  by auto
  moreover have "(r \<cdot> \<sigma>, r \<cdot> \<delta>) \<in> par_rstep R2"
  proof (rule all_ctxt_closed_subst_step)
    fix x
    assume "x \<in> vars_term r"
    show "(\<sigma> x, \<delta> x) \<in> par_rstep R2" using assms by (cases "x \<in> vars_term l") (auto simp: \<delta>_def)
  qed auto
  ultimately show ?thesis using assms par_rstep_rsteps by auto
qed

(* TODO : move*)
lemma ctxt_subst_eq:
  assumes "\<And>x. x \<in> vars_ctxt C \<Longrightarrow> \<sigma> x = \<tau> x"
  shows "C \<cdot>\<^sub>c \<sigma> = C \<cdot>\<^sub>c \<tau>"
using assms
proof (induct C)
  case (More f bef C aft)
  { fix t
    assume t:"t \<in> set bef"
    have "t \<cdot> \<sigma> = t \<cdot> \<tau>" using t More(2) by (auto intro: term_subst_eq)
  }
  moreover
  { fix t
    assume t:"t \<in> set aft"
    have "t \<cdot> \<sigma> = t \<cdot> \<tau>" using t More(2) by (auto intro: term_subst_eq)
  }
  moreover have "C \<cdot>\<^sub>c \<sigma> = C \<cdot>\<^sub>c \<tau>" using More by auto
  ultimately show ?case by auto
qed auto

lemma par_rstep_mctxt_Var_diamond:
  fixes u t :: "('f, 'v) term"
  assumes "(Var x, t) \<in> par_rstep_mctxt R\<^sub>1 C ss ts"
    and "(Var x, u) \<in> par_rstep_mctxt R\<^sub>2 D vs us"
    and left_lin: "left_linear_trs R\<^sub>1" "left_linear_trs R\<^sub>2"
  shows "\<exists>v. (t, v) \<in> par_rstep R\<^sub>2 \<and> (u, v) \<in> par_rstep R\<^sub>1"
proof -
  from assms have s: "Var x =\<^sub>f (C, ss)" and t: "t =\<^sub>f (C ,ts)"
    and s2: "Var x =\<^sub>f (D, vs)" and u:"u =\<^sub>f (D, us)"
    and steps1: "\<forall>i<length ss. (ss ! i, ts ! i) \<in> rrstep R\<^sub>1"
      unfolding par_rstep_mctxt_def by (auto dest!: eqfE)
  then have "Var x = fill_holes C ss" and "length ss = num_holes C" and "Var x = fill_holes D vs"
    by (auto dest: eqfE)
  then consider (MVar1) "C = MVar x" | (MVar2) "D = MVar x"
    | (MHole) "C = MHole" and "ss = [Var x]" and "D = MHole"
    by (cases C, auto) (cases D, auto, cases ss, auto)
  then show ?thesis
  proof (cases)
    case MVar1
    with assms t have "(t, u) \<in> par_rstep_mctxt R\<^sub>2 D vs us" by (auto dest: eqfE)
    then show ?thesis using par_rstep_par_rstep_mctxt_conv by blast
  next
    case MVar2
    with assms u have "(u, t) \<in> par_rstep_mctxt R\<^sub>1 C ss ts" by (auto dest: eqfE)
    then show ?thesis using par_rstep_par_rstep_mctxt_conv by blast
  next
    case MHole
    moreover have "ts = [t]" using t  MHole by (auto elim: eqf_MHoleE)
    ultimately obtain l r \<sigma> where lr: "(l, r) \<in> R\<^sub>1" and l:"Var x = l \<cdot> \<sigma>" and r:"t = r \<cdot> \<sigma>"
      using rrstep_imp_rule_subst steps1 by fastforce
    from lr left_lin have "linear_term l" by (auto simp: left_linear_trs_def)
    from par_rstep_mctxt_linear_subst[OF this assms(2)[unfolded l]]
    obtain \<tau> where \<tau>:"u = l \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term l. (\<sigma> x, \<tau> x) \<in> par_rstep R\<^sub>2) \<or>
          (\<exists>E s'. l = E\<langle>s'\<rangle> \<and> is_Fun s' \<and> hole_pos E \<in> hole_poss D)"
      by blast
    then show ?thesis
    proof
      assume "u = l \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term l. (\<sigma> x, \<tau> x) \<in> par_rstep R\<^sub>2)"
      with step_in_subst_imp_par_diamond r lr show ?thesis using par_rstep_rsteps by fastforce
    next
      assume "\<exists>E s'. l = E\<langle>s'\<rangle> \<and> is_Fun s' \<and> hole_pos E \<in> hole_poss D"
      then obtain E l'' where E:"l = E\<langle>l''\<rangle>" and l'':"is_Fun l''" and "hole_pos E \<in> hole_poss D" by auto
      with MHole have "E = \<box>" by (cases E) auto
      with E l'' l show ?thesis by auto
    qed
  qed
qed


lemma parallel_closed_strongly_commute:
  assumes closed_1:"\<And>p q b. (b, p, q) \<in> critical_pairs R2 R1 \<Longrightarrow> \<exists>v. (q, v) \<in> par_rstep R2 \<and> (p, v) \<in> (rstep R1)\<^sup>*"
    and closed_2: "\<And>p q. (False, p, q) \<in> critical_pairs R1 R2 \<Longrightarrow> (q, p) \<in> par_rstep R1"
    and left_lin: "left_linear_trs R1" "left_linear_trs R2"
  shows "strongly_commute (par_rstep R1) (par_rstep R2)"
proof (rule strongly_commuteI)
  fix s t u
  let ?pR1 = "par_rstep R1"
  let ?pR2 = "par_rstep R2"
  assume "(s, t) \<in> ?pR1" and "(s, u) \<in> ?pR2"
  with par_rstep_par_rstep_mctxt_conv obtain C ss ts and D us vs
    where "(s, t) \<in> par_rstep_mctxt R1 C ss ts" and "(s, u) \<in> par_rstep_mctxt R2 D vs us"
    by metis
  then have "\<exists>v. (t, v) \<in> ?pR2 \<and> (u, v) \<in> (rstep R1)\<^sup>*"
  proof (induct "overlap' (C,ss) (D, vs)" arbitrary: C D ss vs s t us ts u rule: wf_induct_rule[OF wf_mult[OF SN_imp_wf[OF SN_supt]]])
    case 1
    then show ?case
    proof (induct s arbitrary: C D ss vs t ts u us)
      case (Var x)
      with par_rstep_mctxt_Var_diamond left_lin par_rstep_rsteps show ?case by fast
    next
      case (Fun f ss')
      then have s: "Fun f ss' =\<^sub>f (C, ss)" and t: "t =\<^sub>f (C ,ts)"
        and s2: "Fun f ss' =\<^sub>f (D, vs)" and u:"u =\<^sub>f (D, us)"
        and len1: "length ss = length ts" and len2 : "length vs = length us"
        and steps1: "\<forall>i<length ss. (ss ! i, ts ! i) \<in> rrstep R1"
        and steps2: "\<forall>i<length vs. (vs ! i, us ! i) \<in> rrstep R2"
        unfolding par_rstep_mctxt_def by (auto dest!: eqfE)
      show ?case
      proof (cases C)
        case MHole
        then have ss:"ss = [Fun f ss']" using s by (auto elim: eqf_MHoleE)
        moreover have "ts = [t]" using t MHole by (auto elim: eqf_MHoleE)
        ultimately have "(Fun f ss', t) \<in> rrstep R1" using steps1 by auto
        then obtain l r \<sigma> where lr: "(l, r) \<in> R1" and l:"Fun f ss' = l \<cdot> \<sigma>" and r:" t = r \<cdot> \<sigma> "
          using rrstep_imp_rule_subst by fastforce
        from lr left_lin have linl:"linear_term l" by (auto simp: left_linear_trs_def)
        show ?thesis
        proof (cases D)
          case (MFun g Ds)
          then have [simp]: "g = f" using s2 by (auto dest: eqfE)
          obtain vss where lDs:"length ss' = length Ds" and lvss:"length vss = length Ds"
            and vss:"vs = concat vss" and vs'iD:"\<And> i. i < length Ds \<Longrightarrow> ss' ! i =\<^sub>f (Ds ! i, vss ! i)"
            using MFun s2 by (auto elim: eqf_MFunE)
          from u MFun obtain us' where [simp]: "u = Fun f us'" by (auto elim: eqf_MFunE)
          obtain uss where lDs2:"length us' = length Ds" and luss:"length uss = length Ds"
            and uss:"us = concat uss" and us'iD:"\<And> i. i < length Ds \<Longrightarrow> us' ! i =\<^sub>f (Ds ! i, uss ! i)"
          using MFun u by (auto elim: eqf_MFunE)
          { fix i
            assume i:"i < length vss"
            then have "ss' ! i =\<^sub>f (Ds ! i, vss ! i)" using lvss vs'iD by auto
            moreover have "us' ! i =\<^sub>f (Ds ! i, uss ! i)" using luss us'iD i lvss by auto
            ultimately have "length (vss ! i) = length (uss ! i)"  by (auto dest!: eqfE)
          } note lvssuss = this
          from par_rstep_mctxt_linear_subst[OF linl Fun(4)[unfolded l]]
          obtain \<tau> where \<tau>:"u = l \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term l. (\<sigma> x, \<tau> x) \<in> ?pR2) \<or>
            (\<exists>E s' l' r' j D'. l = E\<langle>s'\<rangle> \<and> is_Fun s' \<and> (l', r') \<in> R2 \<and> s' \<cdot> \<sigma> = l' \<cdot> \<tau>
              \<and> hole_pos E \<in> hole_poss D \<and> j < length vs
              \<and> (E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle> =\<^sub>f (D', remove_idx j vs) \<and> u =\<^sub>f (D', remove_idx j us))"
            by blast
          then show ?thesis
          proof
            assume "u = l \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term l. (\<sigma> x, \<tau> x) \<in> ?pR2)"
            with step_in_subst_imp_par_diamond r lr show ?thesis using par_rstep_rsteps by fastforce
          next
            assume "\<exists>E s' l' r' j D'. l = E\<langle>s'\<rangle> \<and> is_Fun s' \<and> (l', r') \<in> R2 \<and> s' \<cdot> \<sigma> = l' \<cdot> \<tau>
              \<and> hole_pos E \<in> hole_poss D \<and> j < length vs
              \<and> (E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle> =\<^sub>f (D', remove_idx j vs) \<and> u =\<^sub>f (D', remove_idx j us)"
            then obtain E l'' l' r' j D' where E:"l = E\<langle>l''\<rangle>" and l'':"is_Fun l''" and lr':"(l', r') \<in> R2"
              and unifiable:"l'' \<cdot> \<sigma> = l' \<cdot> \<tau>" and hpos:"hole_pos E \<in> hole_poss D" and j:"j < length vs"
              and E\<sigma>r:"(E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle> =\<^sub>f (D', remove_idx j vs)"
              and uD:"u =\<^sub>f (D', remove_idx j us)"
              by auto
            obtain \<mu>1 \<mu>2 \<delta> where mgu:"mgu_var_disjoint_string l'' l' = Some (\<mu>1, \<mu>2)"
              and \<sigma>:"\<sigma> = \<mu>1 \<circ>\<^sub>s \<delta>" and \<tau>:"\<tau> = \<mu>2 \<circ>\<^sub>s \<delta>" and \<mu>:"l'' \<cdot> \<mu>1 = l' \<cdot> \<mu>2"
              using mgu_var_disjoint_string_complete[OF unifiable] by auto
            have inner:"hole_pos E \<noteq> \<epsilon>" using hpos MFun by auto
            have "(False, r \<cdot> \<mu>1, (E \<cdot>\<^sub>c \<mu>1)\<langle>r' \<cdot> \<mu>2\<rangle>) \<in> critical_pairs R1 R2"
              using lr lr' E l'' mgu inner by (force intro: critical_pairsI)
            from closed_2[OF this] have closed:"((E \<cdot>\<^sub>c \<mu>1)\<langle>r' \<cdot> \<mu>2\<rangle>, r \<cdot> \<mu>1) \<in> par_rstep R1" by auto
            from closed have "((E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>, r \<cdot> \<sigma>) \<in> par_rstep R1" unfolding \<sigma> \<tau>
              using subst_closed_par_rstep by fastforce
            from par_rstep_mctxtE[OF this] obtain F cs ds where EF:"(E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle> =\<^sub>f (F, cs)" and "r \<cdot> \<sigma> =\<^sub>f (F, ds)"
              and "length ds = length cs" and "\<forall>i < length cs. (cs ! i, ds ! i) \<in> rrstep R1" by metis
            then have Er: "((E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>, r \<cdot> \<sigma>) \<in> par_rstep_mctxt R1 F cs ds"
              unfolding par_rstep_mctxt_def by auto
            let ?ol' = "overlap' (F, cs) (D', remove_idx j vs)"
            have "(?ol', mset (remove_idx j vs)) \<in> (mult {\<lhd>})\<^sup>="
              using overlap'_bounded[OF E\<sigma>r EF] overlap'_symmetric[OF EF E\<sigma>r] by auto
            moreover have "(mset (remove_idx j vs), mset vs) \<in> mult {\<lhd>}"
              using remove_idx_mult j by auto
            moreover have "overlap' (C, ss) (D, vs) = mset vs" using MFun MHole ss by auto
            ultimately have ol:"(?ol', overlap' (C, ss) (D, vs)) \<in> mult {\<lhd>}"
               unfolding mult_def by simp
            from len2 have lenrem:"length (remove_idx j us) = length (remove_idx j vs)"
              by (simp add: remove_idx_def)
            have stepsrem:"\<forall>i<length (remove_idx j vs). (remove_idx j vs ! i, remove_idx j us ! i) \<in> rrstep R2"
              using steps2 len2 j by (metis adjust_idx_length adjust_idx_nth)
            from E\<sigma>r uD lenrem stepsrem have Eu:"((E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>, u) \<in> par_rstep_mctxt R2 D' (remove_idx j vs) (remove_idx j us)"
              unfolding par_rstep_mctxt_def by auto
            from Fun(2)[OF ol Er Eu] show ?thesis using r by auto
          qed
        next
          case MHole
          then have "vs = [Fun f ss']" using s2 by (auto elim: eqf_MHoleE)
          moreover have "us = [u]" using u MHole by (auto elim: eqf_MHoleE)
          ultimately have "(Fun f ss', u) \<in> rrstep R2" using steps2 by auto
          then obtain l' r' \<sigma>' where lr': "(l', r') \<in> R2" and l':"Fun f ss' = l' \<cdot> \<sigma>'"
            and r':" u = r' \<cdot> \<sigma>'"
            using rrstep_imp_rule_subst by fastforce
          with l have eq:"l' \<cdot> \<sigma>' = l \<cdot> \<sigma>" by auto
          from lr' left_lin have "linear_term l'" by (auto simp: left_linear_trs_def)
          from par_rstep_mctxt_linear_subst[OF this Fun(3)[unfolded l']]
          obtain \<tau> where \<tau>:"t = l' \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term l'. (\<sigma>' x, \<tau> x) \<in> par_rstep R1) \<or>
            (\<exists>E s'. l' = E\<langle>s'\<rangle> \<and> is_Fun s' \<and> hole_pos E \<in> hole_poss C)"
            by blast
          then show ?thesis
          proof
            assume "t = l' \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term l'. (\<sigma>' x, \<tau> x) \<in> par_rstep R1)"
            with step_in_subst_imp_par_diamond r' lr' show ?thesis using par_rstep_rsteps by fastforce
          next
            assume "\<exists>E s'. l' = E\<langle>s'\<rangle> \<and> is_Fun s' \<and> hole_pos E \<in> hole_poss C"
            then obtain E s' where "l' = E\<langle>s'\<rangle>" and "is_Fun s'" and "hole_pos E \<in> hole_poss C" by blast
            with \<open>C = MHole\<close> have isFun:"is_Fun l'" by (cases E) auto
            from mgu_var_disjoint_string_complete[OF eq] obtain \<mu>1 \<mu>2 \<delta>
              where unif:"mgu_var_disjoint_string l' l = Some (\<mu>1, \<mu>2)"
              and \<sigma>:"\<sigma>' = \<mu>1 \<circ>\<^sub>s \<delta>" and \<sigma>':"\<sigma> = \<mu>2 \<circ>\<^sub>s \<delta>"
              by auto
            from critical_pairsI[OF lr' lr _ _ unif] isFun
            have "(True, r' \<cdot> \<mu>1, r \<cdot> \<mu>2) \<in> critical_pairs R2 R1" by force
            from closed_1[OF this] obtain w
              where "(r \<cdot> \<mu>2, w) \<in> par_rstep R2" and "(r' \<cdot> \<mu>1, w) \<in> (rstep R1)\<^sup>*" by auto
            with \<sigma> \<sigma>' have "(r \<cdot> \<sigma>, w \<cdot> \<delta>) \<in> par_rstep R2" "(r' \<cdot> \<sigma>', w \<cdot> \<delta>) \<in> (rstep R1)\<^sup>*"
              by (auto simp: subst_closed_par_rstep rsteps_closed_subst)
            with r r' show ?thesis using par_rsteps_rsteps by auto
          qed
        qed (insert s2, auto dest: eqfE)
      next
        case (MFun g Cs) note C = this
        then have [simp]: "g = f" using s by (auto dest: eqfE)
        obtain sss where lCs:"length ss' = length Cs" and lsss:"length sss = length Cs"
          and sss:"ss = concat sss" and ss'iC:"\<And> i. i < length Cs \<Longrightarrow> ss' ! i =\<^sub>f (Cs ! i, sss ! i)"
        using C s by (auto elim: eqf_MFunE)
        from s have "Fun f ss' = fill_holes C ss" "num_holes C = length ss" by (auto dest: eqfE)
        from t C obtain ts' where [simp]: "t = Fun f ts'" by (auto elim: eqf_MFunE)
        obtain tss where lCs2:"length ts' = length Cs" and ltss:"length tss = length Cs"
          and tss:"ts = concat tss" and ts'iC:"\<And> i. i < length Cs \<Longrightarrow> ts' ! i =\<^sub>f (Cs ! i, tss ! i)"
        using MFun t by (auto elim: eqf_MFunE)
        { fix i
          assume i:"i < length sss"
          then have "ss' ! i =\<^sub>f (Cs ! i, sss ! i)" using lsss ss'iC by auto
          moreover have "ts' ! i =\<^sub>f (Cs ! i, tss ! i)" using ltss ts'iC i lsss by auto
          ultimately have "length (sss ! i) = length (tss ! i)"  by (auto dest!: eqfE)
        } note lssstss = this
        show ?thesis
        proof (cases D)
          case MHole
          then have vs:"vs = [Fun f ss']" using s2 by (auto elim: eqf_MHoleE)
          moreover have "us = [u]" using u MHole by (auto elim: eqf_MHoleE)
          ultimately have "(Fun f ss', u) \<in> rrstep R2" using steps2 by auto
          then obtain l r \<sigma> where lr: "(l, r) \<in> R2" and l:"Fun f ss' = l \<cdot> \<sigma>" and r: "u = r \<cdot> \<sigma>"
            using rrstep_imp_rule_subst by fastforce
          from lr left_lin have "linear_term l" by (auto simp: left_linear_trs_def)
          from par_rstep_mctxt_linear_subst[OF this Fun(3)[unfolded l]]
          obtain \<tau> where \<tau>:"t = l \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term l. (\<sigma> x, \<tau> x) \<in> ?pR1) \<or>
            (\<exists>D s' l' r' j C'. l = D\<langle>s'\<rangle> \<and> is_Fun s' \<and> (l', r') \<in> R1 \<and> s' \<cdot> \<sigma> = l' \<cdot> \<tau>
              \<and> hole_pos D \<in> hole_poss C \<and> j < length ss
              \<and> (D \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle> =\<^sub>f (C', remove_idx j ss) \<and> t =\<^sub>f (C', remove_idx j ts))"
            by blast
          then show ?thesis
          proof
            assume *:"t = l \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term l. (\<sigma> x, \<tau> x) \<in> ?pR1)"
            with step_in_subst_imp_par_diamond r lr show ?thesis using par_rstep_rsteps by fastforce
          next
            assume "\<exists>D s' l' r' j C'. l = D\<langle>s'\<rangle> \<and> is_Fun s' \<and> (l', r') \<in> R1 \<and> s' \<cdot> \<sigma> = l' \<cdot> \<tau>
              \<and> hole_pos D \<in> hole_poss C \<and> j < length ss
              \<and> (D \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle> =\<^sub>f (C', remove_idx j ss) \<and> t =\<^sub>f (C', remove_idx j ts)"
            then obtain E l'' l' r' j C' where E:"l = E\<langle>l''\<rangle>" and l'':"is_Fun l''" and lr':"(l', r') \<in> R1"
              and unifiable:"l'' \<cdot> \<sigma> = l' \<cdot> \<tau>" and hpos:"hole_pos E \<in> hole_poss C" and j:"j < length ss"
              and E\<sigma>r:"(E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle> =\<^sub>f (C', remove_idx j ss)" and tCtsj:"t =\<^sub>f (C', remove_idx j ts)"
              by auto
            obtain \<mu>1 \<mu>2 \<delta> where mgu:"mgu_var_disjoint_string l'' l' = Some (\<mu>1, \<mu>2)"
              and \<sigma>:"\<sigma> = \<mu>1 \<circ>\<^sub>s \<delta>" and \<tau>:"\<tau> = \<mu>2 \<circ>\<^sub>s \<delta>" and \<mu>:"l'' \<cdot> \<mu>1 = l' \<cdot> \<mu>2"
              using mgu_var_disjoint_string_complete[OF unifiable] by auto
            have inner:"hole_pos E \<noteq> \<epsilon>" using hpos C by auto
            have "(False, r \<cdot> \<mu>1, (E \<cdot>\<^sub>c \<mu>1)\<langle>r' \<cdot> \<mu>2\<rangle>) \<in> critical_pairs R2 R1"
              using lr lr' E l'' mgu inner by (force intro: critical_pairsI)
            with closed_1[OF this] obtain v where closed1:"((E \<cdot>\<^sub>c \<mu>1)\<langle>r' \<cdot> \<mu>2\<rangle>, v) \<in> ?pR2"
              and closed2:"(r \<cdot> \<mu>1, v) \<in> (rstep R1)\<^sup>*" by auto
            from closed1 have "((E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>, v \<cdot> \<delta>) \<in> par_rstep R2" unfolding \<sigma> \<tau>
              using subst_closed_par_rstep by fastforce
            from par_rstep_mctxtE[OF this] obtain F cs ds where EF:"(E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle> =\<^sub>f (F, cs)" and "v \<cdot> \<delta> =\<^sub>f (F, ds)"
              and "length ds = length cs" and "\<forall>i < length cs. (cs ! i, ds ! i) \<in> rrstep R2" by metis
            then have Ev:"((E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>, v \<cdot> \<delta>) \<in> par_rstep_mctxt R2 F cs ds"
               unfolding par_rstep_mctxt_def by auto
            let ?ol' = "overlap' (C', remove_idx j ss) (F, cs)"
            have "(?ol', mset (remove_idx j ss)) \<in> (mult {\<lhd>})\<^sup>=" using overlap'_bounded[OF E\<sigma>r EF] by auto
            moreover have "(mset (remove_idx j ss), mset ss) \<in> mult {\<lhd>}" using remove_idx_mult j by auto
            moreover have "overlap' (C, ss) (D, vs) = mset ss" using C MHole vs by auto
            ultimately have ol:"(?ol', overlap' (C, ss) (D, vs)) \<in> mult {\<lhd>}" unfolding mult_def by simp
            from len1 have lenrem:"length (remove_idx j ts) = length (remove_idx j ss)"
              by (simp add: remove_idx_def)
            have stepsrem:"\<forall>i<length (remove_idx j ss). (remove_idx j ss ! i, remove_idx j ts ! i) \<in> rrstep R1"
              using len1 steps1 adjust_idx_length adjust_idx_nth j by metis
            from E\<sigma>r tCtsj lenrem stepsrem
            have Et:"((E \<cdot>\<^sub>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>, t) \<in> par_rstep_mctxt R1 C' (remove_idx j ss) (remove_idx j ts)"
               unfolding par_rstep_mctxt_def by auto
            from Fun(2)[OF ol  Et Ev]
            show ?thesis using r rsteps_closed_subst[OF closed2, of \<delta>] unfolding \<sigma> by auto
          qed
        next
          case (MFun h Ds)
          then have [simp]: "h = f" using s2 by (auto dest: eqfE)
          obtain vss where lDs:"length ss' = length Ds" and lvss:"length vss = length Ds"
            and vss:"vs = concat vss" and vs'iD:"\<And> i. i < length Ds \<Longrightarrow> ss' ! i =\<^sub>f (Ds ! i, vss ! i)"
            using MFun s2  by (auto elim: eqf_MFunE)
          from u MFun obtain us' where [simp]: "u = Fun f us'" by (auto elim: eqf_MFunE)
          obtain uss where lDs2:"length us' = length Ds" and luss:"length uss = length Ds"
            and uss:"us = concat uss" and us'iD:"\<And> i. i < length Ds \<Longrightarrow> us' ! i =\<^sub>f (Ds ! i, uss ! i)"
          using MFun u by (auto elim: eqf_MFunE)
          { fix i
            assume i:"i < length vss"
            then have "ss' ! i =\<^sub>f (Ds ! i, vss ! i)" using lvss vs'iD by auto
            moreover have "us' ! i =\<^sub>f (Ds ! i, uss ! i)" using luss us'iD i lvss by auto
            ultimately have "length (vss ! i) = length (uss ! i)"  by (auto dest!: eqfE)
          } note lvssuss = this
          { fix i
            assume i:"i < length ss'"
            then have s: "ss' ! i \<in> set ss'" using nth_mem by blast
            from i ts'iC lCs lCs2 have ts'i:"ts' ! i =\<^sub>f (Cs ! i, tss ! i)" by auto
            from i us'iD lDs lDs2 have us'i:"us' ! i =\<^sub>f (Ds ! i, uss ! i)" by auto
            have "sss ! i = partition_holes ss Cs ! i"
              by (metis eqfE(2) lsss partition_holes_concat_id ss'iC sss)
            moreover have " vss ! i = partition_holes vs Ds ! i"
              by (metis eqfE(2) lvss partition_holes_concat_id vs'iD vss)
            ultimately have "overlap' (Cs ! i, sss ! i) (Ds ! i, vss ! i) \<in>#
              mset (map (\<lambda>i. overlap' (Cs ! i, partition_holes ss Cs ! i) (Ds ! i, partition_holes vs Ds ! i)) [0..<length Cs])"
              unfolding sss vss using i lCs by fastforce
            then have ol:"overlap' (Cs ! i, sss ! i) (Ds ! i, vss ! i) \<subseteq># overlap' (C, ss) (D, vs)"
              unfolding C MFun by (simp add: in_mset_subset_Union)
            { fix C' D' ss' vs' s' t' ts' u' us'
              assume 1:"(overlap' (C', ss') (D', vs'), overlap' (Cs ! i, sss ! i) (Ds ! i, vss ! i)) \<in> mult {\<lhd>}"
              and 2: "(s', t') \<in> par_rstep_mctxt R1 C' ss' ts'" "(s', u') \<in> par_rstep_mctxt R2 D' vs' us'"
              have "trans {\<lhd>}" unfolding trans_def using supt_trans by auto
              with mult_subset_mult ol 1 have "(overlap' (C', ss') (D', vs'), overlap' (C, ss) (D, vs)) \<in> mult {\<lhd>}" by blast
              from Fun(2)[OF this 2]  have "\<exists>v. (t', v) \<in> ?pR2 \<and> (u', v) \<in> (rstep R1)\<^sup>*" .
            } note m = this
            { fix j
              assume j:"j < length (sss ! i)"
              let ?k = "sum_list (map length (take i sss)) + j"
              from lssstss have *:"map length (take i tss) = map length (take i sss)"
                by (simp add: lsss ltss nth_map_conv)
              from concat_nth_length have "?k < length ss"
                  using sss i j lCs lsss by metis
              moreover have "ss ! ?k = sss ! i ! j" using sss concat_nth[OF _ j] i lsss lCs by auto
              moreover have "ts ! ?k = tss ! i ! j" using tss concat_nth[of i tss j ?k]
                j i ltss lsss lCs * lssstss by simp
              ultimately have "(sss ! i ! j, tss ! i ! j) \<in> rrstep R1" using steps1 by auto
            }
            then have "\<forall>j<length (sss ! i). (sss ! i ! j, tss ! i ! j) \<in> rrstep R1" by blast
            with ss'iC i lCs ts'i
            have ssts:"(ss' ! i, ts' ! i) \<in> par_rstep_mctxt R1 (Cs ! i) (sss ! i) (tss ! i)"
              unfolding par_rstep_mctxt_def by auto
            { fix j
              assume j:"j < length (vss ! i)"
              let ?k = "sum_list (map length (take i vss)) + j"
              from lvssuss have *:"map length (take i uss) = map length (take i vss)"
                by (simp add: luss lvss nth_map_conv)
              from concat_nth_length have "?k < length vs"
                using i j lDs lvss vss by metis
              moreover have "vs ! ?k = vss ! i ! j" using vss concat_nth[OF _ j] i lvss lDs
                by auto
              moreover have "us ! ?k = uss ! i ! j" using uss concat_nth[of i uss j ?k]
                j i luss lvss lDs * lvssuss by simp
              ultimately have "(vss ! i ! j, uss ! i ! j) \<in> rrstep R2" using steps2 by auto
            }
            then have "\<forall>j<length (vss ! i). (vss ! i ! j, uss ! i ! j) \<in> rrstep R2" by blast
            with vs'iD i lDs us'i
            have "(ss' ! i, us' ! i) \<in> par_rstep_mctxt R2 (Ds ! i) (vss ! i) (uss ! i)"
              unfolding par_rstep_mctxt_def by auto
            from Fun(1)[OF s m ssts this]
            have "\<exists>v. (ts' ! i, v) \<in> ?pR2 \<and> (us' ! i, v) \<in> (rstep R1)\<^sup>*" by auto
          }
          then obtain v
          where v: "\<And>i. i < length ts' \<Longrightarrow> (ts' ! i, v i) \<in> ?pR2 \<and> (us' ! i, v i) \<in> (rstep R1)\<^sup>*"
            using lCs lCs2 by metis
          let ?vs' = "map v [0..<length ts']"
          from v have v1:"(Fun f ts', Fun f ?vs') \<in> ?pR2" by auto
          from v have *:"\<And>i. i < length us' \<Longrightarrow> (us' ! i, v i) \<in> (rstep R1)\<^sup>*"
            using par_rsteps_rsteps lCs lCs2 lDs lDs2  by auto
          have "length us' = length ?vs'" using lCs lCs2 lDs lDs2 by auto
          with * args_rsteps_imp_rsteps[OF this] have v2:"(Fun f us', Fun f ?vs') \<in> (rstep R1)\<^sup>*"
            by auto
          from v1 v2 show ?thesis by auto
        qed (insert s2, auto dest: eqfE)
      qed (insert s, auto dest: eqfE)
    qed
  qed
  then show "\<exists>v. (t, v) \<in> ?pR2\<^sup>= \<and> (u, v) \<in> ?pR1\<^sup>*" by (auto simp: par_rsteps_rsteps)
qed

corollary parallel_closed_commute:
  assumes closed_1:"\<And>p q b. (b, p, q) \<in> critical_pairs R2 R1 \<Longrightarrow> \<exists>v. (q, v) \<in> par_rstep R2 \<and> (p, v) \<in> (rstep R1)\<^sup>*"
    and closed_2: "\<And>p q. (False, p, q) \<in> critical_pairs R1 R2 \<Longrightarrow> (q, p) \<in> par_rstep R1"
    and left_lin: "left_linear_trs R1" "left_linear_trs R2"
  shows "commute (rstep R1) (rstep R2)"
proof
  fix x y\<^sub>1 y\<^sub>2
  assume "(x, y\<^sub>1) \<in> (rstep R1)\<^sup>*" and "(x, y\<^sub>2) \<in> (rstep R2)\<^sup>*"
  then have "(x, y\<^sub>1) \<in> (par_rstep R1)\<^sup>*" and "(x, y\<^sub>2) \<in> (par_rstep R2)\<^sup>*"
    using rtrancl_mono[OF rstep_par_rstep] by auto
  from commuteE[OF strongly_commute_imp_commute[OF parallel_closed_strongly_commute[OF assms]] this]
  obtain z where "(y\<^sub>1, z) \<in> (par_rstep R2)\<^sup>* \<and> (y\<^sub>2, z) \<in> (par_rstep R1)\<^sup>*" by fast
  then show "\<exists>z. (y\<^sub>1, z) \<in> (rstep R2)\<^sup>* \<and> (y\<^sub>2, z) \<in> (rstep R1)\<^sup>*"
    using rtrancl_mono[OF par_rstep_rsteps] rtrancl_idemp by auto
qed

corollary parallel_closed_imp_CR:
  assumes "\<And>p q. (False, p, q) \<in> critical_pairs R R \<Longrightarrow> (q, p) \<in> par_rstep R"
    and "\<And>p q. (True, p, q) \<in> critical_pairs R R \<Longrightarrow> \<exists>v. (q, v) \<in> par_rstep R \<and> (p, v) \<in> (rstep R)\<^sup>*"
    and "left_linear_trs R"
  shows "CR (rstep R)"
proof -
  { fix p q b
    assume "(b, p, q) \<in> critical_pairs R R"
    then have "\<exists>v. (q, v) \<in> par_rstep R \<and> (p, v) \<in> (rstep R)\<^sup>*"
      using assms(1,2) by (cases b) auto
  } moreover
  { fix p q
    assume "(False, p, q) \<in> critical_pairs R R"
    with assms(1) have "(q, p) \<in> par_rstep R" by auto
  }
  ultimately have "commute (rstep R) (rstep R)"
    using parallel_closed_commute assms(3) by blast
  then show ?thesis using CR_iff_self_commute by auto
qed

end
