theory Lambda_Z
imports
  Main
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
begin

datatype 'a "term" =
  Var 'a
| App "'a term" "'a term"
| Abs 'a "'a term"

fun subst
where
  "subst x s (Var y) = (if x = y then s else Var y)"
| "subst x s (App t u) = App (subst x s t) (subst x s u)"
| "subst x s (Abs y t) = (if x = y then Abs y t else Abs y (subst x s t))"

fun frees
where
  "frees (Var x) = {x}"
| "frees (App s t) = frees s \<union> frees t"
| "frees (Abs x t) = frees t - {x}"

fun free_for
where
  "free_for s x (Var y) \<longleftrightarrow> True"
| "free_for s x (App t u) \<longleftrightarrow> free_for s x t \<and> free_for s x u"
| "free_for s x (Abs y t) \<longleftrightarrow> y \<notin> frees s \<and> free_for s x t"

lemma not_free_subst_id [simp]:
  "x \<notin> frees t \<Longrightarrow> subst x s t = t"
by (induct t) auto

lemma subst_subst:  
  assumes "x \<noteq> y" and "x \<notin> frees u" and "free_for s x t"
  shows "subst y u (subst x s t) = subst x (subst y u s) (subst y u t)"
using assms by (induct t) auto

inductive_set Beta
where
  "free_for t x s \<Longrightarrow> (App (Abs x s) t, subst x t s) \<in> Beta"
| "(s, t) \<in> Beta \<Longrightarrow> (App s u, App t u) \<in> Beta"
| "(s, t) \<in> Beta \<Longrightarrow> (App u s, App u t) \<in> Beta"
| "(s, t) \<in> Beta \<Longrightarrow> (Abs x s, Abs x t) \<in> Beta"

abbreviation beta (infix "\<rightarrow>\<^sub>\<beta>" 50)
where
  "s \<rightarrow>\<^sub>\<beta> t \<equiv> (s, t) \<in> Beta"

abbreviation betas (infix "\<rightarrow>\<^sub>\<beta>\<^sup>*" 50)
where
  "s \<rightarrow>\<^sub>\<beta>\<^sup>* t \<equiv> (s, t) \<in> Beta\<^sup>*"

fun lambda_bullet :: "'a term \<Rightarrow> 'a term" ("_\<^sup>\<bullet>" [1000] 999)
where
  "(Var x)\<^sup>\<bullet> = Var x"
| "(Abs x t)\<^sup>\<bullet> = Abs x (t\<^sup>\<bullet>)"
| "(App s t)\<^sup>\<bullet> =
  (case s\<^sup>\<bullet> of
    Abs x s' \<Rightarrow> if free_for (t\<^sup>\<bullet>) x s' then subst x (t\<^sup>\<bullet>) s' else App (s\<^sup>\<bullet>) (t\<^sup>\<bullet>)
  | _ \<Rightarrow> App (s\<^sup>\<bullet>) (t\<^sup>\<bullet>))"

lemma App_betas:
  assumes "s \<rightarrow>\<^sub>\<beta>\<^sup>* t" and "u \<rightarrow>\<^sub>\<beta>\<^sup>* v"
  shows "App s u \<rightarrow>\<^sub>\<beta>\<^sup>* App t v"
sorry

lemma Abs_betas:
  assumes "s \<rightarrow>\<^sub>\<beta>\<^sup>* t"
  shows "Abs x s \<rightarrow>\<^sub>\<beta>\<^sup>* Abs x t"
sorry

lemma self:
  "t \<rightarrow>\<^sub>\<beta>\<^sup>* t\<^sup>\<bullet>"
apply (induct t)
apply (auto split: term.splits elim: Beta.cases intro: Beta.intros App_betas Abs_betas)
  subgoal for t\<^sub>1 t\<^sub>2 x u
    apply (subgoal_tac "App t\<^sub>1 t\<^sub>2 \<rightarrow>\<^sub>\<beta>\<^sup>* App (Abs x u) (t\<^sub>2\<^sup>\<bullet>)")
    apply (meson Beta.intros(1) r_into_rtrancl rtrancl_trans)
    apply (auto intro: App_betas)
  done
done

locale z_property =
  fixes bullet :: "'a \<Rightarrow> 'a" ("_\<^sup>\<bullet>")
    and R :: "'a rel"
  assumes Z: "(a, b) \<in> R \<Longrightarrow> (b, a\<^sup>\<bullet>) \<in> R\<^sup>* \<and> (a\<^sup>\<bullet>, b\<^sup>\<bullet>) \<in> R\<^sup>*"
begin

lemma CR: "CR R" sorry

end

end
