theory XTC_Parser
  imports XTC_Problem "../LTS/Xmlt2"
begin

definition "xml2claim \<equiv>
  XMLdo ''terminating'' {
    ubo \<leftarrow>? XMLdo ''upperbound'' {
      ret \<leftarrow> XMLdo ''polynomial'' {
        deg_s \<leftarrow>att ''degree'';
        case nat_of_string deg_s of Inl err \<Rightarrow> xml_error err
        | Inr deg \<Rightarrow> xml_return (Upperbound deg)
      };
      xml_return ret
    };
    case ubo of None \<Rightarrow>
      xml_return Terminating
    | Some ub \<Rightarrow> xml_return ub
  }
  XMLor xml_leaf ''nonterminating'' Nonterminating
  XMLor xml_leaf ''confluent'' Confluent
  XMLor xml_leaf ''nonconfluent'' Nonconfluent"

definition parse_claim
where "parse_claim f \<equiv> parse_xml_string xml2claim"



definition plain_var :: "string xmlt2" where "plain_var \<equiv> xml_text ''var''"
definition var :: "('a, string) term xmlt2"
where "var \<equiv> xml_change plain_var (xml_return \<circ> Var)"
hide_const (open) plain_var
hide_const (open) var


context fixes xml2name :: "'a :: show xmlt2"
begin

definition signature where "signature \<equiv> XMLdo ''signature'' {
  ret \<leftarrow>* XMLdo ''funcsym'' {
    name \<leftarrow> xml2name;
    arity \<leftarrow> xml_nat ''arity'';
    theory \<leftarrow>? xml_text ''theory'';
    case theory of None \<Rightarrow>
      xml_return (name,arity,False,False)
    | Some th \<Rightarrow>
      if th = ''A'' then
        if arity = 2 then xml_return (name,arity,True,False)
        else xml_error (''wrong arity '' @ show arity @ '' for associative symbol \"'' @ show name @ ''\"'')
      else if th = ''AC'' then
        if arity = 2 then xml_return (name,arity,True,True)
        else xml_error (''wrong arity '' @ show arity @ '' for AC symbol \"'' @ show name @ ''\"'')
      else if th = ''C'' then
        if arity = 2 then xml_return (name,arity,False,True)
        else xml_error (''wrong arity '' @ show arity @ '' for commutative symbol \"'' @ show name @ ''\"'')
      else xml_error (''unknown theory \"'' @ th @ ''\"'')
  };
  xml_return ret
}"

primrec signature_to_AC :: "_ \<Rightarrow> _ \<Rightarrow> _ \<Rightarrow> 'a list \<times> 'a list" where
  "signature_to_AC as cs [] = (as,cs)"
| "signature_to_AC as cs (fnac#sig) = (case fnac of (f,n,a,c) \<Rightarrow>
   signature_to_AC (if a then f#as else as) (if c then f#cs else cs) sig)"


function
  "term" :: "('a, string) term xmlt2"
where
 "term x = (
    XTC_Parser.var
    XMLor XMLdo ''funapp'' {
      name \<leftarrow> xml2name;
      args \<leftarrow>* term
        XMLor xml_do ''arg'' (xml_take (term) xml_return) (* ''arg'' tag is redundant *);
      xml_return (Fun name args)
    }
  ) x" by (pat_completeness,auto)

termination by (relation "measure size_xml_state", auto)

definition conditions :: "tag \<Rightarrow> ('a, string) rule list xmlt2" where
  "conditions tag \<equiv> XMLdo tag {
    ret \<leftarrow>* XMLdo ''condition'' {
      l \<leftarrow> XMLdo ''lhs'' {l \<leftarrow> term; xml_return l};
      r \<leftarrow> XMLdo ''rhs'' {r \<leftarrow> term; xml_return r};
      xml_return (l,r)
    };
    xml_return ret
  }"

definition crule :: "tag \<Rightarrow> ('a, string) crule xmlt2" where
  "crule tag \<equiv> XMLdo tag {
    l \<leftarrow> XMLdo ''lhs'' {l \<leftarrow> term; xml_return l};
    r \<leftarrow> XMLdo ''rhs'' {r \<leftarrow> term; xml_return r};
    conds \<leftarrow>[[]] conditions ''conditions'';
    xml_return ((l,r),conds)
  }"

definition rule where "rule \<equiv>
  xml_change (crule ''rule'') (\<lambda>(lr,conds).
    if conds = [] then xml_return lr else xml_error (''conditional rule is not allowed here'')
  )"

definition rules where "rules \<equiv> XMLdo ''rules'' {
  ret \<leftarrow>* rule;
  xml_return ret
}"

primrec strategy_to_fp :: "('f,string)fp_strategy \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)forb_pattern list" where
  "strategy_to_fp (Forbidden_Patterns p) r = p"
| "strategy_to_fp Outermost r = o_to_fp_impl (map fst r)"
| "strategy_to_fp (Context_Sensitive \<mu>) r = mu_to_fp_impl \<mu>"

definition xml2ctrs_input
where
  "xml2ctrs_input \<equiv> XMLdo ''ctrsInput'' {
    a \<leftarrow> XMLdo ''rules'' { ret \<leftarrow>* crule ''rule''; xml_return ret };
    xml_return (CTRS_input a)
}"

definition xml2completion_input
where
  "xml2completion_input \<equiv> XMLdo ''completionInput'' {
    a \<leftarrow> XMLdo ''equations'' {ret \<leftarrow> rules; xml_return ret};
    b \<leftarrow> XMLdo ''trs'' {ret \<leftarrow> rules; xml_return ret};
    xml_return (COMP_input a b)
  }"

definition xml2equational_input
where
  "xml2equational_input \<equiv> XMLdo  ''equationalReasoningInput'' {
    eqs \<leftarrow> XMLdo ''equations'' {ret \<leftarrow> rules; xml_return ret};
    goal \<leftarrow> XMLdo ''equation'' {
      a \<leftarrow> term;
      b \<leftarrow> term;
      xml_return (Equation (a,b))
    }
    XMLor XMLdo ''inequality'' {
      a \<leftarrow> term;
      b \<leftarrow> term;
      xml_return (Inequality (a,b))
    };
    xml_return (EQ_input eqs goal)
  }"

definition complexity_class :: "complexity_class xmlt2"
where
  "complexity_class = xml_change (xml_nat ''polynomial'') (xml_return \<circ> Comp_Poly)"

definition symbols :: "string \<Rightarrow> ('a \<times> nat) list xmlt2"
where
  "symbols tagname =
    XMLdo tagname {ret \<leftarrow>* XMLdo ''symbol'' {a \<leftarrow> xml2name; b \<leftarrow> xml_nat ''arity''; xml_return (a,b)}; xml_return ret}"

definition "CPFsignature = symbols ''signature''"

definition complexity_measure :: "('a,string)complexity_measure xmlt2"
  where "complexity_measure \<equiv>
   XMLdo ''derivationalComplexity'' {a \<leftarrow> CPFsignature; xml_return (Derivational_Complexity a)}
   XMLor XMLdo ''runtimeComplexity'' {a \<leftarrow> CPFsignature; b \<leftarrow> CPFsignature; xml_return (Runtime_Complexity a b)}"

definition position :: "nat xmlt2" where
  "position \<equiv> xml_do ''position'' (xml_take_nat (\<lambda>n. xml_return (n - 1)))"

primrec list2position :: "nat list \<Rightarrow> pos" where
  "list2position [] = \<epsilon>"
| "list2position (n # ns) = n <# list2position ns"

definition pos :: "pos xmlt2" where
  "pos \<equiv> XMLdo ''positionInTerm'' {
    lst \<leftarrow>* XTC_Parser.position;
    xml_return (list2position lst)
  }"

definition
  replacement_map :: "'a muL xmlt2"
where
  "replacement_map \<equiv> XMLdo ''contextSensitive'' {
    ret \<leftarrow>* XMLdo ''replacementMapEntry'' {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat ''arity'';
      is \<leftarrow>* position;
      xml_return ((f,a),is)
    };
    xml_return ret
  }"

definition forbidden_pattern :: "((('a, string) ctxt \<times> ('a, string) term \<times> location) xmlt2)"
where
"forbidden_pattern \<equiv> XMLdo ''forbiddenPattern'' {
  t \<leftarrow> term;
  p \<leftarrow> pos;
  l \<leftarrow>
    xml_leaf ''here'' Forbidden_Patterns.H
    XMLor xml_leaf ''above'' Forbidden_Patterns.A
    XMLor xml_leaf ''below'' Forbidden_Patterns.B
    XMLor xml_leaf ''below'' Forbidden_Patterns.R;
  if p \<in> poss t then xml_return (ctxt_of_pos_term p t, t |_ p, l)
  else xml_error ''position does not exist in term''
}"

definition
  forbidden_patterns :: "((('a, string) ctxt \<times> ('a, string) term \<times> Forbidden_Patterns.location) list) xmlt2"
where
  "forbidden_patterns \<equiv> XMLdo ''forbiddenPatterns'' {
    ret \<leftarrow>* forbidden_pattern;
    xml_return ret
  }"

definition innermostLhss :: "('a, string) term list xmlt2"
where
  "innermostLhss \<equiv> XMLdo ''innermostLhss'' {ret \<leftarrow>* term; xml_return ret}"

definition strategy :: "('a,string)strategy xmlt2" where
  "strategy \<equiv> XMLdo ''strategy'' {
    ret \<leftarrow> xml_leaf ''innermost'' Innermost
      XMLor xml_change (innermostLhss) (xml_return \<circ> Innermost_Q);
    xml_return ret
  }"

definition inn_fp_strategy :: "('a,string)inn_fp_strategy xmlt2" where
  "inn_fp_strategy = XMLdo ''strategy'' {
    ret \<leftarrow> xml_leaf ''innermost'' (Inl Innermost)
      XMLor xml_leaf ''outermost'' (Inr Outermost)
      XMLor xml_change (replacement_map) (\<lambda> p. xml_return (Inr (Context_Sensitive p)))
      XMLor xml_change (forbidden_patterns) (\<lambda> p. xml_return (Inr (Forbidden_Patterns p)))
      XMLor xml_change (innermostLhss) (\<lambda> q. xml_return (Inl (Innermost_Q q)));
    xml_return ret
  }"

definition
  xml2_trs_input :: "('a,string)input xmlt2"
where
  "xml2_trs_input \<equiv> XMLdo ''trsInput'' {
    r \<leftarrow> XMLdo ''trs'' { ret \<leftarrow> rules; xml_return ret };
    str \<leftarrow>[Inl No_Strategy] inn_fp_strategy;
(* TODO: is this supported ??
    eqs \<leftarrow>? xml_singleton ''equations'' (CPF_Parser.rules xml2name) id;
*)
    rel \<leftarrow>[[]] XMLdo ''relativeRules'' { ret \<leftarrow> rules; xml_return ret };
    case str
    of Inl istrat \<Rightarrow> xml_return (Inn_TRS_input istrat r rel start_term.Full)
    |  Inr fpstrat \<Rightarrow>
      if rel = [] then xml_return (FP_TRS_input fpstrat r)
      else xml_error (''the combination of relative rules with strategies is only supported for innermost'')
  }"

definition strategy_of_string where
  "strategy_of_string str \<equiv>
   if str = ''FULL'' then xml_return (Inl No_Strategy)
   else if str = ''INNERMOST'' then xml_return (Inl Innermost)
   else if str = ''OUTERMOST'' then xml_return (Inr Outermost)
   else xml_error (''unknown strategy \"'' @ str @ ''\"'')"

definition
  problem :: "('a,string)input xmlt2"
where
  "problem \<equiv> XMLdo ''problem'' {
    ty \<leftarrow>att ''type'';
    (crules, rel, a, c) \<leftarrow> XMLdo ''trs'' {
      (crules,rel) \<leftarrow> XMLdo ''rules'' {
        crules \<leftarrow>* crule ''rule'';
        rel \<leftarrow>[[]] XMLdo ''relrules'' { ret \<leftarrow>* rule; xml_return ret };
        xml_return (crules, rel)
      };
      sig \<leftarrow> signature;
      let (a,c) = signature_to_AC [] [] sig;
      comment \<leftarrow>? XMLdo ''comment'' {
        _ \<leftarrow>* xml_any;
        xml_return ()
      };
      conditiontype \<leftarrow>? XMLdo ''conditiontype'' {
        _ \<leftarrow>* xml_any;
        xml_return ()
      };
      xml_return (crules,rel,a,c)
    };
    strat \<leftarrow> xml_change (xml_text ''strategy'') strategy_of_string;
    start \<leftarrow>[Full] XMLdo ''startterm'' {
      ret \<leftarrow> XMLdo ''constructor-based'' {xml_return Constructor_Based}
          XMLor XMLdo ''full'' {xml_return Full};
      xml_return ret
    };
    status \<leftarrow>? XMLdo ''status'' {
      _ \<leftarrow>* xml_any;
      xml_return ()
    };
    metainfo \<leftarrow>? XMLdo ''metainformation'' {
      _ \<leftarrow>* xml_any;
      xml_return ()
    };
    let relative = (rel \<noteq> []);
    let (conditional,r) =
      (if list_all (\<lambda>(lr,cs). cs = []) crules then (False, map fst crules)
       else (True, []));
    let equational = (a \<noteq> [] \<or> c \<noteq> []);
    case (conditional, equational, relative, strat, start) of
      (False, False, _    , Inl istrat, _) \<Rightarrow> xml_return (Inn_TRS_input istrat r rel start)
    | (False, False, False, Inr fpstrat, Full) \<Rightarrow> xml_return (FP_TRS_input fpstrat r)
    | (True,  False, False, Inl No_Strategy, Full) \<Rightarrow> xml_return (CTRS_input crules)
    | (False, True,  False, Inl No_Strategy, Full) \<Rightarrow> xml_return (AC_input r a c)
    | (True,  True, _) \<Rightarrow> xml_error ''equational conditional system''
    | (True, _, True, _) \<Rightarrow> xml_error ''relative conditional system''
    | (_, True, True, _) \<Rightarrow> xml_error ''relative equational system''
    | _ \<Rightarrow> xml_error ''unsupported combination''
  }"

definition xml2unknown_input :: "unknown_info xmlt2"
  where "xml2unknown_input \<equiv>
       xml_text ''unknownInput''"

abbreviation parse_xtc where "parse_xtc \<equiv> parse_xmlfile problem"

end

hide_const (open) "term"
hide_const (open) crule (*hide more*)hide_const (open) replacement_map
hide_const (open) forbidden_pattern
hide_const (open) forbidden_patterns




(* ... to xml *)
primrec xml_lab :: "('f :: show, ('l :: show)list)lab \<Rightarrow> xml" where
  "xml_lab (UnLab x) = XML ''name'' [] [XML_text (show x)]"
| "xml_lab (Sharp x) = XML ''sharp'' [] [xml_lab x]"
| "xml_lab (FunLab x l) = XML ''labeledSymbol'' [] [xml_lab x, XML ''symbolLabel'' [] (map xml_lab l)]"
| "xml_lab (Lab x l) = XML ''labeledSymbol'' [] [xml_lab x, XML ''numberLabel'' [] (map (\<lambda> n. XML ''number'' [] [XML_text (show n)]) l)]"

definition xml_signature :: "(('f :: show,('l :: show)list)lab \<times> nat)list \<Rightarrow> xml" where
  "\<And>fs. xml_signature fs \<equiv> XML ''signature'' [] (map (\<lambda> (f,n). XML ''symbol'' [] [xml_lab f, XML ''arity'' [] [XML_text (shows n Nil)]]) fs)"

primrec xml_complexity_measure :: "(('f :: show,('l :: show)list)lab,'v)complexity_measure \<Rightarrow> xml" where
  "xml_complexity_measure (Derivational_Complexity F) = XML ''derivationalComplexity'' [] [xml_signature F]"
 | "xml_complexity_measure (Runtime_Complexity C D) = XML ''runtimeComplexity'' [] [xml_signature C, xml_signature D]"

primrec xml_complexity_class :: "complexity_class \<Rightarrow> xml" where
  "xml_complexity_class (Comp_Poly n) = XML ''polynomial'' [] [XML_text (shows n Nil)]"

definition xml_single_pos :: "nat \<Rightarrow> xml" where
  "xml_single_pos i \<equiv> XML ''position'' [] [XML_text (show (Suc i))]"

definition xml_pos :: "pos \<Rightarrow> xml" where
  "xml_pos p \<equiv>  XML ''positionInTerm'' [] (map xml_single_pos (to_list p))"

primrec xml_term :: "(('f :: show,('l :: show)list)lab, 'v :: show)term \<Rightarrow> xml" where
  "xml_term (Var x) = XML ''var'' [] [XML_text (show x)]"
| "xml_term (Fun f ts) = XML ''funapp'' []
    (xml_lab f #
      (map (\<lambda> t. XML ''arg'' [] [xml_term t]) ts))"

definition xml_condition :: "(('f :: show,('l :: show)list)lab,'c :: show) rule \<Rightarrow> xml" where
  "xml_condition \<equiv> \<lambda> (l, r). XML ''condition'' [] [
    XML ''lhs'' [] [xml_term l] ,
    XML ''rhs'' [] [xml_term r] ]"

definition xml_crule :: "(('f :: show,('l :: show)list)lab,'c :: show) crule \<Rightarrow> xml" where
  "xml_crule \<equiv> \<lambda> ((l,r), conds). XML ''rule'' [] (
    XML ''lhs'' [] [xml_term l] #
    XML ''rhs'' [] [xml_term r] #
    (if conds = [] then [] else [XML ''conditions'' [] (map xml_condition conds)]))"

definition xml_rule :: "(('f :: show,('l :: show)list)lab,'c :: show) rule \<Rightarrow> xml" where
  "xml_rule \<equiv> \<lambda> (l, r). XML ''rule'' [] [
    XML ''lhs'' [] [xml_term l] ,
    XML ''rhs'' [] [xml_term r] ]"

definition xml_rules :: "string \<Rightarrow> (('f :: show,('l :: show)list)lab,'c :: show) rules \<Rightarrow> xml" where
  "xml_rules tag rls = XML tag Nil [XML ''rules'' [] (map xml_rule rls)]"

definition xml_forbidden_pattern :: "(('f :: show,('l :: show) list)lab,'v :: show) forb_pattern \<Rightarrow> xml" where
  "xml_forbidden_pattern \<equiv> \<lambda> (C,t,l). XML ''forbiddenPattern'' [] [
    xml_term C\<langle>t\<rangle>,
    xml_pos (hole_pos C),
    (case l of Forbidden_Patterns.H \<Rightarrow> XML ''here'' [] []
             | Forbidden_Patterns.A \<Rightarrow> XML ''above'' [] []
             | Forbidden_Patterns.B \<Rightarrow> XML ''below'' [] []
             | Forbidden_Patterns.R \<Rightarrow> XML ''right'' [] [] )
   ]"

definition xml_repl_map :: "(('f :: show,('l :: show) list)lab \<times> nat) \<times> (nat list) \<Rightarrow> xml" where
  "xml_repl_map \<equiv> \<lambda> ((f,a),l). XML ''replacementMapEntry'' [] (
    xml_lab f #
    XML ''arity'' [] [XML_text (show a)] #
    map xml_single_pos l
    )"

definition xml_strategy :: "(('f :: show,('l :: show) list)lab,'v :: show)inn_fp_strategy \<Rightarrow> xml list"
  where "xml_strategy x \<equiv> case x of
    Inl No_Strategy \<Rightarrow> []
  | Inl Innermost  \<Rightarrow> [XML ''strategy'' [] [XML ''innermost'' [] []]]
  | Inr Outermost \<Rightarrow> [XML ''strategy'' [] [XML ''outermost'' [] []]]
  | Inr (Forbidden_Patterns p) \<Rightarrow> [XML ''strategy'' [] [XML ''forbiddenPatterns'' [] (map xml_forbidden_pattern p)]]
  | Inr (Context_Sensitive \<mu>) \<Rightarrow> [XML ''strategy'' [] [XML ''contextSensitive'' [] (map xml_repl_map \<mu>)]]
  | Inl (Innermost_Q Q) \<Rightarrow> [XML ''strategy'' [] [XML ''innermostLhss'' [] (map xml_term Q)]]"

fun xml_trs_input :: "(('f :: show,('l :: show) list)lab,'v :: show)inn_fp_strategy \<Rightarrow> (('f,'l list)lab,'v)rules
  \<Rightarrow> (('f,'l list)lab,'v)rules \<Rightarrow> xml"
  where "xml_trs_input strat R [] =  XML ''trsInput'' [] ([xml_rules ''trs'' R] @ xml_strategy strat)"
  |     "xml_trs_input strat R S = XML ''trsInput'' [] (xml_rules ''trs'' R # xml_strategy strat @ [xml_rules ''relativeRules'' S])"

definition xml_state :: "string \<Rightarrow> xml" where
  "xml_state s = XML ''state'' [] [XML_text s]"

fun xml_eps_transition :: "string \<times> string \<Rightarrow> xml" where
  "xml_eps_transition (s,t) = XML ''transition'' [] [
    XML ''lhs'' [] [xml_state s],
    XML ''rhs'' [] [xml_state t]]"

fun xml_transition :: "(string,('f :: show,('l :: show) list)lab)ta_rule \<Rightarrow> xml" where
  "xml_transition (TA_rule f qs q) = XML ''transition'' [] [
    XML ''lhs'' [] (xml_lab f # map xml_state qs),
    XML ''rhs'' [] [xml_state q]]"

fun xml_ta :: "(string,('f :: show,('l :: show) list)lab)tree_automaton \<Rightarrow> xml" where
  "xml_ta (Tree_Automaton fin tran eps) =  XML ''treeAutomaton'' [] [
    XML ''finalStates'' [] (map xml_state fin),
    XML ''transitions'' [] (map xml_transition tran @ map xml_eps_transition eps)]"

definition xml_ta_input :: "(string,('f :: show,('l :: show) list)lab)tree_automaton \<Rightarrow> (('f,'l list)lab,'v :: show)rules
  \<Rightarrow> xml"
  where "xml_ta_input ta R =  XML ''treeAutomatonProblem'' [] ([xml_ta ta, xml_rules ''trs'' R])"

definition xml_ctrs_input :: "(('f :: show,('l :: show) list)lab,'v :: show)crules \<Rightarrow> xml"
  where "xml_ctrs_input ctrs =  XML ''ctrsInput'' [] [XML ''rules'' [] (map xml_crule ctrs)]"

definition xml_ac_input :: "(('f :: show,('l :: show)list)lab,'v::show)rules
  \<Rightarrow> ('f,'l list)lab list \<Rightarrow> ('f,'l list)lab list \<Rightarrow> xml" where
  "xml_ac_input r a c \<equiv> XML ''acRewriteSystem'' [] [
     xml_rules ''trs'' r,
     XML ''Asymbols'' [] (map xml_lab a),
     XML ''Csymbols'' [] (map xml_lab c)
   ]"

definition xml_location_id :: "string \<Rightarrow> xml" where
  "xml_location_id x = XML ''locationId'' [] [XML_text x]"

definition xml_transition_id :: "string \<Rightarrow> xml" where
  "xml_transition_id x = XML ''transitionId'' [] [XML_text x]"

fun xml_trans_var where
  "xml_trans_var (Pre x) = XML ''variableId'' [] [XML_text (show x)]"
| "xml_trans_var (Post x) = XML ''post'' [] [XML ''variableId'' [] [XML_text (show x)]]"
| "xml_trans_var _ = Code.abort (STR ''problem in xml_trans_var'') (\<lambda> _. XML [] [] [])"

fun xml_expr :: "(IA.sig,_,_)exp \<Rightarrow> xml" where
  "xml_expr (Var (x,ty)) = xml_trans_var x"
| "xml_expr (Fun (IA.Sum n) ls) = XML ''sum'' [] (map xml_expr ls)"
| "xml_expr (Fun (IA.Const i) []) = XML ''constant'' [] [XML_text (show i)]"
| "xml_expr (Fun (IA.Mult n) ls) = XML ''product'' [] (map xml_expr ls)"
| "xml_expr _ = Code.abort (STR ''problem in xml_expr for LIA'') (\<lambda> _. XML [] [] [])"

fun xml_atom :: "(IA.sig,_,_)exp \<Rightarrow> xml" where
  "xml_atom (Fun IA.Less [l,r]) = XML ''less'' [] [xml_expr l, xml_expr r]"
| "xml_atom (Fun IA.Lesseq [l,r]) = XML ''leq'' [] [xml_expr l, xml_expr r]"
| "xml_atom (Fun IA.Equals [l,r]) = XML ''eq'' [] [xml_expr l, xml_expr r]"
| "xml_atom _ = Code.abort (STR ''problem in xml_atom for IA'') (\<lambda> _. XML [] [] [])"

fun xml_formula :: "_ \<Rightarrow> xml" where
  "xml_formula (Conjunction \<phi>s) = XML ''conjunction'' [] (map xml_formula \<phi>s)"
| "xml_formula (Disjunction \<phi>s) = XML ''disjunction'' [] (map xml_formula \<phi>s)"
| "xml_formula (Atom \<phi>) = xml_atom \<phi>"
| "xml_formula (NegAtom \<phi>) = Code.abort (STR ''negated atom not supported in xml_formula'') (\<lambda> _. XML [] [] [])"

fun xml_transition_rule :: "string \<times> (_,_,_,_) transition_rule \<Rightarrow> xml" where
  "xml_transition_rule (i, Transition l r \<phi>) = XML ''transition'' [] [
      xml_transition_id i,
      XML ''source'' [] [xml_location_id l],
      XML ''target'' [] [xml_location_id r],
      XML ''formula'' [] [xml_formula \<phi>]
   ]"

fun xml_lts :: "(_,_,_,_,_) lts_impl \<Rightarrow> xml" where
  "xml_lts (Lts_Impl I tau lc) = XML ''lts'' [] (
      XML ''initial'' [] (map xml_location_id I) #
      map xml_transition_rule tau
   )"

definition xml_lts_safety :: "(_,_,_,_,_) lts_impl \<Rightarrow> string list \<Rightarrow> xml" where
  "xml_lts_safety ltsi err = XML ''ltsSafetyInput'' [] (
      xml_lts ltsi  #
      map (\<lambda> e. XML ''error'' [] [xml_location_id e]) err
   )"


end
