(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016)
License: LGPL (see file COPYING.LESSER)
*)
theory Ordered_Rewriting_Old
imports
  Trs
  "../Orderings/Reduction_Order"
  "../Rewriting/Renaming_Interpretations"
  "../Rewriting/Multihole_Context"
begin

(*TODO: move*)
lemma rtrancl_Restr:
  assumes "(x, y) \<in> (Restr r A)\<^sup>*"
  shows "(x, y) \<in> r\<^sup>*"
using assms by (induct) auto

(*TODO: move*)
lemma join_mono:
  assumes "r \<subseteq> s"
  shows "r\<^sup>\<down> \<subseteq> s\<^sup>\<down>"
  using rtrancl_mono [OF assms] by (auto simp: join_def rtrancl_converse)

(*TODO: move*)
lemma subst_subst_ctxt:
  "C \<cdot>\<^sub>c \<sigma> \<cdot>\<^sub>c \<tau> = C \<cdot>\<^sub>c (\<sigma> \<circ>\<^sub>s \<tau>)"
  by auto

(*TODO: move*)
lemma join_subst:
  "subst.closed r \<Longrightarrow> (s, t) \<in> r\<^sup>\<down> \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> r\<^sup>\<down>"
  by (simp add: join_def subst.closedD subst.closed_comp subst.closed_converse subst.closed_rtrancl)

(*TODO: move*)
lemma step_imp_conversion:
  "(x, y) \<in> r \<Longrightarrow> (x, y) \<in> r\<^sup>\<leftrightarrow>\<^sup>*"
by (auto)

(*TODO: move*)
lemma step_imp_conversion':
  "(y, x) \<in> r \<Longrightarrow> (x, y) \<in> r\<^sup>\<leftrightarrow>\<^sup>*"
by (auto)

(*TODO: generalization of Trs.replace_at_subst_rsteps; move*)
lemma replace_at_subst_steps:
  fixes \<sigma> \<tau> :: "('f, 'v) subst"
  assumes acc: "all_ctxt_closed UNIV r"
    and refl: "refl r"
    and *: "\<And>x. (\<sigma> x, \<tau> x) \<in> r"
    and "p \<in> poss t"
    and "t |_ p = Var x"
  shows "(replace_at (t \<cdot> \<sigma>) p (\<tau> x), t \<cdot> \<tau>) \<in> r"
using assms(4-)
proof (induction t arbitrary: p)
  case (Var x)
  then show ?case using refl by (simp add: refl_on_def)
next
  case (Fun f ts)
  then obtain i q where [simp]: "p = i <# q" and i: "i < length ts"
    and q: "q \<in> poss (ts ! i)" and [simp]: "ts ! i |_ q = Var x" by (cases p) auto
  let ?C = "ctxt_of_pos_term q (ts ! i \<cdot> \<sigma>)"
  let ?ts = "map (\<lambda>t. t \<cdot> \<tau>) ts"
  let ?ss = "take i (map (\<lambda>t. t \<cdot> \<sigma>) ts) @ ?C\<langle>\<tau> x\<rangle> # drop (Suc i) (map (\<lambda>t. t \<cdot> \<sigma>) ts)"
  have "\<forall>j<length ts. (?ss ! j, ?ts ! j) \<in> r"
  proof (intro allI impI)
    fix j
    assume j: "j < length ts"
    moreover
    { assume [simp]: "j = i"
      have "?ss ! j = ?C\<langle>\<tau> x\<rangle>" using i by (simp add: nth_append_take)
      with Fun.IH [of "ts ! i" q]
      have "(?ss ! j, ?ts ! j) \<in> r" using q and i by simp }
    moreover
    { assume "j < i"
      with i have "?ss ! j = ts ! j \<cdot> \<sigma>"
        and "?ts ! j = ts ! j \<cdot> \<tau>" by (simp_all add: nth_append_take_is_nth_conv)
      then have "(?ss ! j, ?ts ! j) \<in> r" by (auto simp: * all_ctxt_closed_subst_step [OF acc]) }
    moreover
    { assume "j > i"
      with i and j have "?ss ! j = ts ! j \<cdot> \<sigma>"
        and "?ts ! j = ts ! j \<cdot> \<tau>" by (simp_all add: nth_append_drop_is_nth_conv)
      then have "(?ss ! j, ?ts ! j) \<in> r" by (auto simp: * all_ctxt_closed_subst_step [OF acc]) }
    ultimately show "(?ss ! j, ?ts ! j) \<in> r" by arith
  qed
  moreover have "i < length ts" by fact
  ultimately show ?case
    by (auto intro: all_ctxt_closedD [OF acc])
qed

(*TODO: move*)
subsection \<open>Conversions\<close>

inductive conv for r
where
  empty [intro, simp]: "conv r [x]" |
  stepr [intro, simp]: "(x, y) \<in> r \<Longrightarrow> conv r [x, y]" |
  stepl [intro, simp]: "(x, y) \<in> r \<Longrightarrow> conv r [y, x]" |
  trans [intro, simp]: "conv r xs \<Longrightarrow> conv r ys \<Longrightarrow> last xs = hd ys \<Longrightarrow> zs = tl ys \<Longrightarrow> conv r (xs @ zs)"

abbreviation "CONV r \<equiv> {(x, y). \<exists>xs. hd xs = x \<and> last xs = y \<and> conv r xs}"

lemma conv_Nil [simp]:
  "conv r [] = False"
proof -
  { fix xs presume "conv r xs" and "xs = []"
    then have False by (induct) auto }
  then show ?thesis by auto
qed

lemma conv_map:
  assumes "\<And>x y. (x, y) \<in> r \<Longrightarrow> (f x, f y) \<in> r"
    and "conv r xs"
  shows "conv r (map f xs)"
using assms(2)
proof (induct)
  case (trans xs ys zs)
  then have "conv r (map f xs @ map f (tl ys))"
    by (intro conv.trans [of r "map f xs" "map f ys"]; cases xs; cases ys; force simp: last_map)
  then show ?case using trans by simp
qed (auto dest: assms(1))

lemma CR_on_conv:
  assumes "CR_on r A" and "conv r xs" and "set xs \<subseteq> A"
  shows "(hd xs, last xs) \<in> r\<^sup>\<down>"
using assms(2-)
proof (induct)
  case (trans xs ys zs)
  moreover then have "set xs \<subseteq> A" and "set ys \<subseteq> A" by (auto) (cases xs; cases ys, auto split: if_splits)
  ultimately obtain y and z where "(hd xs, y) \<in> r\<^sup>*" and "(last xs, y) \<in> r\<^sup>*"
    and "(hd ys, z) \<in> r\<^sup>*" and "(last ys, z) \<in> r\<^sup>*" by blast
  moreover have "hd ys \<in> A" using trans and \<open>set ys \<subseteq> A\<close> by (cases ys) auto
  ultimately have "(y, z) \<in> r\<^sup>\<down>" using \<open>CR_on r A\<close> by (auto simp: CR_defs \<open>last xs = hd ys\<close>)
  with \<open>(hd xs, y) \<in> r\<^sup>*\<close> and \<open>(last ys, z) \<in> r\<^sup>*\<close> have "(hd xs, last ys) \<in> r\<^sup>\<down>"
    by (blast intro: join_rtrancl_join rtrancl_join_join)
  moreover have "hd (xs @ zs) = hd xs" and "last (xs @ zs) = last ys"
    using trans by (auto simp: hd_append) (cases ys, auto)
  ultimately show ?case by simp
qed auto

lemma conversion_iff_conv:
  "(x, y) \<in> r\<^sup>\<leftrightarrow>\<^sup>* \<longleftrightarrow> (x, y) \<in> CONV r"
proof
  assume "(x, y) \<in> r\<^sup>\<leftrightarrow>\<^sup>*"
  then show "(x, y) \<in> CONV r"
    unfolding conversion_def
  proof (induct)
    case (base)
    show ?case by (auto intro!: exI [of _ "[x]"])
  next
    case (step y z)
    then obtain xs where "conv r xs" and "hd xs = x" and "last xs = y"
      and "conv r [y, z]" by auto
    then show ?case by (auto intro!: exI [of _ "xs @ [z]"] simp: hd_append)
  qed
next
  assume "(x, y) \<in> CONV r"
  then obtain xs where "conv r xs" and "hd xs = x" and "last xs = y" by auto
  then show "(x, y) \<in> r\<^sup>\<leftrightarrow>\<^sup>*"
  proof (induct arbitrary: x y)
    case (trans xs ys zs x y)
    then have "(x, last xs) \<in> r\<^sup>\<leftrightarrow>\<^sup>*" by (auto simp: hd_append)
    moreover have "(last xs, last ys) \<in> r\<^sup>\<leftrightarrow>\<^sup>*" using trans by auto
    moreover have "last ys = y" using trans by (cases ys; cases zs) auto
    ultimately show ?case by (auto simp: conversion_def)
  qed auto
qed

lemma conv_imp_conv:
  assumes "r\<^sup>\<leftrightarrow>\<^sup>* = s\<^sup>\<leftrightarrow>\<^sup>*"
    and "conv r xs"
  shows "(hd xs, last xs) \<in> CONV s"
using assms(2, 1)
proof (induct)
  case (empty x)
  then show ?case by (auto intro: exI [of _ "[x]"])
next
  case (trans us vs ws)
  moreover then obtain xs and ys
    where "hd xs = hd us" and "last xs = last us" and "conv s xs"
      and "hd ys = hd vs" and "last ys = last vs" and "conv s ys" by auto
  ultimately show ?case
    apply (auto intro!: exI [of _ "xs @ tl ys"])
    apply (cases xs; cases us; simp)
    apply (cases ys; cases vs; simp split: if_splits)
    done
qed (force simp: conversion_iff_conv)+

lemma conversion_eq_iff_conv_eq:
  "r\<^sup>\<leftrightarrow>\<^sup>* = s\<^sup>\<leftrightarrow>\<^sup>* \<longleftrightarrow> CONV r = CONV s"
using conv_imp_conv by (auto simp: conversion_iff_conv)

inductive_set
  ordstep :: "_ \<Rightarrow> _ \<Rightarrow> ('f, 'v) term rel" for ord :: "('f, 'v) term rel" and R :: "('f, 'v) trs"
where
  "(l, r) \<in> R \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ord \<Longrightarrow> (s, t) \<in> ordstep ord R"

lemma ordstep_Un:
  "ordstep ord (R \<union> R') = ordstep ord R \<union> ordstep ord R'"
by (auto intro: ordstep.intros elim: ordstep.cases)

lemma ordstep_rstep_conv:
  assumes "subst.closed ord" and "R \<subseteq> ord"
  shows "ordstep ord R = rstep R"
using assms by (fast elim: ordstep.cases intro: ordstep.intros)

lemma ordstep_subst:
  "subst.closed ord \<Longrightarrow> (s, t) \<in> ordstep ord R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ordstep ord R"
  by (elim ordstep.cases, intro ordstep.intros [where C = "C \<cdot>\<^sub>c \<sigma>" and \<sigma> = "\<tau> \<circ>\<^sub>s \<sigma>" for C \<tau>]) auto

lemma subst_closed_ordstep [intro]:
  "subst.closed ord \<Longrightarrow> subst.closed (ordstep ord R)"
by (auto simp: ordstep_subst)

lemma ordstep_ctxt:
  "(s, t) \<in> ordstep ord R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> ordstep ord R"
  by (elim ordstep.cases, intro ordstep.intros [where C = "C \<circ>\<^sub>c D" for D]) auto

lemma ordsteps_ctxt:
  assumes "(s, t) \<in> (ordstep ord R)\<^sup>*"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (ordstep ord R)\<^sup>*"
  using assms by (induct) (auto simp: ordstep_ctxt rtrancl_into_rtrancl)

lemma ctxt_closed_ordstep [intro]:
  "ctxt.closed (ordstep ord R)"
  by (auto simp: ordstep_ctxt)

lemma all_ctxt_closed_ordsteps [simp]:
  "all_ctxt_closed UNIV ((ordstep ord R)\<^sup>*)"
  by (rule trans_ctxt_imp_all_ctxt_closed) (auto simp: trans_def refl_on_def)

lemma ordstep_mono:
  "R \<subseteq> R' \<Longrightarrow> ord \<subseteq> ord' \<Longrightarrow> ordstep ord R \<subseteq> ordstep ord' R'"
by (auto elim!: ordstep.cases intro: ordstep.intros)

lemma ordstep_imp_ord:
  "ctxt.closed ord \<Longrightarrow> (s, t) \<in> ordstep ord R \<Longrightarrow> (s, t) \<in> ord"
by (elim ordstep.cases) auto

lemma ordsteps_imp_ordeq:
  assumes "ctxt.closed ord" and "trans ord" and "(s, t) \<in> (ordstep ord R)\<^sup>*"
  shows "(s, t) \<in> ord\<^sup>="
using assms(3) by (induct) (auto dest: ordstep_imp_ord [OF assms(1)] assms(2) [THEN transD])

lemma ordstep_rstep_conv':
  assumes "subst.closed ord"
  shows "ordstep ord R = rstep {(l \<cdot> \<sigma>, r \<cdot> \<sigma>) | l r \<sigma>. (l, r) \<in> R \<and> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ord}"
proof -
  { fix l r \<sigma> \<tau> C assume "(l, r) \<in> R" and "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ord"
    then have "(l \<cdot> (\<sigma> \<circ>\<^sub>s \<tau>), r \<cdot> (\<sigma> \<circ>\<^sub>s \<tau>)) \<in> ord" using \<open>subst.closed ord\<close> by auto
    with \<open>(l, r) \<in> R\<close> have "(C\<langle>l \<cdot> (\<sigma> \<circ>\<^sub>s \<tau>)\<rangle>, C\<langle>r \<cdot> (\<sigma> \<circ>\<^sub>s \<tau>)\<rangle>) \<in> ordstep ord R"
      by (intro ordstep.intros [where C = C and l = l and r = r and \<sigma> = "\<sigma> \<circ>\<^sub>s \<tau>"]) auto }
  then show ?thesis
    by (auto elim!: ordstep.cases) blast
qed

definition "Req (R :: ('f, 'v) trs) eq true false s t =
  {(Fun eq [Var (SOME x. True), Var (SOME x. True)], Fun true [])} \<union>
  {(Fun eq [s, t], Fun false [])} \<union> R"

lemma rstep_Req_true:
  "(Fun eq [u, u], Fun true []) \<in> rstep (Req R eq true false s t)" (is "_ \<in> rstep ?R")
using rstepI [of "Fun eq [Var (SOME x. True), Var (SOME x. True)]" "Fun true []" ?R, OF _ refl refl, of \<box> "subst (SOME x. True) u"]
by (simp add: Req_def)

lemma rstep_Req_false:
  "(Fun eq [s \<cdot> \<sigma>, t \<cdot> \<sigma>], Fun false []) \<in> rstep (Req R eq true false s t)" (is "_ \<in> rstep ?R")
using rstepI [of "Fun eq [s, t]" "Fun false []" ?R, OF _ refl refl, of \<box> \<sigma>]
by (simp add: Req_def)

lemma subset_Req: "R \<subseteq> Req R eq true false s t" by (auto simp: Req_def)

lemma subst_apply_set_const [simp]:
  "A \<cdot>set (\<lambda>x. Fun f []) \<subseteq> {t. ground t}"
by auto

text \<open>Ground confluence\<close>
abbreviation "GCR r \<equiv> CR_on r {t. ground t}"

abbreviation "GCONV r \<equiv> {(x, y). \<exists>xs. set xs \<subseteq> {t. ground t} \<and> hd xs = x \<and> last xs = y \<and> conv r xs}"

lemma conv_mono:
  assumes "r \<subseteq> s" and "conv r xs"
  shows "conv s xs"
using assms(2, 1) by (induct) auto

lemma conv_symclD:
  assumes "conv (r\<^sup>\<leftrightarrow>) xs"
  shows "conv r xs"
using assms by (induct) auto

lemma conv_symcl [simp]:
  "conv (r\<^sup>\<leftrightarrow>) xs = conv r xs"
by (auto intro: conv_mono [of "r" "r\<^sup>\<leftrightarrow>"] dest: conv_symclD)

lemma conv_rev:
  assumes "conv r (rev xs)" shows "conv r xs"
using assms
proof (induct _ xs\<equiv>"rev xs" arbitrary: xs rule: conv.induct)
  case (stepl x y xs)
  moreover then have "xs = [x, y]" by (cases xs) auto
  ultimately show ?case by auto
next
  case (stepr x y xs)
  moreover then have "xs = [y, x]" by (cases xs) auto
  ultimately show ?case by auto
next
  case (trans xs ys zs us)
  moreover then have "xs \<noteq> []" and "ys \<noteq> []" by force+
  moreover have "us = rev zs @ rev xs" using trans(7) by (metis List.rev_append List.rev_rev_ident)
  ultimately have us: "us = rev ys @ tl (rev xs)"
    apply auto
    by (metis List.append_assoc List.rev.simps(2) List.rev_append List.rev_rev_ident append_Nil2 hd_rev list.exhaust_sel)
  show ?case
    using trans and \<open>xs \<noteq> []\<close> and \<open>ys \<noteq> []\<close>
    unfolding us by (intro conv.trans) (auto simp: last_rev hd_rev)
qed simp

lemma conv_rev':
  assumes "conv r xs" shows "conv r (rev xs)"
using conv_rev [of r "rev xs"] and assms by simp

lemma conv_rtrancl':
  assumes "conv r xs"
  shows "conv (r\<^sup>*) xs"
using assms by (induct) auto

lemma rtrancl_imp_conv:
  assumes "(x, y) \<in> r\<^sup>*"
  shows "\<exists>xs. hd xs = x \<and> last xs = y \<and> conv r xs"
using assms
apply (induct)
apply auto
apply (rule_tac x = "[x]" in exI; simp)
apply (rule_tac x = "xs @ [z]" in exI; simp)
apply (auto simp: hd_append)
done

lemma conv_rtrancl:
  assumes "conv (r\<^sup>*) xs" and "set xs \<subseteq> A"
    and [dest!]: "\<And>x y. (x, y) \<in> r\<^sup>* \<Longrightarrow> x \<in> A \<Longrightarrow> y \<in> A \<Longrightarrow> \<exists>ys. set ys \<subseteq> A \<and> hd ys = x \<and> last ys = y \<and> conv r ys"
  shows "\<exists>ys. set ys \<subseteq> A \<and> hd ys = hd xs \<and> last ys = last xs \<and> conv r ys"
using assms(1, 2)
apply (induct)
apply auto
apply (rule_tac x = "[x]" in exI; simp)
apply (rule_tac x = "rev ys" in exI; simp)
apply (case_tac ys; simp add: hd_rev)
apply (drule conv_rev'; simp)
apply (subgoal_tac "set (xs @ ys) \<subseteq> A")
apply auto
apply (rule_tac x = "ysa @ tl ysb" in exI)
apply auto
apply (metis contra_subsetD list.set_sel(2) tl_Nil)
apply (metis conv_Nil hd_append2)
apply (metis conv_Nil hd_conv_nth last_append last_conv_nth last_tl length_tl list.size(3))
by (metis conv_Nil hd_Cons_tl last_in_set set_ConsD set_mp)

lemma ground_map_imp_rtrancl:
  assumes "\<And>x y. (x, y) \<in> R \<Longrightarrow> (f x, f y) \<in> R"
    and "\<And>x. ground (f x)"
    and "\<And>x. ground x \<Longrightarrow> f x = x"
  shows "\<forall>x y. (x, y) \<in> R\<^sup>* \<and> ground x \<and> ground y \<longrightarrow> (\<exists>ys. set ys \<subseteq> {t. ground t} \<and> hd ys = x \<and> last ys = y \<and> conv R ys)"
using assms
apply (auto dest!: rtrancl_imp_conv)
apply (rule_tac x = "map f xs" in exI)
apply (auto simp: conv_map [OF assms(1)])
apply (case_tac [!] "xs = []")
apply (auto simp: hd_map last_map)
done

lemma GCONV_rtrancl:
  assumes "\<And>x y. (x, y) \<in> R\<^sup>* \<Longrightarrow> ground x \<Longrightarrow> ground y \<Longrightarrow> \<exists>ys. set ys \<subseteq> {t. ground t} \<and> hd ys = x \<and> last ys = y \<and> conv R ys"
  shows "GCONV (R\<^sup>*) = GCONV R"
by (auto simp: conv_rtrancl') (insert conv_rtrancl [OF _ _ assms], blast)

lemma GCONV_rtrancl':
  assumes "\<And>x y. (x, y) \<in> R \<Longrightarrow> (f x, f y) \<in> R"
    and "\<And>x. ground (f x)"
    and "\<And>x. ground x \<Longrightarrow> f x = x"
  shows "GCONV (R\<^sup>*) = GCONV R"
using GCONV_rtrancl [of R] and ground_map_imp_rtrancl [OF assms] by auto

lemma conversion_subset_imp_GCONV_subset:
  assumes f: "\<And>x. ground (f x)" "\<And>x. ground x \<Longrightarrow> f x = x"
    and R: "\<And>x y. (x, y) \<in> R \<Longrightarrow> (f x, f y) \<in> R"
    and S: "\<And>x y. (x, y) \<in> S \<Longrightarrow> (f x, f y) \<in> S"
    and *: "R\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> S\<^sup>\<leftrightarrow>\<^sup>*"
  shows "GCONV R \<subseteq> GCONV S"
proof
  have R: "\<And>x y. (x, y) \<in> R\<^sup>\<leftrightarrow> \<Longrightarrow> (f x, f y) \<in> R\<^sup>\<leftrightarrow>" using R by blast
  have S: "\<And>x y. (x, y) \<in> S\<^sup>\<leftrightarrow> \<Longrightarrow> (f x, f y) \<in> S\<^sup>\<leftrightarrow>" using S by blast
  fix s t
  assume "(s, t) \<in> GCONV R"
  then have "(s, t) \<in> GCONV (R\<^sup>\<leftrightarrow>\<^sup>*)"
    using GCONV_rtrancl' [OF R f] by (auto simp: conversion_def)
  then obtain ts where "set ts \<subseteq> {t. ground t}" and "hd ts = s" and "last ts = t"
    and "conv (R\<^sup>\<leftrightarrow>\<^sup>*) ts" by auto
  moreover with conv_mono [OF *] have "conv (S\<^sup>\<leftrightarrow>\<^sup>*) ts" by auto
  ultimately have "(s, t) \<in> GCONV (S\<^sup>\<leftrightarrow>\<^sup>*)" by blast
  then show "(s, t) \<in> GCONV S"
    using GCONV_rtrancl' [OF S f, simplified] by (auto simp: conversion_def)
qed

lemma conversion_subset_imp_GCONV_subset':
  assumes "subst.closed R" and "subst.closed S"
    and "R\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> S\<^sup>\<leftrightarrow>\<^sup>*"
  shows "GCONV R \<subseteq> GCONV S"
proof -
  let ?\<gamma> = "\<lambda>t. t \<cdot> (\<lambda>x. Fun (SOME f. True) [])"
  have "\<And>t. ground (?\<gamma> t)" and "\<And>t. ground t \<Longrightarrow> ?\<gamma> t = t"
    and "\<And>s t. (s, t) \<in> R \<Longrightarrow> (?\<gamma> s, ?\<gamma> t) \<in> R"
    and "\<And>s t. (s, t) \<in> S \<Longrightarrow> (?\<gamma> s, ?\<gamma> t) \<in> S"
    using assms by (auto simp: ground_subst_apply)
  from conversion_subset_imp_GCONV_subset [OF this assms(3)] show ?thesis by simp
qed

lemma
  assumes *: "\<And>x y. (x, y) \<in> R\<^sup>\<leftrightarrow>\<^sup>* \<Longrightarrow> ground x \<Longrightarrow> ground y \<Longrightarrow> \<exists>ys. set ys \<subseteq> {t. ground t} \<and> hd ys = x \<and> last ys = y \<and> conv (R\<^sup>\<leftrightarrow>) ys"
    and **: "\<And>x y. (x, y) \<in> S\<^sup>\<leftrightarrow>\<^sup>* \<Longrightarrow> ground x \<Longrightarrow> ground y \<Longrightarrow> \<exists>ys. set ys \<subseteq> {t. ground t} \<and> hd ys = x \<and> last ys = y \<and> conv (S\<^sup>\<leftrightarrow>) ys"
    and "R\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> S\<^sup>\<leftrightarrow>\<^sup>*"
  shows "GCONV R \<subseteq> GCONV S"
proof
  fix s t
  assume "(s, t) \<in> GCONV R"
  then have "(s, t) \<in> GCONV (R\<^sup>\<leftrightarrow>\<^sup>*)"
    using GCONV_rtrancl [OF *] by (auto simp: conversion_def)
  then obtain ts where "set ts \<subseteq> {t. ground t}" and "hd ts = s" and "last ts = t"
    and "conv (R\<^sup>\<leftrightarrow>\<^sup>*) ts" by auto
  moreover with conv_mono [OF assms(3)] have "conv (S\<^sup>\<leftrightarrow>\<^sup>*) ts" by auto
  ultimately have "(s, t) \<in> GCONV (S\<^sup>\<leftrightarrow>\<^sup>*)" by blast
  then show "(s, t) \<in> GCONV S"
    using GCONV_rtrancl [OF **] by (auto simp: conversion_def)
qed

lemma conversion_imp_ground_NF_eq:
  assumes GCR: "GCR (ordstep ord S)"
    and "subst.closed ord"
    and gconv: "GCONV (rstep (Req R eq true false s t)) \<subseteq> GCONV (ordstep ord S)" (is "GCONV ?R \<subseteq> GCONV ?S")
    and "\<exists>\<sigma>. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "\<forall>u v. (Fun true [], u) \<in> (ordstep ord S)\<^sup>! \<and> ground u \<and>
    (Fun false [], v) \<in> (ordstep ord S)\<^sup>! \<and> ground v \<longrightarrow> u = v"
proof (intro allI impI; elim conjE)
  fix u v assume "ground u" and "(Fun true [], u) \<in> ?S\<^sup>!" (is "(?true, _) \<in> _")
    and "ground v" and "(Fun false [], v) \<in> ?S\<^sup>!" (is "(?false, _) \<in> _")
  then have NFs: "u \<in> NF ?S" "v \<in> NF ?S" by auto
  let ?\<gamma> = "\<lambda>t. t \<cdot> (\<lambda>x. ?true)"
  obtain \<sigma> where "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?R\<^sup>\<leftrightarrow>\<^sup>*"
    using assms and conversion_mono [of _ ?R, OF subset_Req [THEN rstep_mono]] by blast
  then have "(Fun eq [s \<cdot> \<sigma>, s \<cdot> \<sigma>], Fun eq [s \<cdot> \<sigma>, t \<cdot> \<sigma>]) \<in> ?R\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding conversion_def by (intro args_steps_imp_steps) (auto simp: nth_Cons')
  moreover have "(?true, Fun eq [s \<cdot> \<sigma>, s \<cdot> \<sigma>]) \<in> ?R\<^sup>\<leftrightarrow>\<^sup>*" and "(Fun eq [s \<cdot> \<sigma>, t \<cdot> \<sigma>], ?false) \<in> ?R\<^sup>\<leftrightarrow>\<^sup>*"
    using rstep_Req_true [of eq "s \<cdot> \<sigma>"] and rstep_Req_false
    by (force simp: conversion_inv [of "Fun true []"])+
  ultimately have "(?true, ?false) \<in> ?R\<^sup>\<leftrightarrow>\<^sup>*" by (blast dest: rtrancl_trans)
  then obtain ts where "conv ?R ts" and "hd ts = ?true" and "last ts = ?false"
    by (auto simp: conversion_iff_conv)
  moreover then have conv: "conv ?R (map ?\<gamma> ts)"
    by (intro conv_map ordstep_subst [OF \<open>subst.closed ord\<close>]) auto
  ultimately have "hd (map ?\<gamma> ts) = ?true" "last (map ?\<gamma> ts) = ?false"
    by (case_tac [!] ts) (auto simp: last_map)
  moreover have "conv ?R (map ?\<gamma> ts)" using conv by simp
  ultimately have "(?true, ?false) \<in> GCONV ?R" by (auto intro!: exI [of _ "map ?\<gamma> ts"])
  from gconv [THEN subsetD, OF this] obtain us
    where [simp]: "set us \<subseteq> {t. ground t}" "hd us = ?true" "last us = ?false"
    and "conv ?S us" by auto
  from CR_on_conv [OF GCR \<open>conv ?S us\<close>] obtain w
    where "(?true, w) \<in> ?S\<^sup>*" and "(?false, w) \<in> ?S\<^sup>*" by auto
  moreover have "ground ?true" and "ground ?false" by auto
  ultimately have "(w, u) \<in> ?S\<^sup>*" and "(w, v) \<in> ?S\<^sup>*"
    using GCR and \<open>(?true, u) \<in> ?S\<^sup>!\<close> and \<open>(?false, v) \<in> ?S\<^sup>!\<close>
    by (auto simp: CR_defs normalizability_def NF_join_imp_reach simp del: ground.simps)
  then have "(?\<gamma> w, ?\<gamma> u) \<in> ?S\<^sup>*" and "(?\<gamma> w, ?\<gamma> v) \<in> ?S\<^sup>*"
    by (auto simp: \<open>subst.closed ord\<close> ordstep_subst subst.closedD subst.closedI subst.closed_rtrancl)
  then have "(?\<gamma> w, u) \<in> ?S\<^sup>*" and "(?\<gamma> w, v) \<in> ?S\<^sup>*"
    using \<open>ground u\<close> and \<open>ground v\<close> by (auto simp: ground_subst_apply)
  moreover have "ground (?\<gamma> w)" by simp
  ultimately show "u = v"
    using GCR and NFs by (auto simp: CR_defs dest: join_NF_imp_eq simp del: ground_subst)
qed

lemma infeasibility_via_GCR:
  assumes "GCR (ordstep ord S)"
    and "subst.closed ord"
    and "GCONV (rstep (Req R eq true false s t)) \<subseteq> GCONV (ordstep ord S)"
    and "(Fun true [], u) \<in> (ordstep ord S)\<^sup>! \<and> ground u \<and>
      (Fun false [], v) \<in> (ordstep ord S)\<^sup>! \<and> ground v \<and> u \<noteq> v"
  shows "\<not> (\<exists>\<sigma>. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep R)\<^sup>*)"
using conversion_imp_ground_NF_eq [OF assms(1-3)] and assms(4) by blast

lemma ground_conv_rstep_imp_conv_ordstep:
  assumes "subst.closed ord" and "R \<subseteq> ord"
    and gtotal: "\<And>s t. ground s \<Longrightarrow> ground t \<Longrightarrow> s = t \<or> (s, t) \<in> ord \<or> (t, s) \<in> ord"
    and "set ts \<subseteq> {t. ground t}" and "conv (rstep (R \<union> E)) ts"
  shows "\<exists>us. set us \<subseteq> {t. ground t} \<and>
    hd ts = hd us \<and> last ts = last us \<and> conv (ordstep ord (R \<union> E\<^sup>\<leftrightarrow>)) us"
using assms(5, 4)
proof (induct)
  case (empty t) then show ?case by (intro exI [of _ "[t]"]) auto
next
  case (stepr s t)
  moreover
  { fix l r C \<sigma> assume "(l, r) \<in> E" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
    with stepr have "ground (l \<cdot> \<sigma>)" and "ground (r \<cdot> \<sigma>)" by auto
    then consider "l \<cdot> \<sigma> = r \<cdot> \<sigma>" | "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ord" | "(r \<cdot> \<sigma>, l \<cdot> \<sigma>) \<in> ord"
      using gtotal by blast
    then have ?case
      using \<open>(l, r) \<in> E\<close> and stepr
      apply (cases) apply (auto simp: s t)
      apply (rule_tac x = "[s]" in exI; simp add: s)
      apply (rule_tac x = "[s, t]" in exI; auto simp: s t intro: ordstep.intros)
      apply (rule_tac x = "[s, t]" in exI; auto simp: s t intro: ordstep.intros)
      done }
  moreover {
    assume "(s, t) \<in> rstep R"
    then have ?case
      using assms(1, 2) and stepr
      by (auto elim!: rstepE intro!: exI [of _ "[s, t]"] intro: ordstep.intros) }
  ultimately show ?case by blast
next
  case (stepl s t)
  moreover
  { fix l r C \<sigma> assume "(l, r) \<in> E" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
    with stepl have "ground (l \<cdot> \<sigma>)" and "ground (r \<cdot> \<sigma>)" by auto
    then consider "l \<cdot> \<sigma> = r \<cdot> \<sigma>" | "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ord" | "(r \<cdot> \<sigma>, l \<cdot> \<sigma>) \<in> ord"
      using gtotal by blast
    then have ?case
      using \<open>(l, r) \<in> E\<close> and stepl
      apply (cases) apply (auto simp: s t)
      apply (rule_tac x = "[s]" in exI; simp add: s)
      apply (rule_tac x = "[t, s]" in exI; auto simp: s t intro: ordstep.intros)
      apply (rule_tac x = "[t, s]" in exI; auto simp: s t intro: ordstep.intros)
      done }
  moreover {
    assume "(s, t) \<in> rstep R"
    then have ?case
      using assms(1, 2) and stepl by (auto elim!: rstepE intro!: exI [of _ "[t, s]"] intro: ordstep.intros) }
  ultimately show ?case by blast
next
  case (trans ss ts us)
  moreover then have "set (ss @ ts) \<subseteq> {t. ground t}"
    apply (cases ss; cases ts; simp split: if_splits)
    by (metis contra_subsetD last_in_set mem_Collect_eq)
  ultimately show ?case apply auto
    apply (rule_tac x = "usa @ tl usaa" in exI)
    apply (cases ss; cases ts; simp)
    by (metis (no_types, hide_lams) conv_Nil hd_Cons_tl hd_append2 insert_subset last_append last_snoc last_tl list.set(2))
qed

lemma GCONV_rstep_imp_GCONV_ordstep:
  assumes "subst.closed ord" and "R \<subseteq> ord"
    and "\<And>s t. ground s \<Longrightarrow> ground t \<Longrightarrow> s = t \<or> (s, t) \<in> ord \<or> (t, s) \<in> ord"
  shows "GCONV (rstep (R \<union> E)) \<subseteq> GCONV (ordstep ord (R \<union> E\<^sup>\<leftrightarrow>))"
using ground_conv_rstep_imp_conv_ordstep [OF assms, of _ E] by fastforce

lemma ordstep_imp_rstep:
  "(s, t) \<in> ordstep ord R \<Longrightarrow> (s, t) \<in> rstep R"
by (auto elim: ordstep.cases)

definition "GROUND R = Restr R {t. ground t}"

lemma GROUND_subset: "GROUND R \<subseteq> R" by (auto simp: GROUND_def)

lemma GCONF_iff_GROUND_conversion:
  "(s, t) \<in> GCONV R \<longleftrightarrow> ground s \<and> (s, t) \<in> (GROUND R)\<^sup>\<leftrightarrow>\<^sup>*"
proof
  fix s t assume "(s, t) \<in> GCONV R"
  then obtain ts where "conv R ts" and "set ts \<subseteq> {t. ground t}"
    and "hd ts = s" and "last ts = t" by auto
  then show "ground s \<and> (s, t) \<in> (GROUND R)\<^sup>\<leftrightarrow>\<^sup>*"
    apply (induct arbitrary: s t)
    apply auto
    prefer 4
    apply (subgoal_tac "set ys \<subseteq> {t. ground t}")
    apply (case_tac xs)
    apply (auto simp: last_append last_tl hd_append split: if_splits)
    apply force
    apply (auto simp: GROUND_def)
    by (metis (no_types, lifting) conv_Nil hd_Cons_tl last_in_set mem_Collect_eq set_ConsD subsetCE)
next
  fix s t presume "(s, t) \<in> (GROUND R)\<^sup>\<leftrightarrow>\<^sup>*" and "ground s"
  then show "(s, t) \<in> GCONV R"
    unfolding conversion_def
    apply (induct)
    apply auto
    apply (rule exI [of _ "[s]"]; force)
    apply (rule_tac x = "xs @ [z]" in exI)
    apply (auto simp: GROUND_def hd_append)
    apply (rule_tac x = "xs @ [z]" in exI)
    apply (auto simp: hd_append)
    done
qed auto

lemma GROUND_not_ground:
  assumes "(s, t) \<in> (GROUND R)\<^sup>*" and "\<not> ground s"
  shows "t = s"
using assms by (induct) (auto simp: GROUND_def)

lemma GROUND_rtrancl:
  assumes "(s, t) \<in> (GROUND R)\<^sup>*" and "ground s"
  shows "ground t \<and> (s, t) \<in> R\<^sup>*"
using assms by (induct) (auto simp: GROUND_def)

context reduction_order
begin

lemma ground_less:
  assumes "ground s" and "s \<succ> t"
  shows "ground t"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain x where "x \<in> vars_term t" by (induct t) auto
  then obtain C where "t = C\<langle>Var x\<rangle>" by (auto dest: supteq_Var)
  then have [simp]: "s \<succ> (C \<cdot>\<^sub>c subst x s)\<langle>s\<rangle>"
    using \<open>s \<succ> t\<close> [THEN subst, of "subst x s"] and \<open>ground s\<close> by (simp add: ground_subst_apply)
  def f \<equiv> "\<lambda>n. ((C \<cdot>\<^sub>c subst x s) ^ n)\<langle>s\<rangle>"
  then have "f i \<succ> f (Suc i)" for i
    by (induct i) (simp_all add: ctxt)
  with SN_less show False by (auto simp: SN_defs)
qed

lemma ordstep_GROUND:
  "ground s \<Longrightarrow> (s, t) \<in> ordstep {\<succ>} R \<Longrightarrow> (s, t) \<in> GROUND (ordstep {\<succ>} R)"
by (auto simp: GROUND_def dest: ordstep_imp_ord [OF ctxt_closed_less] ground_less)

lemma ordsteps_GROUND:
  assumes "(s, t) \<in> (ordstep {\<succ>} R)\<^sup>*" and "ground s"
  shows "ground t \<and> (s, t) \<in> (GROUND (ordstep {\<succ>} R))\<^sup>*"
using assms
by (induct rule: rtrancl_induct)
   (auto dest!: ordstep_GROUND, auto simp: GROUND_def)

lemma ordstep_join_GROUND:
  assumes "ground s" and "ground t" and "(s, t) \<in> (ordstep {\<succ>} R)\<^sup>\<down>"
  shows "(s, t) \<in> (GROUND (ordstep {\<succ>} R))\<^sup>\<down>"
using assms by (auto simp: join_def rtrancl_converse dest: ordsteps_GROUND)

lemma GCR_eq_CR_GROUND:
  "GCR (ordstep {\<succ>} R) = CR (GROUND (ordstep {\<succ>} R))"
apply (auto simp: CR_defs dest!: ordsteps_GROUND)
apply (auto intro: join_mono [THEN subsetD, OF GROUND_subset])
apply (case_tac "ground a")
apply (auto dest: GROUND_not_ground)
apply (blast dest: GROUND_rtrancl intro: ordstep_join_GROUND)
done

lemma infeasibility_via_CR_GROUND:
  assumes "CR (GROUND (ordstep {\<succ>} S))"
    and "GCONV (rstep (Req R eq true false s t)) \<subseteq> GCONV (ordstep {\<succ>} S)"
    and "(Fun true [], u) \<in> (ordstep {\<succ>} S)\<^sup>!"
    and "(Fun false [], v) \<in> (ordstep {\<succ>} S)\<^sup>!"
    and "u \<noteq> v"
  shows "\<not> (\<exists>\<sigma>. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep R)\<^sup>*)"
proof -
  have "(Fun true [], u) \<in> (ordstep {\<succ>} S)\<^sup>*"
    and "(Fun false [], v) \<in> (ordstep {\<succ>} S)\<^sup>*" using assms by auto
  then have "ground u" and "ground v" by (auto dest: ordsteps_GROUND)
  with infeasibility_via_GCR [OF assms(1) [folded GCR_eq_CR_GROUND] subst_closed_less assms(2)]
    and assms(3-)
    show ?thesis by simp
qed

end

definition "ooverlap ord R r r' p \<mu> u v \<longleftrightarrow>
  (\<exists>p. p \<bullet> r \<in> R) \<and> (\<exists>p. p \<bullet> r' \<in> R) \<and> vars_rule r \<inter> vars_rule r' = {} \<and>
  p \<in> funposs (fst r) \<and> (\<exists>\<sigma>. mgu (fst r |_ p) (fst r') = Some \<sigma>) \<and>
  \<mu> = the (mgu (fst r |_ p) (fst r')) \<and>
  (snd r \<cdot> \<mu>, fst r \<cdot> \<mu>) \<notin> ord \<and> (snd r' \<cdot> \<mu>, fst r' \<cdot> \<mu>) \<notin> ord \<and>
  u = snd r \<cdot> \<mu> \<and>
  v = replace_at (fst r \<cdot> \<mu>) p (snd r' \<cdot> \<mu>)"

lemma ooverlapI:
  "\<pi> \<bullet> (l, r) \<in> R \<Longrightarrow> \<pi>' \<bullet> (l', r') \<in> R \<Longrightarrow> vars_rule (l, r) \<inter> vars_rule (l', r') = {} \<Longrightarrow>
    p \<in> funposs l \<Longrightarrow> mgu (l |_ p) l' = Some \<mu> \<Longrightarrow>
    (r \<cdot> \<mu>, l \<cdot> \<mu>) \<notin> ord \<Longrightarrow> (r' \<cdot> \<mu>, l' \<cdot> \<mu>) \<notin> ord \<Longrightarrow>
    ooverlap ord R (l, r) (l', r') p \<mu> (r \<cdot> \<mu>) (replace_at (l \<cdot> \<mu>) p (r' \<cdot> \<mu>))"
  by (force simp: ooverlap_def)

locale gtotal_reduction_order =
  reduction_order +
  assumes gtotal: "ground s \<Longrightarrow> ground t \<Longrightarrow> s = t \<or> s \<succ> t \<or> t \<succ> s"

(*TODO: move*)
locale gtotal_reduction_order_inf =
  gtotal_reduction_order less
  for less :: "('a, 'b::infinite) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
begin

lemma ordstep_join:
  assumes ground: "ground s" "ground t"
    and R: "\<And>s t. (s, t) \<in> R \<Longrightarrow> s \<succ> t \<or> (t, s) \<in> R"
    and *: "(s, t) \<in> (ordstep {\<succ>} R)\<^sup>\<down> \<or> (\<exists>l r \<sigma>. (l, r) \<in> R\<^sup>\<leftrightarrow> \<and> s = l \<cdot> \<sigma> \<and> t = r \<cdot> \<sigma>)"
  shows "(s, t) \<in> (ordstep {\<succ>} R)\<^sup>\<down>"
  using *
proof (elim disjE exE conjE)
  fix l r \<sigma> assume "(l, r) \<in> R\<^sup>\<leftrightarrow>" and "s = l \<cdot> \<sigma>" and "t = r \<cdot> \<sigma>"
  then have "(s, t) \<in> ((ordstep {\<succ>} R)\<^sup>\<leftrightarrow>)\<^sup>="
    using gtotal [OF ground]
    apply (auto simp del: ground_subst intro: ordstep.intros [where C = \<box> and \<sigma> = \<sigma>] simp: subst)
     apply (frule R)
     apply (auto intro: ordstep.intros [where C = \<box> and \<sigma> = \<sigma>] simp: subst)
    apply (frule R)
    apply (auto intro: ordstep.intros [where C = \<box> and \<sigma> = \<sigma>] simp: subst)
    done
  then show "(s, t) \<in> (ordstep {\<succ>} R)\<^sup>\<down>" by blast
qed

lemma CR_GROUND_ordstep:
  assumes ooverlaps: "\<And>r r' p \<mu> u v. ooverlap {\<succ>} R r r' p \<mu> u v \<Longrightarrow>
    (u, v) \<in> (ordstep {\<succ>} R)\<^sup>\<down> \<or> (\<exists>l r \<sigma>. (l, r) \<in> R\<^sup>\<leftrightarrow> \<and> u = l \<cdot> \<sigma> \<and> v = r \<cdot> \<sigma>)"
    and R: "\<And>s t. (s, t) \<in> R \<Longrightarrow> s \<succ> t \<or> (t, s) \<in> R"
  shows "CR (GROUND (ordstep {\<succ>} R))" (is "CR (GROUND ?R)" is "CR ?G")
proof
  fix s t' u' assume st: "(s, t') \<in> ?G\<^sup>*" and su: "(s, u') \<in> ?G\<^sup>*"
  then show "(t', u') \<in> ?G\<^sup>\<down>"
  proof (induct s arbitrary: t' u' rule: SN_induct [OF SN_less])
    case less: (1 s)
    have *: "r\<^sup>* = r O r\<^sup>* \<union> Id" for r :: "('a, 'b) term rel" by regexp
    consider t and u where "ground s"
      and "(s, t) \<in> ?G" and "(t, t') \<in> ?G\<^sup>*" and "ground t"
      and "(s, u) \<in> ?G" and "(u, u') \<in> ?G\<^sup>*" and "ground u" | "s = t'" | "s = u'"
      using less.prems by (subst (asm) (3 4) *) (auto simp: GROUND_def)
    then show ?case using less.prems
    proof (cases)
      case 1
      moreover have "s \<succ> t" and "s \<succ> u"
        using 1 by (auto simp: GROUND_def dest: ordstep_imp_ord [OF ctxt_closed_less])
      moreover obtain v where tv: "(t, v) \<in> ?G\<^sup>*" and "(u, v) \<in> ?G\<^sup>*"
      proof -
        obtain C\<^sub>1 and C\<^sub>2 and \<sigma> and \<sigma>\<^sub>2 and l and l\<^sub>2 and r and r\<^sub>2
          where "(l, r) \<in> R" and "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" and s1: "s = C\<^sub>1\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<^sub>1\<langle>r \<cdot> \<sigma>\<rangle>"
          and "(l\<^sub>2, r\<^sub>2) \<in> R" and "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 \<succ> r\<^sub>2 \<cdot> \<sigma>\<^sub>2" and s2: "s = C\<^sub>2\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>" and u: "u = C\<^sub>2\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
          using 1(2, 5) by (auto simp: GROUND_def) (fast elim: ordstep.cases)
        then have lr: "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ?R" and lr\<^sub>2: "(l\<^sub>2 \<cdot> \<sigma>\<^sub>2, r\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> ?R"
          by (auto intro: ordstep.intros [where C = \<box>])

        obtain \<pi> where "vars_rule (\<pi> \<bullet> (l, r)) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}"
          using vars_rule_disjoint by blast
        moreover define l\<^sub>1 and r\<^sub>1 and \<sigma>\<^sub>1 where "l\<^sub>1 = \<pi> \<bullet> l" and "r\<^sub>1 = \<pi> \<bullet> r" and "\<sigma>\<^sub>1 = sop (-\<pi>) \<circ>\<^sub>s \<sigma>"
        ultimately have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}" by (auto simp: eqvt)
        have "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 \<succ> r\<^sub>1 \<cdot> \<sigma>\<^sub>1" using \<open>l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>\<close> by (simp add: l\<^sub>1_def \<sigma>\<^sub>1_def r\<^sub>1_def)
        have t': "t = C\<^sub>1\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>" by (auto simp: t r\<^sub>1_def \<sigma>\<^sub>1_def)

        have rule\<^sub>1: "-\<pi> \<bullet> (l\<^sub>1, r\<^sub>1) \<in> R" and rule\<^sub>2: "0 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> R"
          using \<open>(l, r) \<in> R\<close> and \<open>(l\<^sub>2, r\<^sub>2) \<in> R\<close> unfolding l\<^sub>1_def r\<^sub>1_def by simp_all

        define \<tau> where "\<tau> x = (if x \<in> vars_rule (l\<^sub>1, r\<^sub>1) then \<sigma>\<^sub>1 x else \<sigma>\<^sub>2 x)" for x
        then have \<tau>: "l\<^sub>1 \<cdot> \<tau> = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" "l\<^sub>2 \<cdot> \<tau> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" "r\<^sub>1 \<cdot> \<tau>  = r\<^sub>1 \<cdot> \<sigma>\<^sub>1" "r\<^sub>2 \<cdot> \<tau> = r\<^sub>2 \<cdot> \<sigma>\<^sub>2"
          using disj by (auto simp: vars_defs term_subst_eq_conv)

        have s_mctxt: "s = fill_holes (mctxt_of_ctxt C\<^sub>1) [l \<cdot> \<sigma>]" by (simp add: s1)
        have t_mctxt: "t = fill_holes (mctxt_of_ctxt C\<^sub>1) [r \<cdot> \<sigma>]" by (simp add: t)
        have u_mctxt: "u = fill_holes (mctxt_of_ctxt C\<^sub>2) [r\<^sub>2 \<cdot> \<sigma>\<^sub>2]" by (simp add: u)

        show ?thesis using s1 and s2
        proof (cases rule: two_subterms_cases)
          case eq
          then have "l\<^sub>1 \<cdot> \<tau> = l\<^sub>2 \<cdot> \<tau>" unfolding \<tau> by (auto simp: l\<^sub>1_def \<sigma>\<^sub>1_def)
          then obtain \<mu> where mgu: "mgu l\<^sub>1 l\<^sub>2 = Some \<mu>"
            using mgu_complete by (auto simp: unifiers_def simp del: mgu.simps)
          then have "is_mgu \<mu> {(l\<^sub>1, l\<^sub>2)}" by (simp add: mgu_sound is_imgu_imp_is_mgu)
          with \<open>l\<^sub>1 \<cdot> \<tau> = l\<^sub>2 \<cdot> \<tau>\<close> obtain \<delta> where \<tau>': "\<tau> = \<mu> \<circ>\<^sub>s \<delta>" by (auto simp: is_mgu_def unifiers_def)
          then have \<tau>'': "t \<cdot> \<mu> \<cdot> \<delta> = t \<cdot> \<tau>" for t by simp
          have not_less: "(r\<^sub>1 \<cdot> \<mu>, l\<^sub>1 \<cdot> \<mu>) \<notin> {\<succ>}" "(r\<^sub>2 \<cdot> \<mu>, l\<^sub>2 \<cdot> \<mu>) \<notin> {\<succ>}"
            using \<open>l\<^sub>1 \<cdot> \<sigma>\<^sub>1 \<succ> r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<close> and \<open>l\<^sub>2 \<cdot> \<sigma>\<^sub>2 \<succ> r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> and irrefl
            by (auto dest!: subst [of "r\<^sub>1 \<cdot> \<mu>" "l\<^sub>1 \<cdot> \<mu>" \<delta>] subst [of "r\<^sub>2 \<cdot> \<mu>" "l\<^sub>2 \<cdot> \<mu>" \<delta>] simp: \<tau>'' \<tau> dest: trans)
          show ?thesis
          proof (cases l\<^sub>1)
            case (Fun f ls)
            with ooverlapI [OF rule\<^sub>1 rule\<^sub>2 disj _ _ not_less, of \<epsilon>] and mgu
            have "ooverlap {\<succ>} R (l\<^sub>1, r\<^sub>1) (l\<^sub>2, r\<^sub>2) \<epsilon> \<mu> (r\<^sub>1 \<cdot> \<mu>) (r\<^sub>2 \<cdot> \<mu>)" by simp
            from ooverlaps [OF this]
            have "(r\<^sub>1 \<cdot> \<tau>, r\<^sub>2 \<cdot> \<tau>) \<in> ?R\<^sup>\<down> \<or> (\<exists>l r \<sigma>. (l, r) \<in> R\<^sup>\<leftrightarrow> \<and> r\<^sub>1 \<cdot> \<tau> = l \<cdot> \<sigma> \<and> r\<^sub>2 \<cdot> \<tau> = r \<cdot> \<sigma>)"
               using join_subst [OF subst_closed_ordstep [OF subst_closed_less], of _ _ R]
               apply (auto simp: \<tau>')
               using subst_subst apply blast+
               done
            from ordstep_join [OF _ _ R this] and 1 have "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> ?R\<^sup>\<down>" by (auto simp: \<tau> t' u)
            then have "(t, u) \<in> ?R\<^sup>\<down>" apply (auto simp: t' u)
              by (simp add: eq(1) join_ctxt ordstep_rstep_conv' subst_closed_less)
            then obtain v where "(t, v) \<in> ?G\<^sup>*" and "(u, v) \<in> ?G\<^sup>*"
              using 1 by (auto dest!: ordsteps_GROUND)
            then show ?thesis by (intro that)
          next
            case (Var x)
            show ?thesis
            proof (cases "x \<in> vars_term r\<^sub>1")
              case True
              then obtain C where "r\<^sub>1 = C\<langle>Var x\<rangle>" by (auto dest: vars_term_supteq)
              then have [simp]: "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 \<succ> (C \<cdot>\<^sub>c \<sigma>\<^sub>1)\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>"
                using \<open>l\<^sub>1 \<cdot> \<sigma>\<^sub>1 \<succ> r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<close> by (simp add: Var)
              define h where "h n = ((C \<cdot>\<^sub>c \<sigma>\<^sub>1) ^ n)\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>" for n
              then have "h i \<succ> h (Suc i)" for i by (induct i) (simp_all add: ctxt)
              with SN_less show ?thesis by (auto simp: SN_defs)
            next
              case False
              have "ground (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)" using \<open>ground t\<close> by (simp add: r\<^sub>1_def \<sigma>\<^sub>1_def t)
              moreover have "ground (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)" using \<open>ground u\<close> by (simp add: u)
              ultimately consider "r\<^sub>1 \<cdot> \<sigma>\<^sub>1 = r\<^sub>2 \<cdot> \<sigma>\<^sub>2" | "r\<^sub>1 \<cdot> \<sigma>\<^sub>1 \<succ> r\<^sub>2 \<cdot> \<sigma>\<^sub>2" | "r\<^sub>2 \<cdot> \<sigma>\<^sub>2 \<succ> r\<^sub>1 \<cdot> \<sigma>\<^sub>1"
                by (blast dest: gtotal)
              then show ?thesis
              proof (cases)
                case 1 then show ?thesis by (intro that) (auto simp: t u r\<^sub>1_def \<sigma>\<^sub>1_def eq)
              next
                case 2
                define \<tau> where "\<tau> y = (if y = x then r\<^sub>2 \<cdot> \<sigma>\<^sub>2 else \<sigma>\<^sub>1 y)" for y
                then have *: "l\<^sub>1 \<cdot> \<tau> = r\<^sub>2 \<cdot> \<sigma>\<^sub>2" "r\<^sub>1 \<cdot> \<tau> = r\<^sub>1 \<cdot> \<sigma>\<^sub>1"
                  using False by (auto simp: term_subst_eq_conv Var)
                then have **: "r \<cdot> sop \<pi> \<circ>\<^sub>s \<tau> \<succ> l \<cdot> sop \<pi> \<circ>\<^sub>s \<tau>"
                  using 2 by (simp add: l\<^sub>1_def [symmetric] r\<^sub>1_def [symmetric])
                consider "(r, l) \<in> R" | "l \<succ> r" using \<open>(l, r) \<in> R\<close> and assms by blast
                then show ?thesis
                proof (cases)
                  case 1
                  from ordstep.intros [OF this refl refl, of "sop \<pi> \<circ>\<^sub>s \<tau>" "{\<succ>}" \<box>] and **
                    have "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> ?R" by (fold *) (auto simp: l\<^sub>1_def r\<^sub>1_def)
                  then have "(C\<^sub>1\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>, C\<^sub>2\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>) \<in> ?G"
                    using \<open>ground u\<close> and \<open>ground t\<close> by (auto simp: eq ordstep_ctxt t' u GROUND_def)
                  then show ?thesis by (intro that; force simp: t' u)
                next
                  case 2
                  then have "l \<cdot> sop \<pi> \<circ>\<^sub>s \<tau> \<succ> r \<cdot> sop \<pi> \<circ>\<^sub>s \<tau>" by (blast dest: subst)
                  with ** show ?thesis by (auto simp: irrefl dest: trans)
                qed
              next
                case 3
                define \<tau> where "\<tau> y = (if y = x then r\<^sub>2 \<cdot> \<sigma>\<^sub>2 else \<sigma>\<^sub>1 y)" for y
                then have *: "l\<^sub>1 \<cdot> \<tau> = r\<^sub>2 \<cdot> \<sigma>\<^sub>2" "r\<^sub>1 \<cdot> \<tau> = r\<^sub>1 \<cdot> \<sigma>\<^sub>1"
                  using False by (auto simp: term_subst_eq_conv Var)
                then have "l \<cdot> sop \<pi> \<circ>\<^sub>s \<tau> \<succ> r \<cdot> sop \<pi> \<circ>\<^sub>s \<tau>"
                  using 3 by (simp add: l\<^sub>1_def [symmetric] r\<^sub>1_def [symmetric])
                with ordstep.intros [OF \<open>(l, r) \<in> R\<close> refl refl, of "sop \<pi> \<circ>\<^sub>s \<tau>" "{\<succ>}" \<box>]
                  have "(r\<^sub>2 \<cdot> \<sigma>\<^sub>2, r\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> ?R" by (fold *) (auto simp: l\<^sub>1_def r\<^sub>1_def)
                then have "(C\<^sub>2\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>, C\<^sub>1\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>) \<in> ?G"
                  using \<open>ground u\<close> and \<open>ground t\<close> by (auto simp: eq ordstep_ctxt t' u GROUND_def)
                then show ?thesis by (intro that) (auto simp: t' u)
              qed
            qed
          qed
        next
          case [simp]: (parallel1 C)
          have "fill_holes C [r \<cdot> \<sigma>, l\<^sub>2 \<cdot> \<sigma>\<^sub>2] =\<^sub>f (mctxt_of_ctxt C\<^sub>1, concat [[r \<cdot> \<sigma>],[]])"
            unfolding parallel1 by (intro fill_holes_mctxt_sound) (auto, case_tac i, auto)
          then have t: "t = fill_holes C [r \<cdot> \<sigma>, l\<^sub>2 \<cdot> \<sigma>\<^sub>2]" using t_mctxt by (auto dest: eqfE)
          have "fill_holes C [l \<cdot> \<sigma>, r\<^sub>2 \<cdot> \<sigma>\<^sub>2] =\<^sub>f (mctxt_of_ctxt C\<^sub>2, concat [[], [r\<^sub>2 \<cdot> \<sigma>\<^sub>2]])"
            unfolding parallel1 by (intro fill_holes_mctxt_sound) (auto, case_tac i, auto)
          then have u: "u = fill_holes C [l \<cdot> \<sigma>, r\<^sub>2 \<cdot> \<sigma>\<^sub>2]" using u_mctxt by (auto dest: eqfE)

          from ctxt_imp_mctxt [OF _ lr, of C "[]" "[r\<^sub>2 \<cdot> \<sigma>\<^sub>2]"] and ordstep_ctxt [of _ _ "{\<succ>}" R]
          have "(u, fill_holes C [r \<cdot> \<sigma>, r\<^sub>2 \<cdot> \<sigma>\<^sub>2]) \<in> ?G\<^sup>*" using 1 by (auto simp: u ordstep_GROUND)
          moreover from ctxt_imp_mctxt [OF _ lr\<^sub>2, of C "[r \<cdot> \<sigma>]" "[]"] and ordstep_ctxt [of _ _ "{\<succ>}" R]
          have "(t, fill_holes C [r \<cdot> \<sigma>, r\<^sub>2 \<cdot> \<sigma>\<^sub>2]) \<in> ?G\<^sup>*" using 1 by (auto simp: t ordstep_GROUND)
          ultimately show ?thesis by (intro that)
        next
          case [simp]: (parallel2 C)
          have "fill_holes C [l\<^sub>2 \<cdot> \<sigma>\<^sub>2, r \<cdot> \<sigma>] =\<^sub>f (mctxt_of_ctxt C\<^sub>1, concat [[], [r \<cdot> \<sigma>]])"
            unfolding parallel2 by (intro fill_holes_mctxt_sound) (auto, case_tac i, auto)
          then have t: "t = fill_holes C [l\<^sub>2 \<cdot> \<sigma>\<^sub>2, r \<cdot> \<sigma>]" using t_mctxt by (auto dest: eqfE)
          have "fill_holes C [r\<^sub>2 \<cdot> \<sigma>\<^sub>2, l \<cdot> \<sigma>] =\<^sub>f (mctxt_of_ctxt C\<^sub>2, concat [[r\<^sub>2 \<cdot> \<sigma>\<^sub>2], []])"
            unfolding parallel2 by (intro fill_holes_mctxt_sound) (auto, case_tac i, auto)
          then have u: "u = fill_holes C [r\<^sub>2 \<cdot> \<sigma>\<^sub>2, l \<cdot> \<sigma>]" using u_mctxt by (auto dest: eqfE)

          from ctxt_imp_mctxt [OF _ lr, of C "[r\<^sub>2 \<cdot> \<sigma>\<^sub>2]" "[]"] and ordstep_ctxt [of _ _ "{\<succ>}" R]
          have "(u, fill_holes C [r\<^sub>2 \<cdot> \<sigma>\<^sub>2, r \<cdot> \<sigma>]) \<in> ?G\<^sup>*" using 1 by (auto simp: u ordstep_GROUND)
          moreover from ctxt_imp_mctxt [OF _ lr\<^sub>2, of C "[]" "[r \<cdot> \<sigma>]"] and ordstep_ctxt [of _ _ "{\<succ>}" R]
          have "(t, fill_holes C [r\<^sub>2 \<cdot> \<sigma>\<^sub>2, r \<cdot> \<sigma>]) \<in> ?G\<^sup>*" using 1 by (auto simp: t ordstep_GROUND)
          ultimately show ?thesis by (intro that)
        next
          case [simp]: (nested1 C)
          have eq: "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = C\<langle>l \<cdot> \<sigma>\<rangle>" using s1 and s2 by simp
          define p where "p = hole_pos C"
          have C: "C = ctxt_of_pos_term p (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" by (simp add: p_def eq)

          show ?thesis
          proof (cases "p \<in> funposs l\<^sub>2")
            case False
            moreover have "p \<in> poss (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" by (simp add: eq p_def)
            ultimately obtain q\<^sub>1 and q\<^sub>2 and x where p: "p = q\<^sub>1 <#> q\<^sub>2" and q\<^sub>1: "q\<^sub>1 \<in> poss l\<^sub>2"
                and lq\<^sub>1: "l\<^sub>2 |_ q\<^sub>1 = Var x" and q\<^sub>2: "q\<^sub>2 \<in> poss (\<sigma>\<^sub>2 x)"
              by (blast intro: poss_subst_apply_term)
            moreover have [simp]: "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 |_ p = l \<cdot> \<sigma>" by (simp add: eq p_def) 
            ultimately have [simp]: "\<sigma>\<^sub>2 x |_ q\<^sub>2 = l \<cdot> \<sigma>" by simp

            define \<delta> where "\<delta> y = (if y = x then replace_at (\<sigma>\<^sub>2 x) q\<^sub>2 (r \<cdot> \<sigma>) else \<sigma>\<^sub>2 y)" for y

            have \<delta>_x: "\<delta> x = replace_at (\<sigma>\<^sub>2 x) q\<^sub>2 (r \<cdot> \<sigma>)" by (simp add: \<delta>_def)
            have "(\<sigma>\<^sub>2 x, \<delta> x) \<in> ?R"
              using ordstep_ctxt [OF lr, of "ctxt_of_pos_term q\<^sub>2 (\<sigma>\<^sub>2 x)"] and q\<^sub>2
              by (simp add: \<delta>_def replace_at_ident)
            then have *: "\<And>x. (\<sigma>\<^sub>2 x, \<delta> x) \<in> ?R\<^sup>*" by (auto simp: \<delta>_def)
            then have "(r\<^sub>2 \<cdot> \<sigma>\<^sub>2, r\<^sub>2 \<cdot> \<delta>) \<in> ?R\<^sup>*"
              using all_ctxt_closed_subst_step [OF all_ctxt_closed_ordsteps] by metis
            then have u1: "(u, C\<^sub>2\<langle>r\<^sub>2 \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>*" by (auto simp: u ordsteps_ctxt)
            have t1: "(t, C\<^sub>2\<langle>l\<^sub>2 \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>*"
            proof -
              have "C\<langle>r \<cdot> \<sigma>\<rangle> = replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) q\<^sub>1 (\<delta> x)"
                using q\<^sub>1 and q\<^sub>2 by (simp add: \<delta>_def ctxt_of_pos_term_append lq\<^sub>1 C p)
              moreover
              have "(replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) q\<^sub>1 (\<delta> x), l\<^sub>2 \<cdot> \<delta>) \<in> ?R\<^sup>*"
                by (rule replace_at_subst_steps [OF all_ctxt_closed_ordsteps refl_rtrancl * q\<^sub>1 lq\<^sub>1])
              ultimately show ?thesis by (simp add: t ordsteps_ctxt)
            qed
            consider "(C\<^sub>2\<langle>l\<^sub>2 \<cdot> \<delta>\<rangle>, C\<^sub>2\<langle>r\<^sub>2 \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>=" | "(C\<^sub>2\<langle>r\<^sub>2 \<cdot> \<delta>\<rangle>, C\<^sub>2\<langle>l\<^sub>2 \<cdot> \<delta>\<rangle>) \<in> ?R"
            proof -
              have "ground (l\<^sub>2 \<cdot> \<delta>)" and "ground (r\<^sub>2 \<cdot> \<delta>)"
                using \<open>ground t\<close> and \<open>ground u\<close> and t1 and u1 by (auto dest: ordsteps_GROUND)
              from gtotal [OF this]
              show ?thesis
                using \<open>(l\<^sub>2, r\<^sub>2) \<in> R\<close> and R [OF \<open>(l\<^sub>2, r\<^sub>2) \<in> R\<close>]
                by cases (auto dest: subst intro: that ordstep.intros)
            qed
            then show ?thesis
            proof (cases)
              case 1
              with t1 have "(t, C\<^sub>2\<langle>r\<^sub>2 \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>*" by auto
              with u1 show ?thesis using \<open>ground u\<close> and \<open>ground t\<close> by (intro that) (auto dest!: ordsteps_GROUND)
            next
              case 2
              with u1 have "(u, C\<^sub>2\<langle>l\<^sub>2 \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>*" by auto
              with t1 show ?thesis using \<open>ground u\<close> and \<open>ground t\<close> by (intro that) (auto dest!: ordsteps_GROUND)
            qed
          next
            case True
            have "l\<^sub>2 |_ p \<cdot> \<tau> = l\<^sub>1 \<cdot> \<tau>"
              unfolding funposs_imp_poss [OF True, THEN subt_at_subst, symmetric] \<tau> eq
              by (simp add: l\<^sub>1_def \<sigma>\<^sub>1_def p_def)
            then obtain \<mu> where mgu: "mgu (l\<^sub>2 |_ p) l\<^sub>1 = Some \<mu>"
              using mgu_complete by (auto simp: unifiers_def simp del: mgu.simps)
            then have "is_mgu \<mu> {(l\<^sub>2 |_ p, l\<^sub>1)}" by (simp add: mgu_sound is_imgu_imp_is_mgu)
            with \<open>l\<^sub>2 |_ p \<cdot> \<tau> = l\<^sub>1 \<cdot> \<tau>\<close> obtain \<delta> where \<tau>': "\<tau> = \<mu> \<circ>\<^sub>s \<delta>" by (auto simp: is_mgu_def unifiers_def)
            then have \<tau>'': "t \<cdot> \<mu> \<cdot> \<delta> = t \<cdot> \<tau>" for t by auto
            have not_less: "(r\<^sub>1 \<cdot> \<mu>, l\<^sub>1 \<cdot> \<mu>) \<notin> {\<succ>}" "(r\<^sub>2 \<cdot> \<mu>, l\<^sub>2 \<cdot> \<mu>) \<notin> {\<succ>}"
              using \<open>l\<^sub>1 \<cdot> \<sigma>\<^sub>1 \<succ> r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<close> and \<open>l\<^sub>2 \<cdot> \<sigma>\<^sub>2 \<succ> r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> and irrefl
              by (auto dest!: subst [of "r\<^sub>1 \<cdot> \<mu>" "l\<^sub>1 \<cdot> \<mu>" \<delta>] subst [of "r\<^sub>2 \<cdot> \<mu>" "l\<^sub>2 \<cdot> \<mu>" \<delta>] simp: \<tau>'' \<tau> dest: trans)

            let ?u = "r\<^sub>2 \<cdot> \<mu>" and ?v = "(ctxt_of_pos_term p (l\<^sub>2 \<cdot> \<mu>))\<langle>r\<^sub>1 \<cdot> \<mu>\<rangle>"
            from ooverlapI [OF rule\<^sub>2 rule\<^sub>1 _ True mgu not_less(2,1), THEN ooverlaps] and disj
            have *: "(?u \<cdot> \<delta>, ?v \<cdot> \<delta>) \<in> ?R\<^sup>\<down> \<or> (\<exists>l r \<sigma>. (l, r) \<in> R\<^sup>\<leftrightarrow> \<and> ?u \<cdot> \<delta> = l \<cdot> \<sigma> \<and> ?v \<cdot> \<delta> = r \<cdot> \<sigma>)"
               using join_subst [OF subst_closed_ordstep [OF subst_closed_less], of _ _ R]
               by (metis (no_types, lifting) inf_commute subst_subst_compose)
             have "(?u \<cdot> \<delta>, ?v \<cdot> \<delta>) \<in> ?R\<^sup>\<down>"
               using 1 and \<tau>' and t' and u  and \<tau> and funposs_imp_poss [OF True]
               apply (intro ordstep_join [OF _ _ R *])
               apply (auto simp: C)
               by (metis ctxt_of_pos_term_subst poss_imp_subst_poss)
             then have "(u, t) \<in> ?R\<^sup>\<down>"
               using funposs_imp_poss [OF True]
               by (auto simp: t' u \<tau>'' ctxt_of_pos_term_subst [symmetric] \<tau> C join_ctxt ordstep_rstep_conv' subst_closed_less)
            then obtain v where "(t, v) \<in> ?G\<^sup>*" and "(u, v) \<in> ?G\<^sup>*"
              using 1 by (auto dest!: ordsteps_GROUND)
            then show ?thesis by (intro that)
          qed
        next
          case [simp]: (nested2 C)
          have eq: "l \<cdot> \<sigma> = C\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>" using s1 and s2 by (simp)
          have eq\<^sub>1: "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 = C\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>" using s1 and s2 by (simp add: l\<^sub>1_def \<sigma>\<^sub>1_def)
          define p where "p = hole_pos C"
          have C: "C = ctxt_of_pos_term p (l \<cdot> \<sigma>)" by (simp add: p_def eq)
          have C': "C = ctxt_of_pos_term p (l\<^sub>1 \<cdot> \<sigma>\<^sub>1)" by (simp add: p_def eq l\<^sub>1_def \<sigma>\<^sub>1_def)

          show ?thesis
          proof (cases "p \<in> funposs l")
            case False
            moreover have "p \<in> poss (l \<cdot> \<sigma>)" by (simp add: eq p_def)
            ultimately obtain q\<^sub>1 and q\<^sub>2 and x where p: "p = q\<^sub>1 <#> q\<^sub>2" and q\<^sub>1: "q\<^sub>1 \<in> poss l"
                and lq\<^sub>1: "l |_ q\<^sub>1 = Var x" and q\<^sub>2: "q\<^sub>2 \<in> poss (\<sigma> x)"
              by (blast intro: poss_subst_apply_term)
            moreover have [simp]: "l \<cdot> \<sigma> |_ p = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" by (simp add: eq p_def) 
            ultimately have [simp]: "\<sigma> x |_ q\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" by simp

            define \<delta> where "\<delta> y = (if y = x then replace_at (\<sigma> x) q\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2) else \<sigma> y)" for y

            have \<delta>_x: "\<delta> x = replace_at (\<sigma> x) q\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)" by (simp add: \<delta>_def)
            have "(\<sigma> x, \<delta> x) \<in> ?R"
              using ordstep_ctxt [OF lr\<^sub>2, of "ctxt_of_pos_term q\<^sub>2 (\<sigma> x)"] and q\<^sub>2
              by (simp add: \<delta>_def replace_at_ident)
            then have *: "\<And>x. (\<sigma> x, \<delta> x) \<in> ?R\<^sup>*" by (auto simp: \<delta>_def)
            then have "(r \<cdot> \<sigma>, r \<cdot> \<delta>) \<in> ?R\<^sup>*"
              using all_ctxt_closed_subst_step [OF all_ctxt_closed_ordsteps] by metis
            then have t1: "(t, C\<^sub>1\<langle>r \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>*" by (auto simp: t ordsteps_ctxt)
            have u1: "(u, C\<^sub>1\<langle>l \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>*"
            proof -
              have "C\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle> = replace_at (l \<cdot> \<sigma>) q\<^sub>1 (\<delta> x)"
                using q\<^sub>1 and q\<^sub>2 by (simp add: \<delta>_def ctxt_of_pos_term_append lq\<^sub>1 C p)
              moreover
              have "(replace_at (l \<cdot> \<sigma>) q\<^sub>1 (\<delta> x), l \<cdot> \<delta>) \<in> ?R\<^sup>*"
                by (rule replace_at_subst_steps [OF all_ctxt_closed_ordsteps refl_rtrancl * q\<^sub>1 lq\<^sub>1])
              ultimately show ?thesis by (simp add: u ordsteps_ctxt)
            qed
            consider "(C\<^sub>1\<langle>l \<cdot> \<delta>\<rangle>, C\<^sub>1\<langle>r \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>=" | "(C\<^sub>1\<langle>r \<cdot> \<delta>\<rangle>, C\<^sub>1\<langle>l \<cdot> \<delta>\<rangle>) \<in> ?R"
            proof -
              have "ground (l \<cdot> \<delta>)" and "ground (r \<cdot> \<delta>)"
                using \<open>ground t\<close> and \<open>ground u\<close> and t1 and u1 by (auto dest: ordsteps_GROUND)
              from gtotal [OF this]
              show ?thesis
                using \<open>(l, r) \<in> R\<close> and R [OF \<open>(l, r) \<in> R\<close>]
                by cases (auto dest: subst intro: that ordstep.intros)
            qed
            then show ?thesis
            proof (cases)
              case 1
              with u1 have "(u, C\<^sub>1\<langle>r \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>*" by auto
              with t1 show ?thesis using \<open>ground u\<close> and \<open>ground t\<close> by (intro that) (auto dest!: ordsteps_GROUND)
            next
              case 2
              with t1 have "(t, C\<^sub>1\<langle>l \<cdot> \<delta>\<rangle>) \<in> ?R\<^sup>*" by auto
              with u1 show ?thesis using \<open>ground u\<close> and \<open>ground t\<close> by (intro that) (auto dest!: ordsteps_GROUND)
            qed
          next
            case True
            then have funposs: "p \<in> funposs l\<^sub>1" by (auto simp: l\<^sub>1_def)
            have "l\<^sub>1 |_ p \<cdot> \<tau> = l\<^sub>2 \<cdot> \<tau>"
              unfolding funposs_imp_poss [OF funposs, THEN subt_at_subst, symmetric] \<tau> eq\<^sub>1
              by (simp add: p_def)
            then obtain \<mu> where mgu: "mgu (l\<^sub>1 |_ p) l\<^sub>2 = Some \<mu>"
              using mgu_complete by (auto simp: unifiers_def simp del: mgu.simps)
            then have "is_mgu \<mu> {(l\<^sub>1 |_ p, l\<^sub>2)}" by (simp add: mgu_sound is_imgu_imp_is_mgu)
            with \<open>l\<^sub>1 |_ p \<cdot> \<tau> = l\<^sub>2 \<cdot> \<tau>\<close> obtain \<delta> where \<tau>': "\<tau> = \<mu> \<circ>\<^sub>s \<delta>" by (auto simp: is_mgu_def unifiers_def)
            then have \<tau>'': "t \<cdot> \<mu> \<cdot> \<delta> = t \<cdot> \<tau>" for t by auto
            have not_less: "(r\<^sub>1 \<cdot> \<mu>, l\<^sub>1 \<cdot> \<mu>) \<notin> {\<succ>}" "(r\<^sub>2 \<cdot> \<mu>, l\<^sub>2 \<cdot> \<mu>) \<notin> {\<succ>}"
              using \<open>l\<^sub>1 \<cdot> \<sigma>\<^sub>1 \<succ> r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<close> and \<open>l\<^sub>2 \<cdot> \<sigma>\<^sub>2 \<succ> r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<close> and irrefl
              by (auto dest!: subst [of "r\<^sub>1 \<cdot> \<mu>" "l\<^sub>1 \<cdot> \<mu>" \<delta>] subst [of "r\<^sub>2 \<cdot> \<mu>" "l\<^sub>2 \<cdot> \<mu>" \<delta>] simp: \<tau>'' \<tau> dest: trans)

            let ?u = "r\<^sub>1 \<cdot> \<mu>" and ?v = "(ctxt_of_pos_term p (l\<^sub>1 \<cdot> \<mu>))\<langle>r\<^sub>2 \<cdot> \<mu>\<rangle>"
            from ooverlapI [OF rule\<^sub>1 rule\<^sub>2 disj funposs mgu not_less, THEN ooverlaps]
            have *: "(?u \<cdot> \<delta>, ?v \<cdot> \<delta>) \<in> ?R\<^sup>\<down> \<or> (\<exists>l r \<sigma>. (l, r) \<in> R\<^sup>\<leftrightarrow> \<and> ?u \<cdot> \<delta> = l \<cdot> \<sigma> \<and> ?v \<cdot> \<delta> = r \<cdot> \<sigma>)"
               using join_subst [OF subst_closed_ordstep [OF subst_closed_less], of _ _ R]
               by (metis (no_types, lifting) subst_subst_compose)
             have "(?u \<cdot> \<delta>, ?v \<cdot> \<delta>) \<in> ?R\<^sup>\<down>"
               using 1 and \<tau>' and t' and u  and \<tau> and funposs_imp_poss [OF funposs]
               apply (intro ordstep_join [OF _ _ R *])
               apply (auto simp: C)
               by (metis ctxt_of_pos_term_subst eq eq\<^sub>1 poss_imp_subst_poss)
             then have "(u, t) \<in> ?R\<^sup>\<down>"
               using funposs_imp_poss [OF funposs]
               by (auto simp: t' u \<tau>'' ctxt_of_pos_term_subst [symmetric] \<tau> C' join_ctxt ordstep_rstep_conv' subst_closed_less)
            then obtain v where "(t, v) \<in> ?G\<^sup>*" and "(u, v) \<in> ?G\<^sup>*"
              using 1 by (auto dest!: ordsteps_GROUND)
            then show ?thesis by (intro that)
          qed
        qed
      qed
      ultimately have "(t', v) \<in> ?G\<^sup>\<down>" and "(v, u') \<in> ?G\<^sup>\<down>" by (auto simp: less.hyps)
      then obtain t'' and u'' where t't'': "(t', t'') \<in> ?G\<^sup>*" and "(v, t'') \<in> ?G\<^sup>*"
        and "(v, u'') \<in> ?G\<^sup>*" and u'u'': "(u', u'') \<in> ?G\<^sup>*" by blast
      moreover have "s \<succ> v"
        using 1 and tv
        by (auto simp: GROUND_def dest!: ordstep_imp_ord [OF ctxt_closed_less] ordsteps_imp_ordeq [OF ctxt_closed_less trans_less] rtrancl_Restr)
           (auto dest: trans)
      ultimately have "(t'', u'') \<in> ?G\<^sup>\<down>" by (auto simp: less.hyps)
      then show ?thesis using t't'' and u'u'' by (blast dest: join_rtrancl_join rtrancl_join_join)
    qed auto
  qed
qed

end

end
