(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016, 2017)
Author:  Sarah Winkler <sarah.winkler@uibk.ac.at> (2017)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Ordered Completion\<close>

theory Ordered_Completion
  imports
    "../Rewriting/Ordered_Rewriting"
    "../Normalization_Equivalence/Encompassment"
    "../Normalization_Equivalence/Normalization_Equivalence"
    "../Orderings/Lexicographic_Extension"
    "../Orderings/KBO"
    "../Auxiliaries/Multiset_Extension2"
    "../Abstract_Completion/Abstract_Completion"
    "../Abstract_Completion/Prime_Critical_Pairs"
    "../Rewriting/Q_Restricted_Rewriting"
begin

locale ordered_completion = reduction_order
begin

definition "ostep E R = rstep R \<union> ordstep {\<succ>} (E\<^sup>\<leftrightarrow>)"

lemma ostep_iff_ordstep:
  assumes "R \<subseteq> {\<succ>}"
  shows "ostep E R = ordstep {\<succ>} (R \<union> E\<^sup>\<leftrightarrow>)"
  using subst_closed_less and assms by (simp add: ordstep_Un ostep_def ordstep_rstep_conv)

lemma ostep_cases [consumes 1, cases pred: ostep]:
  assumes "(s, t) \<in> ostep E R"
  obtains (rstep) l r C \<sigma> where "(l, r) \<in> R" and "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  | (ordstep) l r C \<sigma> where "(l, r) \<in> E\<^sup>\<leftrightarrow>" and "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" and "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  using assms by (fastforce simp: ostep_def elim: ordstep.cases)

lemma ostep_mono:
  "E \<subseteq> E' \<Longrightarrow> R \<subseteq> R'\<Longrightarrow> ostep E R \<subseteq> ostep E' R'"
  by (auto simp: ostep_def rstep_mono [THEN subsetD]
      intro: ordstep_mono [THEN subsetD, OF _ subset_refl, of "E\<^sup>\<leftrightarrow>" "E'\<^sup>\<leftrightarrow>"])

lemma rstep_imp_ostep: "(s, t) \<in> rstep R \<Longrightarrow> (s, t) \<in> ostep E R" by (auto simp: ostep_def)

lemma ostep_imp_rstep_sym:
  "(s, t) \<in> ostep E R \<Longrightarrow> (s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>)"
  unfolding ostep_def using ordstep_imp_rstep by fast

lemma E_imp_ostep:
  "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<Longrightarrow> (l, r) \<in> E\<^sup>\<leftrightarrow> \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> (s, t) \<in> ostep E R"
  by (auto simp: ostep_def intro: ordstep.intros)

lemma ostep_subst:
  "(s, t) \<in> ostep E R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ostep E R"
  using ordstep_subst [of "{\<succ>}", OF subst_closed_less] and rstep.subst
  by (auto simp: ostep_def)

lemma ostep_ctxt:
  "(s, t) \<in> ostep E R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> ostep E R"
  using ordstep_ctxt [where ord = "{\<succ>}"] and rstep.ctxt
  by (auto simp: ostep_def)

lemma ostep_imp_less:
  "R \<subseteq> {\<succ>} \<Longrightarrow> (s, t) \<in> ostep E R \<Longrightarrow> s \<succ> t"
  by (force simp: ostep_def subst ctxt elim: ordstep.cases)

inductive
  oKB :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> bool" (infix "\<turnstile>\<^sub>o\<^sub>K\<^sub>B" 55)
  where
    deduce: "(s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (s, u) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E \<union> {(t, u)}, R)" |
    orientl: "s \<succ> t \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E - {(s, t)}, R \<union> {(s, t)})" |
    orientr: "t \<succ> s \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E - {(s, t)}, R \<union> {(t, s)})" |
    delete: "(s, s) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E - {(s, s)}, R)" |
    compose: "(t, u) \<in> ostep E (R - {(s, t)}) \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E, (R - {(s, t)}) \<union> {(s, u)})" |
    simplifyl: "(s, u) \<in> ostep (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B ((E - {(s, t)}) \<union> {(u, t)}, R)" |
    simplifyr: "(t, u) \<in> ostep (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B ((E - {(s, t)}) \<union> {(s, u)}, R)" |
    collapse: "(t, u) \<in> ostep E (R - {(t, s)}) \<Longrightarrow> (t, s) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E \<union> {(u, s)}, R - {(t, s)})"

lemma oKB_less:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')" and "R \<subseteq> {\<succ>}"
  shows "R' \<subseteq> {\<succ>}"
proof -
  have [dest]: "(x, y) \<in> R \<Longrightarrow> x \<succ> y" for x y using \<open>R \<subseteq> {\<succ>}\<close> by auto
  note assms(1)
  moreover
  { fix s t u assume "(t, u) \<in> rstep (R - {(s, t)})" and "(s, t) \<in> R"
    then have "s \<succ> u" by (induct t u) (blast intro: subst ctxt dest: trans) }
  moreover
  { fix s t u assume "(t, u) \<in> ordstep {\<succ>} (E\<^sup>\<leftrightarrow>)" and "(s, t) \<in> R"
    then have "s \<succ> u" by (cases) (auto dest: trans ctxt) }
  ultimately show ?thesis by (cases) (fastforce simp: ostep_def)+
qed

lemma oKB_rtrancl_less:
  assumes "oKB\<^sup>*\<^sup>* (E, R) (E', R')" and "R \<subseteq> {\<succ>}"
  shows "R' \<subseteq> {\<succ>}"
  using assms by (induct "(E, R)" "(E', R')" arbitrary: E' R') (auto dest!: oKB_less)

lemma oKB_E:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "E \<subseteq> Id \<union> E' \<union> R'\<^sup>\<leftrightarrow> \<union> (ostep E' R') O E' \<union> E' O (ostep E' R')\<inverse>"
  using assms
proof (cases)
  case (simplifyl s u t)
  then have "(s, u) \<in> ostep E' R'"
    by (intro ostep_mono [THEN subsetD, OF _ _ simplifyl(3)]) auto
  then have "(s, t) \<in> (ostep E' R') O E'" using simplifyl by auto
  then show ?thesis using simplifyl by auto
next
  case (simplifyr t u s)
  then have "(t, u) \<in> ostep E' R'"
    by (intro ostep_mono [THEN subsetD, OF _ _ simplifyr(3)]) auto
  then have "(s, t) \<in> E' O (ostep E' R')\<inverse>" using simplifyr by blast
  then show ?thesis using simplifyr by blast
qed auto

lemma oKB_E_rev:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "E' \<subseteq> E \<union> (rstep (R \<union> E\<^sup>\<leftrightarrow>))\<inverse> O (rstep (R \<union> E\<^sup>\<leftrightarrow>)) \<union> (ostep E R)\<inverse> O E \<union> E O (ostep E R)"
  using assms
proof (cases)
  case (simplifyl s u t)
  have "(s, u) \<in> ostep E R"  by (intro ostep_mono [THEN subsetD, OF _ _ simplifyl(3)]) auto
  with simplifyl(4) have "(u, t) \<in> (ostep E R)\<inverse> O E" by auto
  then show ?thesis using simplifyl by auto
next
  case (simplifyr t u s)
  have "(t, u) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ simplifyr(3)]) auto
  with simplifyr(4) have "(s, u) \<in> E O (ostep E R)" by auto
  then show ?thesis using simplifyr by auto
next
  case (deduce u s t)
  then have "(s, t) \<in> (rstep (R \<union> E\<^sup>\<leftrightarrow>))\<inverse> O (rstep (R \<union> E\<^sup>\<leftrightarrow>))" by blast
  with deduce show ?thesis by auto
next
  case (collapse t u s)
  have "(t, u) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ collapse(3)]) auto
  from ostep_imp_rstep_sym [OF this] subsetD [OF rstep_mono, OF Un_upper1 rstep.rule [OF collapse(4)]]
  have "(u, s) \<in> (rstep (R \<union> E\<^sup>\<leftrightarrow>))\<inverse> O (rstep (R \<union> E\<^sup>\<leftrightarrow>))" by blast
  with collapse show ?thesis by auto
qed auto

lemma oKB_R:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "R \<subseteq> R' \<union> R' O (ostep E' R')\<inverse> \<union> ostep E' R' O E'"
  using assms
proof (cases)
  case (compose t u s)
  then have "(t, u) \<in> ostep E' R'"
    by (intro ostep_mono [THEN subsetD, OF _ _ compose(3)]) auto
  then have "(s, t) \<in> R' O (ostep E' R')\<inverse>" using compose by blast
  then show ?thesis using compose by blast
next
  case (collapse t u s)
  then have "(t, u) \<in> ostep E' R'"
    by (intro ostep_mono [THEN subsetD, OF _ _ collapse(3)]) auto
  then have "(t, s) \<in> ostep E' R' O E'" using collapse by auto
  then show ?thesis using collapse by blast
qed auto

lemma oKB_R_rev:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "R' \<subseteq> R \<union> E\<^sup>\<leftrightarrow> \<union> R O (ostep E R)"
  using assms
proof (cases)
  case (compose s t u)
  have "(s, t) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ compose(3)]) auto
  with compose show ?thesis by fast
qed auto

lemma conversion2:
  "(x, y) \<in> r\<^sup>\<leftrightarrow> \<Longrightarrow> (y, z) \<in> r\<^sup>\<leftrightarrow> \<Longrightarrow> (x, z) \<in> r\<^sup>\<leftrightarrow>\<^sup>*"
  by (auto simp: conversion_def intro: rtrancl_trans)

lemma oKB_rstep_subset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "rstep (E \<union> R) \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  apply (auto elim!: rstepE dest!: oKB_E [OF oKB, THEN subsetD] oKB_R [OF oKB, THEN subsetD])
     apply (auto  dest!: ostep_imp_rstep_sym)
     apply (rule_tac y = "C\<langle>y \<cdot> \<sigma>\<rangle>" in conversion2; fast)+
  done

lemma oKB_conversion_subset:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using conversion_mono [OF oKB_rstep_subset [OF assms]] by simp

lemma oKB_rtrancl_conversion_subset:
  assumes "oKB\<^sup>*\<^sup>* (E, R) (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  by (induct "(E, R)" "(E', R')" arbitrary: E' R')
    (force dest!: oKB_conversion_subset)+

lemma oKB_rstep_subset_rev:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "rstep (E' \<union> R') \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  apply (auto elim!: rstepE)
   apply (auto dest!: oKB_R_rev [OF oKB, THEN subsetD] oKB_E_rev [OF oKB, THEN subsetD])
     apply (auto dest!: ostep_imp_rstep_sym)
     apply (rule conversion2; fast)+
  done

lemma oKB_conversion_supset:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using conversion_mono [OF oKB_rstep_subset_rev [OF assms]] by simp


section \<open>Abstract ordered completion according to proof for FSCD'17\<close>

inductive_set encstep1 for E :: "('a, 'b) trs" and R :: "('a, 'b) trs"
  where
    estep: "(l, r) \<in> E\<^sup>\<leftrightarrow> \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<Longrightarrow> l \<lhd>\<cdot> s \<Longrightarrow> (s, t) \<in> encstep1 E R"
  | rstep: "(s, t) \<in> rstep R \<Longrightarrow> (s, t) \<in> encstep1 E R"

inductive_set encstep2 for E :: "('a, 'b) trs" and R :: "('a, 'b) trs"
  where
    estep: "(l, r) \<in> E\<^sup>\<leftrightarrow> \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<Longrightarrow> l \<lhd>\<cdot> s \<Longrightarrow> (s, t) \<in> encstep2 E R"
  | rstep: "(l, r) \<in> R \<Longrightarrow> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<Longrightarrow> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<Longrightarrow> l \<lhd>\<cdot> s \<Longrightarrow> (s, t) \<in> encstep2 E R"

inductive
  oKBi :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> bool" (infix "\<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity>" 55)
  where
    deduce: "(s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (s, u) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E \<union> {(t, u)}, R)" |
    orientl: "s \<succ> t \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E - {(s, t)}, R \<union> {(s, t)})" |
    orientr: "t \<succ> s \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E - {(s, t)}, R \<union> {(t, s)})" |
    delete: "(s, s) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E - {(s, s)}, R)" |
    compose: "(t, u) \<in> ostep E (R - {(s, t)}) \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E, (R - {(s, t)}) \<union> {(s, u)})" |
    simplifyl: "(s, u) \<in> encstep1 (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> ((E - {(s, t)}) \<union> {(u, t)}, R)" |
    simplifyr: "(t, u) \<in> encstep1 (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> ((E - {(s, t)}) \<union> {(s, u)}, R)" |
    collapse: "(t, u) \<in> encstep2 E (R - {(t, s)}) \<Longrightarrow> (t, s) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E \<union> {(u, s)}, R - {(t, s)})"

lemma encstep2_encstep1:
  assumes "(s, t) \<in> encstep2 E R"
  shows "(s, t) \<in> encstep1 E R"
  unfolding encstep1.simps using assms by (cases) auto

lemma encstep1_ostep:
  assumes "(s, t) \<in> encstep1 E R"
  shows "(s, t) \<in> ostep E R"
  using assms
proof (cases)
  case (estep l r C \<sigma>)
  then show ?thesis unfolding ostep_def using ordstep.simps by fast
next
  case rstep
  then show ?thesis unfolding ostep_def rstep_def by fast
qed

lemma encstep2_ostep: "(s, t) \<in> encstep2 E R \<Longrightarrow> (s, t) \<in> ostep E R"
  using encstep2_encstep1 encstep1_ostep by blast

lemma encstep1_mono:
  "E \<subseteq> E' \<Longrightarrow> R \<subseteq> R'\<Longrightarrow> encstep1 E R \<subseteq> encstep1 E' R'"
  using rstep_mono [of R R']
  by (auto elim!: encstep1.cases intro: encstep1.intros)

lemma encstep2_mono:
  "E \<subseteq> E' \<Longrightarrow> R \<subseteq> R'\<Longrightarrow> encstep2 E R \<subseteq> encstep2 E' R'"
  by (auto elim!: encstep2.cases intro: encstep2.intros)

lemma encstep1_imp_rstep_sym:
  "(s, t) \<in> encstep1 E R \<Longrightarrow> (s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>)"
  using encstep1_ostep ostep_imp_rstep_sym by force

lemma encstep2_imp_rstep_sym:
  "(s, t) \<in> encstep2 E R \<Longrightarrow> (s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>)"
  using encstep2_ostep ostep_imp_rstep_sym by force

lemma oKBi_oKB:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  using assms
proof (cases)
  case (simplifyl s u t)
  then show ?thesis using oKB.simplifyl and encstep1_ostep by blast
next
  case (simplifyr t u s)
  then show ?thesis using oKB.simplifyr and encstep1_ostep by blast
next
  case (collapse t u s)
  then show ?thesis using oKB.collapse encstep2_ostep [OF collapse(3)] by blast
qed (blast intro: oKB.intros)+

lemma oKBi_subset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "E' \<union> R' \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_oKB [OF assms] and oKB_rstep_subset_rev subset_rstep [of "E' \<union> R'"] by blast

lemma oKBi_rstep_subset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "rstep(E' \<union> R') \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_oKB [OF assms] and oKB_rstep_subset_rev subset_rstep [of "E' \<union> R'"] by blast

lemma oKBi_conversion_subset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_rstep_subset [OF assms] and conversion_mono and conversion_conversion_idemp by metis

lemma oKBi_E_supset:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "E - E' \<subseteq> (encstep1 E' R' O E'\<^sup>\<leftrightarrow>)\<^sup>\<leftrightarrow> \<union> R'\<^sup>\<leftrightarrow> \<union> Id"
  using assms by (cases) (use encstep1_mono [of "E - {(s, t)}" E' R R' for s t] in blast)+

lemma oKBi_E_supset':
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "E - E' \<subseteq> (encstep1 E' R' O E') \<union> (E' O (encstep1 E' R')\<inverse>) \<union> R'\<^sup>\<leftrightarrow> \<union> Id"
  using assms by (cases) (use encstep1_mono [of "E - {(s, t)}" E' R R' for s t] in blast)+

lemma oKBi_R_supset:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "R - R' \<subseteq> encstep2 E' R' O E' \<union> R' O (ostep E' R')\<inverse>"
  using assms
proof (cases)
  case (compose s t u)
  then have "(s, t) \<in> ostep E' R'"
    using ostep_mono [of E' E "R - {(u, s)}" R'] by auto
  with compose show ?thesis by fast
next
  case (collapse s t u)
  with encstep2_mono [of E "E \<union> {(t, u)}"] show ?thesis by fast
qed auto

lemma oKBi_supset:
  assumes step: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "E \<union> R \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms subset_rstep encstep1_imp_rstep_sym encstep2_imp_rstep_sym
  by (smt Diff_partition oKB_rstep_subset oKBi_oKB ordered_completion_axioms sup.boundedE)

lemma oKBi_rstep_supset:
  assumes step: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "rstep(E \<union> R) \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_supset [OF assms]
  by (simp add: step oKB_rstep_subset oKBi_oKB ordered_completion_axioms)

lemma oKBi_conversion_supset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKBi_rstep_supset [OF assms] and conversion_mono and conversion_conversion_idemp by metis

lemma oKBi_conversion:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms and oKBi_conversion_supset and oKBi_conversion_subset by blast

lemma oKBi_less:
  assumes "R \<subseteq> {\<succ>}" and "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E', R')"
  shows "R' \<subseteq> {\<succ>}"
  using assms(1) and oKBi_oKB [OF assms(2)] and oKB_less by auto

lemma oKBi_rtrancl_less:
  assumes "oKBi\<^sup>*\<^sup>* (E, R) (E', R')" and "R \<subseteq> {\<succ>}"
  shows "R' \<subseteq> {\<succ>}"
  using assms oKB_less
proof (induct "(E, R)" "(E', R')" arbitrary: E' R')
  case rtrancl_refl
  then show ?case by fast
next
  case (rtrancl_into_rtrancl S)
  obtain E2 R2 where S: "S = (E2, R2)" by force
  from rtrancl_into_rtrancl show ?case using oKBi_less unfolding S by metis
qed

end

locale okb_irun = ordered_completion +
  fixes R E
  assumes R0: "R 0 = {}"
    and irun: "\<And>i. (E i, R i) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E (Suc i), R (Suc i))"
begin

sublocale kb less enc ..

abbreviation Rinf ("R\<^sub>\<infinity>") where "R\<^sub>\<infinity> \<equiv> (\<Union>i. R i)"
abbreviation Einf ("E\<^sub>\<infinity>") where "E\<^sub>\<infinity> \<equiv> (\<Union>i. E i)"

definition "R\<^sub>\<omega> = (\<Union>i. \<Inter>j\<in>{j. j\<ge>i}. R j)"
definition "E\<^sub>\<omega> = (\<Union>i. \<Inter>j\<in>{j. j\<ge>i}. E j)"

lemma Rw_subset_Rinf: "R\<^sub>\<omega> \<subseteq> R\<^sub>\<infinity>"
  by (auto simp: R\<^sub>\<omega>_def)

lemma Ew_subset_Einf: "E\<^sub>\<omega> \<subseteq> E\<^sub>\<infinity>"
  by (auto simp: E\<^sub>\<omega>_def)

lemma oKBi_rtrancl_i: "oKBi\<^sup>*\<^sup>* (E 0, R 0) (E i, R i)"
  using irun by (induct i) (auto intro: rtranclp.rtrancl_into_rtrancl)

lemma oKBi_conversion_ERi: "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E i \<union> R i))\<^sup>\<leftrightarrow>\<^sup>*"
  using irun and oKBi_conversion and R0 by (induct i) auto

lemma Ri_less: "R i \<subseteq> {\<succ>}"
  using oKBi_rtrancl_less [OF oKBi_rtrancl_i] and R0 by auto

lemma Rinf_less: "R\<^sub>\<infinity> \<subseteq> {\<succ>}"
  using Ri_less by auto

lemma Rw_less: "R\<^sub>\<omega> \<subseteq> {\<succ>}"
  unfolding R\<^sub>\<omega>_def by (use Ri_less in auto)

lemma encstep2D:
  assumes "(s, t) \<in> encstep2 E\<^sub>\<infinity> R\<^sub>\<infinity>"
  shows "\<exists>C \<tau> l r. (l, r) \<in> R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow> \<and> s = C\<langle>l \<cdot> \<tau>\<rangle> \<and> t = C\<langle>r \<cdot> \<tau>\<rangle> \<and> l \<cdot> \<tau> \<succ> r \<cdot> \<tau> \<and> l \<lhd>\<cdot> s"
  using assms
proof (cases)
  case (estep l r C \<sigma>)
  then show ?thesis by blast
next
  case (rstep l r C \<sigma>)
  from rstep(1) and Ri_less and subst [of l r] have "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" by auto
  with rstep show ?thesis by blast
qed

definition lessencp (infix "\<cdot>\<succ>" 50) where "lessencp s t \<longleftrightarrow> (s, t) \<in> (relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+"
abbreviation lessenc ("{\<cdot>\<succ>}") where "lessenc \<equiv> {(s, t). s \<cdot>\<succ> t}"

definition leqencp (infix "\<cdot>\<succeq>" 50) where "leqencp s t \<longleftrightarrow> (s, t) \<in> (relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+ \<union> {\<cdot>\<unrhd>}\<^sup>*"
abbreviation leqenc ("{\<cdot>\<succeq>}") where "leqenc \<equiv> {(s, t). s \<cdot>\<succeq> t}"

lemma SN_lessenc: "SN {\<cdot>\<succ>}"
  using reduction_order.SN_encomp_Un_less_relto_encompeq
    SN_encomp_Un_less_relto_encompeq SN_trancl_SN_conv unfolding lessencp_def by auto

lemma lessenc_compat1: "{\<cdot>\<succ>} O {\<cdot>\<unrhd>} \<subseteq> {\<cdot>\<succ>}"
proof
  fix s t
  let ?R = "{\<cdot>\<unrhd>}\<^sup>* O ({\<cdot>\<rhd>} \<union> {\<succ>}) O {\<cdot>\<unrhd>}\<^sup>*"
  assume "(s, t) \<in> {\<cdot>\<succ>} O {\<cdot>\<unrhd>}"
  then obtain u where u: "(u, t) \<in> {\<cdot>\<unrhd>}" "s \<cdot>\<succ> u" by auto
  then obtain v where v: "(v, u) \<in> ?R" "(s, v) \<in> ?R\<^sup>*" using tranclD2 unfolding lessencp_def by metis
  from u(1) v(1) have "(v, t) \<in> ?R O {\<cdot>\<unrhd>}" by blast
  with u(1) v(1) have "(v, t) \<in> ?R" using O_assoc O_mono2 [of "{\<cdot>\<unrhd>} O {\<cdot>\<unrhd>}\<^sup>*" "{\<cdot>\<unrhd>}\<^sup>*"]
    by (smt r_into_rtrancl relcomp.relcompI rtrancl_idemp_self_comp u(1) v(1))
  with v(2) have "s \<cdot>\<succ> t" unfolding trancl_unfold_right lessencp_def by blast
  then show "(s, t) \<in> {\<cdot>\<succ>}" by auto
qed

lemma leqenc_compat1: "{\<cdot>\<succeq>} O {\<cdot>\<unrhd>} \<subseteq> {\<cdot>\<succeq>}"
proof
  fix s t
  let ?R = "({\<cdot>\<unrhd>}\<^sup>* O ({\<cdot>\<rhd>} \<union> {\<succ>}) O {\<cdot>\<unrhd>}\<^sup>*) \<union> {\<cdot>\<unrhd>}\<^sup>*"
  assume "(s, t) \<in> {\<cdot>\<succeq>} O {\<cdot>\<unrhd>}"
  then obtain u where u: "s \<cdot>\<succeq> u" "(u, t) \<in> {\<cdot>\<unrhd>}" by auto
  then consider "s \<cdot>\<succ> u" | "(s, u) \<in> {\<cdot>\<unrhd>}\<^sup>*"
    unfolding leqencp_def lessencp_def Un_iff by blast
  then show "(s, t) \<in> {\<cdot>\<succeq>}"
  proof (cases)
    case 1
    with lessenc_compat1 u(2) show ?thesis unfolding leqencp_def lessencp_def by auto
  next
    case 2
    with u(2) rtrancl_comp_trancl_conv [of "{\<cdot>\<unrhd>}"] have "(s, t) \<in> {\<cdot>\<unrhd>}\<^sup>*" by auto
    then show ?thesis unfolding leqencp_def Un_iff mem_Collect_eq split by auto
  qed
qed

lemma lessleqenc_compat1: "{\<cdot>\<succ>} O {\<cdot>\<succeq>} \<subseteq> {\<cdot>\<succ>}"
proof
  fix x z
  assume "(x, z) \<in> {\<cdot>\<succ>} O {\<cdot>\<succeq>}"
  then obtain y where y: "(x, y) \<in> {\<cdot>\<succ>}" "(y, z) \<in> {\<cdot>\<succeq>}" by auto
  then consider "y \<cdot>\<succ> z" | "(y, z) \<in> {\<cdot>\<unrhd>}\<^sup>*"
    unfolding leqencp_def lessencp_def Un_iff by blast
  then show "(x, z) \<in> {\<cdot>\<succ>}"
  proof (cases)
    case 1
    with y(1) show ?thesis using trancl_trans unfolding lessencp_def by auto
  next
    case 2
    from encompeq_trans trans_inv_image transI have t: "trans {\<cdot>\<unrhd>}"
      by (smt Pair_inject case_prodE case_prodI2 mem_Collect_eq)
    have "{\<cdot>\<unrhd>}\<^sup>* = {\<cdot>\<unrhd>}" by(rule trans_refl_imp_rtrancl_id, insert t encompeq_refl refl_O_iff, auto)
    with 2 [unfolded this] y(1) lessenc_compat1 show ?thesis by auto
  qed
qed

lemma lessenc_compat2: "{\<cdot>\<unrhd>} O {\<cdot>\<succ>} \<subseteq> {\<cdot>\<succ>}"
proof
  fix s t
  let ?R = "{\<cdot>\<unrhd>}\<^sup>* O ({\<cdot>\<rhd>} \<union> {\<succ>}) O {\<cdot>\<unrhd>}\<^sup>*"
  assume "(s, t) \<in> {\<cdot>\<unrhd>} O {\<cdot>\<succ>}"
  then obtain u where u: "(s, u) \<in> {\<cdot>\<unrhd>}" "u \<cdot>\<succ> t" by auto
  then obtain v where v: "(u, v) \<in> ?R" "(v, t) \<in> ?R\<^sup>*" using tranclD unfolding lessencp_def by metis
  from u(1) v(1) have "(s, v) \<in> {\<cdot>\<unrhd>} O ?R" by blast
  with u(1) v(1) have "(s, v) \<in> ?R" using O_assoc O_mono2 [of "{\<cdot>\<unrhd>} O {\<cdot>\<unrhd>}\<^sup>*" "{\<cdot>\<unrhd>}\<^sup>*"]
    by (smt r_into_rtrancl relcomp.relcompI rtrancl_idemp_self_comp u(1) v(1))
  with v(2) have "s \<cdot>\<succ> t" unfolding trancl_unfold_left lessencp_def by blast
  then show "(s, t) \<in> {\<cdot>\<succ>}" by auto
qed

lemma encstep1_less:
  assumes "(s, t) \<in> encstep1 (E i) (R i)"
  shows "s \<succ> t"
  using assms
proof (cases)
  case (estep l r C \<sigma>)
  with ctxt [OF estep(4)] show ?thesis unfolding lessencp_def by fast
next
  case rstep
  with compatible_rstep_imp_less [OF Ri_less] show ?thesis unfolding lessencp_def by blast
qed

lemma encstep1_lessenc:
  assumes "(s, t) \<in> encstep1 (E i) (R i)"
  shows "s \<cdot>\<succ> t"
  using assms encstep1_less unfolding lessencp_def by fast

lemma trans_less_enc: "trans {\<cdot>\<succ>}"
proof -
  from commutes_rewrel_encomp [of "{\<succ>}"] and ctxt and subst have "{\<cdot>\<unrhd>} O {\<succ>} \<subseteq>  {\<succ>} O {\<cdot>\<unrhd>}" by simp
  with trans_relto [of "{\<succ>}" "{\<cdot>\<unrhd>}"] and trans show ?thesis unfolding lessencp_def by auto
qed

lemma refl_encompeq: "refl {\<cdot>\<unrhd>}"
  using encompeq_refl reflpI reflp_refl_eq by fastforce

abbreviation "mulless \<equiv> s_mul_ext {\<cdot>\<unrhd>} {\<cdot>\<succ>}"
abbreviation "mulleq \<equiv> ns_mul_ext {\<cdot>\<unrhd>} {\<cdot>\<succ>}"

sublocale lessenc_op: SN_order_pair mulless "ns_mul_ext {\<cdot>\<unrhd>} {\<cdot>\<succ>}"
proof (intro SN_order_pair.mul_ext_SN_order_pair)
  show "SN_order_pair {\<cdot>\<succ>} {\<cdot>\<unrhd>}"
  proof
    show "trans {\<cdot>\<succ>}" using trans_less_enc by auto
  next
    show "refl {\<cdot>\<unrhd>}" using refl_encompeq .
  next
    show "trans {\<cdot>\<unrhd>}" using encomp_order.order_trans transpI transp_trans by metis
  next
    from SN_lessenc show "SN {\<cdot>\<succ>}" by auto
  next
    from lessenc_compat2 show "{\<cdot>\<unrhd>} O {\<cdot>\<succ>} \<subseteq> {\<cdot>\<succ>}" by auto
  next
    from lessenc_compat1 show "{\<cdot>\<succ>} O {\<cdot>\<unrhd>} \<subseteq> {\<cdot>\<succ>}" by auto
  qed
qed

lemma mset_two:
  assumes "s \<cdot>\<succ> u"
  shows "({#s, t#}, {#u, t#}) \<in> mulless"
proof -
  from assms s_mul_ext_singleton have su: "({#s#}, {#u#}) \<in> mulless" by auto
  from ns_mul_ext_refl [OF refl_encompeq] have "({#t#}, {#t#}) \<in> mulleq" by auto
  from s_ns_mul_ext_union_compat [OF su this] show ?thesis
    using add.commute add_mset_add_single by metis
qed

lemma mset_two2:
  assumes "s \<cdot>\<succ> u"
  shows "({#t, s#}, {#t, u#}) \<in> mulless"
  using mset_two [OF assms] add_mset_commute by (metis (no_types, lifting))

definition Sw ("S\<^sub>\<omega>") where "S\<^sub>\<omega> \<equiv> R\<^sub>\<omega> \<union> {(s \<cdot> \<sigma>, t \<cdot> \<sigma>) | s t \<sigma>. (s, t) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow> \<and> s \<cdot> \<sigma> \<succ> t \<cdot> \<sigma>}"
  
lemma Sw_less: "S\<^sub>\<omega> \<subseteq> {\<succ>}"
  using Ri_less unfolding Sw_def R\<^sub>\<omega>_def by fast
  
lemma Sw_step_less: "rstep S\<^sub>\<omega> \<subseteq> {\<succ>}"
  using compatible_rstep_imp_less[OF Sw_less] by auto
    
lemma Sw_implies_REw: "rstep S\<^sub>\<omega> \<subseteq> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>"
  unfolding Sw_def rstep_mono by fastforce
    
lemma Sw_implies_REw': "rstep S\<^sub>\<omega> \<subseteq> rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
  unfolding Sw_def rstep_mono by fastforce

lemma SN_Sw_step: "SN (rstep S\<^sub>\<omega>)"
  using Sw_step_less SN_subset [OF SN_less] by fastforce
    
lemma rstep_imp_mstep:
  assumes "(t, u) \<in> rstep RR" and "\<exists>t' \<in># S. t' \<succeq> t" and "\<exists>u' \<in># S. u' \<succeq> u"
  shows "(t, u) \<in> mstep S RR"
  using assms by (auto simp: mstep_iff dest: trans)
    
lemma mstep_succeq_mono:
  assumes "\<forall>m \<in># M. \<exists>n \<in># N. n \<succeq> m"
  shows "mstep M RR \<subseteq> mstep N RR"
proof
  fix s t
  assume mstep: "(s, t) \<in> mstep M RR"
  with assms trans show "(s, t) \<in> mstep N RR" unfolding mstep_def mem_Collect_eq split
    by (smt sup2E sup2I1)
qed

lemma msteps_succeq_mono:
  assumes "\<forall>m \<in># M. \<exists>n \<in># N. n \<succeq> m"
  shows "(mstep M RR)\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (mstep N RR)\<^sup>\<leftrightarrow>\<^sup>*"
  by (rule mstep_succeq_mono [OF assms, THEN conversion_mono])

lemma mstep_subst_ctxt:
  assumes "(s, t) \<in> mstep S RR"
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> mstep {#C\<langle>u \<cdot> \<sigma>\<rangle>. u \<in># S #} RR"
proof -
  let ?S = "{#C\<langle>u \<cdot> \<sigma>\<rangle>. u \<in># S #}"
  from assms have step: "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> rstep RR" unfolding mstep_def by auto
  from assms have gt:"\<exists>s' \<in># S. s' \<succeq> s" "\<exists>t' \<in># S. t' \<succeq> t" unfolding mstep_def by auto
  with subst ctxt have s':"\<exists>s' \<in># ?S. s' \<succeq> C\<langle>s \<cdot> \<sigma>\<rangle>" by fast
  from gt(2) subst ctxt have "\<exists>t' \<in># ?S. t' \<succeq> C\<langle>t \<cdot> \<sigma>\<rangle>" by fast
  with step s' show ?thesis unfolding mstep_def by blast
qed

lemma msteps_subst_ctxt:
  assumes "(s, t) \<in> (mstep S RR)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (mstep {#C\<langle>u \<cdot> \<sigma>\<rangle>. u \<in># S #} RR)\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  let ?S = "{#C\<langle>u \<cdot> \<sigma>\<rangle>. u \<in># S #}"
  from mstep_subst_ctxt have x: "\<And> s t. (s, t) \<in> (mstep S RR)\<^sup>\<leftrightarrow> \<Longrightarrow> (C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (mstep ?S RR)\<^sup>\<leftrightarrow>" by auto
  from assms rtrancl_imp_relpow obtain n where "(s, t) \<in> (mstep S RR)\<^sup>\<leftrightarrow> ^^ n" by fast
  then have "(C\<langle>s \<cdot> \<sigma>\<rangle>, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> (mstep ?S RR)\<^sup>\<leftrightarrow> ^^ n"
  proof (induct n arbitrary: s)
    case (Suc n)
    then obtain s' where s': "(s, s') \<in> (mstep S RR)\<^sup>\<leftrightarrow>" "(s', t) \<in> (mstep S RR)\<^sup>\<leftrightarrow> ^^ n"
      unfolding relpow_Suc by blast
    from x [OF s'(1)] Suc(1) [OF s'(2)] show ?case unfolding relpow_Suc by blast
  qed simp
  with relpow_imp_rtrancl show ?thesis by blast
qed

lemma Einf_without_Ew:
  assumes "(l,r) \<in> E\<^sub>\<infinity>" and "(l,r) \<notin>  E\<^sub>\<omega>"
  shows "\<exists>j. (l, r) \<in> (E j) \<and> (l, r) \<notin> (E (Suc j))"
proof-
  from assms obtain j where "(l, r) \<in> E j" and "\<exists>i>j. (l, r) \<notin> E i"
    unfolding E\<^sub>\<omega>_def le_eq_less_or_eq by blast
  moreover define i where "i = (LEAST i. i > j \<and> (l, r) \<notin> E i)"
  ultimately have "i > j" and not_i: "(l, r) \<notin> E i" by (metis (lifting) LeastI)+
  then have "i - 1 < i" and [simp]: "Suc (i - Suc 0) = i" by auto
  from not_less_Least [OF this(1) [unfolded i_def], folded i_def] and \<open>i > j\<close> and \<open>(l, r) \<in> E j\<close>
  have pred_i: "(l, r) \<in> E (i - 1)" by (cases "i = Suc j") auto
  have i_simp: "Suc (i - 1) = i" by auto
  show ?thesis by (rule exI[of _ "i - 1"], insert not_i pred_i i_simp, auto)
qed

lemma Rinf_without_Rw:
  assumes "(l,r) \<in> R\<^sub>\<infinity>" and "(l,r) \<notin>  R\<^sub>\<omega>"
  shows "\<exists>j. (l, r) \<in> (R j) \<and> (l, r) \<notin> (R (Suc j))"
proof-
  from assms obtain j where "(l, r) \<in> R j" and "\<exists>i>j. (l, r) \<notin> R i"
    unfolding R\<^sub>\<omega>_def le_eq_less_or_eq by blast
  moreover define i where "i = (LEAST i. i > j \<and> (l, r) \<notin> R i)"
  ultimately have "i > j" and not_i: "(l, r) \<notin> R i" by (metis (lifting) LeastI)+
  then have "i - 1 < i" and [simp]: "Suc (i - Suc 0) = i" by auto
  from not_less_Least [OF this(1) [unfolded i_def], folded i_def] and \<open>i > j\<close> and \<open>(l, r) \<in> R j\<close>
  have pred_i: "(l, r) \<in> R (i - 1)" by (cases "i = Suc j") auto
  have i_simp: "Suc (i - 1) = i" by auto
  show ?thesis by (rule exI[of _ "i - 1"], insert not_i pred_i i_simp, auto)
qed
  
lemma Ei_subset_EwRi: "mstep S E\<^sub>\<infinity> \<subseteq> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  { fix t0 u0
    assume "(t0, u0) \<in> rstep E\<^sub>\<infinity>" and St: "\<exists>s \<in># S. s \<succeq> t0" and Su: "\<exists>s \<in># S. s \<succeq> u0"
    with rstep_imp_C_s_r [OF this(1)] obtain C \<sigma> t u
      where Cs: "t0 = C\<langle>t \<cdot> \<sigma>\<rangle>" "u0 = C\<langle>u \<cdot> \<sigma>\<rangle>" and tu: "(t, u) \<in> E\<^sub>\<infinity>" by blast
    from tu St Su have "(C\<langle>t \<cdot> \<sigma>\<rangle>, C\<langle>u \<cdot> \<sigma>\<rangle>) \<in> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding Cs
    proof (induct "{#t, u#}" arbitrary: t u S C \<sigma> rule: SN_induct [OF lessenc_op.SN])
      case 1
      note sc = this
      let ?t = "C\<langle>t \<cdot> \<sigma>\<rangle>" and ?u = "C\<langle>u \<cdot> \<sigma>\<rangle>"
      have comm: "\<And>t u. {#t, u#} =  {#u, t#}" using add_mset_commute by auto
      note r_imp_msteps = rstep_imp_mstep [OF _ sc(4) sc(3)] rstep_imp_mstep [OF _ sc(3) sc(4)]
      show ?case
      proof (cases "(?t, ?u) \<in> rstep E\<^sub>\<omega>")
        case True
        with r_imp_msteps mstep_Un [of _ "R\<^sub>\<infinity>" "E\<^sub>\<omega>"] show ?thesis by auto
      next
        case False
        with rstep_imp_C_s_r have "(t, u) \<notin> E\<^sub>\<omega>" by fast
        from Einf_without_Ew[OF sc(2) this]  obtain i where
          pred_i:"(t, u) \<in> E i" and not_i: "(t, u) \<notin> E (Suc i)" by auto
        let ?enc = "encstep1 (E (Suc i)) (R (Suc i)) O (E (Suc i))\<^sup>\<leftrightarrow>"
        from pred_i oKBi_E_supset [OF irun [of i]] and not_i
        consider "(t, u) \<in> (R (Suc i))\<^sup>\<leftrightarrow>" | "u = t" | "(t, u) \<in> ?enc\<^sup>\<leftrightarrow>" by auto blast
        then show ?thesis
        proof (cases)
          case 1
          then have "(?t, ?u) \<in> rstep (R\<^sub>\<infinity>\<^sup>\<leftrightarrow>)" by auto
          with r_imp_msteps [of "R\<^sub>\<infinity>"] converse_iff have "(?t, ?u) \<in> (mstep S R\<^sub>\<infinity>) \<or> (?u, ?t) \<in> (mstep S R\<^sub>\<infinity>)"
            by fastforce
          with mstep_Un conversion_inv show ?thesis by blast
        next
          case 3
          then have "(t, u) \<in> ?enc \<or> (u, t) \<in> ?enc" by auto
          with sc obtain t' u' where tu': "(t', u') \<in> ?enc" "(t=t' \<and> u=u') \<or> (t=u' \<and> u=t')" by blast
          then have tu_tu': "{# t, u #} = {# t', u' #}" by force
          from sc(3) sc(4) tu'(2) trans have Stu: "\<exists>s \<in># S. s \<succeq> C\<langle>u' \<cdot> \<sigma>\<rangle> \<and> (\<exists>s \<in># S. s \<succeq> C\<langle>t' \<cdot> \<sigma>\<rangle>)" by meson
          from tu'(1) obtain v where
            v: "(t', v) \<in> encstep1 (E (Suc i)) (R (Suc i))" "(v, u') \<in> (E (Suc i))\<^sup>\<leftrightarrow>" by auto
          from tu_tu' mset_two [OF encstep1_lessenc [OF v(1)]]
          have dec: "({# t, u #}, {# v, u' #}) \<in> mulless" by auto
          then have dec': "({# t, u #}, {# u', v #}) \<in> mulless" using add_mset_commute by metis
          from encstep1_ostep [OF v(1), THEN ostep_imp_less [OF Ri_less]] have tv: "t' \<succeq> v" by blast
          have vu: "(C\<langle>v\<cdot> \<sigma>\<rangle>, C\<langle>u'\<cdot> \<sigma>\<rangle>) \<in> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
          proof -
            from trans subst ctxt tv Stu have Sv: "\<exists>s \<in># S. s \<succeq> C\<langle>v\<cdot> \<sigma>\<rangle>" by fast
            from v(2) have "(v, u') \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" by auto
            with sc(1) [OF dec] sc(1) [OF dec'] Sv Stu show ?thesis
              unfolding Un_iff conversion_inv by blast
          qed
          from v(1) have "(C\<langle>t' \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<sigma>\<rangle>) \<in> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
          proof (cases)
            case (estep l r D \<tau>)
            then have lr: "(l, r) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" by auto
            with encompeq.intros [of t' D "l \<cdot> \<tau>" Var] estep(2) have tl: "l \<cdot> \<tau> \<unlhd>\<cdot> t'" by auto
            with encompeq.intros [of "r \<cdot> \<tau>" \<box>] estep(2) have "r \<unlhd>\<cdot> r \<cdot> \<tau>" by auto
            with estep(4) tl have tr: "t' \<cdot>\<succ> r" unfolding lessencp_def by blast
            from estep(5) have "t' \<cdot>\<succ> l" unfolding lessencp_def by blast
            with tr ns_s_mul_ext_union_multiset_l [OF ns_mul_ext_bottom [of "{#u'#}"], of _ "{#l, r#}"]
            have "({#u', t'#}, {#l, r#}) \<in> mulless" using multi_member_last by auto
            with tu_tu' have dec: "({#t, u#}, {#l, r#}) \<in> mulless" using comm by metis
            with comm [of l] have dec': "({#t, u#}, {#r, l#}) \<in> mulless" by presburger
            note ih = sc(1) [OF dec, of "{#l, r#}" \<box> Var] sc(1) [OF dec', of "{#l, r#}" \<box> Var]
            from lr [unfolded Un_iff] ih have "(l, r) \<in> (mstep {#l, r#} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
              unfolding conversion_inv [of r] ctxt.cop_nil subst_apply_term_empty by force
            from msteps_subst_ctxt [OF this] estep(2) estep(3)
            have mstep: "(t', v) \<in> (mstep {#t', v#} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
            from Stu tv trans ctxt [OF subst] have "\<exists>s\<in>#S. s \<succeq> C\<langle>v \<cdot> \<sigma>\<rangle>" by fast
            with Stu tv have S:"\<forall>m\<in>#{#C\<langle>t' \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<sigma>\<rangle>#}. \<exists>n\<in>#S. n \<succeq> m" by auto
            have mset:"{#C\<langle>u \<cdot> \<sigma>\<rangle>. u \<in># {#t', v#}#} = {#C\<langle>t' \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<sigma>\<rangle>#}" by simp
            with msteps_subst_ctxt [OF mstep, of C \<sigma>] conversion_mono [OF mstep_succeq_mono [OF S]]
              show ?thesis unfolding mset by blast
          next
            case rstep
            with rstep_subset_less [OF Ri_less] have tv': "t' \<succ> v" by blast
            from rstep have "(t', v) \<in> rstep (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>)" by fast
            from rstep_imp_mstep [OF this] tv'  have "(t', v) \<in> mstep {#t', t'#} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>)" by auto
            from mstep_subst_ctxt [OF this, of C \<sigma>] mstep_succeq_mono [of "{#C\<langle>t' \<cdot> \<sigma>\<rangle>, C\<langle>t' \<cdot> \<sigma>\<rangle>#}" S] Stu
            show ?thesis by fastforce
          qed
          from transD [OF conversion_trans, OF this vu] tu'(2) symD [OF conversion_sym]
          show ?thesis by metis
        qed auto
      qed
    qed
    with Cs have "(t0, u0) \<in> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  }
  then show ?thesis unfolding mstep_def by blast
qed

lemma Einf_sym_mstep: "mstep S (E\<^sub>\<infinity>\<^sup>\<leftrightarrow>) \<subseteq> (mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  have inv: "mstep S (E\<^sub>\<infinity>\<inverse>) = (mstep S E\<^sub>\<infinity>)\<inverse>"
    unfolding mstep_def reverseTrs [symmetric] converse_iff by auto
  from Ei_subset_EwRi show ?thesis unfolding mstep_Un inv Un_subset_iff
    by (metis converse_mono conversion_converse)
qed

abbreviation "lexless \<equiv> lex_two {\<cdot>\<succ>} Id {\<cdot>\<succ>}"

lemma SN_lexless: "SN lexless"
  using SN_lessenc
  by (intro lex_two) (auto simp only: SN_trancl_SN_conv intro: SN_encomp_Un_less_relto_encompeq)

lemma Ri_subset_ERw: "mstep S R\<^sub>\<infinity> \<subseteq> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  { fix t u
    assume "(t, u) \<in> rstep R\<^sub>\<infinity>" and St: "\<exists>s \<in># S. s \<succeq> t" and Su: "\<exists>s \<in># S. s \<succeq> u"
    with rstep_imp_C_s_r [OF this(1)] obtain C \<sigma> l r
      where Cs: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" "u = C\<langle>r \<cdot> \<sigma>\<rangle>" and tu: "(l, r) \<in> R\<^sub>\<infinity>" by blast
    from tu St Su have "(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding Cs
    proof (induct "(l, r)" arbitrary: l r S C \<sigma> rule: SN_induct [OF SN_lexless])
      case 1
      note sc = this
      with Ri_less have lr: "l \<succ> r" by auto
      let ?t = "C\<langle>l \<cdot> \<sigma>\<rangle>" and ?u = "C\<langle>r \<cdot> \<sigma>\<rangle>"
      show ?case
      proof (cases "(l, r) \<in> R\<^sub>\<omega>")
        case True
          (* rule l \<rightarrow> r is persistent *)
        then have "(?t, ?u) \<in> rstep R\<^sub>\<omega>" by auto
        with rstep_imp_mstep [OF this sc(3) sc(4)] mstep_Un show ?thesis by auto
      next
        case False
        (* rule l \<rightarrow> r was removed in step (E i, R i) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sub>\<infinity> (E (i+1), R (i+1)) *)
        from Rinf_without_Rw[OF sc(2) False] obtain i where
          pred_i: "(l, r) \<in> R i" and not_i: "(l, r) \<notin> R (Suc i)" by auto
        (* first verify that (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>) conversion below l corresponds to a (R\<^sub>\<omega> \<union> E\<^sub>\<omega>) conversion *)
        { fix S'
          assume below_l: "\<forall> s' \<in># S'. l \<cdot>\<succ> s'"
          { fix s t
            assume "(s, t) \<in> mstep S' R\<^sub>\<infinity>"
            then have mstep: "(s, t) \<in> rstep R\<^sub>\<infinity>" "\<exists>s' \<in># S'. s' \<succeq> s" "\<exists>s' \<in># S'. s' \<succeq> t"
              unfolding mstep_def by auto
            then obtain C \<sigma> l1 r1 where l1r1: "s = C\<langle>l1 \<cdot> \<sigma>\<rangle>" "t = C\<langle>r1 \<cdot> \<sigma>\<rangle>" "(l1, r1) \<in> R\<^sub>\<infinity>" by fast
            from mstep obtain s' where s': "s' \<in># S'" "s' \<cdot>\<succeq> s" unfolding leqencp_def by blast
            from l1r1 encompeq.intros have sl1: "s \<cdot>\<unrhd> l1" by auto
            from s' below_l have ls': "l \<cdot>\<succ> s'" by auto
            with s' lessleqenc_compat1 have "l \<cdot>\<succ> s" by auto
            with sl1 lessenc_compat1 have "l \<cdot>\<succ> l1" by auto
            then have "((l, r), (l1, r1)) \<in> lexless" by force
            from sc(1) [OF this l1r1(3)] mstep have "(s, t) \<in> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
              unfolding l1r1 by meson
          } note step = this
          from r_into_rtrancl have "mstep S' E\<^sub>\<omega> \<subseteq> (mstep S' E\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*" by auto
          with step mstep_Un have "mstep S' (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>) \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          from conversion_mono [OF this] have R: "(mstep S' (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
            by simp
        } note mstep_Rinf_msteps_ERw = this
        (* derive equation (1): (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>) steps below l correspond to a (R\<^sub>\<omega> \<union> E\<^sub>\<omega>) conversion *)
        { fix S'
          assume below_l: "\<forall> s' \<in># S'. l \<cdot>\<succ> s'"
          note R = mstep_Rinf_msteps_ERw [OF this]
          with Einf_sym_mstep have "mstep S' (E\<^sub>\<infinity>\<^sup>\<leftrightarrow>) \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          with R mstep_Un have "mstep S' (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>) \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          from conversion_mono [OF this] have "(mstep S' (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (mstep S' (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
            by simp
        } note msteps_ERinf_mstep_Rw = this
        (* case distinction according to Lemma 31 *)
        let ?Ei = "E (Suc i)" and ?Ri = "R (Suc i)"
        from pred_i oKBi_R_supset [OF irun [of i]] and not_i
          consider "(l, r) \<in> encstep2 ?Ei ?Ri O ?Ei" | "(l, r) \<in> ?Ri O (ostep ?Ei ?Ri)\<inverse>" by blast
        then have "(l, r) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
        proof (cases)
          (* collapse step *)
          case 1
          then obtain u where u: "(l, u) \<in> encstep2 ?Ei ?Ri" "(u, r) \<in> ?Ei" by auto
          (* l \<longleftrightarrow>\<^sub>R\<^sub>\<omega> \<^sub>\<union> \<^sub>E\<^sub>\<omega>\<^sup>{\<^sup>l\<^sup>, \<^sup>r\<^sup>}\<^sup>* u *)
          with encstep2_mono [of ?Ei "E\<^sub>\<infinity>" ?Ri "R\<^sub>\<infinity>"] have lu: "(l, u) \<in> encstep2 E\<^sub>\<infinity> R\<^sub>\<infinity>" by auto
          from encstep2D [OF this] obtain D \<tau> l0 r0 where
            D: "(l0, r0) \<in> R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" "l = D\<langle>l0 \<cdot> \<tau>\<rangle>" "u = D\<langle>r0 \<cdot> \<tau>\<rangle>" " l0 \<cdot> \<tau> \<succ> r0 \<cdot> \<tau>" "l0 \<lhd>\<cdot> l" by blast
          then have l_l0: "l \<cdot>\<succ> l0" unfolding lessencp_def by auto
          from D encompeq.intros [of l D _ Var] have l_l0\<tau>: "l \<cdot>\<unrhd> l0 \<cdot> \<tau>" by auto
          from encompeq.intros [of _ \<box>] have "r0 \<cdot> \<tau> \<cdot>\<unrhd> r0" by auto
          with l_l0\<tau> D(4) have l_r0: "l \<cdot>\<succ> r0" unfolding lessencp_def by blast
          have "(l0, r0) \<in> (mstep {#l0, r0#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
          proof (cases "(l0, r0) \<in> R\<^sub>\<infinity>")
            case True
            with Ri_less have l_gt_r: "l0 \<succ> r0" by auto
            from l_l0 have "((l, r), (l0, r0)) \<in> lexless" by auto
            from sc(1) [OF this True, of "{#l0, r0#}" \<box> Var, simplified] l_gt_r show ?thesis by auto
          next
            case False
            with D(1) have "(l0, r0) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" by auto
            then have "(l0, r0) \<in> mstep {#l0, r0#} (E\<^sub>\<infinity>\<^sup>\<leftrightarrow>)" unfolding mstep_def by auto
            with in_mono [OF Einf_sym_mstep] have l0r0: "(l0, r0) \<in> (mstep {#l0, r0#} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
            from l_l0 l_r0 have "\<forall> s' \<in># {#l0, r0#}. l \<cdot>\<succ> s'" by auto
            from l0r0 mstep_Un mstep_Rinf_msteps_ERw [OF this] show ?thesis by auto
          qed
          from D msteps_subst_ctxt [OF this] have lu: "(l, u) \<in> (mstep {#l, u#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          from u(1) encstep1_less [OF encstep2_encstep1] have l_succeq_u: "l \<succeq> u" by auto
          then have "\<forall>m\<in>#{#l, u#}. \<exists>n\<in>#{#l, r#}. n \<succeq> m" by auto
          from lu msteps_succeq_mono [OF this] have lu: "(l, u) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by blast
              (* u \<longleftrightarrow>\<^sub>R\<^sub>\<omega> \<^sub>\<union> \<^sub>E\<^sub>\<omega>\<^sup>{\<^sup>l\<^sup>, \<^sup>r\<^sup>}\<^sup>* r *)
          from u(1) encstep1_lessenc [OF encstep2_encstep1] have "l \<cdot>\<succ> u" by auto
          with lr have lt_l: "\<forall> s' \<in># {#u, r#}. l \<cdot>\<succ> s'" unfolding lessencp_def by force
          from u(2) have "(u, r) \<in> mstep {#u, r#} (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>)" unfolding mstep_def by auto
          from r_into_rtrancl [OF this] msteps_ERinf_mstep_Rw [OF lt_l]
          have ur: "(u, r) \<in> (mstep {#u, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
          from l_succeq_u have "\<forall>m\<in>#{#u, r#}. \<exists>n\<in>#{#l, r#}. n \<succeq> m" by auto
          from ur msteps_succeq_mono [OF this] have "(u, r) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by blast
          with lu transD [OF conversion_trans] show ?thesis by auto
        next
          (* compose step *)
          case 2
          then obtain u where u: "(l, u) \<in> ?Ri" "(u, r) \<in> (ostep ?Ei ?Ri)\<inverse>" by auto
              (* l \<longleftrightarrow>\<^sub>R\<^sub>\<omega> \<^sub>\<union> \<^sub>E\<^sub>\<omega>\<^sup>{\<^sup>l\<^sup>, \<^sup>r\<^sup>}\<^sup>* u*)
          with ostep_imp_less [OF Ri_less] have ru: "r \<succ> u" by auto
          then have "r \<cdot>\<succ> u" unfolding lessencp_def by blast
          then have "((l, r), (l, u)) \<in> lexless" unfolding lex_two.simps mem_Collect_eq split by auto
          from trans [OF lr ru] sc(1) [OF this, of "{#l, r#}" \<box> Var, simplified] u(1)
          have lu: "(l, u) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
              (* u \<longleftrightarrow>\<^sub>R\<^sub>\<omega> \<^sub>\<union> \<^sub>E\<^sub>\<omega>\<^sup>{\<^sup>l\<^sup>, \<^sup>r\<^sup>}\<^sup>* r*)
          from ostep_imp_rstep_sym u(2) have "(r, u) \<in> rstep (?Ri \<union> ?Ei\<^sup>\<leftrightarrow>)" by auto
          with rstep_UN rstep_union have rstep: "(r, u) \<in> rstep (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>)" by fast
          from rstep have mstep: "(r, u) \<in> mstep {#u, r#} (R\<^sub>\<infinity> \<union> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>)"
            unfolding mstep_def mem_Collect_eq split by auto
          from lr trans [OF lr ru] have lt_l: "\<forall> s' \<in># {#u, r#}. l \<cdot>\<succ> s'"
            unfolding lessencp_def by force
          from msteps_ERinf_mstep_Rw [OF this] r_into_rtrancl [OF mstep]
          have ur: "(u, r) \<in> (mstep {#u, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_inv [of u] by auto
          have "(u, r) \<in> (mstep {#l, r#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
            by(rule subsetD [OF msteps_succeq_mono], insert trans [OF lr ru] ur, auto)
          with lu transD [OF conversion_trans] show ?thesis by auto
        qed
        from msteps_subst_ctxt [OF this] have tu: "(?t, ?u) \<in> (mstep {#?t, ?u#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
        show ?thesis by(rule subsetD [OF msteps_succeq_mono], insert sc(3) sc(4) tu, auto)
      qed
    qed
    with Cs have "(t, u) \<in> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  }
  then show ?thesis using mstep_def by auto
qed

lemma RiEw_subset_ERw: "(mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  from r_into_rtrancl have "mstep S E\<^sub>\<omega> \<subseteq> (mstep S E\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*" by auto
  with Ri_subset_ERw mstep_Un have "mstep S (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>) \<subseteq> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  from conversion_mono [OF this] show ?thesis by auto
qed

lemma ERi_subset_ERw: "mstep S (E\<^sub>\<infinity> \<union> R\<^sub>\<infinity>) \<subseteq> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
  using Ei_subset_EwRi Ri_subset_ERw RiEw_subset_ERw unfolding mstep_Un [of _ "E\<^sub>\<infinity>"] by blast

lemma oKBi_conversion_ERw: "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof
  have "rstep (E 0) \<subseteq> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
  proof
    fix s t
    assume "(s, t) \<in> rstep (E 0)"
    with ERi_subset_ERw [of "{#s, t#}"] mstep_def have st: "(s, t) \<in> (mstep {#s, t#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by force
    have "mstep {#s, t#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>) \<subseteq> rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>)" using mstep_def by force
    from conversion_mono [OF this] st show "(s, t) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  qed
  from conversion_mono [OF this] show "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by simp
next
  have "rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>) \<subseteq> (rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>*"
  proof
    fix s t
    assume "(s, t) \<in> rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>)"
    then obtain k where "(s, t) \<in> rstep (R k \<union> E k)" unfolding E\<^sub>\<omega>_def R\<^sub>\<omega>_def by blast
    with oKBi_conversion_ERi show "(s, t) \<in> (rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>*" by blast
  qed
  from conversion_mono [OF this] show "(rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>*" by simp
qed
  
end


locale gtotal_okb_irun = okb_irun + gtotal_reduction_order less
begin

lemma gmstep_ERw_Sw: "GROUND ((mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>) \<subseteq> GROUND (((mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>)\<^sup>=)"
proof
  fix s t
  assume "(s, t) \<in> GROUND ((mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>)"
  then have ground: "ground s" "ground t" and mstep: "(s, t) \<in> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>" unfolding GROUND_def by auto
  then have dom: "\<exists>s' t'. s' \<in># S \<and> t' \<in># S \<and> s' \<succeq> s \<and> t' \<succeq> t" unfolding mstep_def by auto
  from mstep reverseTrs have mstep: "(s, t) \<in> mstep S (R\<^sub>\<omega>\<^sup>\<leftrightarrow> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
    unfolding mstep_def Un_iff converse_iff mem_Collect_eq split rstep_union by blast
  from mstep mstep_Un consider "(s, t) \<in> mstep S (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" | "(s, t) \<in> mstep S (R\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
  then show "(s, t) \<in> GROUND (((mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>)\<^sup>=)"
  proof (cases)
    case 1
    from ground gtotal have "s \<succ> t \<or> t \<succ> s \<or> s = t" by auto
    from 1 rstepE obtain C \<sigma> l r where C: "(l, r) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      unfolding mstep_def by blast
    from ground [unfolded C ground_ctxt_apply] have "ground (l \<cdot> \<sigma>)" and "ground (r \<cdot> \<sigma>)" by auto
    from gtotal [OF this] have oriented: "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<or> r \<cdot> \<sigma> \<succ> l \<cdot> \<sigma> \<or> l \<cdot> \<sigma> = r \<cdot> \<sigma>" by auto
    from C(1) have rl: "(r, l) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" by auto
    have eq: "l \<cdot> \<sigma> = r \<cdot> \<sigma> \<Longrightarrow> s = t" unfolding C by auto
    have lr: "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<Longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S\<^sub>\<omega>" unfolding Sw_def
      by(rule UnI2, unfold mem_Collect_eq split, insert 1 C(1), auto)
    have rl: "r \<cdot> \<sigma> \<succ> l \<cdot> \<sigma> \<Longrightarrow> (r \<cdot> \<sigma>, l \<cdot> \<sigma>) \<in> S\<^sub>\<omega>" unfolding Sw_def
      by(rule UnI2, unfold mem_Collect_eq split, insert 1 C(1), auto)
    { assume "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma> \<or> r \<cdot> \<sigma> \<succ> l \<cdot> \<sigma>"
      with lr rl have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S\<^sub>\<omega>\<^sup>\<leftrightarrow>" by auto
      with rstepI have "(s, t) \<in> (rstep S\<^sub>\<omega>)\<^sup>\<leftrightarrow>" unfolding C by auto
      with dom have "(s, t) \<in> (mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>" unfolding mstep_def mem_Collect_eq split by auto
      with ground have ?thesis unfolding GROUND_def by auto
    }
    with oriented eq ground show ?thesis using GROUND_def by blast
  next
    case 2
    with mstep [unfolded mstep_def] have "(s, t) \<in> (mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>"
      unfolding Sw_def mstep_Un mstep_def Un_iff converse_iff mem_Collect_eq split by auto
    with ground show ?thesis unfolding GROUND_def by auto
  qed
qed
  
lemma ground_Einf_Sw:
  assumes "(s, t) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>" and gs: "ground s" and gt: "ground t"
  shows "(s, t) \<in> (mstep {#s, t#} S\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  { fix s t
    assume a: "(s, t) \<in> rstep E\<^sub>\<infinity>" and gs: "ground s" and gt: "ground t"
    define S where "S = {#s, t#}"
    with a have "(s, t) \<in> rstep (E\<^sub>\<infinity> \<union> R\<^sub>\<infinity>)" unfolding rstep_union by auto
    then have "(s, t) \<in> mstep {#s, t#} (E\<^sub>\<infinity> \<union> R\<^sub>\<infinity>)" unfolding mstep_def by auto
    with ERi_subset_ERw have "(s, t) \<in> (mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding S_def by auto
    from this [unfolded conversion_def] have "(s, t) \<in> (mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*"
    proof (induct)
      case base
      then show ?case by auto
    next
      case (step u v)
      note rsteps = step(2) [unfolded mstep_def Un_iff converse_iff mem_Collect_eq split]
      then have st_uv: "(s \<succeq> u \<or> t \<succeq> u) \<and> (s \<succeq> v \<or> t \<succeq> v)" unfolding S_def by force
      from ground_less have ground_leq: "\<And> s t. ground s \<Longrightarrow> s \<succeq> t \<Longrightarrow> ground t" by blast
      from st_uv ground_leq [OF gs] ground_leq [OF gt] have "ground u \<and> ground v" by metis
      with step(2) have "(u, v) \<in> GROUND ((mstep S (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>)" unfolding GROUND_def by auto
      with gmstep_ERw_Sw have uv: "(u, v) \<in> ((mstep S S\<^sub>\<omega>)\<^sup>\<leftrightarrow>)\<^sup>=" unfolding GROUND_def by auto
      with step(3) rtrancl_trans [OF step(3) [unfolded conversion_def] r_into_rtrancl] show ?case
        unfolding conversion_def by fast
    qed
    then have "(s, t) \<in> (mstep {#s, t#} S\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*" unfolding S_def by auto
  } note estep = this
  from assms(1) estep [OF _ gs gt] estep [OF _ gt gs] show ?thesis
    unfolding Un_iff add_mset_commute [of t s "{#}"] converse_iff conversion_inv by auto
qed

end

locale ordered_completion_inf =
  ordered_completion less for less :: "('a, 'b::infinite) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
begin

lemma less_set_permute:
  "\<pi> \<bullet> {\<succ>} = {\<succ>}"
  apply (auto simp: )
   apply (metis inv_rule_mem_trs_simps(1) local.subst mem_Collect_eq old.prod.case term_apply_subst_Var_Rep_perm term_pt.permute_minus_cancel(1))
  by (metis case_prodI inv_rule_mem_trs_simps(1) mem_Collect_eq reduction_order.subst reduction_order_axioms term_apply_subst_Var_Rep_perm)

lemma ostep_permute:
  "(\<pi> \<bullet> s, \<pi> \<bullet> t) \<in> ostep (\<pi> \<bullet> E) (\<pi> \<bullet> R) \<longleftrightarrow> (s, t) \<in> ostep E R"
  using ordstep_permute [of \<pi> s t "{\<succ>}" "E\<^sup>\<leftrightarrow>"]
  unfolding less_set_permute
  by (simp add: eqvt [symmetric] ostep_def)

lemma oKB_permute:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (E', R')"
  shows "(\<pi> \<bullet> E, \<pi> \<bullet> R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B (\<pi> \<bullet> E', \<pi> \<bullet> R')"
  using assms
proof (cases)
  case (deduce s t u)
  with oKB.deduce [of "\<pi> \<bullet> s" "\<pi> \<bullet> t" "\<pi> \<bullet> R" "\<pi> \<bullet> E" "\<pi> \<bullet> u"] show ?thesis
    by (simp add: eqvt) (metis (no_types, lifting) rstep_permute rstep_simps(5) rstep_union)
next
  case (orientl s t)
  with oKB.orientl [of "\<pi> \<bullet> s" "\<pi> \<bullet> t" "\<pi> \<bullet> E" "\<pi> \<bullet> R"] show ?thesis
    by (simp add: eqvt) (metis local.subst term_apply_subst_Var_Rep_perm)
next
  case (orientr t s)
  with oKB.orientr [of "\<pi> \<bullet> t" "\<pi> \<bullet> s" "\<pi> \<bullet> E" "\<pi> \<bullet> R"] show ?thesis
    by (simp add: eqvt) (metis local.subst term_apply_subst_Var_Rep_perm)
next
  case (delete s)
  with oKB.delete [of "\<pi> \<bullet> s" "\<pi> \<bullet> E" "\<pi> \<bullet> R"] show ?thesis by (simp add: eqvt)
next
  case (compose t u s)
  with oKB.compose [of "\<pi> \<bullet> t" "\<pi> \<bullet> u" "\<pi> \<bullet> E" "\<pi> \<bullet> R" "\<pi> \<bullet> s"] show ?thesis
    using ostep_permute [of \<pi> t u E "R - {(s, t)}"] by (simp add: eqvt)
next
  case (simplifyl s u t)
  with oKB.simplifyl [of "\<pi> \<bullet> s" "\<pi> \<bullet> u" "\<pi> \<bullet> E" "\<pi> \<bullet> t" "\<pi> \<bullet> R"] show ?thesis
    using ostep_permute [of \<pi> s u "E - {(s, t)}" R] by (simp add: eqvt)
next
  case (simplifyr t u s)
  with oKB.simplifyr [of "\<pi> \<bullet> t" "\<pi> \<bullet> u" "\<pi> \<bullet> E" "\<pi> \<bullet> s" "\<pi> \<bullet> R"] show ?thesis
    using ostep_permute [of \<pi> t u "E - {(s, t)}" R] by (simp add: eqvt)
next
  case (collapse t u s)
  with oKB.collapse [of "\<pi> \<bullet> t" "\<pi> \<bullet> u" "\<pi> \<bullet> E" "\<pi> \<bullet> R" "\<pi> \<bullet> s"] show ?thesis
    using ostep_permute [of \<pi> t u "E" "R - {(t, s)}"] by (simp add: eqvt)
qed

inductive
  oKB' :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> bool"
  where
    deduce: "(s, t) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> (s, u) \<in> rstep (R \<union> E\<^sup>\<leftrightarrow>) \<Longrightarrow> oKB' (E, R) (E \<union> {(t, u)}, R)" |
    orientl: "s \<succ> t \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> oKB' (E, R) (E - {(s, t)}, R \<union> {(p \<bullet> s, p \<bullet> t)})" |
    orientr: "t \<succ> s \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> oKB' (E, R) (E - {(s, t)}, R \<union> {(p \<bullet> t, p \<bullet> s)})" |
    delete: "(s, s) \<in> E \<Longrightarrow> oKB' (E, R) (E - {(s, s)}, R)" |
    compose: "(t, u) \<in> ostep E (R - {(s, t)}) \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> oKB' (E, R) (E, (R - {(s, t)}) \<union> {(p \<bullet> s, p \<bullet> u)})" |
    simplifyl: "(s, u) \<in> ostep (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> oKB' (E, R) ((E - {(s, t)}) \<union> {(p \<bullet> u, p \<bullet> t)}, R)" |
    simplifyr: "(t, u) \<in> ostep (E - {(s, t)}) R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> oKB' (E, R) ((E - {(s, t)}) \<union> {(p \<bullet> s, p \<bullet> u)}, R)" |
    collapse: "(t, u) \<in> ostep E (R - {(t, s)}) \<Longrightarrow> (t, s) \<in> R \<Longrightarrow> oKB' (E, R) (E \<union> {(p \<bullet> u, p \<bullet> s)}, R - {(t, s)})"

end

definition "E_ord ord E = {(s \<cdot> \<sigma>, t \<cdot> \<sigma>) | s t \<sigma>. (s, t) \<in> E\<^sup>\<leftrightarrow> \<and> ord (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)}"

locale gtotal_okb_irun_inf =
  ordered_completion_inf less +
  gtotal_okb_irun less for less :: "('a, 'b::infinite) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
begin

sublocale gtotal_reduction_order_inf ..

definition "PCP_ext RR = {(u, v) | r r' p \<mu> u v.
    ooverlap {\<succ>} RR r r' p \<mu> u v \<and>
    (\<forall>u \<lhd> fst r \<cdot> \<mu>. u \<in> NF (ordstep {\<succ>} RR))}"

context
  fixes \<R> \<E> :: "('a, 'b) trs"
  assumes rules: "\<R> \<subseteq> {\<succ>}"
begin

lemma perm_R_less: "\<pi> \<bullet> (l, r) \<in> \<R> \<Longrightarrow> l \<succ> r"
  using rules by (auto simp: eqvt perm_less)

abbreviation "\<S> \<equiv> \<R> \<union> E_ord (op \<succ>) \<E>"

lemma rstep_S_ordstep:
  assumes "(s, t) \<in> rstep \<S>"
  shows "(s, t) \<in> ordstep {\<succ>} \<S>"
  using assms
proof (cases rule: rstepE)
  case (rstep C \<sigma> l r)
  then show ?thesis
    using rules
    by (intro ordstep.intros [where C = C and l = l and r = r and \<sigma> = \<sigma>])
      (auto dest: subst simp: E_ord_def)
qed

lemma ordstep_subset_S:
  "ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) \<subseteq> ordstep {\<succ>} \<S>"
proof -
  { fix s t assume "(s, t) \<in> ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
    then obtain C l r \<sigma> where "(l, r) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>"
      and [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" by (cases) auto
    then consider "(l, r) \<in> \<R>" | "(l, r) \<in> \<E>\<^sup>\<leftrightarrow>" by blast
    then have "(s, t) \<in> ordstep {\<succ>} \<S>"
    proof (cases)
      case 1
      then show ?thesis by (auto intro: ordstep.intros)
    next
      case 2
      then show ?thesis
        by (intro ordstep.intros [where C = C and \<sigma> = Var and l = "l \<cdot> \<sigma>" and r = "r \<cdot> \<sigma>"])
          (force simp: E_ord_def)+
    qed }
  then show ?thesis by auto
qed

lemma NF_ordstep_S_subset: "NF (ordstep {\<succ>} \<S>) \<subseteq> NF (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))"
  using ordstep_subset_S by blast

lemma ordstep_S_conv [simp]:
  "ordstep {\<succ>} \<S> = rstep \<S>"
  using rstep_S_ordstep by (auto elim: ordstep.cases)

lemma perm_S_ordstep:
  assumes "\<pi> \<bullet> (s, t) \<in> \<S>"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ordstep {\<succ>} \<S>"
  using assms
  by (intro ordstep.intros [where C = \<box> and \<sigma> = "sop (-\<pi>) \<circ>\<^sub>s \<sigma>" and l = "\<pi> \<bullet> s" and r = "\<pi> \<bullet> t"])
    (auto simp: eqvt perm_R_less E_ord_def subst, (metis perm_less subst)+)

lemma perm_R_ordstep:
  assumes "\<pi> \<bullet> (s, t) \<in> \<R>"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
  using assms
  by (intro ordstep.intros [where C = \<box> and \<sigma> = "sop (-\<pi>) \<circ>\<^sub>s \<sigma>" and l = "\<pi> \<bullet> s" and r = "\<pi> \<bullet> t"])
    (auto simp: eqvt perm_R_less subst)

lemma E_ordstep:
  assumes "(s, t) \<in> \<E>\<^sup>\<leftrightarrow>" and "s \<cdot> \<sigma> \<succ> t \<cdot> \<sigma>"
  shows "(s \<cdot> \<sigma> \<circ>\<^sub>s \<tau>, t \<cdot> \<sigma> \<circ>\<^sub>s \<tau>) \<in> ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
  using assms
  by (intro ordstep.intros [where C = \<box> and \<sigma> = "\<sigma> \<circ>\<^sub>s \<tau>" and l = "s" and r = "t"])
    (auto simp: subst)

lemma PCP_xPCP:
  assumes "ground s" and "ground t"
    and "(s, t) \<in> (rstep (PCP \<S>))\<^sup>\<leftrightarrow>"
  shows "(s, t) \<in> (rstep (PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)))\<^sup>\<leftrightarrow> \<union> (rstep \<S>)\<^sup>\<down>"
    (is "_ \<in> ?R")
proof -
  have S: "\<And>s t. (s, t) \<in> \<S> \<Longrightarrow> s \<succ> t \<or> (t, s) \<in> \<S>"
    using rules by (auto simp: E_ord_def)
  have RE: "\<And>s t. (s, t) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow> \<Longrightarrow> s \<succ> t \<or> (t, s) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>"
    using rules by (auto)
  let ?S = "ordstep {\<succ>} \<S>"
  let ?Eo = "E_ord (op \<succ>) \<E>"
  let ?RE = "ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
  { fix s t and \<sigma> :: "('a, 'b) subst"
    assume "(s, t) \<in> PCP \<S>" and ground: "ground (s \<cdot> \<sigma>)" "ground (t \<cdot> \<sigma>)"
    then obtain \<mu> l\<^sub>1 r\<^sub>1 p l\<^sub>2 r\<^sub>2 where o: "overlap \<S> \<S> (l\<^sub>1, r\<^sub>1) p (l\<^sub>2, r\<^sub>2)"
      and "the_mgu l\<^sub>1 (l\<^sub>2 |_ p) = \<mu>"
      and s: "s = replace_at l\<^sub>2 p r\<^sub>1 \<cdot> \<mu>"
      and t: "t = r\<^sub>2 \<cdot> \<mu>"
      and NF: "\<forall>u \<lhd> l\<^sub>1 \<cdot> \<mu>. u \<in> NF (rstep \<S>)"
      by (auto simp: PCP_def the_mgu_def split: option.splits)
    then have mgu: "mgu l\<^sub>1 (l\<^sub>2 |_ p) = Some \<mu>"
      by (auto simp: the_mgu_def unifiers_def overlap_def simp del: mgu.simps split: option.splits dest!: mgu_complete)
    from mgu_sound [OF this]
    have "l\<^sub>1 \<cdot> \<mu> = (l\<^sub>2 |_ p) \<cdot> \<mu>" by (auto simp: is_imgu_def)
    moreover
    define C and \<tau> where "C = ctxt_of_pos_term p (l\<^sub>2 \<cdot> \<mu>) \<cdot>\<^sub>c \<sigma>" and "\<tau> = \<mu> \<circ>\<^sub>s \<sigma>"
    ultimately
    have eq: "\<box>\<langle>l\<^sub>2 \<cdot> \<tau>\<rangle> = C\<langle>l\<^sub>1 \<cdot> \<tau>\<rangle>"
      using o
      by simp (metis ctxt_of_pos_term_subst ctxt_supt_id fst_conv overlap_source_eq(1) subst_apply_term_ctxt_apply_distrib)

    have s': "s \<cdot> \<sigma> = C\<langle>r\<^sub>1 \<cdot> \<tau>\<rangle>"
      using o by (simp add: s C_def \<tau>_def overlap_def ctxt_of_pos_term_subst funposs_imp_poss)

    have t': "t \<cdot> \<sigma> = \<box>\<langle>r\<^sub>2 \<cdot> \<tau>\<rangle>" by (simp add: t \<tau>_def)

    have p: "p \<in> funposs l\<^sub>2" using o by (auto simp: overlap_def)
    then have [simp]: "hole_pos C = p" by (auto simp: C_def funposs_imp_poss)

    have disj: "vars_rule (l\<^sub>2, r\<^sub>2) \<inter> vars_rule (l\<^sub>1, r\<^sub>1) = {}" using o by (auto simp: overlap_def)

    { fix \<pi>\<^sub>1 and \<pi>\<^sub>2 assume *: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> \<R>" "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> \<R>"
      then have rule1: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> \<S>" and rule2: "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> \<S>" by auto
      have "(s \<cdot> \<sigma> , t \<cdot> \<sigma>) \<in> ?R"
      proof (cases "hole_pos C \<in> funposs l\<^sub>2")
        case False
        from non_ooverlap_GROUND_joinable [OF S eq refl False _ _ _ s' t' ground rule2]
          and perm_S_ordstep [OF rule1, of \<tau>] and perm_S_ordstep [OF rule2, of \<tau>]
        show ?thesis by (auto) (metis GROUND_subset Un_absorb1 Un_iff join_mono)
      next
        case True
        then have "ooverlap {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) (l\<^sub>1, r\<^sub>1) (l\<^sub>2, r\<^sub>2) p \<mu> s t"
          using * and disj and p [THEN funposs_imp_poss] and irrefl
          by (auto simp: ooverlap_def mgu s t ctxt_of_pos_term_subst simp del: mgu.simps)
            (auto dest!: perm_R_less dest: subst trans)
        moreover have "\<forall>u \<lhd> l\<^sub>1 \<cdot> \<mu>. u \<in> NF (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))"
          using NF and NF_ordstep_S_subset by auto
        ultimately have "(s, t) \<in> PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)"
          by  (force simp: PCP_ext_def)
        then show ?thesis by blast
      qed }
    moreover
    { fix \<pi>\<^sub>1 and \<pi>\<^sub>2 assume rule1: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> \<R>" and "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> ?Eo"
      then obtain u\<^sub>2 v\<^sub>2 \<tau>\<^sub>2' where rule2: "(u\<^sub>2, v\<^sub>2) \<in> \<E>\<^sup>\<leftrightarrow>"
        and "\<pi>\<^sub>2 \<bullet> l\<^sub>2 = u\<^sub>2 \<cdot> \<tau>\<^sub>2'" and "\<pi>\<^sub>2 \<bullet> r\<^sub>2 = v\<^sub>2 \<cdot> \<tau>\<^sub>2'"
        and "\<pi>\<^sub>2 \<bullet> l\<^sub>2 \<succ> \<pi>\<^sub>2 \<bullet> r\<^sub>2" and "u\<^sub>2 \<cdot> \<tau>\<^sub>2' \<succ> v\<^sub>2 \<cdot> \<tau>\<^sub>2'"
        by (auto simp: E_ord_def eqvt)
      moreover define \<tau>\<^sub>2 where "\<tau>\<^sub>2 = \<tau>\<^sub>2' \<circ>\<^sub>s sop (-\<pi>\<^sub>2)"
      ultimately have [simp]: "l\<^sub>2 = u\<^sub>2 \<cdot> \<tau>\<^sub>2" "r\<^sub>2 = v\<^sub>2 \<cdot> \<tau>\<^sub>2"
        and gr2: "u\<^sub>2 \<cdot> \<tau>\<^sub>2 \<succ> v\<^sub>2 \<cdot> \<tau>\<^sub>2" by (auto simp: term_pt.permute_flip perm_less)
      from eq have eq': "\<box>\<langle>u\<^sub>2 \<cdot> \<tau>\<^sub>2 \<circ>\<^sub>s \<tau>\<rangle> = C\<langle>l\<^sub>1 \<cdot> \<tau>\<rangle>" by simp

      have rule2': "0 \<bullet> (u\<^sub>2, v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" using \<open>(u\<^sub>2, v\<^sub>2) \<in> \<E>\<^sup>\<leftrightarrow>\<close> by (auto)
      have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?R"
      proof (cases "hole_pos C \<in> funposs u\<^sub>2")
        case False

        from t' have t'': "t \<cdot> \<sigma> = \<box>\<langle>v\<^sub>2 \<cdot> \<tau>\<^sub>2 \<circ>\<^sub>s \<tau>\<rangle>" by simp
        from non_ooverlap_GROUND_joinable [OF RE eq' refl False _ _ _ s' t'' ground rule2']
          and perm_R_ordstep [OF rule1, of \<tau>] and E_ordstep [OF rule2 gr2, of \<tau>]
        have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (GROUND (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)))\<^sup>\<down>" by auto
        then show ?thesis
          using join_mono [OF ordstep_subset_S] and GROUND_subset
          by (metis (no_types, lifting) UnI2 join_mono ordstep_S_conv set_mp)
      next
        case True
        obtain \<pi> where disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_rule (\<pi> \<bullet> u\<^sub>2, \<pi> \<bullet> v\<^sub>2) = {}"
          using vars_rule_disjoint unfolding eqvt [symmetric] by blast
        have rule2'': "-\<pi> \<bullet> (\<pi> \<bullet> u\<^sub>2, \<pi> \<bullet> v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>"
          using rule2' by (simp)
        then have rule2'': "\<exists>p. p \<bullet> (\<pi> \<bullet> u\<^sub>2, \<pi> \<bullet> v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" ..
        define \<nu> where "\<nu> x = (if x \<in> vars_rule (l\<^sub>1, r\<^sub>1) then Var x else (sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2) x)" for x
        have 1: "\<pi> \<bullet> u\<^sub>2 \<cdot> \<nu> = u\<^sub>2 \<cdot> \<tau>\<^sub>2"
        proof -
          have "\<pi> \<bullet> u\<^sub>2 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>2 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2"
            using disj by (unfold term_subst_eq_conv) (auto simp add: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed
        have 2: "\<pi> \<bullet> v\<^sub>2 \<cdot> \<nu> = v\<^sub>2 \<cdot> \<tau>\<^sub>2"
        proof -
          have "\<pi> \<bullet> v\<^sub>2 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>2 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2"
            using disj by (unfold term_subst_eq_conv) (auto simp add: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed
        have [simp]: "r\<^sub>1 \<cdot> \<nu> = r\<^sub>1"
        proof -
          have "r\<^sub>1 \<cdot> \<nu> = r\<^sub>1 \<cdot> Var" by (unfold term_subst_eq_conv) (simp add: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed
        have r\<^sub>2: "r\<^sub>2 = \<pi> \<bullet> v\<^sub>2 \<cdot> \<nu>"
        proof -
          have "\<pi> \<bullet> v\<^sub>2 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>2 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis using p and True
            by (auto simp: eqvt dest!: funposs_imp_poss)
        qed
        have l\<^sub>2: "l\<^sub>2 = \<pi> \<bullet> u\<^sub>2 \<cdot> \<nu>"
        proof -
          have "\<pi> \<bullet> u\<^sub>2 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>2 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>2"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis using p and True
            by (auto simp: eqvt dest!: funposs_imp_poss)
        qed
        then have "l\<^sub>2 |_ p = (\<pi> \<bullet> u\<^sub>2 |_ p) \<cdot> \<nu>"
          using True by (auto dest!: funposs_imp_poss)
        moreover
        have "l\<^sub>1 \<cdot> \<nu> \<circ>\<^sub>s \<mu> = l\<^sub>1 \<cdot> \<mu>"
          by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def)
        ultimately
        have *: "l\<^sub>1 \<cdot> \<nu> \<circ>\<^sub>s \<mu> = (\<pi> \<bullet> u\<^sub>2 |_ p) \<cdot> \<nu> \<circ>\<^sub>s \<mu>" using \<open>l\<^sub>1 \<cdot> \<mu> = (l\<^sub>2 |_ p) \<cdot> \<mu>\<close> by simp
        then obtain \<mu>' where mgu: "mgu l\<^sub>1 (\<pi> \<bullet> u\<^sub>2 |_ p) = Some \<mu>'"
          using mgu_complete by (auto simp: unifiers_def simp del: subst_subst_compose)
        from mgu_sound [OF this, THEN is_imgu_imp_is_mgu] and * obtain \<delta> where \<mu>': " \<nu> \<circ>\<^sub>s \<mu> = \<mu>' \<circ>\<^sub>s \<delta>"
          by (auto simp: is_mgu_def unifiers_def simp del: subst_subst_compose)
        have l\<^sub>1: "l\<^sub>1 \<cdot> \<mu> = l\<^sub>1 \<cdot> \<mu>' \<circ>\<^sub>s \<delta>"
          by (unfold term_subst_eq_conv, fold \<mu>') (simp add: subst_compose \<nu>_def vars_defs)
        let ?s = "(ctxt_of_pos_term p (\<pi> \<bullet> u\<^sub>2 \<cdot> \<mu>'))\<langle>r\<^sub>1 \<cdot> \<mu>'\<rangle>"
        let ?t = "\<pi> \<bullet> v\<^sub>2 \<cdot> \<mu>'"
        have 3: "t \<cdot> \<mu>' \<cdot> \<delta> = t \<cdot> \<nu> \<cdot> \<mu>" for t by (fold subst_subst_compose) (simp add: \<mu>')
        have "p \<in> poss (\<pi> \<bullet> u\<^sub>2)" using True by (simp add: eqvt funposs_imp_poss)
        then have 4: "s \<cdot> \<sigma> = ?s \<cdot> \<delta> \<cdot> \<sigma>"
          by (simp add: s ctxt_of_pos_term_subst [symmetric] 1 3)
        have 5: "t \<cdot> \<sigma> = ?t \<cdot> \<delta> \<cdot> \<sigma>" by (simp add: t 2 3)

        have "\<not> \<pi> \<bullet> v\<^sub>2 \<cdot> \<mu>' \<succ> \<pi> \<bullet> u\<^sub>2 \<cdot> \<mu>'" (is "\<not> ?v \<cdot> _ \<succ> ?u \<cdot> _")
        proof
          presume "\<not> ?thesis"
          then have "?v \<cdot> \<nu> \<circ>\<^sub>s \<mu> \<succ> ?u \<cdot> \<nu> \<circ>\<^sub>s \<mu>"
            unfolding \<mu>' by (simp add: subst)
          then have "r\<^sub>2 \<cdot> \<mu> \<succ> l\<^sub>2 \<cdot> \<mu>" unfolding l\<^sub>2 r\<^sub>2 by simp
          moreover have "l\<^sub>2 \<succ> r\<^sub>2" using \<open>\<pi>\<^sub>2 \<bullet> l\<^sub>2 \<succ> \<pi>\<^sub>2 \<bullet> r\<^sub>2\<close> by (simp add: perm_less)
          ultimately show False using irrefl by (auto dest: subst trans)
        qed simp
        then have "ooverlap {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) (l\<^sub>1, r\<^sub>1) (\<pi> \<bullet> u\<^sub>2, \<pi> \<bullet> v\<^sub>2) p \<mu>' ?s ?t"
          using mgu and disj and rule1 and rule2'' and True and irrefl
          by (auto simp: ooverlap_def perm_less simp del: mgu.simps)
            (auto dest!: perm_R_less dest: subst trans simp del: mgu.simps)
        moreover have "\<forall>u \<lhd> l\<^sub>1 \<cdot> \<mu>'. u \<in> NF (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))"
          using NF and NF_ordstep_S_subset
          unfolding l\<^sub>1
          by auto (meson NF_instance contra_subsetD instance_no_supt_imp_no_supt)
        ultimately have "(?s, ?t) \<in> PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)" by (force simp: PCP_ext_def)
        then show ?thesis
          unfolding 4 and 5 by blast
      qed }
    moreover
    { fix \<pi>\<^sub>1 and \<pi>\<^sub>2 assume *: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> ?Eo" and rule2: "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> \<R>"
      then have rule1: "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> \<S>" and rule2': "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> \<S>" by auto
      have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?R"
      proof (cases "hole_pos C \<in> funposs l\<^sub>2")
        case False
        from non_ooverlap_GROUND_joinable [OF S eq refl False _ _ _ s' t' ground rule2']
          and perm_S_ordstep [OF rule1, of \<tau>] and perm_S_ordstep [OF rule2', of \<tau>]
        show ?thesis by (auto) (metis GROUND_subset Un_absorb1 Un_iff join_mono)
      next
        from * obtain u\<^sub>1 v\<^sub>1 \<tau>\<^sub>1' where rule1: "(u\<^sub>1, v\<^sub>1) \<in> \<E>\<^sup>\<leftrightarrow>"
          and "\<pi>\<^sub>1 \<bullet> l\<^sub>1 = u\<^sub>1 \<cdot> \<tau>\<^sub>1'" and "\<pi>\<^sub>1 \<bullet> r\<^sub>1 = v\<^sub>1 \<cdot> \<tau>\<^sub>1'"
          and "\<pi>\<^sub>1 \<bullet> l\<^sub>1 \<succ> \<pi>\<^sub>1 \<bullet> r\<^sub>1" and "u\<^sub>1 \<cdot> \<tau>\<^sub>1' \<succ> v\<^sub>1 \<cdot> \<tau>\<^sub>1'"
          by (auto simp: E_ord_def eqvt)
        moreover define \<tau>\<^sub>1 where "\<tau>\<^sub>1 = \<tau>\<^sub>1' \<circ>\<^sub>s sop (-\<pi>\<^sub>1)"
        ultimately have [simp]: "l\<^sub>1 = u\<^sub>1 \<cdot> \<tau>\<^sub>1" "r\<^sub>1 = v\<^sub>1 \<cdot> \<tau>\<^sub>1"
          and gr2: "u\<^sub>1 \<cdot> \<tau>\<^sub>1 \<succ> v\<^sub>1 \<cdot> \<tau>\<^sub>1" by (auto simp: term_pt.permute_flip perm_less)
        case True
        obtain \<pi> where disj: "vars_rule (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}"
          using vars_rule_disjoint unfolding eqvt [symmetric] by blast
        have rule1'': "-\<pi> \<bullet> (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>"
          using rule1 by (simp)
        then have rule1'': "\<exists>p. p \<bullet> (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" ..
        define \<nu> where "\<nu> x = (if x \<in> vars_rule (l\<^sub>2, r\<^sub>2) then Var x else (sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1) x)" for x
        have [simp]: "l\<^sub>2 \<cdot> \<nu> = l\<^sub>2 \<cdot> Var" by (unfold term_subst_eq_conv) (simp add: \<nu>_def vars_defs)
        then have [simp]: "(l\<^sub>2 |_ p) \<cdot> \<nu> = l\<^sub>2 |_ p"
          using True by (auto dest!: funposs_imp_poss) (metis \<open>l\<^sub>2 \<cdot> \<nu> = l\<^sub>2 \<cdot> Var\<close> subst.cop_nil subt_at_subst)

        have [simp]: "r\<^sub>2 \<cdot> \<nu> = r\<^sub>2 \<cdot> Var" by (unfold term_subst_eq_conv) (simp add: \<nu>_def vars_defs)
        have [simp]: "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = u\<^sub>1 \<cdot> \<tau>\<^sub>1"
        proof -
          have "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis by simp
        qed
        have [simp]: "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = v\<^sub>1 \<cdot> \<tau>\<^sub>1"
        proof -
          have "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis by simp
        qed
        have l\<^sub>1: "l\<^sub>1 = \<pi> \<bullet> u\<^sub>1 \<cdot> \<nu>"
        proof -
          have "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis using p and True
            by (auto simp: eqvt dest!: funposs_imp_poss)
        qed
        have r\<^sub>1: "r\<^sub>1 = \<pi> \<bullet> v\<^sub>1 \<cdot> \<nu>"
        proof -
          have "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis using p and True
            by (auto simp: eqvt dest!: funposs_imp_poss)
        qed
        have *: "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> \<circ>\<^sub>s \<mu> = (l\<^sub>2 |_ p) \<cdot> \<nu> \<circ>\<^sub>s \<mu>" using \<open>l\<^sub>1 \<cdot> \<mu> = (l\<^sub>2 |_ p) \<cdot> \<mu>\<close> by (simp)
        then obtain \<mu>' where mgu: "mgu (\<pi> \<bullet> u\<^sub>1) (l\<^sub>2 |_ p) = Some \<mu>'"
          using mgu_complete by (auto simp: unifiers_def simp del: subst_subst_compose)
        from mgu_sound [OF this, THEN is_imgu_imp_is_mgu] and * obtain \<delta> where \<mu>': " \<nu> \<circ>\<^sub>s \<mu> = \<mu>' \<circ>\<^sub>s \<delta>"
          by (auto simp: is_mgu_def unifiers_def simp del: subst_subst_compose)

        have 1: "t \<cdot> \<mu>' \<cdot> \<delta> = t \<cdot> \<nu> \<cdot> \<mu>" for t
          by (fold subst_subst_compose) (simp add: \<mu>')
        let ?s = "(ctxt_of_pos_term p (l\<^sub>2 \<cdot> \<mu>'))\<langle>\<pi> \<bullet> v\<^sub>1 \<cdot> \<mu>'\<rangle>"
        let ?t = "r\<^sub>2 \<cdot> \<mu>'"
        have 2: "s \<cdot> \<sigma> = ?s \<cdot> \<delta> \<cdot> \<sigma>"
          using True [THEN funposs_imp_poss]
          by (simp add: s ctxt_of_pos_term_subst [symmetric] 1)
        have 3: "t \<cdot> \<sigma> = ?t \<cdot> \<delta> \<cdot> \<sigma>" by (simp add: t 1)

        have "\<not> \<pi> \<bullet> v\<^sub>1 \<cdot> \<mu>' \<succ> \<pi> \<bullet> u\<^sub>1 \<cdot> \<mu>'"
          (is "\<not> ?v \<cdot> _ \<succ> ?u \<cdot> _")
        proof
          presume "\<not> ?thesis"
          then have "?v \<cdot> \<nu> \<circ>\<^sub>s \<mu> \<succ> ?u \<cdot> \<nu> \<circ>\<^sub>s \<mu>"
            unfolding \<mu>' by (simp add: subst)
          then have "r\<^sub>1 \<cdot> \<mu> \<succ> l\<^sub>1 \<cdot> \<mu>" unfolding l\<^sub>1 r\<^sub>1 by simp
          moreover have "l\<^sub>1 \<succ> r\<^sub>1" using \<open>\<pi>\<^sub>1 \<bullet> l\<^sub>1 \<succ> \<pi>\<^sub>1 \<bullet> r\<^sub>1\<close> by (simp add: perm_less)
          ultimately show False using irrefl by (auto dest: subst trans)
        qed simp
        then have "ooverlap {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) (l\<^sub>2, r\<^sub>2) p \<mu>' ?s ?t"
          using mgu and disj and rule1'' and rule2 and True and irrefl
          by (auto simp: ooverlap_def perm_less) (auto dest!: perm_R_less dest: subst trans)
        moreover have "\<forall>u \<lhd> \<pi> \<bullet> u\<^sub>1 \<cdot> \<mu>'. u \<in> NF (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))"
          using NF and NF_ordstep_S_subset
          unfolding l\<^sub>1 and 1 [symmetric]
          by auto (meson NF_instance contra_subsetD instance_no_supt_imp_no_supt)
        ultimately have "(?s, ?t) \<in> PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)" by (force simp: PCP_ext_def)
        then show ?thesis
          unfolding 2 and 3 by blast
      qed }
    moreover
    { fix \<pi>\<^sub>1 and \<pi>\<^sub>2 assume "\<pi>\<^sub>1 \<bullet> (l\<^sub>1, r\<^sub>1) \<in> ?Eo" and "\<pi>\<^sub>2 \<bullet> (l\<^sub>2, r\<^sub>2) \<in> ?Eo"
      then obtain u\<^sub>1 v\<^sub>1 \<tau>\<^sub>1' and u\<^sub>2 v\<^sub>2 \<tau>\<^sub>2'
        where rule1: "(u\<^sub>1, v\<^sub>1) \<in> \<E>\<^sup>\<leftrightarrow>" and "\<pi>\<^sub>1 \<bullet> l\<^sub>1 = u\<^sub>1 \<cdot> \<tau>\<^sub>1'" and "\<pi>\<^sub>1 \<bullet> r\<^sub>1 = v\<^sub>1 \<cdot> \<tau>\<^sub>1'"
          and "\<pi>\<^sub>1 \<bullet> l\<^sub>1 \<succ> \<pi>\<^sub>1 \<bullet> r\<^sub>1"
          and rule2: "(u\<^sub>2, v\<^sub>2) \<in> \<E>\<^sup>\<leftrightarrow>" and "\<pi>\<^sub>2 \<bullet> l\<^sub>2 = u\<^sub>2 \<cdot> \<tau>\<^sub>2'" and "\<pi>\<^sub>2 \<bullet> r\<^sub>2 = v\<^sub>2 \<cdot> \<tau>\<^sub>2'"
          and "\<pi>\<^sub>2 \<bullet> l\<^sub>2 \<succ> \<pi>\<^sub>2 \<bullet> r\<^sub>2"
        by (force simp: E_ord_def eqvt)
      moreover
      define \<tau>\<^sub>1 and \<tau>\<^sub>2 where "\<tau>\<^sub>1 = \<tau>\<^sub>1' \<circ>\<^sub>s sop (-\<pi>\<^sub>1)" and "\<tau>\<^sub>2 = \<tau>\<^sub>2' \<circ>\<^sub>s sop (-\<pi>\<^sub>2)"
      ultimately
      have [simp]: "u\<^sub>1 \<cdot> \<tau>\<^sub>1 = l\<^sub>1" "v\<^sub>1 \<cdot> \<tau>\<^sub>1 = r\<^sub>1" "u\<^sub>2 \<cdot> \<tau>\<^sub>2 = l\<^sub>2" "v\<^sub>2 \<cdot> \<tau>\<^sub>2 = r\<^sub>2"
        and gr1: "u\<^sub>1 \<cdot> \<tau>\<^sub>1 \<succ> v\<^sub>1 \<cdot> \<tau>\<^sub>1" and gr2: "u\<^sub>2 \<cdot> \<tau>\<^sub>2 \<succ> v\<^sub>2 \<cdot> \<tau>\<^sub>2"
        by (auto simp: term_pt.permute_flip [symmetric] perm_less)

      have rule2': "0 \<bullet> (u\<^sub>2, v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" using rule2 by simp

      have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?R"
      proof (cases "hole_pos C \<in> funposs u\<^sub>2")
        case False

        from eq have eq': "\<box>\<langle>u\<^sub>2 \<cdot> \<tau>\<^sub>2 \<circ>\<^sub>s \<tau>\<rangle> = C\<langle>u\<^sub>1 \<cdot> \<tau>\<^sub>1 \<circ>\<^sub>s \<tau>\<rangle>" by simp
        from s' have s'': "s \<cdot> \<sigma> = C\<langle>v\<^sub>1 \<cdot> \<tau>\<^sub>1 \<circ>\<^sub>s \<tau>\<rangle>" by simp
        from t' have t'': "t \<cdot> \<sigma> = \<box>\<langle>v\<^sub>2 \<cdot> \<tau>\<^sub>2 \<circ>\<^sub>s \<tau>\<rangle>" by simp
        from non_ooverlap_GROUND_joinable [OF RE eq' refl False _ _ _ s'' t'' ground rule2']
          and E_ordstep [OF rule1 gr1, of \<tau>] and E_ordstep [OF rule2 gr2, of \<tau>]
        have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (GROUND (ordstep {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)))\<^sup>\<down>" by auto
        then show ?thesis
          using join_mono [OF ordstep_subset_S] and GROUND_subset
          by (metis (no_types, lifting) UnI2 join_mono ordstep_S_conv set_mp)
      next
        case True
        obtain \<pi> where disj: "vars_rule (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<inter> vars_rule (u\<^sub>2, v\<^sub>2) = {}"
          using vars_rule_disjoint unfolding eqvt [symmetric] by blast
        have "-\<pi> \<bullet> (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" using rule1 by simp
        then have rule1': "\<exists>p. p \<bullet> (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" ..
        have rule2'': "\<exists>p. p \<bullet> (u\<^sub>2, v\<^sub>2) \<in> \<R> \<union> \<E>\<^sup>\<leftrightarrow>" using rule2' ..

        define \<nu> where "\<nu> x = (if x \<in> vars_rule (u\<^sub>2, v\<^sub>2) then \<tau>\<^sub>2 x else (sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1) x)" for x

        have l\<^sub>1: "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = l\<^sub>1"
        proof -
          have "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> u\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis by simp
        qed
        have [simp]: "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = r\<^sub>1"
        proof -
          have "\<pi> \<bullet> v\<^sub>1 \<cdot> \<nu> = \<pi> \<bullet> v\<^sub>1 \<cdot> sop (-\<pi>) \<circ>\<^sub>s \<tau>\<^sub>1"
            using disj
            by (unfold term_subst_eq_conv) (auto simp: subst_compose vars_defs \<nu>_def eqvt)
          then show ?thesis by simp
        qed
        have [simp]: "u\<^sub>2 \<cdot> \<nu> = l\<^sub>2"
        proof -
          have "u\<^sub>2 \<cdot> \<nu> = u\<^sub>2 \<cdot> \<tau>\<^sub>2" by (unfold term_subst_eq_conv) (auto simp: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed
        then have [simp]: "u\<^sub>2 |_ p \<cdot> \<nu> = l\<^sub>2 |_ p"
          using True [THEN funposs_imp_poss] by (auto) (metis \<open>u\<^sub>2 \<cdot> \<nu> = l\<^sub>2\<close> subt_at_subst)

        have [simp]: "v\<^sub>2 \<cdot> \<nu> = r\<^sub>2"
        proof -
          have "v\<^sub>2 \<cdot> \<nu> = v\<^sub>2 \<cdot> \<tau>\<^sub>2" by (unfold term_subst_eq_conv) (auto simp: \<nu>_def vars_defs)
          then show ?thesis by simp
        qed

        have *: "\<pi> \<bullet> u\<^sub>1 \<cdot> \<nu> \<circ>\<^sub>s \<mu> = (u\<^sub>2 |_ p) \<cdot> \<nu> \<circ>\<^sub>s \<mu>" using \<open>l\<^sub>1 \<cdot> \<mu> = (l\<^sub>2 |_ p) \<cdot> \<mu>\<close> by (simp add: l\<^sub>1)
        then obtain \<mu>' where mgu: "mgu (\<pi> \<bullet> u\<^sub>1) (u\<^sub>2 |_ p) = Some \<mu>'"
          using mgu_complete by (auto simp: unifiers_def simp del: subst_subst_compose)
        from mgu_sound [OF this, THEN is_imgu_imp_is_mgu] and * obtain \<delta> where \<mu>': " \<nu> \<circ>\<^sub>s \<mu> = \<mu>' \<circ>\<^sub>s \<delta>"
          by (auto simp: is_mgu_def unifiers_def simp del: subst_subst_compose)

        have 1: "t \<cdot> \<mu>' \<cdot> \<delta> = t \<cdot> \<nu> \<cdot> \<mu>" for t
          by (fold subst_subst_compose) (simp add: \<mu>')
        let ?s = "(ctxt_of_pos_term p (u\<^sub>2 \<cdot> \<mu>'))\<langle>\<pi> \<bullet> v\<^sub>1 \<cdot> \<mu>'\<rangle>"
        let ?t = "v\<^sub>2 \<cdot> \<mu>'"
        have 2: "s \<cdot> \<sigma> = ?s \<cdot> \<delta> \<cdot> \<sigma>"
          using True [THEN funposs_imp_poss] and p [THEN funposs_imp_poss]
          by (simp add: s ctxt_of_pos_term_subst [symmetric] 1)
        have 3: "t \<cdot> \<sigma> = ?t \<cdot> \<delta> \<cdot> \<sigma>" by (simp add: t 1)

        have "\<not> \<pi> \<bullet> v\<^sub>1 \<cdot> \<mu>' \<succ> \<pi> \<bullet> u\<^sub>1 \<cdot> \<mu>'" (is "\<not> ?v \<cdot> _ \<succ> ?u \<cdot> _")
        proof
          presume "\<not> ?thesis"
          then have "?v \<cdot> \<nu> \<circ>\<^sub>s \<mu> \<succ> ?u \<cdot> \<nu> \<circ>\<^sub>s \<mu>" unfolding \<mu>' by (simp add: subst)
          then have "r\<^sub>1 \<cdot> \<mu> \<succ> l\<^sub>1 \<cdot> \<mu>" by (simp add: l\<^sub>1)
          moreover have "l\<^sub>1 \<succ> r\<^sub>1" using \<open>\<pi>\<^sub>1 \<bullet> l\<^sub>1 \<succ> \<pi>\<^sub>1 \<bullet> r\<^sub>1\<close> by (simp add: perm_less)
          ultimately show False using irrefl by (auto dest: subst trans)
        qed simp
        moreover have "\<not> v\<^sub>2 \<cdot> \<mu>' \<succ> u\<^sub>2 \<cdot> \<mu>'" (is "\<not> ?v \<cdot> _ \<succ> ?u \<cdot> _")
        proof
          presume "\<not> ?thesis"
          then have "?v \<cdot> \<nu> \<circ>\<^sub>s \<mu> \<succ> ?u \<cdot> \<nu> \<circ>\<^sub>s \<mu>" unfolding \<mu>' by (simp add: subst)
          then have "r\<^sub>2 \<cdot> \<mu> \<succ> l\<^sub>2 \<cdot> \<mu>" by simp
          moreover have "l\<^sub>2 \<succ> r\<^sub>2" using \<open>\<pi>\<^sub>2 \<bullet> l\<^sub>2 \<succ> \<pi>\<^sub>2 \<bullet> r\<^sub>2\<close> by (simp add: perm_less)
          ultimately show False using irrefl by (auto dest: subst trans)
        qed simp
        ultimately
        have "ooverlap {\<succ>} (\<R> \<union> \<E>\<^sup>\<leftrightarrow>) (\<pi> \<bullet> u\<^sub>1, \<pi> \<bullet> v\<^sub>1) (u\<^sub>2, v\<^sub>2) p \<mu>' ?s ?t"
          using mgu and disj and rule1' and rule2'' and True
          by (auto simp: ooverlap_def perm_less)
        moreover
        have "\<forall>u \<lhd> \<pi> \<bullet> u\<^sub>1 \<cdot> \<mu>'. u \<in> NF ?RE"
          using NF and NF_ordstep_S_subset
          unfolding l\<^sub>1 [symmetric] and 1 [symmetric]
          by auto (meson NF_instance contra_subsetD instance_no_supt_imp_no_supt)
        ultimately have "(?s, ?t) \<in> PCP_ext (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)" by (force simp: PCP_ext_def)
        then show ?thesis
          unfolding 2 and 3 by blast
      qed }
    ultimately have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ?R" using o by (auto simp: overlap_def) }
  then show ?thesis
    using assms by (auto elim!: rstepE) blast+
qed

abbreviation "\<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" where "\<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c \<equiv> \<E>\<^sup>\<leftrightarrow> \<inter> {\<succ>}"
definition \<R>':: "('a, 'b) term rel" where "\<R>' \<equiv> {(l, r) | l r. (l, r) \<in> dot (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c) \<and> (\<nexists>u l0 r0. (l, u) \<in> rstep {(l0, r0)} \<and> (l0, r0) \<in> \<S> \<and> l0 \<lhd>\<cdot> l)}"
definition \<E>' :: "('a, 'b) term rel" where "\<E>' \<equiv> {(some_NF (rstep \<R>') s, some_NF (rstep \<R>') t) | s t. (s, t) \<in> \<E> \<and> some_NF (rstep \<R>') s \<noteq> some_NF (rstep \<R>') t }"
definition "\<S>'":: "('a, 'b) term rel" where "\<S>' = \<R>' \<union> E_ord (op \<succ>) \<E>'"


lemma Esucc_E_ord: "\<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c \<subseteq> E_ord (op \<succ>) \<E>"
proof
  fix s t
  assume "(s, t) \<in> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c"
  note ord = this [unfolded Int_iff split]
  show "(s, t) \<in> E_ord (op \<succ>) \<E>" unfolding E_ord_def mem_Collect_eq
    by(rule exI [of _ s], rule exI [of _ t], rule exI [of _ Var], insert ord, auto)
qed

lemma encomp_perm: "u \<unlhd>\<cdot> s = (u \<unlhd>\<cdot> (\<pi> \<bullet> s))"
proof
  from term_apply_subst_Var_Rep_perm obtain \<sigma> where sub: "\<pi> \<bullet> s = s \<cdot> \<sigma>" by metis
  show "u \<unlhd>\<cdot> s \<Longrightarrow> u \<unlhd>\<cdot> \<pi> \<bullet> s" unfolding encompeq.simps sub
    by (metis subst_apply_term_ctxt_apply_distrib subst_subst)
next
  have "s = ( -\<pi>) \<bullet> \<pi> \<bullet> s" by auto
  with term_apply_subst_Var_Rep_perm obtain \<sigma> where sub: "s = (\<pi> \<bullet> s) \<cdot> \<sigma>" by metis
  show "u \<unlhd>\<cdot> \<pi> \<bullet> s \<Longrightarrow> u \<unlhd>\<cdot> s" unfolding encompeq.simps using sub
    by (metis subst_apply_term_ctxt_apply_distrib subst_subst)
qed

abbreviation "lexencomp \<equiv> lex_two {\<cdot>\<rhd>} {\<cdot>\<unrhd>}  {(a, b :: bool) . a \<and> \<not> b}"

lemma SN_lexencomp: "SN lexencomp"
proof-
  have t: "SN {(a, b). a \<and> \<not> b}" using SN_inv_image [OF SN_nat_gt, of "\<lambda>b. if b then 1 else 0"]
    by (metis (no_types, lifting) SN_onI mem_Collect_eq split_conv)
  then show ?thesis using SN_encomp encomp_order.less_le_trans encomp_order.le_less_trans
    by(intro lex_two, auto)
qed
  
lemma gterms_ER_conv_implies_S_conv:
  assumes RR_less: "RR \<subseteq> {\<succ>}"
  assumes ground: "ground s" "ground t"
  assumes conv: "(s, t) \<in> (rstep (RR \<union> EE\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" (is "_ \<in> ?R\<^sup>\<leftrightarrow>\<^sup>*")
  shows "(s, t) \<in> (GROUND (rstep (RR \<union> E_ord (op \<succ>) EE)))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  have step_subset: "GROUND(rstep (RR \<union> EE\<^sup>\<leftrightarrow>)) \<subseteq> (GROUND (rstep (RR \<union> E_ord (op \<succ>) EE)))\<^sup>\<leftrightarrow>\<^sup>*"
  proof
    fix s t
    assume step: "(s, t) \<in> GROUND(rstep (RR \<union> EE\<^sup>\<leftrightarrow>))"
    then obtain C l r \<sigma> where C: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" "(l, r) \<in> RR \<union> EE\<^sup>\<leftrightarrow>" "ground s" "ground t"
      unfolding GROUND_def by auto
    show "(s, t) \<in> (GROUND (rstep (RR \<union> E_ord (op \<succ>) EE)))\<^sup>\<leftrightarrow>\<^sup>*"
    proof (cases "(l, r) \<in> RR")
      case True
      with step have "(s, t) \<in> (rstep (RR \<union> E_ord (op \<succ>) EE))" unfolding C rstep_union by blast
      with C(4) C(5) show ?thesis unfolding GROUND_def by auto
    next
      case False
      with step rstep_union C(3) have in_E: "(l, r) \<in> EE\<^sup>\<leftrightarrow>" unfolding C by auto
      from C(4) C(5) have g': "ground (l \<cdot> \<sigma>)" "ground (r \<cdot> \<sigma>)" unfolding C by auto
      from gtotal [OF g'] consider " l \<cdot> \<sigma> = r \<cdot> \<sigma>" | "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" | "r \<cdot> \<sigma> \<succ> l \<cdot> \<sigma>" by auto
      then show ?thesis
      proof (cases)
        case 1
        then show ?thesis unfolding C by auto
      next
        case 2
        with in_E have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> E_ord (op \<succ>) EE" unfolding E_ord_def mem_Collect_eq by auto
        then have "(s, t) \<in> rstep (RR \<union> E_ord op \<succ> EE)" unfolding C by auto
        with C(4) C(5) show ?thesis unfolding GROUND_def by auto
      next
        case 3
        with in_E have "(r \<cdot> \<sigma>, l \<cdot> \<sigma>) \<in> E_ord (op \<succ>) EE" unfolding E_ord_def mem_Collect_eq by auto
        then have "(t, s) \<in> rstep (RR \<union> E_ord op \<succ> EE)" unfolding C by auto
        with C(4) C(5) show ?thesis unfolding GROUND_def by auto
      qed
    qed
  qed
  from gterm_conv_GROUND_conv assms have "(s, t) \<in> (GROUND (rstep (RR \<union> EE\<^sup>\<leftrightarrow>)))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  with conversion_mono [OF step_subset] show ?thesis unfolding conversion_conversion_idemp by auto
qed

lemma reduced_ground_complete:
  assumes "\<S> \<subseteq> {\<succ>}" and "GCR (rstep \<S>) \<and> SN (GROUND (rstep \<S>))"
  shows "GCR (rstep \<S>') \<and>
         SN (GROUND (rstep \<S>')) \<and>
         (GROUND (rstep \<S>))\<^sup>! = (GROUND (rstep \<S>'))\<^sup>!"
proof-
  let ?S = "\<R>' \<union> (E_ord (op \<succ>) \<E>')"
  have sub_dot: "\<R>' \<subseteq> dot (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" unfolding \<R>'_def by fast
  with rstep_subset_less \<open>\<R> \<subseteq> {\<succ>}\<close> have **: "\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c \<subseteq> {\<succ>}" unfolding rstep_union by auto
  with rstep_subset_less \<open>\<R> \<subseteq> {\<succ>}\<close> have *: "rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c) \<subseteq> {\<succ>}" unfolding rstep_union by auto
  with SN_less SN_subset have "SN (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c))" by auto
  interpret SN_trs "\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" by standard fact

  { fix l r
    assume a: "(l, r) \<in> dot' (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)"
    then have "\<exists> r'. (l, r') \<in> (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c) \<and> r = some_NF (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)) r'"
      unfolding dot'_def mem_Collect_eq by fast
    with some_NF_rtrancl rstep.rule [OF a] have "(l, r) \<in> (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c))\<^sup>+"
      by (meson rstep.rule rtrancl_into_trancl2)
  }
  with rsteps_subset_less [OF **] rm_variants_subset have "dot (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c) \<subseteq> {\<succ>}"
    unfolding dot_def by blast
  with sub_dot rstep_subset_less have R'succ: "\<R>' \<subseteq> {\<succ>}" by auto
  with rstep_subset_less have "rstep \<R>' \<subseteq> {\<succ>}" unfolding E_ord_def by blast
  with SN_less SN_subset have "SN (rstep \<R>')" unfolding \<S>'_def by metis
  from R'succ rstep_subset_less have "rstep ?S \<subseteq> {\<succ>}" unfolding E_ord_def by blast
  with SN_less SN_subset GROUND_subset have x: "SN (GROUND (rstep \<S>'))" unfolding \<S>'_def by metis

  { fix s t
    assume "(s, t) \<in> GROUND (rstep \<S>)"
    then have g: "ground s" "ground t" "(s, t) \<in> rstep \<S>" unfolding GROUND_def by auto
    { fix l r
      assume in_S: "(l, r) \<in> \<S>"
      let ?orig = "\<lambda> rl. \<not> rl \<in> \<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c"
      from in_S have "\<exists> r'. (l, r') \<in> rstep \<S>'"
      proof (induct "(l, ?orig (l, r))" arbitrary: l r rule: SN_induct [OF SN_lexencomp])
        case (1 l)
        then show ?case
        proof (cases "(l, r) \<in> \<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c")
          case True
          note REsucc = this
          show ?thesis
          proof (cases "\<exists>u l0 r0. (l, u) \<in> rstep {(l0, r0)} \<and> (l0, r0) \<in> \<S> \<and> l0 \<lhd>\<cdot> l")
            case True
            then obtain u l0 r0 where step: "(l, u) \<in> rstep {(l0, r0)}" "(l0, r0) \<in> \<S>" and ll0: "l0 \<lhd>\<cdot> l" by blast
            then obtain C \<sigma> where C: "l = C\<langle>l0 \<cdot> \<sigma>\<rangle>" "(l0, r0) \<in> \<S>" using rstep.simps by auto
            from ll0 have "((l, ?orig (l, r)), (l0, ?orig (l0, r0))) \<in> lexencomp" by force
            with 1(1) [OF this] step have "\<exists> r0'. (l0, r0') \<in> rstep \<S>'" by auto
            with C(1) rstep.ctxt show ?thesis by force
          next
            case False
            from REsucc obtain u where in_dot: "(l, u) \<in> dot' (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" unfolding dot'_def by auto
            from rm_variants_representative [OF this]
            obtain \<pi> l' r' where var: "(l', r') \<in> dot(\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" "\<pi> \<bullet> (l, u) = (l', r')"
              unfolding dot_def by blast
            then have "\<pi> \<bullet> l = l' \<and> \<pi> \<bullet> u = r'" by (simp add: rule_pt.permute_prod.simps)
            then have lu: "l = ( - \<pi>) \<bullet> l'" "u = ( - \<pi>) \<bullet> r'" by auto
            { assume "\<exists>u l0 r0. (l', u) \<in> rstep {(l0, r0)} \<and> (l0, r0) \<in> \<S> \<and> l0 \<lhd>\<cdot> l'"
              then obtain v l0 r0 where step': "(l', v) \<in> rstep {(l0, r0)} "
                and in_S: "(l0, r0) \<in> \<S>" and enc: "l0 \<lhd>\<cdot> l'" by blast
              with rstep_r_p_s_imp_rstep have step: "(l', v) \<in> rstep {(l0, r0)}" by auto
              with permute_rstep lu have step: "(l, ( - \<pi>) \<bullet> v) \<in> rstep {(l0, r0)}" by auto
              from encomp_perm enc lu have enc: "l0 \<lhd>\<cdot> l" unfolding encomp_def
                by (metis encomp_def encomp_order.le_less_trans encompeq_refl)
              with False enc step in_S have False by blast
            }
            with var have "(l', r') \<in> \<R>'" unfolding \<R>'_def mem_Collect_eq by blast
            with var(2) perm_rstep_conv rstep.rule have "(l, u) \<in> rstep \<R>'" by metis
            then show ?thesis unfolding rstep_union \<S>'_def by fast
          qed
        next
          case False
          note unorientable = this
          with 1(2) obtain u v \<sigma> where uv: "(u, v) \<in> \<E>\<^sup>\<leftrightarrow>" "l = u \<cdot> \<sigma>" "r = v \<cdot> \<sigma>" "u \<cdot> \<sigma> \<succ> v \<cdot> \<sigma>"
            unfolding E_ord_def by blast
          let ?u = "some_NF (rstep \<R>') u"
          let ?v = "some_NF (rstep \<R>') v"
          interpret SN_R': SN_ars "rstep \<R>'" by standard fact
          from \<open>SN (rstep \<R>')\<close> have uu: "(u, ?u) \<in> (rstep \<R>')\<^sup>*"
            by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
          from \<open>SN (rstep \<R>')\<close> have vv: "(v, ?v) \<in> (rstep \<R>')\<^sup>*"
            by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
          show ?thesis
          proof (cases "?u = u")
            case True
            from vv rsteps_subset_less [OF \<open>\<R>' \<subseteq> {\<succ>}\<close>] have "v \<succeq> ?v"
              unfolding rtrancl_eq_or_trancl by auto
            with subst [of v ?v \<sigma>] trans [OF uv(4)] uv(4) have gt: "u \<cdot> \<sigma> \<succ> ?v \<cdot> \<sigma>" by auto
            with irrefl have "u \<noteq> ?v" by auto
            with uv(1) True have in_E': "(u, ?v) \<in> \<E>'\<^sup>\<leftrightarrow>"
              unfolding \<E>'_def Un_iff mem_Collect_eq converse_iff by metis
            from gt uv in_E' have "(l, ?v \<cdot> \<sigma>) \<in> E_ord op \<succ> \<E>'"
              unfolding E_ord_def mem_Collect_eq by blast
            with uv show ?thesis unfolding rstep_union \<S>'_def by blast
          next
            case False
            with uu have "(u, ?u) \<in> (rstep \<R>')\<^sup>+" by (simp add: rtrancl_eq_or_trancl)
            with tranclD obtain w where "(u, w) \<in> (rstep \<R>')" "(w, ?u) \<in> (rstep \<R>')\<^sup>*"
              by (meson tranclD)
            then obtain l0 r0 C \<tau> where l0r0: "u = C\<langle>l0 \<cdot> \<tau>\<rangle>" "w = C\<langle>r0 \<cdot> \<tau>\<rangle>" "(l0, r0) \<in> \<R>'" by auto
            then have "(l0, r0) \<in> dot (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" unfolding \<R>'_def by auto
            with lhss_dot obtain r0' where oriented: "(l0, r0') \<in> \<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" by fastforce
            from Esucc_E_ord oriented have in_S: "(l0, r0') \<in> \<S>" unfolding Un_iff by auto
            from uv(2) have "l \<cdot>\<unrhd> l0" unfolding encompeq.simps l0r0(1)
                subst_apply_term_ctxt_apply_distrib subst_subst by fast
            with unorientable oriented have "((l, ?orig (l, r)), (l0, ?orig (l0, r0'))) \<in> lexencomp" by simp
            from 1(1) [OF this in_S] show ?thesis unfolding uv l0r0 by fast
          qed
        qed
      qed
    }
    with g have "\<exists>t'. (s, t') \<in> rstep \<S>'" by fast
    then obtain t' where step: "(s, t') \<in> rstep \<S>'" by auto
    with \<open>rstep ?S \<subseteq> {\<succ>}\<close> ground_less [OF g(1)] have "ground t'" unfolding \<S>'_def by auto
    with g step have "\<exists>t'.(s, t') \<in> GROUND (rstep \<S>')" unfolding GROUND_def by auto
  }
  then have NFs: "NF (GROUND (rstep \<S>')) \<subseteq> NF (GROUND (rstep \<S>))" unfolding NF_def by auto

  have step_conv: "GROUND(rstep \<S>') \<subseteq> (GROUND (rstep \<S>))\<^sup>\<leftrightarrow>\<^sup>*"
  proof
    fix s t
    assume "(s, t) \<in> GROUND(rstep \<S>')"
    then have g: "ground s" "ground t" "(s, t) \<in> rstep \<S>'" unfolding GROUND_def by auto
    then obtain l r \<sigma> C where C: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" "(l, r) \<in> \<S>'" by auto
    { fix l r
      assume "(l, r) \<in> \<R>'"
      with rm_variants_subset have "(l, r) \<in> dot' (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" unfolding \<R>'_def mem_Collect_eq dot_def by fast
      then obtain r' where r: "r = some_NF (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)) r'" "(l, r') \<in> \<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" unfolding dot'_def  by auto
      interpret REsucc: SN_ars "rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c)" by standard fact
      from \<open>SN (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c))\<close> have "(r', r) \<in> (rstep (\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c))\<^sup>*" unfolding r
        by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
      with rstep_mono [of "\<R> \<union> \<E>\<^sub>s\<^sub>u\<^sub>c\<^sub>c" "\<R> \<union> \<E>\<^sup>\<leftrightarrow>", THEN rtrancl_mono] have rr: "(r', r) \<in> (rstep  (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>*" by blast
      from r(2) have "(l, r') \<in> rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>)" by fastforce
      with rr have "(l, r) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_def
        by (meson converse_rtrancl_into_rtrancl in_rtrancl_UnI)
    } note R'_RE = this
    {  fix s t
      assume "(s, t) \<in> rstep \<R>'"
      then obtain l r C \<sigma> where "(l, r) \<in> \<R>'" "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
      with R'_RE [OF this(1)] ctxt.closed_conversion [OF ctxt_closed_rstep, THEN ctxt.closedD]
        subst.closed_conversion [OF subst_closed_rstep, THEN subst.closedD] have "(s, t) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by metis
    } note R'_step_RE = this
    then have "rstep \<R>' \<subseteq> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
    from conversion_conversion_idemp conversion_mono [OF this]
    have R'_steps_RE: "\<And> s t.(s, t) \<in> (rstep \<R>')\<^sup>* \<Longrightarrow> (s, t) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
    show "(s, t) \<in> (GROUND (rstep \<S>))\<^sup>\<leftrightarrow>\<^sup>*"
    proof (cases "(l, r) \<in> \<R>'")
      case True
      from R'_RE [OF this] ctxt.closed_conversion [OF ctxt_closed_rstep, THEN ctxt.closedD]
        subst.closed_conversion [OF subst_closed_rstep, THEN subst.closedD]
      have "(s, t) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding C by metis
      with gterms_ER_conv_implies_S_conv [OF \<open>\<R> \<subseteq> {\<succ>}\<close> g(1) g(2)] show ?thesis .
    next
      case False
      with C(3) [unfolded \<S>'_def] have "(l, r) \<in> E_ord op \<succ> \<E>'" by auto
      then obtain u v \<tau> where uv: "l = u\<cdot> \<tau>" "r = v \<cdot> \<tau>" "(u, v) \<in> \<E>'\<^sup>\<leftrightarrow>" "u\<cdot> \<tau> \<succ> v \<cdot> \<tau>"
        unfolding E_ord_def mem_Collect_eq by blast
      from uv(3) obtain u' v' where uv': "(u', v') \<in> \<E>\<^sup>\<leftrightarrow>" "u = some_NF (rstep \<R>') u'" "v = some_NF (rstep \<R>') v'"
        unfolding \<E>'_def Un_iff mem_Collect_eq by auto
      interpret SN_R': SN_ars "rstep \<R>'" by standard fact
      from \<open>SN (rstep \<R>')\<close> have "(u', u) \<in> (rstep \<R>')\<^sup>*" unfolding uv'
        by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
      with rsteps_closed_subst rsteps_closed_ctxt have "(C\<langle>u' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>, C\<langle>u \<cdot> \<tau> \<cdot> \<sigma>\<rangle>) \<in> (rstep \<R>')\<^sup>*" by metis
      with R'_steps_RE conversion_inv have uconv: "(C\<langle>u \<cdot> \<tau> \<cdot> \<sigma>\<rangle>, C\<langle>u' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by metis
      from \<open>SN (rstep \<R>')\<close> uv' have "(v', v) \<in> (rstep \<R>')\<^sup>*"
        by (simp add: SN_ars.some_NF_rtrancl Normalization_Equivalence.SN_ars.intro)
      with rsteps_closed_subst rsteps_closed_ctxt have "(C\<langle>v' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<tau> \<cdot> \<sigma>\<rangle>) \<in> (rstep \<R>')\<^sup>*" by metis
      with R'_steps_RE have vconv: "(C\<langle>v' \<cdot> \<tau> \<cdot> \<sigma>\<rangle>, C\<langle>v \<cdot> \<tau> \<cdot> \<sigma>\<rangle>) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
      from uv' uv rstepI have "(C\<langle>u' \<cdot> \<tau> \<cdot>\<sigma>\<rangle>, C\<langle>v' \<cdot> \<tau> \<cdot>\<sigma>\<rangle>) \<in> rstep (\<E>\<^sup>\<leftrightarrow>)" by auto
      then have "(C\<langle>u' \<cdot> \<tau> \<cdot>\<sigma>\<rangle>, C\<langle>v' \<cdot> \<tau> \<cdot>\<sigma>\<rangle>) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_def using r_into_rtrancl
        by fastforce
      from conversion_trans' [OF uconv this, THEN conversion_trans'] vconv
      have "(s, t) \<in> (rstep (\<R> \<union> \<E>\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding C uv by auto
      with gterms_ER_conv_implies_S_conv [OF \<open>\<R> \<subseteq> {\<succ>}\<close> g(1) g(2)] show ?thesis .
    qed
  qed

  from assms have CR_SN: "CR (GROUND (rstep \<S>)) \<and> SN (GROUND (rstep \<S>))"
    by (meson GROUND_subset SN_subset)
  interpret complete_ars "GROUND (rstep \<S>)" by (standard, insert CR_SN, auto)
  from complete_NE_intro1 [OF step_conv x NFs] complete_ars.CR show ?thesis
    unfolding complete_ars_def Normalization_Equivalence.SN_ars_def by auto
qed
end
    
context
  assumes fair: "\<forall>(s, t)\<in> PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>). (s, t) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>"
begin
  
lemma msteps_imp_source_steps:
  assumes "\<forall> t \<in># M. s \<succ> t"
  shows "(mstep M \<R>)\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (\<Union>z \<in> {z. s \<succ> z}. slab \<R> z)\<^sup>\<leftrightarrow>\<^sup>*"
proof-
  have "mstep M \<R> \<subseteq> (\<Union>z \<in> {z. s \<succ> z}. slab \<R> z)"
  proof
    fix t u
    assume a: "(t, u) \<in> mstep M \<R>"
    then have tu: "(t, u) \<in> rstep \<R>" unfolding mstep_def mem_Collect_eq by auto
    from a have "\<exists> v \<in># M. v \<succeq> t" unfolding mstep_def mem_Collect_eq by auto
    with assms transD [OF trans_less] have st: "s \<succ> t" by blast
    from tu have "(t, u) \<in> slab \<R> t" unfolding source_step_def by auto
    with st show "(t, u) \<in> (\<Union>z \<in> {z. s \<succ> z}. slab \<R> z)" by auto
  qed
  from conversion_mono [OF this] show ?thesis by auto
qed

text \<open>Establish peak decreasingness of \<open>S\<^sub>\<omega>\<close> steps on ground terms for correctness result below.\<close>
lemma ground_Sw_peak_decreasing:
  assumes "ground s" and "(s, t) \<in> slab S\<^sub>\<omega> s" and "(s, u) \<in> slab S\<^sub>\<omega> s"
  shows "(t, u) \<in> (\<Union> v \<in> {v. ground v \<and> s \<succ> v}. slab S\<^sub>\<omega> v)\<^sup>\<leftrightarrow>\<^sup>*"
proof-
  from assms have rsteps: "(s, t) \<in> rstep S\<^sub>\<omega>" "(s, u) \<in> rstep S\<^sub>\<omega>" unfolding source_step_def by auto
  note vc = SN_imp_variable_condition [OF SN_Sw_step]
  from vc peak_imp_nabla2 [OF _ rsteps] have tu: "(t, u) \<in> nabla S\<^sub>\<omega> s ^^ 2" by auto
  have Sw_less: "S\<^sub>\<omega> \<subseteq> {\<succ>}" using Ri_less unfolding Sw_def R\<^sub>\<omega>_def by fast
  from compatible_rstep_imp_less [OF Sw_less] have Sw_step_less: "rstep S\<^sub>\<omega> \<subseteq> {\<succ>}" by blast
  let ?less_s_conv = "(\<Union>z\<in>{z. s \<succ> z}. slab S\<^sub>\<omega> z)\<^sup>\<leftrightarrow>\<^sup>*"
  let ?leq_conv = "\<lambda> v. (\<Union>z\<in>{z. v \<succeq> z}. slab S\<^sub>\<omega> z)\<^sup>\<leftrightarrow>\<^sup>*"
  { fix v w
    assume vw: "(v, w) \<in> nabla S\<^sub>\<omega> s" and "ground v" and "ground w" and "s \<succ> v" and "s \<succ> w"
    then have vw: "(v, w) \<in> (rstep S\<^sub>\<omega>)\<^sup>\<down> \<or> (v, w) \<in> (rstep (PCP S\<^sub>\<omega>))\<^sup>\<leftrightarrow>" unfolding nabla_def by auto
    { assume "(v, w) \<in> (rstep S\<^sub>\<omega>)\<^sup>\<down>"
      then obtain x where vx: "(v, x) \<in> (rstep S\<^sub>\<omega>)\<^sup>*" and wx: "(w, x) \<in> (rstep S\<^sub>\<omega>)\<^sup>*"
        unfolding join_def rtrancl_converse by blast
      from rsteps_slabI [OF vx _ Sw_step_less, of v] have "(v, x) \<in> ?leq_conv v" by auto
      from slab_conv_less_label [OF this] \<open>s \<succ> v\<close> have vx: "(v, x) \<in> ?less_s_conv" by auto
      from rsteps_slabI [OF wx _ Sw_step_less, of w] have "(w, x) \<in> ?leq_conv w" by auto
      from slab_conv_less_label [OF this] \<open>s \<succ> w\<close> have "(x, w) \<in> ?less_s_conv"
        unfolding conversion_inv by auto
      from conversion_trans' [OF vx this] have "(v, w) \<in> ?less_s_conv" by auto
    } note if_joinable = this
    { assume "(v, w) \<in> (rstep (PCP S\<^sub>\<omega>))\<^sup>\<leftrightarrow>"
      from PCP_xPCP [OF Rw_less \<open>ground v\<close> \<open>ground w\<close> this [unfolded Sw_def, folded E_ord_def]]
      have vw: "(v, w) \<in> (rstep S\<^sub>\<omega>)\<^sup>\<down> \<or> (v, w) \<in> (rstep (PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)))\<^sup>\<leftrightarrow>"
        unfolding Sw_def E_ord_def by auto
      { assume "(v, w) \<in> (rstep (PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)))\<^sup>\<leftrightarrow>"
        with fair rstep_rstep have "(v, w) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>" by fast
        from ground_Einf_Sw [OF this \<open>ground v\<close> \<open>ground w\<close>] have vw: "(v, w) \<in> (mstep {#v, w#} S\<^sub>\<omega>)\<^sup>\<leftrightarrow>\<^sup>*" .
        from \<open>s \<succ> v\<close> \<open>s \<succ> w\<close> have "\<forall> t \<in># {#v, w#}. s \<succ> t" by auto
        from vw msteps_imp_source_steps [OF this] have "(v, w) \<in> ?less_s_conv" by auto
      }
      with vw if_joinable have "(v, w) \<in> ?less_s_conv" by auto
    }
    with vw if_joinable have "(v, w) \<in> ?less_s_conv" by auto
  } note nabla_slab = this
  from tu obtain v where v: "(t, v) \<in> nabla S\<^sub>\<omega> s" "(v, u) \<in> nabla S\<^sub>\<omega> s" by auto
  note gimp = rstep_ground [OF _ \<open>ground s\<close>]
  from this [OF _ rsteps(1)] this [OF _ rsteps(2)] vc have "ground t" "ground u" by auto
  from v [unfolded nabla_def] rstep_ground [OF _ \<open>ground s\<close>] have sv: "(s, v) \<in> (rstep S\<^sub>\<omega>)\<^sup>+" by auto
  from rsteps_ground [OF _ \<open>ground s\<close> trancl_into_rtrancl [OF this]] vc have "ground v" by fast
  from sv rsteps_subset_less [OF Sw_less] have sv: "s \<succ> v" by auto
  from Sw_step_less rsteps have st: "s \<succ> t" and su: "s \<succ> u" by auto
  note tv = nabla_slab [OF v(1) \<open>ground t\<close> \<open>ground v\<close> st sv]
  note vu = nabla_slab [OF v(2) \<open>ground v\<close> \<open>ground u\<close> sv su]
  from ground_less [OF \<open>ground s\<close>] have "{v. ground v \<and> s \<succ> v} = {v. s \<succ> v}" by auto
  with conversion_trans' [OF tv vu] show ?thesis by auto
qed

lemma correctness_okb: "GCR (rstep S\<^sub>\<omega>) \<and> SN (rstep S\<^sub>\<omega>) \<and> (rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof -
  interpret ars_peak_decreasing "slab S\<^sub>\<omega>" "{t. ground t}" "op \<succ>"
    by (unfold_locales, insert ground_Sw_peak_decreasing SN_less, auto)
  from CR have CR: "CR (\<Union>t \<in> {t. ground t}. slab S\<^sub>\<omega> t)" by auto
  from rstep_ground SN_imp_variable_condition [OF SN_Sw_step]
  have "\<And> s t. ground s \<Longrightarrow> (s, t) \<in> rstep S\<^sub>\<omega> \<Longrightarrow> ground t" by fastforce
  then have "(\<Union>t \<in> {t. ground t}. slab S\<^sub>\<omega> t) = GROUND (rstep S\<^sub>\<omega>)" unfolding source_step_def GROUND_def
    by blast
  with CR have "GCR (rstep S\<^sub>\<omega>)" unfolding GROUND_def source_step_def by auto
  with SN_Sw_step oKBi_conversion_ERw show ?thesis by auto
qed
  
end
end
  

locale homomorphism =
  fixes h\<^sub>F :: "'a \<Rightarrow> 'c"
begin
definition h where "h \<equiv> map_funs_term h\<^sub>F"
definition h\<^sub>R where "h\<^sub>R \<equiv> map_funs_trs h\<^sub>F"
definition h\<^sub>s where "h\<^sub>s \<equiv> map_funs_subst h\<^sub>F"
definition h\<^sub>C where "h\<^sub>C \<equiv> map_funs_ctxt h\<^sub>F"

lemma [simp]: "h = map_funs_term h\<^sub>F" unfolding h_def by auto
lemma [simp]: "h\<^sub>C = map_funs_ctxt h\<^sub>F" unfolding h\<^sub>C_def by auto
lemma [simp]: "h\<^sub>s = map_funs_subst h\<^sub>F" unfolding h\<^sub>s_def by auto
lemma [simp]: "h\<^sub>R = map_funs_trs h\<^sub>F" unfolding h\<^sub>R_def by auto
  
lemma ctxt_apply_imp:"t = C\<langle>u\<rangle> \<Longrightarrow> (h t = (h\<^sub>C C)\<langle>h u\<rangle>)"
using map_funs_term_ctxt_distrib by auto

lemma hR_union [simp]: "h\<^sub>R (R1 \<union> R2) = h\<^sub>R R1 \<union> h\<^sub>R R2"
 by (simp add: map_funs_trs_union)

lemma hR_sym [simp]: "h\<^sub>R (R1\<^sup>\<leftrightarrow>) = (h\<^sub>R R1)\<^sup>\<leftrightarrow>"
  unfolding h\<^sub>R_def hR_union map_funs_trs.simps map_funs_rule.simps image_Un by force
    
lemma subst_apply_h:"h (t \<cdot> \<sigma>) = (h t) \<cdot> (h\<^sub>s \<sigma>)"
  using map_funs_subst_distrib by fastforce
end
  
locale inj_homomorphism = homomorphism "h\<^sub>F"
  for h\<^sub>F :: "'a \<Rightarrow> 'c" +
  assumes injF:"inj h\<^sub>F"
begin
  
lemma inj: "inj h"
proof-
  { fix s t::"('a, 'b) term"
    assume st:"s \<noteq> t"
    hence "h s \<noteq> h t" proof(induct s arbitrary:t)
      case (Var x)
      with term.simps show ?case by (cases t, auto)
    next
      case (Fun f ss)
      note Funs = this
      thus ?case proof(cases t)
        case (Var x)
        then show ?thesis by auto
      next
        case (Fun g ts)
        then show ?thesis proof(cases "f=g")
          case True
          from Funs[unfolded Fun True] have "ss \<noteq> ts" by auto
          hence diff:"(\<exists>i < length ss. ss ! i \<noteq> ts ! i) \<or> (length ss \<noteq> length ts)"
            using nth_equalityI by blast
          then show ?thesis proof(cases "length ss \<noteq> length ts")
            case True
            then show ?thesis unfolding Fun using map_eq_imp_length_eq by auto
          next
            case False
            with diff obtain i where i:"i < length ss \<and> ss ! i \<noteq> ts ! i" by auto
            hence "ss ! i \<in> set ss" by auto
            with Funs(1)[OF this] i have "h (ss ! i) \<noteq> h (ts ! i)" by auto
            with i show ?thesis unfolding Fun using map_nth_conv by fastforce
          qed
        next
          case False
          with Funs injF[unfolded inj_on_def] show ?thesis unfolding Fun by auto
        qed
      qed
        
    qed
  }
  thus ?thesis unfolding inj_on_def by auto
qed
  
definition h_inv ("h\<^sup>-\<^sup>1") where "h\<^sup>-\<^sup>1 \<equiv> map_funs_term (inv h\<^sub>F)"
abbreviation h\<^sub>s_inv ("h\<^sub>s\<^sup>-\<^sup>1") where "h\<^sub>s_inv \<sigma> \<equiv> map_funs_subst (inv h\<^sub>F) \<sigma>"
abbreviation h\<^sub>C_inv ("h\<^sub>C\<^sup>-\<^sup>1") where "h\<^sub>C_inv \<sigma> \<equiv> map_funs_ctxt (inv h\<^sub>F) \<sigma>"
  
lemma ctxt_apply_h:"t = C\<langle>u\<rangle> = (h t = (h\<^sub>C C)\<langle>h u\<rangle>)"
proof
  assume "t = C\<langle>u\<rangle>"
  thus "h t = (h\<^sub>C C)\<langle>h u\<rangle>" using ctxt_apply_imp by auto
next
  assume "h t = (h\<^sub>C C)\<langle>h u\<rangle>"
  thus "t = C\<langle>u\<rangle>" proof(induct C arbitrary:t u)
    case Hole
    with inj[unfolded inj_on_def] show ?case by auto
  next
    case (More f bef C aft)
    note h_t = More(2)[unfolded ctxt_apply_term.simps]
    then obtain g ts where t:"t = Fun g ts" by (cases t, auto)
    from h_t injF[unfolded inj_on_def] have fg:"f = g" unfolding t by fastforce
    from h_t have lists:"map h ts = map h bef @ (h\<^sub>C C)\<langle>h u\<rangle> # map h aft" (is "?L = ?R")
      unfolding t by fastforce
    hence ith:"?L ! (length bef) = ?R ! (length bef)" by auto
    from lists have len:"length bef < length ts" using length_map[of h ts] by auto
    with ith have h_ti:"h (ts ! (length bef)) = (h\<^sub>C C)\<langle>h u\<rangle>" (is "h ?ti = _")
      unfolding nth_map[OF len, of h]
      by (metis length_map nth_append_length)
    with More(1) have ti:"?ti = C\<langle>u\<rangle>" by auto
    from lists[unfolded h_ti[symmetric]] have "map h ts = map h (bef @ (ts ! length bef) # aft)" by auto
    with map_injective[OF _ inj] have "ts = bef @ (?ti # aft)" by blast
    thus ?case using t ti fg by simp
  qed  
qed
  
lemma ctxt_apply_h':"h C\<langle>u\<rangle> = (h\<^sub>C C)\<langle>h u\<rangle>" using ctxt_apply_h by auto
 
lemma inv_h_h: "h\<^sup>-\<^sup>1 (h t) = t"
  unfolding h_def h_inv_def map_funs_term_comp inv_o_cancel[OF injF] by simp

lemma map_funs_subst_compose:
fixes \<sigma>::"('d,'e) subst" and f::"'d \<Rightarrow> 'f"
shows "map_funs_subst f (\<sigma> \<circ>\<^sub>s \<tau>) = (map_funs_subst f \<sigma> \<circ>\<^sub>s (map_funs_subst f \<tau>))"
proof-
  have "\<And>x f \<sigma> \<tau>. map_funs_subst f (\<sigma> \<circ>\<^sub>s \<tau>) x = ((map_funs_subst f \<sigma>) \<circ>\<^sub>s (map_funs_subst f \<tau>)) x"
    unfolding subst_compose_def using subst_apply_h by simp
  thus ?thesis unfolding h\<^sub>s_def by fast
qed

lemma ctxt_image:
  assumes "h t = C\<^sub>h\<langle>u\<^sub>h\<rangle>"
  shows "C\<^sub>h = h\<^sub>C (h\<^sub>C_inv C\<^sub>h) \<and> u\<^sub>h = h (h\<^sup>-\<^sup>1 u\<^sub>h)"
proof-
  from assms have "h\<^sup>-\<^sup>1 (h t) = (h\<^sub>C\<^sup>-\<^sup>1 C\<^sub>h)\<langle>h\<^sup>-\<^sup>1 u\<^sub>h\<rangle>" unfolding h_inv_def by simp
  from this[unfolded inv_h_h] have "t = (h\<^sub>C\<^sup>-\<^sup>1 C\<^sub>h)\<langle>h\<^sup>-\<^sup>1 u\<^sub>h\<rangle>" by auto
  hence "h t = (h\<^sub>C (h\<^sub>C\<^sup>-\<^sup>1 C\<^sub>h))\<langle>h (h\<^sup>-\<^sup>1 u\<^sub>h)\<rangle>" by auto
  with assms have eq:"C\<^sub>h\<langle>u\<^sub>h\<rangle> = (h\<^sub>C (h\<^sub>C\<^sup>-\<^sup>1 C\<^sub>h))\<langle>h (h\<^sup>-\<^sup>1 u\<^sub>h)\<rangle>" by auto
  hence "h\<^sub>C (h\<^sub>C\<^sup>-\<^sup>1 C\<^sub>h) = C\<^sub>h" by (induct C\<^sub>h, auto)
  with eq[unfolded this ctxt_eq] show ?thesis by argo
qed
  
lemma subst_image:
  fixes t::"('a, 'b) term"
  assumes "h t = (h u) \<cdot> \<sigma>\<^sub>h"
  shows "\<forall>x \<in> vars_term (h u). \<sigma>\<^sub>h x = h (h\<^sub>s\<^sup>-\<^sup>1 \<sigma>\<^sub>h x)"
proof -
  from map_funs_subst_distrib[of "inv h\<^sub>F" "h u" \<sigma>\<^sub>h] have t:"t = u \<cdot> (h\<^sub>s\<^sup>-\<^sup>1 \<sigma>\<^sub>h)"
    using map_funs_term_ident unfolding assms[symmetric] h_inv_def[symmetric] inv_h_h by auto
  from assms have "(h u) \<cdot> \<sigma>\<^sub>h = (h u) \<cdot> h\<^sub>s (h\<^sub>s\<^sup>-\<^sup>1 \<sigma>\<^sub>h)" unfolding t subst_apply_h by auto
  thus ?thesis unfolding term_subst_eq_conv by auto
qed
  
lemma ctxt_subst_image:
  fixes t::"('a, 'b) term" and l::"('a, 'b) term"
  assumes "((h t)::('c, 'b) term) = D\<^sub>h\<langle>h l \<cdot> \<tau>\<rangle>"
  shows "t = (h\<^sub>C\<^sup>-\<^sup>1 D\<^sub>h)\<langle>l \<cdot> (h\<^sub>s\<^sup>-\<^sup>1 \<tau>)\<rangle>"
proof-
  from ctxt_image[OF assms] have Du:"D\<^sub>h = h\<^sub>C (h\<^sub>C\<^sup>-\<^sup>1 D\<^sub>h)" "h (h\<^sup>-\<^sup>1 (h l \<cdot> \<tau>)) = h l \<cdot> \<tau>" by auto
  from subst_image[OF] Du(2) have "\<forall>x\<in>vars_term (h l). \<tau> x = h (h\<^sub>s_inv \<tau> x)" by auto
  hence \<sigma>:"(h l) \<cdot> \<tau> = (h l) \<cdot> (h\<^sub>s (\<lambda>x. h\<^sub>s_inv \<tau> x))" unfolding term_subst_eq_conv by force
  from assms Du have "h t = (h\<^sub>C (h\<^sub>C\<^sup>-\<^sup>1 D\<^sub>h))\<langle>h l \<cdot> \<tau>\<rangle>" by auto
  from this[unfolded \<sigma> subst_apply_h[symmetric] ctxt_apply_h[symmetric]] show ?thesis by auto
qed
  
lemma rstep_h:"(s,t) \<in> rstep RR = ((h s, h t) \<in> rstep (h\<^sub>R RR))"
proof
  assume rstep:"(s,t) \<in> rstep RR"
  then obtain C \<sigma> l r where lr:"(l, r) \<in> RR" and sC:"s = C\<langle>l \<cdot> \<sigma>\<rangle>" and tC:"t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  hence rule:"(h l, h r) \<in> h\<^sub>R RR" unfolding h\<^sub>R_def map_funs_trs.simps map_funs_rule.simps by force
  from sC tC ctxt_apply_h have C:"h s = (h\<^sub>C C)\<langle>h (l \<cdot> \<sigma>)\<rangle>" "h t = (h\<^sub>C C)\<langle>h (r \<cdot> \<sigma>)\<rangle>" by auto
  from subst_apply_h have sub:"h (l \<cdot> \<sigma>) = h l \<cdot> (h\<^sub>s \<sigma>)" "h (r \<cdot> \<sigma>) = h r \<cdot> (h\<^sub>s \<sigma>)" by auto
  with rule C show "(h s, h t) \<in> rstep (h\<^sub>R RR)" by auto
next
  assume "(h s, h t) \<in> rstep (h\<^sub>R RR)"
  then obtain Ch \<sigma>h lh rh where lr:"(lh, rh) \<in> h\<^sub>R RR" and sC:"h s = Ch\<langle>lh \<cdot> \<sigma>h\<rangle>" and tC:"h t = Ch\<langle>rh \<cdot> \<sigma>h\<rangle>" by auto
  from lr obtain l r where lr:"(l, r) \<in> RR" "lh = h l" "rh = h r"
    unfolding h\<^sub>R_def map_funs_trs.simps map_funs_rule.simps by force
  with sC[THEN ctxt_image] obtain C u where C:"Ch = h\<^sub>C C" and lu:"h u = h l \<cdot> \<sigma>h" by metis
  from lr tC[THEN ctxt_image] obtain v where rv:"h v = h r \<cdot> \<sigma>h" by metis
  from lu[unfolded lr, THEN subst_image] rv[THEN subst_image] obtain \<sigma> where
    "\<forall>x \<in> vars_term (h l) \<union> vars_term (h r). (\<sigma>h x = h\<^sub>s \<sigma> x)" by force
  hence \<sigma>:"(h l) \<cdot> \<sigma>h = (h l) \<cdot> (h\<^sub>s \<sigma>)" "(h r) \<cdot> \<sigma>h = (h r) \<cdot> (h\<^sub>s \<sigma>)" using term_subst_eq_conv by (fast,fast)
  from sC[unfolded C lr] have s:"s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    unfolding \<sigma> subst_apply_h[symmetric] ctxt_apply_h[symmetric] by fast
  from tC[unfolded C lr] have t:"t = C\<langle>r \<cdot> \<sigma>\<rangle>"
    unfolding \<sigma> subst_apply_h[symmetric] ctxt_apply_h[symmetric] by fast
  with s lr(1) show "(s,t) \<in> rstep RR" by auto
qed

lemma subterm_h: "s \<unrhd> t = (h s \<unrhd> h t)"
proof
  assume "s \<unrhd> t"
  with ctxt_apply_h show "h s \<unrhd> h t" by fast
next
  assume "h s \<unrhd> h t"
  then obtain C\<^sub>h where ht:"h s = C\<^sub>h\<langle>h t\<rangle>" by auto
  from ctxt_image[OF this] obtain C u where C:"C\<^sub>h = h\<^sub>C C" and u:"h u = h s" by force
  from ht[unfolded C ctxt_apply_h[symmetric]] show "s \<unrhd> t" by auto
qed
  
lemma subterm_strict_h: "s \<rhd> t = (h s \<rhd> h t)"
  using subterm_h supt_supteq_set_conv by auto

lemma enc_h: "s \<unlhd>\<cdot> t = (h s \<unlhd>\<cdot> h t)"
proof
  assume "s \<unlhd>\<cdot> t"
  with encompeq_termE obtain C \<sigma> where "t = C\<langle>s \<cdot> \<sigma>\<rangle>" by auto
  with ctxt_apply_h have C:"h t = (h\<^sub>C C)\<langle>h (s \<cdot> \<sigma>)\<rangle>" by auto
  from subst_apply_h have "h (s \<cdot> \<sigma>) = h s \<cdot> (h\<^sub>s \<sigma>)" by auto
  with C show "h s \<unlhd>\<cdot> h t" unfolding encompeq.simps by force
next
  assume "h s \<unlhd>\<cdot> h t"
  with encompeq_termE obtain C\<^sub>h \<sigma>\<^sub>h where ht:"h t = C\<^sub>h\<langle>h s \<cdot> \<sigma>\<^sub>h\<rangle>" by auto
  from ctxt_image[OF this] obtain C u where C:"C\<^sub>h = h\<^sub>C C" and u:"h u = h s \<cdot> \<sigma>\<^sub>h" by metis
  with subst_image[OF u] have \<sigma>:"\<forall>x \<in> vars_term (h s). (\<sigma>\<^sub>h x = h\<^sub>s (h\<^sub>s\<^sup>-\<^sup>1 \<sigma>\<^sub>h) x)" by auto
  hence \<sigma>:"(h s) \<cdot> \<sigma>\<^sub>h = (h s) \<cdot> (h\<^sub>s (h\<^sub>s\<^sup>-\<^sup>1 \<sigma>\<^sub>h))" using term_subst_eq_conv by blast
  from ht[unfolded C] show "s \<unlhd>\<cdot> t"
    unfolding \<sigma> subst_apply_h[symmetric] ctxt_apply_h[symmetric] by fast
qed
  
lemma inv_hR_rule[simp]:
  assumes "(hl,hr) \<in> h\<^sub>R R"
  shows "(h\<^sup>-\<^sup>1 hl, h\<^sup>-\<^sup>1 hr) \<in> R \<and> h (h\<^sup>-\<^sup>1 hl) = hl \<and> h (h\<^sup>-\<^sup>1 hr) = hr"
proof-
  note hlr = assms[unfolded h\<^sub>R_def map_funs_trs.simps map_funs_rule.simps]
  from imageE[OF hlr, unfolded prod.inject] obtain l r where "(l, r) \<in> R" and hlr:"hl = h l" "hr = h r"
    unfolding h_def by (metis prod.collapse)
  thus ?thesis unfolding hlr inv_h_h by auto
qed
end
  
locale gtotal_okb_irun_h = gtotal_okb_irun_inf R E less + inj_homomorphism "h\<^sub>F"
  for R :: "nat \<Rightarrow> (('a, 'b::infinite) Term.term \<times> ('a, 'b) Term.term) set" 
  and E :: "nat \<Rightarrow> (('a, 'b) Term.term \<times> ('a, 'b) Term.term) set" 
  and less :: "('a, 'b) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
  and h\<^sub>F :: "'a \<Rightarrow> 'c" +
  fixes less_h :: "('c, 'b) term \<Rightarrow> ('c, 'b) term \<Rightarrow> bool" (infix "\<succ>\<^sub>h" 50)
  assumes SN_less_h: "SN {(x, y). x \<succ>\<^sub>h y}"
    and ctxt_h: "s \<succ>\<^sub>h t \<Longrightarrow> C\<langle>s\<rangle> \<succ>\<^sub>h C\<langle>t\<rangle>"
    and subst_h: "s \<succ>\<^sub>h t \<Longrightarrow> s \<cdot> \<sigma> \<succ>\<^sub>h t \<cdot> \<sigma>"
    and trans_h: "s \<succ>\<^sub>h t \<Longrightarrow> t \<succ>\<^sub>h u \<Longrightarrow> s \<succ>\<^sub>h u"
    and gtotal_h: "ground s \<and> ground t \<longrightarrow> s = t \<or> s \<succ>\<^sub>h t \<or> t \<succ>\<^sub>h s"
    and compat_h:"s' \<succ> t' \<Longrightarrow> (h s') \<succ>\<^sub>h (h t')"
begin

lemma compat_h':"s' \<succeq> t' \<Longrightarrow> (op \<succ>\<^sub>h)\<^sup>=\<^sup>= (h s') (h t')" using compat_h by auto
  
abbreviation less_sk_set ("{\<succ>\<^sub>h}") where "{\<succ>\<^sub>h} \<equiv> {(x, y). x \<succ>\<^sub>h y}"

abbreviation E\<^sub>h where "E\<^sub>h i \<equiv> h\<^sub>R (E i)"
abbreviation R\<^sub>h where "R\<^sub>h i \<equiv> h\<^sub>R (R i)"
abbreviation oKBi_h (infix "\<turnstile>\<^sub>h" 55)
  where "oKBi_h ER ER' \<equiv> ordered_completion.oKBi less_h ER ER'"
    
sublocale okb_h: ordered_completion less_h
  by (unfold_locales, insert SN_less_h ctxt_h subst_h trans_h gtotal_h, blast+)
  
lemma ordstep_h:
  assumes "(s,t) \<in> ordstep {\<succ>} EE"
  shows "(h s, h t) \<in> ordstep {\<succ>\<^sub>h} (h\<^sub>R EE)"
proof-
  from assms obtain C l r \<sigma> where lr:"(l, r) \<in> EE" and sC:"s = C\<langle>l \<cdot> \<sigma>\<rangle>" and tC:"t = C\<langle>r \<cdot> \<sigma>\<rangle>" and
    less:"l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" unfolding ordstep.simps by blast
  hence rule:"(h l, h r) \<in> h\<^sub>R EE" unfolding h\<^sub>R_def map_funs_trs.simps by force
  from sC tC ctxt_apply_h have C:"h s = (h\<^sub>C C)\<langle>h (l \<cdot> \<sigma>)\<rangle>" "h t = (h\<^sub>C C)\<langle>h (r \<cdot> \<sigma>)\<rangle>" by auto
  from subst_apply_h have sub:"h (l \<cdot> \<sigma>) = h l \<cdot> (h\<^sub>s \<sigma>)" "h (r \<cdot> \<sigma>) = h r \<cdot> (h\<^sub>s \<sigma>)" by auto
  with rule C compat_h[OF less] show "(h s, h t) \<in> ordstep {\<succ>\<^sub>h} (h\<^sub>R EE)"
    unfolding ordstep.simps by auto
qed
    
lemma ostep_h:
  assumes "(s,t) \<in> ostep EE RR"
  shows "(h s, h t) \<in> okb_h.ostep (h\<^sub>R (EE)) (h\<^sub>R RR)"
  using assms rstep_h[of s t RR] ordstep_h[of s t "EE\<^sup>\<leftrightarrow>"]
  unfolding ostep_def okb_h.ostep_def hR_sym by fast

lemma encstep2_h:
  assumes "(s,t) \<in> encstep2 EE RR"
  shows "(h s, h t) \<in> okb_h.encstep2 (h\<^sub>R EE) (h\<^sub>R RR)"
using assms proof(cases)
  case (estep l r C \<sigma>)
  hence rule:"(h l, h r) \<in> (h\<^sub>R EE)\<^sup>\<leftrightarrow>" unfolding h\<^sub>R_def map_funs_trs.simps by force
  from estep ctxt_apply_h have C:"h s = (h\<^sub>C C)\<langle>h (l \<cdot> \<sigma>)\<rangle>" "h t = (h\<^sub>C C)\<langle>h (r \<cdot> \<sigma>)\<rangle>" by auto
  from subst_apply_h have sub:"h (l \<cdot> \<sigma>) = h l \<cdot> (h\<^sub>s \<sigma>)" "h (r \<cdot> \<sigma>) = h r \<cdot> (h\<^sub>s \<sigma>)" by auto
  from enc_h estep(5) have "h l \<lhd>\<cdot> h s" unfolding encomp_def by auto
  with okb_h.encstep2.estep rule sub rule C compat_h[OF estep(4)] show ?thesis by auto
next
  case (rstep l r C \<sigma>)
  hence rule:"(h l, h r) \<in> h\<^sub>R RR" unfolding h\<^sub>R_def map_funs_trs.simps by force
  from rstep ctxt_apply_h have C:"h s = (h\<^sub>C C)\<langle>h (l \<cdot> \<sigma>)\<rangle>" "h t = (h\<^sub>C C)\<langle>h (r \<cdot> \<sigma>)\<rangle>" by auto
  from subst_apply_h have sub:"h (l \<cdot> \<sigma>) = h l \<cdot> (h\<^sub>s \<sigma>)" "h (r \<cdot> \<sigma>) = h r \<cdot> (h\<^sub>s \<sigma>)" by auto
  from enc_h rstep(4) have "h l \<lhd>\<cdot> h s" unfolding encomp_def by auto
  with okb_h.encstep2.rstep rule sub rule C show ?thesis by auto
qed

lemma enc_hstep1:
  assumes "(s,t) \<in> encstep1 EE RR"
  shows "(h s, h t) \<in> okb_h.encstep1 (h\<^sub>R EE) (h\<^sub>R RR)"
using assms proof(cases)
  case (estep l r C \<sigma>)
  hence rule:"(h l, h r) \<in> (h\<^sub>R EE)\<^sup>\<leftrightarrow>" unfolding h\<^sub>R_def map_funs_trs.simps by force
  from estep ctxt_apply_h have C:"h s = (h\<^sub>C C)\<langle>h (l \<cdot> \<sigma>)\<rangle>" "h t = (h\<^sub>C C)\<langle>h (r \<cdot> \<sigma>)\<rangle>" by auto
  from subst_apply_h have sub:"h (l \<cdot> \<sigma>) = h l \<cdot> (h\<^sub>s \<sigma>)" "h (r \<cdot> \<sigma>) = h r \<cdot> (h\<^sub>s \<sigma>)" by auto
  from enc_h estep(5) have "h l \<lhd>\<cdot> h s" unfolding encomp_def by auto
  with okb_h.encstep1.estep rule sub rule C compat_h[OF estep(4)] show ?thesis by auto
next
  case rstep
  from this[unfolded rstep_h] okb_h.encstep1.rstep show ?thesis by simp
qed

lemma oKBi_h: "(E\<^sub>h i, R\<^sub>h i) \<turnstile>\<^sub>h (E\<^sub>h (Suc i), R\<^sub>h (Suc i))"
proof-
  from hR_union hR_sym have s1 [simp]:"rstep (R\<^sub>h i \<union> (E\<^sub>h i)\<^sup>\<leftrightarrow>) = rstep (h\<^sub>R(R i \<union> (E i)\<^sup>\<leftrightarrow>))" by metis
  let ?h_rule = "\<lambda>(s,t).(h s, h t)"
  have h_rule:"\<And>A. h\<^sub>R A = ?h_rule ` A" unfolding h\<^sub>R_def map_funs_trs.simps by force
  from inj have inj':"inj ?h_rule" unfolding inj_on_def by (simp add: inj_on_def, auto)
  have hR_diff [simp]:"\<And>A B. h\<^sub>R (A - B) = h\<^sub>R A - h\<^sub>R B"
    using image_set_diff[OF inj'] unfolding h_rule by metis
  have h_rule [simp]:"\<And>s t. h\<^sub>R {(s,t)} = {(h s, h t)}" unfolding h\<^sub>R_def map_funs_trs.simps by force
  have h_lift [simp]:"\<And>s t S. (s, t) \<in> S \<Longrightarrow> (h s, h t) \<in> h\<^sub>R S" unfolding h\<^sub>R_def map_funs_trs.simps by force
  from irun[of i] show ?thesis proof(cases)
    case (deduce s t u)
    with rstep_h s1 have st:"(h s, h t) \<in> rstep (R\<^sub>h i \<union> (E\<^sub>h i)\<^sup>\<leftrightarrow>)" by blast
    from deduce rstep_h s1 have su:"(h s, h u) \<in> rstep (R\<^sub>h i \<union> (E\<^sub>h i)\<^sup>\<leftrightarrow>)" by blast
    from deduce(1) have "E\<^sub>h (Suc i) = E\<^sub>h i \<union> {(h t, h u)}" unfolding h\<^sub>R_def map_funs_trs.simps by force
    with okb_h.oKBi.deduce[OF st su] deduce(2) show ?thesis by simp
  next
    case (orientl s t)
    with compat_h have st:"h s \<succ>\<^sub>h h t" by auto
    from orientl(4) have st':"(h s, h t) \<in> E\<^sub>h i" using h_lift by force
    have E:"E\<^sub>h (Suc i) = E\<^sub>h i - {(h s, h t)}" unfolding orientl(1) hR_diff h_rule by auto
    have "R\<^sub>h (Suc i) = R\<^sub>h i \<union> {(h s, h t)}" unfolding orientl(2) hR_union h_rule by auto
    with okb_h.oKBi.orientl[OF st st'] E show ?thesis by auto
  next
    case (orientr t s)
    with compat_h have st:"h t \<succ>\<^sub>h h s" by auto
    from orientr(4) have st':"(h s, h t) \<in> E\<^sub>h i" using h_lift by force
    have E:"E\<^sub>h (Suc i) = E\<^sub>h i - {(h s, h t)}" unfolding orientr(1) hR_diff h_rule by auto
    have "R\<^sub>h (Suc i) = R\<^sub>h i \<union> {(h t, h s)}" unfolding orientr(2) hR_union h_rule by auto
    with okb_h.oKBi.orientr[OF st st'] E show ?thesis by auto
  next
    case (delete s)
    have E:"E\<^sub>h (Suc i) = E\<^sub>h i - {(h s, h s)}" unfolding delete(1) hR_diff h_rule by force
    from delete(3) h_lift have "(h s, h s) \<in> E\<^sub>h i" by force
    from okb_h.oKBi.delete[OF this] E delete(2) show ?thesis by auto
  next
    case (compose t u s)
    have R:"R\<^sub>h (Suc i) = R\<^sub>h i - {(h s, h t)} \<union> {(h s, h u)}"
      unfolding compose(2) hR_union hR_diff h_rule by simp
    from compose(3)[THEN ostep_h] h_rule
    have ostep:"(h t, h u) \<in> okb_h.ostep (E\<^sub>h i) (R\<^sub>h i - {(h s, h t)})"
      unfolding hR_diff h_rule by simp
    from okb_h.oKBi.compose[OF ostep h_lift[OF compose(4)]] R compose(1) show ?thesis by auto
  next
    case (simplifyl s u t)
    have E:"E\<^sub>h (Suc i) = E\<^sub>h i - {(h s, h t)} \<union> {(h u, h t)}"
      unfolding simplifyl(1) hR_union hR_diff h_rule by force
    from simplifyl(3)[THEN enc_hstep1]
      have "(h s, h u) \<in> okb_h.encstep1 (E\<^sub>h i - {(h s, h t)}) (R\<^sub>h i)" unfolding hR_diff h_rule by auto
    from okb_h.oKBi.simplifyl[OF this h_lift[OF simplifyl(4)]] E simplifyl(2) show ?thesis by argo
  next
    case (simplifyr t u s)
    have E:"E\<^sub>h (Suc i) = E\<^sub>h i - {(h s, h t)} \<union> {(h s, h u)}"
      unfolding simplifyr(1) hR_union hR_diff h_rule by force
    from simplifyr(3)[THEN enc_hstep1]
      have "(h t, h u) \<in> okb_h.encstep1 (E\<^sub>h i - {(h s, h t)}) (R\<^sub>h i)" unfolding hR_diff h_rule by auto
    from okb_h.oKBi.simplifyr[OF this h_lift[OF simplifyr(4)]] E simplifyr(2) show ?thesis by argo
  next
    case (collapse t u s)
    have E:"E\<^sub>h (Suc i) = E\<^sub>h i \<union> {(h u, h s)}" unfolding collapse(1) hR_union h_rule by force
    from h\<^sub>R_def have R:"R\<^sub>h (Suc i) = R\<^sub>h i - {(h t, h s)}" unfolding collapse(2) hR_diff h_rule by force
    from collapse(3)[THEN encstep2_h]
      have "(h t, h u) \<in> okb_h.encstep2 (E\<^sub>h i) (R\<^sub>h i - {(h t, h s)})" unfolding hR_diff h_rule by auto
    from okb_h.collapse[OF this h_lift[OF collapse(4)]] R E show ?thesis by auto
  qed
qed
  
sublocale irun_h: gtotal_okb_irun_inf R\<^sub>h E\<^sub>h less_h
proof(unfold_locales)
  show "\<And>s t. ground s \<Longrightarrow> ground t \<Longrightarrow> s = t \<or> s \<succ>\<^sub>h t \<or> t \<succ>\<^sub>h s" using gtotal_h by auto
next
  show "\<And>i. okb_h.oKBi (E\<^sub>h i, R\<^sub>h i) (E\<^sub>h (Suc i), R\<^sub>h (Suc i))" using oKBi_h by auto
qed(auto simp:R0 map_funs_trs.simps)
    
lemma hEinf [simp]:"h\<^sub>R E\<^sub>\<infinity> = irun_h.Einf" unfolding h\<^sub>R_def
    by (smt SUP_cong image_UN map_funs_trs.simps)

lemma hEw [simp]:"h\<^sub>R E\<^sub>\<omega> = irun_h.E\<^sub>\<omega>"
proof-
  have eq:"h\<^sub>R E\<^sub>\<omega> = (map_funs_rule h\<^sub>F) ` E\<^sub>\<omega>" unfolding h\<^sub>R_def map_funs_trs.simps by force
  from inj have inj_rule:"inj (map_funs_rule h\<^sub>F)" unfolding h_def map_funs_rule.simps inj_on_def by auto
  { fix j ::nat
    let ?C = "Collect (op \<le> j)"
    have univ:"\<forall>x\<in>Collect (op \<le> j). E x \<subseteq> UNIV" using subset_UNIV by blast
    from le0 have "0 \<in> {i. i \<le> j}" by simp
    with image_INT[OF inj_rule univ] have "map_funs_rule h\<^sub>F ` INTER ?C E = (\<Inter>x\<in>?C. map_funs_rule h\<^sub>F ` E x)" by auto
  }
  with eq have "h\<^sub>R E\<^sub>\<omega> = (\<Union>i. \<Inter>j\<in>Collect (op \<le> i). map_funs_rule h\<^sub>F ` E j)"
    unfolding E\<^sub>\<omega>_def image_UN by presburger
  thus ?thesis unfolding irun_h.E\<^sub>\<omega>_def unfolding h\<^sub>R_def map_funs_trs.simps by simp
qed

lemma hRw [simp]:"h\<^sub>R R\<^sub>\<omega> = irun_h.R\<^sub>\<omega>"
proof-
  have eq:"h\<^sub>R R\<^sub>\<omega> = (map_funs_rule h\<^sub>F) ` R\<^sub>\<omega>" unfolding h\<^sub>R_def map_funs_trs.simps by force
  from inj have inj_rule:"inj (map_funs_rule h\<^sub>F)" unfolding h_def map_funs_rule.simps inj_on_def by auto
  { fix j ::nat
    let ?C = "Collect (op \<le> j)"
    have univ:"\<forall>x\<in>Collect (op \<le> j). R x \<subseteq> UNIV" using subset_UNIV by blast
    from le0 have "0 \<in> {i. i \<le> j}" by simp
    with image_INT[OF inj_rule univ] have "map_funs_rule h\<^sub>F ` INTER ?C R = (\<Inter>x\<in>?C. map_funs_rule h\<^sub>F ` R x)" by auto
  }
  with eq have "h\<^sub>R R\<^sub>\<omega> = (\<Union>i. \<Inter>j\<in>Collect (op \<le> i). map_funs_rule h\<^sub>F ` R j)"
    unfolding R\<^sub>\<omega>_def image_UN by presburger
  thus ?thesis unfolding irun_h.R\<^sub>\<omega>_def unfolding h\<^sub>R_def map_funs_trs.simps by simp
qed
  
lemma hSw: "h\<^sub>R S\<^sub>\<omega> \<subseteq> irun_h.Sw"
proof-
  let ?l = "h\<^sub>R {(s \<cdot> \<sigma>, t \<cdot> \<sigma>) | s t \<sigma>. (s, t) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow> \<and> s \<cdot> \<sigma> \<succ> t \<cdot> \<sigma>}"
  let ?r = "{(s \<cdot> \<sigma>, t \<cdot> \<sigma>) | s t \<sigma>. (s, t) \<in> irun_h.E\<^sub>\<omega>\<^sup>\<leftrightarrow> \<and> s \<cdot> \<sigma> \<succ>\<^sub>h t \<cdot> \<sigma>}"
  have lr:"?l \<subseteq> ?r"
  proof
    fix u v
    assume "(u,v) \<in> ?l"
    then obtain s t \<sigma> where st:"u = h (s \<cdot> \<sigma>)" "v = h (t \<cdot> \<sigma>)" "(s, t) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "s \<cdot> \<sigma> \<succ> t \<cdot> \<sigma>"
      unfolding h\<^sub>R_def map_funs_trs.simps image_Collect by force
    { fix l r
      assume "(l,r) \<in> E\<^sub>\<omega>"
      hence "(h l,h r) \<in> h\<^sub>R E\<^sub>\<omega>" unfolding h\<^sub>R_def h_def map_funs_trs.simps map_funs_rule.simps
        using image_iff by fastforce 
      with hEw have "(h l, h r) \<in> irun_h.E\<^sub>\<omega>" by auto
    }
    with st(3) have "(h s, h t) \<in> irun_h.E\<^sub>\<omega>\<^sup>\<leftrightarrow>" by auto
    with compat_h[OF st(4)] show "(u,v) \<in> ?r" unfolding st subst_apply_h by blast
  qed
  show ?thesis unfolding Sw_def irun_h.Sw_def hR_union by (rule Un_mono, insert hRw lr, auto)
qed

lemma perm_h:"p \<bullet> (h t) = h (p \<bullet> t)"
  by (induct t, unfold h_def, auto)

lemma perm_inv_h:"p \<bullet> (h\<^sup>-\<^sup>1 t) = h\<^sup>-\<^sup>1 (p \<bullet> t)"
  by (induct t, unfold h_inv_def, auto)
  
lemma poss_inv_h:"poss (h\<^sup>-\<^sup>1 t) = poss t" unfolding h_inv_def
  by (induct t, unfold term.map, auto)
      
lemma funposs_inv_h:"funposs (h\<^sup>-\<^sup>1 t) = funposs t" unfolding h_inv_def
  by (induct t, unfold term.map, auto)
    
lemma pos_term_h:"p \<in> poss t \<Longrightarrow> (h t |_ p) = h (t |_ p)" unfolding h_def
proof (induct t arbitrary:p)
  case (Var x)
  hence p:"p = \<epsilon>" by auto
  then show ?case unfolding p term.map subt_at.simps by auto
next
  case (Fun f ts)
  show ?case proof(cases p)
    case Empty
    show ?thesis unfolding Empty term.map subt_at.simps by auto
  next
    case (PCons i ps)
    from Fun(2) have "i < length ts" "ps \<in> poss (ts ! i)" unfolding PCons poss.simps by auto
    with Fun(1)[OF _ this(2)]  show ?thesis unfolding PCons using subt_at.simps by auto
  qed
qed

lemma unify_h:
  fixes es::"(('d, 'e) term \<times> ('d, 'e) term) list" and f::"'d \<Rightarrow> 'f"
  assumes "unify es bs = Some cs"
  defines "fh \<equiv> map_funs_term f"
  shows "unify (map (\<lambda>(u,v).(fh u, fh v)) es) (map (\<lambda>(x,t).(x, fh t)) bs) = Some (map (\<lambda>(x,t).(x, fh t)) cs)"
proof-
  let ?he = "(\<lambda>(u,v).(fh u, fh v))"
  let ?mhe = "map ?he"
  let ?hb = "(\<lambda>(x,t).(x, fh t))"
  let ?mhb = "map ?hb"
  { fix s t::"('d, 'e) term"
    fix us::"(('d, 'e) term \<times> ('d, 'e) term) list"
    assume "decompose s t = Some us"
    hence "decompose (fh s) (fh t) = Some (map ?he us)" proof(induct s arbitrary:t)
      case (Var x)
      then show ?case unfolding decompose_def h_def term.map by simp
    next
      case (Fun f ss)
      from Fun(2)[unfolded decompose_def] obtain g ts where t:"t = Fun g ts" by (cases t, auto)
      from Fun(2)[unfolded decompose_def t, simplified] option.simps(3) have fg:"f = g" by metis
      from Fun(2)[unfolded decompose_def t] have d:"decompose (Fun f ss) t = zip_option ss ts"
        unfolding t fg decompose_def h_def term.map by simp
      from Fun(2)[unfolded this] zip_option_zip_conv have
        l:"length ts = length ss" "length us = length ss" "us = zip ss ts" by auto
      from this(3) zip_map_map have z:"zip (map fh ss) (map fh ts) = map ?he us" by auto
      have zo:"zip_option (map fh ss) (map fh ts) = Some (map ?he us)"
        by (rule zip_option_intros(1), unfold length_map, insert l z, auto)
      have "decompose (fh (Fun f ss)) (fh t) = zip_option (map fh ss) (map fh ts)"
        unfolding t fg decompose_def fh_def term.map by simp
      with zo show ?case by auto
    qed
  } note dec = this
  {fix x::"'e" and t E
    { fix u v
      have "map_funs_subst f (Var(x := t)) = Var(x := fh t)" unfolding fh_def by auto
      hence "\<And>u. fh (u \<cdot> Var(x := t)) = (fh u) \<cdot> Var(x := fh t)"
        unfolding fh_def map_funs_subst_distrib by auto
      hence "(fh (u \<cdot> Var(x := t)), fh (v \<cdot> Var(x := t))) = ((fh u) \<cdot> Var(x := fh t), fh v \<cdot> Var(x := fh t))"
        by auto
    }
    hence "?mhe (subst_list (subst x t) E) = subst_list (subst x (fh t)) (?mhe E)"
      unfolding subst_list_def map_map subst_def o_def split by auto
  } note subst_list = this
show ?thesis using assms(1)
proof (induction es bs arbitrary: cs rule: unify.induct)
  case (1 bs)
  then show ?case by force
next
  case (2 f ss g ts E bs)
  then obtain us where d:"decompose (Fun f ss) (Fun g ts) = Some us"
    and [simp]: "f = g" "length ss = length ts" "us = zip ss ts"
    and u:"unify (us @ E) bs = Some cs" by (auto split: option.splits)
  from dec[OF this(1)] have
   "unify (?mhe ((Fun f ss, Fun g ts) # E)) (?mhb bs) = unify ((?mhe us) @ (?mhe E)) (?mhb bs)"
    unfolding list.map split fh_def term.map unify.simps unfolding fh_def[symmetric] by auto
  with 2(1)[OF d u] show ?case by auto
next
  case (3 x t E bs)
  show ?case
  proof (cases "t = Var x")
    assume t:"t = Var x"
    with 3(3) have u1:"unify E bs = Some cs" by simp
    have u2:"unify (?mhe ((Var x, t) # E)) (?mhb bs) = unify (?mhe E) (?mhb bs)"
      unfolding list.map split fh_def term.map t unify.simps by simp
    with 3(1)[OF t u1] show ?case by argo
  next
    assume t:"t \<noteq> Var x"
    let ?sub = "subst_list (subst x t) E"
    let ?sub' = "subst_list (subst x (fh t)) (?mhe E)"
    from t 3(3) have xt:"x \<notin> vars_term t" unfolding unify.simps by auto
    with 3(3) t have "unify (subst_list (subst x t) E) ((x, t) # bs) = Some cs" by auto
    with 3(2)[OF t xt] have u1:"unify (?mhe ?sub) ((x, fh t) # (?mhb bs)) = Some (?mhb cs)" by auto
    from t have ht:"fh t \<noteq> Var x" unfolding fh_def by (cases t, auto)
    with xt have u2:"unify (?mhe ((Var x, t) # E)) (?mhb bs) = unify ?sub' ((x, fh t) # (?mhb bs))"
      unfolding list.map split fh_def term.map unify.simps vars_term_map_funs_term2 by auto
    with u1 show ?case unfolding list.map subst_list[symmetric] split by argo
  qed
next
  case (4 f ss x E bs)
    let ?t = "Fun f ss"
    let ?sub = "subst_list (subst x ?t) E"
    let ?sub' = "subst_list (subst x (fh ?t)) (?mhe E)"
    from 4(2) have xt:"x \<notin> vars_term ?t" unfolding unify.simps by force
    with 4(2) have "unify (subst_list (subst x ?t) E) ((x, ?t) # bs) = Some cs" by auto
    with 4(1)[OF xt] have u1:"unify (?mhe ?sub) ((x, fh ?t) # (?mhb bs)) = Some (?mhb cs)" by auto
    have ht:"fh ?t \<noteq> Var x" unfolding fh_def by auto
    with xt have u2:"unify (?mhe ((?t, Var x) # E)) (?mhb bs) = unify ?sub' ((x, fh ?t) # (?mhb bs))"
      unfolding list.map split fh_def term.map unify.simps vars_term_map_funs_term2 by auto
    with u1 show ?case unfolding list.map subst_list[symmetric] split by argo
  qed
qed
  
lemma mgu_h:
  fixes s::"('a, 'b) term"
  assumes "mgu (h s) (h t) = Some \<mu>"
  shows "mgu s t = Some ((h\<^sub>s_inv \<mu>)) \<and> h\<^sub>s (h\<^sub>s_inv \<mu>) = \<mu>"
proof-
  let ?bmap = "map (\<lambda>(x,t). (x, h\<^sup>-\<^sup>1 t))"
  let ?bmap' = "map (\<lambda>(x,t). (x, h t))"
  from assms[unfolded mgu.simps] obtain bs where
    uh:"unify [(h s, h t)] [] = Some bs" and mu:"\<mu> = subst_of bs" by (cases "unify [(h s, h t)] []", auto)
  from unify_h[OF this(1), of "inv h\<^sub>F"] have u:"unify [(s, t)] [] = Some (?bmap bs)"
    unfolding list.map split h_def map_funs_term_comp o_def inv_f_f[OF injF] term.map_ident h_inv_def
    by auto
  from unify_h[OF this, of h\<^sub>F] uh have bs:"bs = ?bmap' (?bmap bs)"
    unfolding list.map split h_def[symmetric] by auto
  have x:"\<And> x f t. subst x (map_funs_term f t) = map_funs_subst f (subst x t)" unfolding subst_def by auto
  { fix bs::"('e \<times> ('d, 'e) term) list"
    fix f::"'d \<Rightarrow> 'f"
    have "subst_of (map (\<lambda>(x, t). (x, map_funs_term f t)) bs) = map_funs_subst f (subst_of bs)"
    proof-
      let ?m = "\<lambda>(x, t). (x, map_funs_term f t)"
      show ?thesis proof(induct bs, simp)
        case (Cons xt bs)
        then obtain x t where xt:"xt = (x,t)" by fastforce
        have "subst_of (map ?m (xt # bs)) = subst_of (map ?m bs) \<circ>\<^sub>s (subst x (map_funs_term f t))"
          unfolding xt list.map split by auto
        also from Cons x[of x f] have "\<dots> = (map_funs_subst f (subst_of bs)) \<circ>\<^sub>s (map_funs_subst f (subst x t))" by simp
        also from map_funs_subst_compose[of f] have "\<dots> = map_funs_subst f (subst_of bs \<circ>\<^sub>s (subst x t))" by metis
        finally show ?case unfolding xt by auto
      qed
    qed
  }
  hence "subst_of (map (\<lambda>(x, t). (x, h\<^sup>-\<^sup>1 t)) bs) = h\<^sub>s\<^sup>-\<^sup>1 (subst_of bs)"
    unfolding h_inv_def by auto
  hence mgu:"mgu s t = Some ((h\<^sub>s_inv \<mu>))" unfolding mgu.simps u mu by simp
  { fix y
    from bs have "map_funs_subst h\<^sub>F (map_funs_subst (inv h\<^sub>F) (subst_of bs)) y = (subst_of bs) y" (is "?l y = ?r y")
    proof(induct bs, simp)
      case (Cons xt bs)
      then obtain x t where xt:"xt = (x,t)" by fastforce
      let ?h = "map_funs_subst h\<^sub>F"
      let ?hi = "map_funs_subst (inv h\<^sub>F)"
      let ?xt = "subst x t"
      from Cons(2)[unfolded xt list.map split] have t:"h (h\<^sup>-\<^sup>1 t) = t" by auto
      have "?h (?hi (subst_of (xt # bs))) y = ?h (?hi (subst_of bs \<circ>\<^sub>s ?xt)) y" (is "?t = _ ")
        unfolding xt by auto
      with map_funs_subst_compose[of "inv h\<^sub>F"] have "?t = ?h (?hi (subst_of bs) \<circ>\<^sub>s (?hi ?xt)) y" by metis
      with map_funs_subst_compose[of h\<^sub>F] have "?t = (?h (?hi (subst_of bs)) \<circ>\<^sub>s (?h (?hi ?xt))) y" by metis
      hence "?t = (?h (?hi (subst_of bs)) y) \<cdot> (?h (?hi ?xt))" using subst_compose_def by metis
      with Cons have "?t = (subst_of bs y) \<cdot> (?h (?hi ?xt))" by auto
      with t have "?t = (subst_of bs y) \<cdot> (?h (?hi ?xt))" unfolding subst_of_def by auto
      with x[of x "inv h\<^sub>F"] have "?t = (subst_of bs y) \<cdot> (?h (subst x (h\<^sup>-\<^sup>1 t)))"
        unfolding subst_of_def h_inv_def by auto
      with x[of x h\<^sub>F] have "?t = (subst_of bs y) \<cdot> (subst x (h (h\<^sup>-\<^sup>1 t)))" unfolding subst_of_def by auto
      with t have "?t = (subst_of bs y) \<cdot> ?xt" by auto
      hence "?t = ((subst_of bs) \<circ>\<^sub>s ?xt) y" unfolding subst_compose_def by metis
      then show ?case unfolding xt by auto
    qed
  }
  hence "h\<^sub>s (h\<^sub>s_inv \<mu>) = \<mu>" unfolding h\<^sub>s_def mu by auto
  with mgu show ?thesis by auto
qed
   
lemma PCP_h:
  assumes "(hs,ht) \<in> irun_h.PCP_ext (h\<^sub>R R\<^sub>\<omega> \<union> ((h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>))"
  shows "(h\<^sup>-\<^sup>1 hs, h\<^sup>-\<^sup>1 ht) \<in> PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>) \<and> hs = h (h\<^sup>-\<^sup>1 hs) \<and> ht = h (h\<^sup>-\<^sup>1 ht)"
proof-
  let ?s = "h\<^sup>-\<^sup>1 hs"
  let ?t = "h\<^sup>-\<^sup>1 ht"
  from assms[unfolded irun_h.PCP_ext_def] obtain hrl hrl' pos \<mu> where
    ooverlap:"ooverlap {\<succ>\<^sub>h} (h\<^sub>R R\<^sub>\<omega> \<union> (h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>) hrl hrl' pos \<mu> (hs) (ht)" and
    nf:"\<forall>hs\<lhd>fst hrl \<cdot> \<mu>. hs \<in> NF (ordstep {\<succ>\<^sub>h} (h\<^sub>R R\<^sub>\<omega> \<union> (h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>))" by blast
  (* ooverlap *)
  obtain hl hr hl' hr' where hrl:"hrl = (hl, hr)" "hrl' = (hl', hr')" by force
  { fix p hl hr
    assume p:"\<exists>p. p \<bullet> (hl, hr) \<in> h\<^sub>R R\<^sub>\<omega> \<union> (h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>"
    then obtain p where "(p \<bullet> hl, p \<bullet> hr) \<in> h\<^sub>R R\<^sub>\<omega> \<union> (h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>"
      unfolding rule_pt.permute_prod_eqvt by auto
    hence p:"(p \<bullet> hl, p \<bullet> hr) \<in> h\<^sub>R (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" using hR_union hR_sym by blast
    let ?l = "p \<bullet> hl" and ?r = "p \<bullet> hr"
    from inv_hR_rule[OF p] have x:"(h\<^sup>-\<^sup>1 ?l, h\<^sup>-\<^sup>1 ?r) \<in> (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" "h (h\<^sup>-\<^sup>1 ?l) = ?l \<and> h (h\<^sup>-\<^sup>1 ?r) = ?r"
      unfolding perm_inv_h by auto
    from this(2) have "h (h\<^sup>-\<^sup>1 hl) = hl \<and> h (h\<^sup>-\<^sup>1 hr) = hr"
      unfolding perm_inv_h[symmetric] perm_h[symmetric] term_pt.permute_eq_iff by auto
    with x have "(\<exists>p. p \<bullet> (h\<^sup>-\<^sup>1 hl, h\<^sup>-\<^sup>1 hr) \<in> R\<^sub>\<omega> \<union> (E\<^sub>\<omega>)\<^sup>\<leftrightarrow>) \<and> h (h\<^sup>-\<^sup>1 hl) = hl \<and> h (h\<^sup>-\<^sup>1 hr) = hr"
      unfolding rule_pt.permute_prod_eqvt perm_inv_h by meson
  } note perm_rule = this
  note ooverlap = ooverlap[unfolded ooverlap_def hrl snd_conv fst_conv]
  hence
    p:"\<exists>p. p \<bullet> (hl, hr) \<in> h\<^sub>R R\<^sub>\<omega> \<union> (h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>" and
    p':"\<exists>p'. p' \<bullet> (hl', hr') \<in> h\<^sub>R R\<^sub>\<omega> \<union> (h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>" and
    vars:"vars_rule (hl, hr) \<inter> vars_rule (hl', hr') = {}" and
    pos:"pos \<in> funposs hl'" and
    mgu:"mgu hl (hl' |_ pos) = Some \<mu>" and
    less:"(hr \<cdot> \<mu>, hl \<cdot> \<mu>) \<notin> {\<succ>\<^sub>h}" and less':"(hr' \<cdot> \<mu>, hl' \<cdot> \<mu>) \<notin> {\<succ>\<^sub>h}" and
    cs:"hs = (ctxt_of_pos_term pos (hl' \<cdot> \<mu>))\<langle>hr \<cdot> \<mu>\<rangle>" and ht_mu:"ht = hr' \<cdot> \<mu>" by auto
  note p = perm_rule[OF p] perm_rule[OF p']
  from p have 1:"\<exists>p. p \<bullet> (h\<^sup>-\<^sup>1 hl, h\<^sup>-\<^sup>1 hr) \<in> R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "\<exists>p. p \<bullet> (h\<^sup>-\<^sup>1 hl', h\<^sup>-\<^sup>1 hr') \<in> R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" by auto
  from p have inv:"h (h\<^sup>-\<^sup>1 hl) = hl" "h (h\<^sup>-\<^sup>1 hr) = hr" "h (h\<^sup>-\<^sup>1 hl') = hl'" "h (h\<^sup>-\<^sup>1 hr') = hr'" by auto
  have vt:"\<And>t. vars_term (h\<^sup>-\<^sup>1 t) = vars_term t" by (unfold h_inv_def, auto)
  from vars have 2:"vars_rule (h\<^sup>-\<^sup>1 hl, h\<^sup>-\<^sup>1 hr) \<inter> vars_rule (h\<^sup>-\<^sup>1 hl', h\<^sup>-\<^sup>1  hr') = {}"
    unfolding vars_rule_def fst_conv snd_conv vt by auto
  from pos have 3:"pos \<in> funposs (h\<^sup>-\<^sup>1 hl')" using funposs_inv_h by auto
  from mgu inv(3) have "mgu hl (h (h\<^sup>-\<^sup>1 hl') |_ pos ) = Some \<mu>" by argo
  with pos_term_h[OF funposs_imp_poss[OF 3]] have "mgu (h (h\<^sup>-\<^sup>1 hl)) (h (h\<^sup>-\<^sup>1 hl' |_ pos )) = Some \<mu>"
    using inv(1) by auto
  from mgu_h[OF this] have 4:"mgu (h\<^sup>-\<^sup>1 hl) (h\<^sup>-\<^sup>1 hl' |_ pos) = Some (h\<^sub>s_inv \<mu>)" and \<mu>:"\<mu> = h\<^sub>s (h\<^sub>s_inv \<mu>)"
    unfolding h\<^sub>s_def by auto
  let ?\<mu> = "h\<^sub>s_inv \<mu>"
  from \<mu> less inv have "(h (h\<^sup>-\<^sup>1 hr) \<cdot> (h\<^sub>s ?\<mu>), h (h\<^sup>-\<^sup>1 hl) \<cdot> (h\<^sub>s ?\<mu>)) \<notin> {\<succ>\<^sub>h}" by auto
  with subst_apply_h have "(h (h\<^sup>-\<^sup>1 hr \<cdot> ?\<mu>), h (h\<^sup>-\<^sup>1 hl \<cdot> ?\<mu>)) \<notin> {\<succ>\<^sub>h}" by auto
  with compat_h have 5:"((h\<^sup>-\<^sup>1 hr) \<cdot> ?\<mu>, (h\<^sup>-\<^sup>1 hl) \<cdot> ?\<mu>) \<notin> {\<succ>}" by fast
  from \<mu> less' inv have "(h (h\<^sup>-\<^sup>1 hr') \<cdot> (h\<^sub>s ?\<mu>), h (h\<^sup>-\<^sup>1 hl') \<cdot> (h\<^sub>s ?\<mu>)) \<notin> {\<succ>\<^sub>h}" by auto
  with subst_apply_h have "(h (h\<^sup>-\<^sup>1 hr' \<cdot> ?\<mu>), h (h\<^sup>-\<^sup>1 hl' \<cdot> ?\<mu>)) \<notin> {\<succ>\<^sub>h}" by auto
  with compat_h have 6:"((h\<^sup>-\<^sup>1 hr') \<cdot> ?\<mu>, (h\<^sup>-\<^sup>1 hl') \<cdot> ?\<mu>) \<notin> {\<succ>}" by fast
  let ?C = "ctxt_of_pos_term pos"
  from cs inv \<mu> have "hs = (ctxt_of_pos_term pos (h (h\<^sup>-\<^sup>1 hl') \<cdot> h\<^sub>s ?\<mu>))\<langle>h (h\<^sup>-\<^sup>1 hr) \<cdot> h\<^sub>s ?\<mu>\<rangle>" by auto
  with subst_apply_h have hs:"hs = (?C (h (h\<^sup>-\<^sup>1 hl' \<cdot> ?\<mu>)))\<langle>h (h\<^sup>-\<^sup>1 hr \<cdot> ?\<mu>)\<rangle>" by auto
      
  from ctxt_of_pos_term_map_funs_term_conv funposs_imp_poss[OF 3]
    have "?C (h (h\<^sup>-\<^sup>1 hl' \<cdot> ?\<mu>)) = h\<^sub>C (?C (h\<^sup>-\<^sup>1 hl' \<cdot> ?\<mu>))"
    unfolding h_def h\<^sub>C_def by auto
  with ctxt_apply_h have hs:"hs = h ((?C (h\<^sup>-\<^sup>1 hl' \<cdot> ?\<mu>))\<langle>h\<^sup>-\<^sup>1 hr \<cdot> ?\<mu>\<rangle>)" unfolding hs by auto
  note inv_subst_apply_h = map_funs_subst_distrib[of "inv h\<^sub>F", unfolded h_inv_def[symmetric]]
  note inv_ctxt_apply_h = map_funs_term_ctxt_distrib[of "inv h\<^sub>F", unfolded h_inv_def[symmetric]]
  from funposs_imp_poss[OF 3] have pos':"pos \<in> poss (hl' \<cdot> \<mu>)" unfolding poss_inv_h by auto
  from hs have "hs = h (h\<^sup>-\<^sup>1 (?C (hl' \<cdot> \<mu>))\<langle>hr \<cdot> \<mu>\<rangle>)"
    unfolding inv_subst_apply_h[symmetric] unfolding h_inv_def
    ctxt_of_pos_term_map_funs_term_conv[OF pos', symmetric]
    inv_ctxt_apply_h[symmetric, unfolded h_inv_def] by simp
  with cs have hs':"h (h\<^sup>-\<^sup>1 hs) = hs" by auto
  have 7:"h\<^sup>-\<^sup>1 hs = (ctxt_of_pos_term pos (h\<^sup>-\<^sup>1 hl' \<cdot> h\<^sub>s\<^sup>-\<^sup>1 \<mu>))\<langle>h\<^sup>-\<^sup>1 hr \<cdot> h\<^sub>s\<^sup>-\<^sup>1 \<mu>\<rangle>" unfolding hs inv_h_h by auto
  
  from ht_mu \<mu> inv(4) have "h (h\<^sup>-\<^sup>1 ht) = h (h\<^sup>-\<^sup>1 (h (h\<^sup>-\<^sup>1 hr') \<cdot> h\<^sub>s (h\<^sub>s_inv \<mu>)))" by auto
  with subst_apply_h have "h (h\<^sup>-\<^sup>1 ht) = h (h\<^sup>-\<^sup>1 (h (h\<^sup>-\<^sup>1 hr' \<cdot> h\<^sub>s_inv \<mu>)))" by auto
  hence "h (h\<^sup>-\<^sup>1 ht) = h (h\<^sup>-\<^sup>1 hr') \<cdot> (h\<^sub>s (h\<^sub>s_inv \<mu>))" unfolding inv_h_h unfolding subst_apply_h by auto
  with inv(4) \<mu> ht_mu have ht':"h (h\<^sup>-\<^sup>1 ht) = ht" by auto
  hence "?t = h\<^sup>-\<^sup>1 (hr' \<cdot> \<mu>)" unfolding ht_mu using inv_h_h by auto
  hence 8:"?t = h\<^sup>-\<^sup>1 hr' \<cdot> ?\<mu>" unfolding inv_subst_apply_h by blast   
  from 1 2 3 4 5 6 7 8 have o:"ooverlap {\<succ>} (R\<^sub>\<omega> \<union> (E\<^sub>\<omega>)\<^sup>\<leftrightarrow>) (h\<^sup>-\<^sup>1 hl, h\<^sup>-\<^sup>1 hr) (h\<^sup>-\<^sup>1 hl', h\<^sup>-\<^sup>1 hr') pos ?\<mu> ?s ?t"
    unfolding ooverlap_def fst_conv snd_conv by argo
  have nf:"\<forall>u \<lhd> (h\<^sup>-\<^sup>1 hl) \<cdot> ?\<mu>. u \<in> NF (ordstep {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))"
  proof(rule, rule)
    fix u
    assume "(h\<^sup>-\<^sup>1 hl) \<cdot> ?\<mu> \<rhd> u"
    with subterm_strict_h have "h ((h\<^sup>-\<^sup>1 hl) \<cdot> ?\<mu>) \<rhd> h u" by metis
    with inv(1) \<mu> subst_apply_h have a:"hl \<cdot> \<mu> \<rhd> h u" by auto
    { fix v
      assume "(u,v) \<in> ordstep {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
      hence False proof(cases)
        case (1 l r C \<sigma>)
        from 1(1) have "(h l, h r) \<in> h\<^sub>R (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
          unfolding h\<^sub>R_def h_def map_funs_trs.simps map_funs_rule.simps by force
        with hR_union hR_sym have hlr:"(h l, h r) \<in> h\<^sub>R R\<^sub>\<omega> \<union> (h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>" by blast
        from 1(2) subst_apply_h ctxt_apply_h have 2:"h u = (h\<^sub>C C)\<langle>h l \<cdot> h\<^sub>s \<sigma>\<rangle>" by auto
        from 1(3) subst_apply_h ctxt_apply_h have 3:"h v = (h\<^sub>C C)\<langle>h r \<cdot> h\<^sub>s \<sigma>\<rangle>" by auto
        from 1(4) compat_h subst_apply_h have 4:"(h l \<cdot> h\<^sub>s \<sigma>, h r \<cdot> h\<^sub>s \<sigma>) \<in> {\<succ>\<^sub>h}" by force
        from hlr 2 3 4 have "(h u,h v) \<in> ordstep {\<succ>\<^sub>h} (h\<^sub>R R\<^sub>\<omega> \<union> (h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>)"
          unfolding ordstep.simps by blast
        with nf a show False unfolding hrl fst_conv by auto
      qed
    }
    thus "u \<in> NF (ordstep {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))" by fast
  qed
  from o nf have "(?s,?t) \<in> PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" unfolding PCP_ext_def by force
  with hs' ht' show ?thesis by auto
qed
  
end


datatype ('a,'b) f_ext = FOrig 'a | FFresh 'b
  
text \<open>Towards completeness of ordered completion.\<close>
locale gtotal_okb_irun2 = gtotal_okb_irun_inf R E less
  for R :: "nat \<Rightarrow> (('a, 'b::infinite) Term.term \<times> ('a, 'b) Term.term) set" 
  and E :: "nat \<Rightarrow> (('a, 'b) Term.term \<times> ('a, 'b) Term.term) set" 
  and less :: "('a, 'b) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
  +
  fixes "c"::'a
  assumes c_min:"\<forall>(t::('a, 'b) Term.term). t \<succeq> (Fun c [])"
begin

text \<open>Replace variables in a term by fresh function symbols.\<close>
  
sublocale sk: inj_homomorphism FOrig
  by (unfold_locales, insert inj_on_def, auto)

definition "\<sigma>\<^sub>s\<^sub>k" where "\<sigma>\<^sub>s\<^sub>k \<equiv> (\<lambda>x. Fun (FFresh x) [])"
    
definition sk :: "('a, 'b) term \<Rightarrow> (('a, 'b)f_ext, 'b) term"
  where "sk t = (sk.h t) \<cdot> \<sigma>\<^sub>s\<^sub>k"
  
lemma ground_sk [simp]: "ground (sk t)"
  by (induct t, insert sk_def sk.h_def, unfold \<sigma>\<^sub>s\<^sub>k_def, auto)
  
definition orig_sig :: "(('a,'b) f_ext, 'b) term \<Rightarrow> bool"
  where "orig_sig t = (\<forall>x. FFresh x \<notin> funs_term t)"

definition orig_sig\<^sub>C :: "(('a,'b) f_ext, 'b) ctxt \<Rightarrow> bool"
   where "orig_sig\<^sub>C t = (\<forall>x. FFresh x \<notin> funs_ctxt t)"
  
lemma inj_on_sigma_sk: "inj_on (\<lambda> t. t \<cdot> \<sigma>\<^sub>s\<^sub>k) {t|t. orig_sig t}"
proof-
  { fix s::"(('a,'b) f_ext, 'b) term"
    fix t
    assume st:"s \<noteq> t" and s_dom:"\<forall>x. FFresh x \<notin> funs_term s" and t_dom:"\<forall>x. FFresh x \<notin> funs_term t"
    hence "s \<cdot> \<sigma>\<^sub>s\<^sub>k \<noteq> t \<cdot> \<sigma>\<^sub>s\<^sub>k" proof(induct s arbitrary:t)
      case (Var x)
      thus ?case using subst_apply_term.simps unfolding \<sigma>\<^sub>s\<^sub>k_def by (cases t, auto)
    next
      case (Fun f ss)
      note Funs = this
      thus ?case proof(cases t)
        case (Var x)
        from Funs show ?thesis unfolding Var subst_apply_term.simps \<sigma>\<^sub>s\<^sub>k_def by auto
      next
        case (Fun g ts)
        then show ?thesis proof(cases "f=g")
          case True
          from Funs[unfolded Fun True] have "ss \<noteq> ts" by auto
          hence diff:"(\<exists>i < length ss. ss ! i \<noteq> ts ! i) \<or> (length ss \<noteq> length ts)"
            using nth_equalityI by blast
          then show ?thesis proof(cases "length ss \<noteq> length ts")
            case True
             then show ?thesis unfolding Fun subst_apply_term.simps using map_eq_imp_length_eq by blast
          next
            case False
            with diff obtain i where i:"i < length ss" "ss ! i \<noteq> ts ! i" by auto
            from i have in_set:"ss ! i \<in> set ss" by auto
            from i False have in_set2:"ts ! i \<in> set ts" by auto
            with in_set Funs(3) Funs(4)
              have all:"\<forall>x. FFresh x \<notin> funs_term (ss ! i)" " \<forall>x. FFresh x \<notin> funs_term (ts ! i)"
              unfolding Fun by auto
            from Funs(1)[OF in_set i(2) all]
             i show ?thesis unfolding Fun using map_nth_conv by fastforce
          qed
        next
          case False
          with Funs show ?thesis unfolding Fun by auto
        qed
      qed
    qed
  }
  thus ?thesis unfolding inj_on_def orig_sig_def by auto
qed
    
definition ks where "ks = the_inv_into {t|t. orig_sig t} (\<lambda> t. t \<cdot> \<sigma>\<^sub>s\<^sub>k)"
  
lemma ks_sk:
  assumes "orig_sig s"
  shows "ks (s \<cdot> \<sigma>\<^sub>s\<^sub>k) = s"
 using inj_on_sigma_sk[THEN the_inv_into_f_f, of s] assms unfolding ks_def orig_sig_def by auto
    
lemma sk_ks_sk:
  assumes "orig_sig s"
  shows "(ks (s \<cdot> \<sigma>\<^sub>s\<^sub>k)) \<cdot> \<sigma>\<^sub>s\<^sub>k = s \<cdot> \<sigma>\<^sub>s\<^sub>k"
  using assms ks_sk by metis
    
lemma orig_sig_h:"orig_sig (sk.h t)"
  unfolding orig_sig_def sk.h_def using funs_term_map_funs_term[of FOrig t]
  by blast

lemma ctxt_sk_image:
  fixes t::"(('a, 'b) f_ext, 'b) term"
  fixes u\<^sub>s::"(('a, 'b) f_ext, 'b) term"
  assumes "orig_sig t" and "t \<cdot> \<sigma>\<^sub>s\<^sub>k = C\<^sub>s\<langle>u\<^sub>s\<rangle>"
  shows "\<exists> C u. C\<^sub>s = C \<cdot>\<^sub>c \<sigma>\<^sub>s\<^sub>k \<and> u\<^sub>s = u \<cdot> \<sigma>\<^sub>s\<^sub>k \<and> orig_sig u \<and> orig_sig\<^sub>C C"
using assms
proof (induct C\<^sub>s arbitrary: t)
  case Hole
  show ?case unfolding orig_sig\<^sub>C_def
    by (rule exI[of _ Hole], rule exI[of _ t], insert Hole, auto)
next
  case (More g bef C aft)
  let ?s = "\<lambda>ti. ti \<cdot> \<sigma>\<^sub>s\<^sub>k"
  from More(3) obtain f ts where t: "t = Fun f ts" unfolding \<sigma>\<^sub>s\<^sub>k_def by (cases t, auto)
  with More(3) have "Fun f (map ?s ts) = (More g bef C aft)\<langle>u\<^sub>s\<rangle>" by auto
  hence fg:"f=g" and ts:"map ?s ts = (bef @ C\<langle>u\<^sub>s\<rangle> # aft)" (is "_ = ?bca") by auto
  hence len:"length ts = length ?bca" using length_map by metis
  let ?i = "length bef"
  from len have i: "?i < length ?bca" by auto
  with ts have ti:"ts ! ?i \<cdot> \<sigma>\<^sub>s\<^sub>k = C\<langle>u\<^sub>s\<rangle>" unfolding map_nth_eq_conv[OF len] by auto
  from More(2)[unfolded t] have os:"\<And>ti. ti \<in> set ts \<Longrightarrow> orig_sig ti" unfolding orig_sig_def by auto
  {fix i 
    assume i:"i < length bef"
    with len os have sig: "orig_sig (ts ! i)" by auto
    from i ts have bi:"bef ! i = ?s (ts ! i)" unfolding map_nth_eq_conv[OF len, of ?s] length_append
      using append_Cons_nth_left by fastforce
    from sig ks_sk[OF sig] have "orig_sig (ks (bef ! i))" unfolding bi by metis
    with sk_ks_sk[OF sig] have "(?s \<circ> ks) (bef ! i) = bef ! i \<and> orig_sig (ks (bef ! i))"
      unfolding bi by auto
  } note bi = this
  hence mb:"map (?s \<circ> ks) bef = bef" by (simp add: map_nth_eq_conv)
  from bi have bef_sig:"\<And>bi. bi \<in> set bef \<Longrightarrow> orig_sig (ks bi)" by (metis in_set_conv_nth)
  {fix j 
    assume j:"j < length aft"
    let ?j = "?i + (Suc j)"
    from j len os have sig: "orig_sig (ts ! ?j)" by auto
    from j len have "?j < length ts" by auto
    from ts nth_map[OF this] have "?s (ts ! ?j) = ?bca ! ?j" by metis
    with j have ai:"?s (ts ! ?j) = aft ! j" unfolding nth_append_length_plus by force
    from sig ks_sk[OF sig] have sig':"orig_sig (ks (aft ! j))" unfolding ai[symmetric] by metis
    from sk_ks_sk[OF sig] have "(?s \<circ> ks) (aft ! j) = aft ! j" unfolding ai[symmetric] by force
    with sig' have "(?s \<circ> ks) (aft ! j) = aft ! j \<and> orig_sig (ks (aft ! j))" by auto
  } note aj = this
  hence ma:"map (?s \<circ> ks) aft = aft" by (simp add: map_nth_eq_conv)
  from aj have aft_sig:"\<And>aj. aj \<in> set aft \<Longrightarrow> orig_sig (ks aj)" by (metis in_set_conv_nth)
  have j: "Suc ?i \<le> length ?bca" by auto
  from ti os i len have "orig_sig (ts ! ?i)" by force
  from More(1)[OF this ti] obtain D u where
    D:"C = D \<cdot>\<^sub>c \<sigma>\<^sub>s\<^sub>k" "orig_sig\<^sub>C D" and u:"u\<^sub>s = u \<cdot> \<sigma>\<^sub>s\<^sub>k" "orig_sig u" by auto
  let ?C = "More g (map ks bef) D (map ks aft)"
  from t More(2) fg have "\<forall>x. FFresh x \<noteq> g" unfolding orig_sig_def by auto
  with D(2) bef_sig aft_sig have sig:"orig_sig\<^sub>C ?C"
    unfolding orig_sig\<^sub>C_def orig_sig_def ctxt.set(2) by auto
  have "More g bef C aft = ?C \<cdot>\<^sub>c \<sigma>\<^sub>s\<^sub>k"
    unfolding subst_apply_ctxt.simps D(1)[symmetric] map_map mb ma by simp
  with u sig show ?case by blast
qed
  
lemma subst_sk_image:
  assumes "orig_sig s" and "orig_sig l" and "s \<cdot> \<sigma>\<^sub>s\<^sub>k = l \<cdot> \<tau>"
  shows "\<forall>x \<in> vars_term l. \<tau> x = (ks (\<tau> x)) \<cdot> \<sigma>\<^sub>s\<^sub>k \<and> orig_sig (ks (\<tau> x))"
  using assms
proof(induct l arbitrary:s)
  case (Var x)
  hence tx:"\<tau> x = s \<cdot> \<sigma>\<^sub>s\<^sub>k" by auto
  from Var(1) have sig:"orig_sig (ks (\<tau> x))" unfolding tx ks_sk[OF Var(1)] by auto
  have "(ks (\<tau> x)) \<cdot> \<sigma>\<^sub>s\<^sub>k = \<tau> x" using sk_ks_sk[OF Var(1)] unfolding orig_sig_def tx by auto
  with sig show ?case by auto
next
  case (Fun f ls)
  note Funl = this
  hence eq:"s \<cdot> \<sigma>\<^sub>s\<^sub>k = Fun f (map (\<lambda>ti. ti \<cdot> \<tau>) ls)" by auto
  show ?case proof(cases s)
    case (Var x)
    from eq[unfolded Var] have "f = FFresh x" unfolding \<sigma>\<^sub>s\<^sub>k_def by auto
    with Fun(3) show ?thesis unfolding orig_sig_def by auto
  next
    case (Fun g ss)
    from eq[unfolded Fun subst_apply_term.simps] have
      fg:"f=g" "map (\<lambda>t. t \<cdot> \<sigma>\<^sub>s\<^sub>k) ss = map (\<lambda>ti. ti \<cdot> \<tau>) ls" by auto
    hence len:"length ss = length (map (\<lambda>ti. ti \<cdot> \<tau>) ls)" using length_map by metis
    { fix i
      assume i:"i < length ls"
      with Funl orig_sig_def have li_sig:"orig_sig (ls ! i)" by force
      from i Funl(2) len have si_sig:"orig_sig (ss ! i)" unfolding Fun orig_sig_def by force
      from fg(2) map_nth_eq_conv[OF len] i have "(ss ! i) \<cdot> \<sigma>\<^sub>s\<^sub>k = (ls ! i) \<cdot> \<tau>" by force
      from Funl(1)[OF _ si_sig li_sig this] i have
        "\<forall>x\<in>vars_term (ls ! i). \<tau> x = ks (\<tau> x) \<cdot> \<sigma>\<^sub>s\<^sub>k \<and> orig_sig (ks (\<tau> x))" by auto
    }
    thus ?thesis using var_imp_var_of_arg[of _ f ls] by force
  qed
qed
  
lemma orig_sig_ctxt_apply:"orig_sig\<^sub>C C \<Longrightarrow> orig_sig t \<Longrightarrow> orig_sig C\<langle>t\<rangle>"
  unfolding orig_sig_def orig_sig\<^sub>C_def by (induct C, auto)
    
lemma orig_sig_subst_apply:"orig_sig t \<Longrightarrow> \<forall>x\<in>vars_term t. orig_sig (\<sigma> x) \<Longrightarrow> orig_sig (t \<cdot> \<sigma>)"
  unfolding orig_sig_def orig_sig\<^sub>C_def by (induct t, auto)
   
lemma sk_subst_ctxt_split1:
  fixes l::" (('a,'b)f_ext,'b) term"
  assumes "orig_sig l" and "sk s = C\<langle>l \<cdot> \<sigma>\<rangle>"
  shows "\<exists> D \<tau>. sk.h s = D\<langle>l \<cdot> \<tau>\<rangle>"
proof-
  from assms(2)[unfolded sk_def, THEN ctxt_sk_image[OF orig_sig_h]] obtain D u where
    Du:"l \<cdot> \<sigma> = u \<cdot> \<sigma>\<^sub>s\<^sub>k" "orig_sig u" "orig_sig\<^sub>C D" "C = D \<cdot>\<^sub>c \<sigma>\<^sub>s\<^sub>k" by auto
  define \<tau> where "\<tau> = (\<lambda>x. ks (\<sigma> x))"
  from subst_sk_image[OF Du(2) assms(1) Du(1)[symmetric]] have
    vs:"\<forall>x\<in>vars_term l. \<sigma> x = ks (\<sigma> x) \<cdot> \<sigma>\<^sub>s\<^sub>k" "\<forall>x\<in>vars_term l. orig_sig (ks (\<sigma> x))" by auto
  from this(1) term_subst_eq_conv[of l \<sigma> "(\<tau> \<circ>\<^sub>s \<sigma>\<^sub>s\<^sub>k)"] have eq:"l \<cdot> \<sigma> = l \<cdot> \<tau> \<cdot> \<sigma>\<^sub>s\<^sub>k"
    unfolding \<tau>_def subst_subst_compose subst_compose by auto
  note sig = orig_sig_subst_apply[OF assms(1) vs(2),THEN orig_sig_ctxt_apply[OF Du(3)]]
  note eq = assms[unfolded sk_def Du(4) eq subst_apply_term_ctxt_apply_distrib[symmetric]]
  from eq sig inj_on_sigma_sk[unfolded inj_on_def] orig_sig_h have
    eq:"sk.h s = D\<langle>l \<cdot> \<tau>\<rangle>" unfolding \<tau>_def by fast
  show ?thesis by (rule exI[of _ D], rule exI[of _ \<tau>], insert eq, auto)
qed
  
lemma sk_subst_ctxt_split2:
  fixes l::" ('a,'b) term"
  assumes "sk s = C\<langle>sk.h l \<cdot> \<sigma>\<rangle>"
  shows "\<exists> (D::('a,'b) ctxt) \<tau>. s = D\<langle>l \<cdot> \<tau>\<rangle>"
proof-
  from sk_subst_ctxt_split1[OF orig_sig_h assms] obtain D\<^sub>h :: "(('a,'b)f_ext,'b) ctxt" and \<tau> where
    skh_s:"sk.h s = D\<^sub>h\<langle>sk.h l \<cdot> \<tau>\<rangle>" by auto
  from sk.ctxt_subst_image[OF this] show ?thesis by auto
qed
  
lemma rstep_sk:"(s,t) \<in> rstep RR = ((sk s, sk t) \<in> rstep (sk.h\<^sub>R RR))"
proof
  assume "(s, t) \<in> rstep RR"
  from this sk.rstep_h[of s t RR] rstep.subst show "(sk s, sk t) \<in> rstep (sk.h\<^sub>R RR)"
    unfolding sk_def by blast
next
  assume a:"(sk s, sk t) \<in> rstep (sk.h\<^sub>R RR)"
  then obtain C l r \<sigma> where st:"sk s = C\<langle>l \<cdot> \<sigma>\<rangle>" "sk t = C\<langle>r \<cdot> \<sigma>\<rangle>" and lr:"(l,r) \<in> sk.h\<^sub>R RR"
    unfolding sk_def by auto
  hence h:"(sk.h s) \<cdot> \<sigma>\<^sub>s\<^sub>k = C\<langle>l \<cdot> \<sigma>\<rangle>" "(sk.h t) \<cdot> \<sigma>\<^sub>s\<^sub>k = C\<langle>r \<cdot> \<sigma>\<rangle>" unfolding sk_def by (argo,argo)
  from lr orig_sig_h have sig_l:"orig_sig l" unfolding sk.h\<^sub>R_def map_funs_trs.simps by force
  from lr orig_sig_h have sig_r:"orig_sig r" unfolding sk.h\<^sub>R_def map_funs_trs.simps by force
  let ?\<sigma> = "\<lambda>x. ks (\<sigma> x)"
  from ctxt_sk_image[OF orig_sig_h h(1)] obtain D and u::"(('a, 'b) f_ext, 'b) term"
    where C:"C = D \<cdot>\<^sub>c \<sigma>\<^sub>s\<^sub>k" "orig_sig\<^sub>C D" and u:"l \<cdot> \<sigma> = u \<cdot> \<sigma>\<^sub>s\<^sub>k" "orig_sig u" by blast
  note conv = term_subst_eq_conv[of _ \<sigma> "?\<sigma> \<circ>\<^sub>s \<sigma>\<^sub>s\<^sub>k", unfolded subst_compose]
  from subst_sk_image[OF u(2) sig_l u(1)[symmetric]] have
    vl:"\<forall>x\<in>vars_term l. \<sigma> x = ks (\<sigma> x) \<cdot> \<sigma>\<^sub>s\<^sub>k" "\<forall>x\<in>vars_term l. orig_sig (ks (\<sigma> x))" by auto
  from this(1) have "l \<cdot> \<sigma> = l \<cdot> (?\<sigma> \<circ>\<^sub>s \<sigma>\<^sub>s\<^sub>k)" unfolding conv[symmetric] subst_compose by auto
  hence l:"l \<cdot> \<sigma> = (l \<cdot> ?\<sigma>) \<cdot> \<sigma>\<^sub>s\<^sub>k" using subst_compose by simp
  from h ctxt_sk_image[OF orig_sig_h] obtain v::"(('a, 'b) f_ext, 'b) term"
    where v:"r \<cdot> \<sigma> = v \<cdot> \<sigma>\<^sub>s\<^sub>k" "orig_sig v" by blast
  from subst_sk_image[OF v(2) sig_r v(1)[symmetric]] have
    vr:"\<forall>x\<in>vars_term r. \<sigma> x = ks (\<sigma> x) \<cdot> \<sigma>\<^sub>s\<^sub>k" "\<forall>x\<in>vars_term r. orig_sig (ks (\<sigma> x))" by auto
  hence "r \<cdot> \<sigma> = r \<cdot> (?\<sigma> \<circ>\<^sub>s \<sigma>\<^sub>s\<^sub>k)" unfolding conv[symmetric] subst_compose by simp
  hence r:"r\<cdot> \<sigma> = (r \<cdot> ?\<sigma>) \<cdot> \<sigma>\<^sub>s\<^sub>k" using subst_compose by simp
  from orig_sig_subst_apply[OF sig_l vl(2)] have "orig_sig (l \<cdot> (\<lambda>x. ks (\<sigma> x)))" .
  with orig_sig_ctxt_apply C(2) have sig_Dl:"orig_sig (D\<langle>l \<cdot> (\<lambda>x. ks (\<sigma> x))\<rangle>)" by auto
  from orig_sig_subst_apply[OF sig_r vr(2)] have "orig_sig (r \<cdot> (\<lambda>x. ks (\<sigma> x)))" by auto 
  with orig_sig_ctxt_apply C(2) have sig_Dr:"orig_sig (D\<langle>r \<cdot> (\<lambda>x. ks (\<sigma> x))\<rangle>)" by auto 
  note eqs = h[unfolded C l r subst_apply_term_ctxt_apply_distrib[symmetric]]
  from eqs inj_on_sigma_sk sig_Dl sig_Dr orig_sig_h have
    "sk.h s = D\<langle>l \<cdot> (\<lambda>x. ks (\<sigma> x))\<rangle>" "sk.h t = D\<langle>r \<cdot> (\<lambda>x. ks (\<sigma> x))\<rangle>"
    unfolding inj_on_def by (blast, blast)
  from rstepI[OF lr this] sk.rstep_h show "(s, t) \<in> rstep RR" by auto
qed
    
lemma sk_inj:"sk s = sk t \<Longrightarrow> s = t"
  using sk.inj inj_on_sigma_sk orig_sig_h unfolding sk_def inj_on_def by blast

definition sk_redord :: "((('a,'b) f_ext, 'b) term \<Rightarrow> (('a,'b) f_ext, 'b) term \<Rightarrow> bool) \<Rightarrow> bool"
  where "sk_redord Rel \<equiv>
  ((\<forall> s t. s \<succ> t \<longrightarrow> Rel (sk.h s) (sk.h t)) \<and>
   (\<forall> s t. ground s \<and> ground t \<longrightarrow> s = t \<or> Rel s t \<or> Rel t s) \<and>
   SN {(x, y). Rel x y} \<and>
   (\<forall> s t C. Rel s t \<longrightarrow> Rel C\<langle>s\<rangle> C\<langle>t\<rangle>) \<and>
   (\<forall> s t \<sigma>. Rel s t \<longrightarrow> Rel (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)) \<and>
   (\<forall> s t u. Rel s t \<longrightarrow> Rel t u \<longrightarrow> Rel s u) \<and>
   (\<forall> t. ground t \<longrightarrow> Rel\<^sup>=\<^sup>= t (sk.h (Fun c []))))"
  
text \<open>Assume a ground total extension of \<succ> on image of homomorphism terms .\<close>
  
definition prec\<^sub>o :: "(('a \<times> nat) \<times> ('a \<times> nat)) set" where
  "prec\<^sub>o \<equiv> (SOME Rel. well_order_on (UNIV - {(c,0)}) Rel) \<union> {((c,0),fn) |fn. True}"
  
definition prec\<^sub>f :: "(('b \<times> nat) \<times> ('b \<times> nat)) set" where
  "prec\<^sub>f \<equiv> (SOME Rel. well_order Rel)"
  
definition prec_set where
 "prec_set \<equiv> {((FOrig f, n), (FFresh g, m)) |f g n m. True} \<union>
             {((FFresh f, n), (FFresh g, m)) |f g n m. ((f,n),(g,m)) \<in> prec\<^sub>f - Id} \<union>
             {((FOrig f, n), (FOrig g, m)) |f g n m. ((f,n),(g,m)) \<in> prec\<^sub>o - Id}"

definition prec\<^sub>s\<^sub>k where "prec\<^sub>s\<^sub>k gm fn \<equiv> (fn, gm) \<in> prec_set"

lemma well_order_precf:"well_order prec\<^sub>f"
proof-
  from Zorn.well_order_on obtain r ::"(('b \<times> nat) \<times> ('b \<times> nat)) set" where r:"well_order r" by blast
  have "\<exists>Rel :: (('b \<times> nat) \<times> ('b \<times> nat)) set. well_order Rel" unfolding prec\<^sub>f_def
    by (rule exI[of _r], insert r, auto)
  from someI_ex[OF this] show ?thesis using prec\<^sub>f_def by auto
qed

lemma well_order_preco:"well_order prec\<^sub>o"
proof-  
  define R ::"(('a \<times> nat) \<times> ('a \<times> nat)) set" where "R \<equiv> SOME Rel. well_order_on (UNIV - {(c,0)}) Rel"
  have p:"prec\<^sub>o = R \<union> {((c,0),fn) |fn. True}" unfolding prec\<^sub>o_def R_def by auto
      
  from Zorn.well_order_on obtain r ::"(('a \<times> nat) \<times> ('a \<times> nat)) set" where r:"well_order_on (UNIV - {(c,0)}) r" by blast
  have ex:"\<exists>Rel ::(('a \<times> nat) \<times> ('a \<times> nat)) set. well_order_on (UNIV - {(c,0)}) Rel"
    by (rule exI[of _ r], insert r, auto)
  from someI_ex[OF ex] have x:"well_order_on (UNIV - {(c,0)}) R" unfolding R_def by auto
  hence wf1:"wf (R - Id)" (is "wf ?R1") unfolding well_order_on_def by auto 
  have wf0:"wf ({((c,0),fn) |fn. True} - Id)" (is "wf ?R0") unfolding wf_def
    by (smt CollectD DiffD1 DiffD2 Pair_inject pair_in_Id_conv)
  from well_order_on_Field[OF x, unfolded Field_def]
    have "Domain ({((c,0),fn) |fn. True} - Id) \<inter> Range (R - Id) = {}" by auto
  from wf_Un[OF wf0 wf1 this] have "wf (({((c,0),fn) |fn. True} - Id) \<union> (R - Id))" by metis
  hence wf:"wf (prec\<^sub>o - Id)" unfolding p Un_Diff using Un_commute by metis
  
  let ?U = "UNIV - {(c,0)}"
  from well_order_on_Field[OF x, unfolded Field_def] have c:"(c,0) \<notin> Range R" unfolding R_def by auto
  from x have "linear_order_on?U R" unfolding well_order_on_def by auto
  hence po:"partial_order_on ?U R" and total:"Relation.total_on ?U R"
    unfolding linear_order_on_def by auto
  from po[unfolded partial_order_on_def] have pre:"preorder_on ?U R" and antisym:"antisym R" by auto
  hence refl:"refl_on ?U R" and trans:"trans R" unfolding preorder_on_def by auto
  have total:"total prec\<^sub>o"
  proof
    fix f g :: "('a \<times> nat)"
    assume "f \<noteq> g"
    with total[unfolded Relation.total_on_def] show "(f, g) \<in> prec\<^sub>o \<or> (g, f) \<in> prec\<^sub>o"
      unfolding p by (cases "f = (c,0)", auto)
  qed
  from refl have refl:"refl prec\<^sub>o" unfolding p refl_on_def by auto
  from trans c have trans:"trans prec\<^sub>o" unfolding p trans_def by blast
  from refl trans have pre:"preorder_on UNIV prec\<^sub>o" unfolding preorder_on_def by auto
  have antisym:"antisym prec\<^sub>o" proof
    fix f g :: "('a \<times> nat)"
    assume "(f, g) \<in> prec\<^sub>o" and "(g, f) \<in> prec\<^sub>o"
    with antisym[unfolded antisym_def] c show "f = g" unfolding p by blast
  qed
  from pre antisym have partial:"partial_order_on UNIV prec\<^sub>o" unfolding partial_order_on_def by auto
  from partial total have "linear_order prec\<^sub>o" unfolding linear_order_on_def by auto
  with wf show ?thesis unfolding well_order_on_def by auto
qed
  
lemma SN_prec:"SN {(gm,fn) |fn gm. prec\<^sub>s\<^sub>k gm fn }"
proof-
  let ?R\<^sub>f = "{((FFresh f, n), (FFresh g, m)) |f g n m. ((f,n),(g,m)) \<in> prec\<^sub>f - Id}"
  let ?R\<^sub>o = "{((FOrig f, n), (FOrig g, m)) |f g n m. ((f,n),(g,m)) \<in> prec\<^sub>o - Id}"
  let ?R\<^sub>o\<^sub>f = "{((FOrig f, n), (FFresh g, m)) |f g n m. True}"
  let ?F = "\<lambda>C. (\<lambda>(f,n).(C f, n))"
  have inj:"inj (?F FFresh)" unfolding inj_on_def by auto
  have aux:"\<And>C P.(map_prod (?F C) (?F C) ` (P - Id)) = {(?F C (f,n), ?F C (g,m)) |f g n m. ((f,n), (g,m)) \<in> P - Id}"
    unfolding split by force
  from well_order_precf have "wf (prec\<^sub>f - Id)" using well_order_on_def by auto
  from wf_map_prod_image[OF this inj] have wf\<^sub>f:"wf ?R\<^sub>f" unfolding aux[of FFresh] split by auto
      
  have inj:"inj (?F FOrig)" unfolding inj_on_def by auto
  from well_order_preco have "wf (prec\<^sub>o - Id)" unfolding well_order_on_def by auto
  from wf_map_prod_image[OF this inj] have wf\<^sub>o:"wf ?R\<^sub>o" unfolding aux by simp
      
  have wf:"wf ?R\<^sub>o\<^sub>f" by (smt Pair_inject f_ext.distinct(1) mem_Collect_eq wf_def)
  have "?R\<^sub>o\<^sub>f O ?R\<^sub>f \<subseteq> ?R\<^sub>o\<^sub>f" by blast
  from wf_union_compatible[OF wf wf\<^sub>f this] have wf:"wf (?R\<^sub>o\<^sub>f \<union> ?R\<^sub>f)" by auto
  from wf_union_compatible[OF this wf\<^sub>o] have wf:"wf ((?R\<^sub>o\<^sub>f \<union> ?R\<^sub>f) \<union> ?R\<^sub>o)" by blast
  hence "wf prec_set" unfolding prec_set_def by auto
  thus ?thesis unfolding prec\<^sub>s\<^sub>k_def SN_iff_wf converse_unfold mem_Collect_eq split by simp
qed
  
text \<open>Replace Skolem constants by minimal constant c\<close>
definition D\<^sub>c :: "(('a, 'b) f_ext, 'b) term \<Rightarrow> ('a, 'b) term"
  where "D\<^sub>c t \<equiv> map_funs_term (\<lambda>f. case f of FOrig f \<Rightarrow> f | FFresh z \<Rightarrow> c) t"
    
definition Dc\<^sub>c :: "(('a, 'b) f_ext, 'b) ctxt \<Rightarrow> ('a, 'b) ctxt"
  where "Dc\<^sub>c C \<equiv> map_funs_ctxt (\<lambda>f. case f of FOrig f \<Rightarrow> f | FFresh z \<Rightarrow> c) C"
    
definition Ds\<^sub>c :: "(('a, 'b) f_ext, 'b) subst \<Rightarrow> ('a, 'b) subst"
  where "Ds\<^sub>c \<sigma> \<equiv> map_funs_subst (\<lambda>f. case f of FOrig f \<Rightarrow> f | FFresh z \<Rightarrow> c) \<sigma>"
     
lemma Dc_h: "D\<^sub>c (sk.h t) = t" unfolding D\<^sub>c_def sk.h_def map_funs_term_comp
  by (simp add: funs_term_map_funs_term_id)
    
abbreviation kbo_sk :: "(('a, 'b) f_ext, 'b) Term.term \<Rightarrow> (('a, 'b) f_ext, 'b) Term.term \<Rightarrow> bool \<times> bool"
  where "kbo_sk \<equiv> weight_fun_prc.kbo (*admissible_weight_fun_prc.gtkbo*) (\<lambda>f.1) 1 (\<lambda>f n. 1) (\<lambda>f. f = FOrig c) prec\<^sub>s\<^sub>k prec\<^sub>s\<^sub>k\<^sup>=\<^sup>="

lemma irrefl_precsk:"\<forall>fn. \<not> (fn,fn) \<in> prec_set"
  using SN_prec unfolding refl_on_def  prec\<^sub>s\<^sub>k_def by fastforce
    
lemma trans_preceq_sk:
  assumes "fn = gm \<or> prec\<^sub>s\<^sub>k fn gm" and "gm = hk \<or> prec\<^sub>s\<^sub>k gm hk"
  shows "(fn = gm \<and> gm = hk) \<or> prec\<^sub>s\<^sub>k fn hk"
proof-
  obtain f n g m h k where fn:"fn = (f,n)" and gm:"gm = (g,m)" and hk:"hk = (h,k)" by fastforce
  from assms have a:"(f,n) = (g,m) \<or> prec\<^sub>s\<^sub>k (f,n) (g,m)" "(g,m) = (h,k) \<or> prec\<^sub>s\<^sub>k (g,m) (h,k)"
    unfolding fn gm hk by auto
  have "\<And>R. well_order R \<Longrightarrow> trans R"
    unfolding well_order_on_def linear_order_on_def partial_order_on_def preorder_on_def by auto
  note trans = this[OF well_order_precf] this[OF well_order_preco]
  { assume p:"((h,k), (g,m)) \<in> prec_set" "((g,m),(f,n)) \<in> prec_set" and neq:"(h,k) \<noteq> (f,n)"
    have "((h,k),(f,n)) \<in> prec_set" proof(cases h)
      case (FFresh a)
      from p[unfolded this] prec_set_def obtain b where g:"g = FFresh b" by auto
      from p[unfolded this] prec_set_def obtain c where f:"f = FFresh c" by auto
      from p[unfolded FFresh g f] prec_set_def have "((a,k),(b,m)) \<in> prec\<^sub>f" "((b,m),(c,n)) \<in> prec\<^sub>f" by auto
      note t = trans(1)[unfolded trans_def, rule_format, OF this]
      with neq show ?thesis unfolding FFresh f prec_set_def by auto
    next
      case (FOrig a)
      note h = this
      then show ?thesis proof(cases g)
        case (FFresh b)
        then show ?thesis using neq p unfolding prec_set_def FOrig FFresh by (cases h, auto)
      next
        case (FOrig b)
        show ?thesis using neq trans(2)[unfolded trans_def, rule_format, of "(a,k)" "(b,m)"] p
          unfolding prec_set_def h FOrig by (cases f, auto)
      qed
    qed
  }
  note trans = this
  show ?thesis
  proof (cases "(g, m) = (h, k)")
    case True
    from a show ?thesis unfolding fn gm hk prec\<^sub>s\<^sub>k_def split True by auto
  next
    case False
    with a have p:"((h, k), g, m) \<in> prec_set" unfolding prec\<^sub>s\<^sub>k_def by auto
    show ?thesis proof (cases "(f, n) = (g, m)")
      case True
      with a p show ?thesis unfolding fn gm hk prec\<^sub>s\<^sub>k_def by auto
    next
      case False
      from False a have p':"((g, m), f, n) \<in> prec_set" unfolding prec\<^sub>s\<^sub>k_def by auto
      { assume eq:"(h, k) = (f, n)"
        have aux:"{(gm, fn) |fn gm. (fn, gm) \<in> prec_set}\<inverse> = prec_set" by fast
        from p p' have "(fn,fn) \<in> ({(gm, fn) |fn gm. (fn, gm) \<in> prec_set}\<inverse>)\<^sup>+" unfolding fn eq prec\<^sub>s\<^sub>k_def aux by auto
        hence False using wf_acyclic[OF SN_imp_wf[OF SN_prec]] unfolding prec\<^sub>s\<^sub>k_def
          unfolding acyclic_def by blast
      }
      with p p' trans show ?thesis unfolding fn gm hk prec\<^sub>s\<^sub>k_def by fast
    qed
  qed
qed
 
lemma trans_prec_sk:
  assumes "prec\<^sub>s\<^sub>k fn gm" and "prec\<^sub>s\<^sub>k gm hk"
  shows "prec\<^sub>s\<^sub>k fn hk"
proof-
  from trans_preceq_sk assms have "(fn = gm \<and> gm = hk) \<or> prec\<^sub>s\<^sub>k fn hk" by presburger
  with assms irrefl_precsk show ?thesis unfolding refl_on_def by auto
qed
 
lemma antisym_prec_sk:
  assumes "prec\<^sub>s\<^sub>k fn gm"
  shows "\<not> prec\<^sub>s\<^sub>k gm fn"
using trans_prec_sk[OF assms, of fn] irrefl_precsk unfolding prec\<^sub>s\<^sub>k_def by auto
    
lemma adm:"admissible_weight_fun_prc (\<lambda>f.1) 1 prec\<^sub>s\<^sub>k prec\<^sub>s\<^sub>k\<^sup>=\<^sup>= (\<lambda>f. f = FOrig c) (\<lambda>f n. 1)"
proof-
  have min:"\<forall>g. g \<noteq> FOrig c \<longrightarrow> prec\<^sub>s\<^sub>k (g, 0) (FOrig c, 0)"
  proof(rule, rule)
    fix g :: "('a, 'b) f_ext"
    assume gc:"g \<noteq> FOrig c"
    have "\<And>a. a \<noteq> c \<Longrightarrow> ((FOrig c,0), (FOrig a,0)) \<in> prec_set" unfolding prec\<^sub>o_def prec_set_def by auto
    with gc show "prec\<^sub>s\<^sub>k (g, 0) (FOrig c, 0)" unfolding prec\<^sub>s\<^sub>k_def prec_set_def by (cases g, auto)
  qed
  { fix f
    assume a:"\<forall>g. g = f \<or> prec\<^sub>s\<^sub>k (g, 0) (f, 0)" "f \<noteq> FOrig c"
    hence p:"prec\<^sub>s\<^sub>k (FOrig c, 0) (f, 0)" by auto
    from min a(2) have p':"prec\<^sub>s\<^sub>k (f, 0) (FOrig c, 0)" by auto
    from trans_prec_sk[OF p p'] irrefl_precsk[unfolded refl_on_def] have False
      using SN_on_irrefl SN_prec unfolding prec\<^sub>s\<^sub>k_def by blast
  }
  with min irrefl_precsk[unfolded refl_on_def] have
    min:"\<And>f. (f = (FOrig c)) = (\<forall>g. g = f \<or> prec\<^sub>s\<^sub>k (g, 0) (f, 0))" by auto
  show ?thesis  
  proof(unfold_locales)
    fix fn gm hk
    from trans_preceq_sk show "prec\<^sub>s\<^sub>k\<^sup>=\<^sup>= fn gm \<Longrightarrow> prec\<^sub>s\<^sub>k\<^sup>=\<^sup>= gm hk \<Longrightarrow> prec\<^sub>s\<^sub>k\<^sup>=\<^sup>= fn hk" by fast
  next
    fix fn gm
    from antisym_prec_sk show "prec\<^sub>s\<^sub>k fn gm = strict prec\<^sub>s\<^sub>k\<^sup>=\<^sup>= fn gm" by blast
  next
    show "SN {(x, y). prec\<^sub>s\<^sub>k x y}" using SN_prec by fast
  next
  qed (insert min, auto)
qed
 
lemma precsk_gtotal:"fn = gm \<or> prec\<^sub>s\<^sub>k fn gm \<or> prec\<^sub>s\<^sub>k gm fn"
proof-
  { assume neq:"fn \<noteq> gm"
  obtain f g n m where p:"fn = (f,n)" "gm = (g,m)" by fastforce
  from well_order_preco well_order_precf have t:"total prec\<^sub>o" "total prec\<^sub>f"
    unfolding well_order_on_def linear_order_on_def by auto
  have "prec\<^sub>s\<^sub>k fn gm \<or> prec\<^sub>s\<^sub>k gm fn" proof(cases f)
    case (FOrig a)
    from t[unfolded Relation.total_on_def] neq show "prec\<^sub>s\<^sub>k fn gm \<or> prec\<^sub>s\<^sub>k gm fn"
      unfolding prec\<^sub>s\<^sub>k_def prec_set_def p FOrig by (cases g, auto)
  next
    case (FFresh a)
    note a = this
    show "prec\<^sub>s\<^sub>k fn gm \<or> prec\<^sub>s\<^sub>k gm fn" proof (cases g)
      case (FOrig b)
      show ?thesis unfolding prec\<^sub>s\<^sub>k_def prec_set_def p FFresh FOrig by force
    next
      case (FFresh b)
      from t(2)[unfolded Relation.total_on_def] neq show ?thesis  
        unfolding prec\<^sub>s\<^sub>k_def prec_set_def p FFresh a by auto
    qed
  qed
  }
  thus ?thesis by auto
qed

abbreviation lex_sk
  where "lex_sk \<equiv> lex_two {\<succ>} Id  {(s, t). fst (kbo_sk s t)}"
  
definition less_sk :: "(('a,'b) f_ext, 'b) term \<Rightarrow> (('a,'b) f_ext, 'b) term \<Rightarrow> bool" (infix "\<succ>\<^sub>s\<^sub>k" 50)
  where "less_sk s t = (((D\<^sub>c s, s), (D\<^sub>c t, t)) \<in> lex_sk)"
(*    
definition less_sk :: "(('a,'b) f_ext, 'b) term \<Rightarrow> (('a,'b) f_ext, 'b) term \<Rightarrow> bool" (infix "\<succ>\<^sub>s\<^sub>k" 50)
  where "less_sk s t = (D\<^sub>c s \<succ> D\<^sub>c t \<or> (D\<^sub>c s = D\<^sub>c t \<and> fst (kbo_sk s t)))"*)
  
abbreviation less_sk_set ("{\<succ>\<^sub>s\<^sub>k}") where "less_sk_set \<equiv> {(s, t). s \<succ>\<^sub>s\<^sub>k t}"

abbreviation lesseq_sk (infix "\<succeq>\<^sub>s\<^sub>k" 50) where "s \<succeq>\<^sub>s\<^sub>k t \<equiv> (op \<succ>\<^sub>s\<^sub>k)\<^sup>=\<^sup>= s t"

  
text \<open>There exists a ground total extension of \<succ> on skolemized terms.\<close>
lemma less_sk: "sk_redord less_sk"
proof-
  define \<sigma>\<^sub>c :: "'b \<Rightarrow> ('a, 'b) Term.term" where "\<sigma>\<^sub>c \<equiv> (\<lambda>x. Fun c [])"
  have 1:"\<forall>s t. s \<succ> t \<longrightarrow> sk.h s \<succ>\<^sub>s\<^sub>k sk.h t"
  proof(rule, rule, rule)
    fix s t
    assume "s \<succ> t"
    with subst have "D\<^sub>c (sk.h s) \<succ> D\<^sub>c (sk.h t)" unfolding Dc_h by auto
    thus "sk.h s \<succ>\<^sub>s\<^sub>k sk.h t" unfolding less_sk_def by auto
  qed
    
  have 2:"\<forall>s t. ground s \<and> ground t \<longrightarrow> s = t \<or> s \<succ>\<^sub>s\<^sub>k t \<or> t \<succ>\<^sub>s\<^sub>k s"
  proof(rule, rule, rule)
    fix s t :: "(('a,'b) f_ext, 'b) term"
    assume "ground s \<and> ground t"
    hence g:"ground s" "ground t" by auto
    let ?gtkbo = "admissible_weight_fun_prc.gtkbo (\<lambda>f. 1) 1 prec\<^sub>s\<^sub>k (\<lambda>f n. 1)"
    from admissible_weight_fun_prc.gtkbo_gtotal[OF adm _ precsk_gtotal _ g(1) _ g(2)]
    have kbo_gtotal:"s = t \<or> ?gtkbo s t \<or> ?gtkbo t s" by auto
    note eq = admissible_weight_fun_prc.kbo_eq_gtkbo[OF adm _ precsk_gtotal]
    from kbo_gtotal eq[OF _ g] eq[OF _ g(2) g(1)] have kbo_gtotal:"s = t \<or> fst (kbo_sk s t) \<or> fst (kbo_sk t s)" by force
    from g have g':"ground (D\<^sub>c s)" "ground (D\<^sub>c t)" unfolding D\<^sub>c_def by auto
    from gtotal[OF this] kbo_gtotal show "s = t \<or> s \<succ>\<^sub>s\<^sub>k t \<or> t \<succ>\<^sub>s\<^sub>k s" unfolding less_sk_def by force
  qed 
    
  have 3:"SN {\<succ>\<^sub>s\<^sub>k}"
  proof-
    from precsk_gtotal have t:"total_on prec\<^sub>s\<^sub>k\<inverse>\<inverse> UNIV" unfolding Restricted_Predicates.total_on_def by auto
    have aux:"(\<lambda> fn gm. prec\<^sub>s\<^sub>k gm fn) = prec\<^sub>s\<^sub>k\<inverse>\<inverse>" by auto
  from SN_onE[OF SN_prec] have wfp:"wfp_on (\<lambda> fn gm. prec\<^sub>s\<^sub>k gm fn) UNIV"
    unfolding SN_iff_wf wfp_on_def mem_Collect_eq by metis
  with aux have wfp:"wfp_on prec\<^sub>s\<^sub>k\<inverse>\<inverse> UNIV" by auto
  from trans_prec_sk have "transp_on prec\<^sub>s\<^sub>k\<inverse>\<inverse>\<^sup>=\<^sup>= UNIV" unfolding transp_on_def by blast
  with total_on_and_wfp_on_imp_almost_full_on[OF t wfp] have "wqo_on prec\<^sub>s\<^sub>k\<inverse>\<inverse>\<^sup>=\<^sup>= UNIV"
    unfolding wqo_on_def by auto
  with admissible_weight_fun_prc.SN_S[OF adm] have "SN {(s, t). fst (kbo_sk s t)}" by auto
  from lex_two[OF _ SN_less this, of Id] have "SN {(s,t).(s,t) \<in> lex_sk}" by auto
  then show ?thesis unfolding less_sk_def by fast
  qed
    
  have 4:"\<forall>s t C. s \<succ>\<^sub>s\<^sub>k t \<longrightarrow> C\<langle>s\<rangle> \<succ>\<^sub>s\<^sub>k C\<langle>t\<rangle>"
  proof(rule+)
    fix s t C
    assume "s \<succ>\<^sub>s\<^sub>k t"
    then consider "D\<^sub>c s \<succ> D\<^sub>c t" | "(D\<^sub>c s = D\<^sub>c t \<and> fst (kbo_sk s t))" unfolding less_sk_def by auto
    thus "C\<langle>s\<rangle> \<succ>\<^sub>s\<^sub>k C\<langle>t\<rangle>" proof(cases)
      case 1
      from ctxt[OF 1, of "Dc\<^sub>c C"] show ?thesis
        unfolding D\<^sub>c_def Dc\<^sub>c_def map_funs_term_ctxt_distrib[symmetric] 
        unfolding D\<^sub>c_def[symmetric] less_sk_def by auto
    next
      case 2
      hence eq:"D\<^sub>c C\<langle>s\<rangle> = D\<^sub>c C\<langle>t\<rangle>" unfolding D\<^sub>c_def by auto
      from 2 admissible_weight_fun_prc.S_ctxt[OF adm, of s t C] have "fst (kbo_sk C\<langle>s\<rangle> C\<langle>t\<rangle>)" by blast
      with eq show "C\<langle>s\<rangle> \<succ>\<^sub>s\<^sub>k C\<langle>t\<rangle>" unfolding less_sk_def by force
    qed
  qed
    
  have 5:"\<forall>s t \<sigma>. s \<succ>\<^sub>s\<^sub>k t \<longrightarrow> s \<cdot> \<sigma> \<succ>\<^sub>s\<^sub>k t \<cdot> \<sigma>"
  proof(rule+)
    fix s t \<sigma>
    assume "s \<succ>\<^sub>s\<^sub>k t"
    then consider "D\<^sub>c s \<succ> D\<^sub>c t" | "(D\<^sub>c s = D\<^sub>c t \<and> fst (kbo_sk s t))" unfolding less_sk_def by auto
    thus "s \<cdot> \<sigma> \<succ>\<^sub>s\<^sub>k t \<cdot> \<sigma>" proof(cases)
      case 1
      from subst[OF 1, of "Ds\<^sub>c \<sigma>"] show ?thesis 
        unfolding D\<^sub>c_def Ds\<^sub>c_def map_funs_subst_distrib[symmetric] 
        unfolding D\<^sub>c_def[symmetric] less_sk_def by auto
    next
      case 2
      hence eq:"D\<^sub>c (s \<cdot> \<sigma>) = D\<^sub>c (t \<cdot> \<sigma>)" unfolding D\<^sub>c_def by auto
      from 2 admissible_weight_fun_prc.S_subst[OF adm, of s t \<sigma>] have "fst (kbo_sk (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))" by blast
      with eq show ?thesis  unfolding less_sk_def by auto
    qed
  qed
    
  have 6:"\<forall>s t u. s \<succ>\<^sub>s\<^sub>k t \<longrightarrow> t \<succ>\<^sub>s\<^sub>k u \<longrightarrow> s \<succ>\<^sub>s\<^sub>k u"
  proof(rule+)
    fix s t u
    assume "s \<succ>\<^sub>s\<^sub>k t" and tu:"t \<succ>\<^sub>s\<^sub>k u"
    then consider "D\<^sub>c s \<succ> D\<^sub>c t" | "(D\<^sub>c s = D\<^sub>c t \<and> fst (kbo_sk s t))" unfolding less_sk_def by auto
    thus "s \<succ>\<^sub>s\<^sub>k u" proof(cases)
      case 1
      from tu have "D\<^sub>c t \<succeq> D\<^sub>c u" unfolding less_sk_def by auto
      with trans[unfolded trans_def, OF 1, of "D\<^sub>c u"] 1 show ?thesis unfolding less_sk_def by auto
    next
      case 2
      have S_NS:"fst (kbo_sk t u) \<longrightarrow> snd (kbo_sk t u)"
        using adm admissible_weight_fun_prc.S_imp_NS by blast
      note trans = admissible_weight_fun_prc.kbo_trans[OF adm, THEN conjunct1, rule_format, of s t u]
      with 2 S_NS have "fst (kbo_sk t u) \<Longrightarrow> fst (kbo_sk s u)" by blast
      with 2 tu show ?thesis unfolding less_sk_def by (cases "D\<^sub>c t \<succ> D\<^sub>c u", auto)
    qed
  qed
    
  have "\<forall>t. ground t \<longrightarrow> t \<succeq>\<^sub>s\<^sub>k Fun (FOrig c) []"
  proof(rule+)
    fix t :: "(('a,'b) f_ext, 'b) term"
    assume g:"ground t" and t_neq_c:"t \<noteq> Fun (FOrig c) []"
    have Dc_c:"D\<^sub>c (Fun (FOrig c) []) = Fun c []" unfolding D\<^sub>c_def by auto
    show "t \<succ>\<^sub>s\<^sub>k Fun (FOrig c) []" proof(cases "\<exists>a. t = Fun (FFresh a) []")
      case True
      then obtain a where t:"t = Fun (FFresh a) []" by auto
      have Dc_f:"D\<^sub>c t = Fun c []" unfolding t D\<^sub>c_def Nil by auto
      let ?w = "weight_fun.weight (\<lambda>f. 1) 1 (\<lambda>f n. 1)"
      have w:"?w (Fun (FOrig c) []) = ?w t" unfolding t Nil weight_fun.weight.simps by auto
      have p:"prec\<^sub>s\<^sub>k (FFresh a, 0) (FOrig c, 0)" unfolding prec\<^sub>s\<^sub>k_def prec_set_def by auto
      note kbo = weight_fun_prc.kbo.simps[of _ _ _ _ _ _ t "Fun (FOrig c) []", simplified]
      from p w have "fst (kbo_sk t (Fun (FOrig c) []))"
        unfolding kbo unfolding t Nil by force
      with Dc_f show ?thesis unfolding less_sk_def Dc_c by auto
    next
      case False
      from g obtain f ts where t:"t = Fun f ts" by (cases t, auto)
      from False t_neq_c have "D\<^sub>c t \<noteq> Fun c []" unfolding D\<^sub>c_def t by (cases f, auto)
      with c_min[rule_format, of "D\<^sub>c t"] have "D\<^sub>c t \<succ> Fun c []" unfolding Dc_c by auto
      thus ?thesis unfolding less_sk_def Dc_c by auto
    qed
  qed
  hence 7:"\<forall>t. ground t \<longrightarrow> op \<succ>\<^sub>s\<^sub>k\<^sup>=\<^sup>= t (sk.h (Fun c []))" by auto
  from 1 2 3 4 5 6 7 show ?thesis unfolding sk_redord_def by force
qed
    
 
lemma c_min_less_sk:"ground t \<longrightarrow> less_sk\<^sup>=\<^sup>= t (sk.h (Fun c []))"
  using less_sk[unfolded sk_redord_def] by fast
    
sublocale sk_run:gtotal_okb_irun_h R E "op \<succ>" FOrig less_sk
proof (unfold_locales)
  show "\<And>s t u. less_sk s t \<Longrightarrow> less_sk t u \<Longrightarrow> less_sk s u" using less_sk[unfolded sk_redord_def] by fast
qed (insert less_sk[unfolded sk_redord_def], auto)
    
lemma rstep_sk_exists:
  assumes "(sk s, t) \<in> rstep (sk.h\<^sub>R RR)" and "ground t"
  shows "\<exists>u. (s, u) \<in> rstep RR \<and> t \<succeq>\<^sub>s\<^sub>k sk u"
proof -
  from assms obtain C \<sigma> l r where lr:"(l,r) \<in> sk.h\<^sub>R RR" and sl:"(sk.h s) \<cdot> \<sigma>\<^sub>s\<^sub>k = C\<langle>l\<cdot>\<sigma>\<rangle>" and tr:"t = C\<langle>r\<cdot>\<sigma>\<rangle>"
    unfolding sk_def by auto
  from assms(2)[unfolded tr] have g_rsigma:"ground (r\<cdot>\<sigma>)" by auto
  from lr orig_sig_h have l_sig:"orig_sig l"
    unfolding sk.h\<^sub>R_def sk.h_def map_funs_trs.simps map_funs_rule.simps by blast
  from ctxt_sk_image[OF orig_sig_h sl] obtain D v where
    C:"C = D \<cdot>\<^sub>c \<sigma>\<^sub>s\<^sub>k" and lsigma:"v \<cdot> \<sigma>\<^sub>s\<^sub>k = l \<cdot> \<sigma>" and v_sig:"orig_sig v" and D_sig:"orig_sig\<^sub>C D" by auto
  let ?\<sigma>\<^sub>k\<^sub>s = "\<lambda>x. ks (\<sigma> x)"
  from subst_sk_image[OF v_sig l_sig lsigma] have
    vl:"\<forall>x\<in>vars_term l. \<sigma> x = ?\<sigma>\<^sub>k\<^sub>s x \<cdot> \<sigma>\<^sub>s\<^sub>k" "\<forall>x\<in>vars_term l. orig_sig (ks (\<sigma> x))" by auto
  define \<tau> where "\<tau> = ext_subst \<sigma> l (FOrig c)"
  let ?\<tau>\<^sub>k\<^sub>s = "\<lambda>x. ks (\<tau> x)"
  have vlt:"\<forall>x\<in>vars_term l. \<sigma> x = \<tau> x" unfolding \<tau>_def by (simp add: ext_subst_def)
  hence lst:"l \<cdot> \<sigma> = l \<cdot> \<tau>" using term_subst_eq_conv by auto
  from vl(2) vlt have v_sig:"\<forall>x\<in>vars_term l. orig_sig (ks (\<tau> x))" by auto
  note conv = term_subst_eq_conv[of _ \<sigma> "?\<tau>\<^sub>k\<^sub>s \<circ>\<^sub>s \<sigma>\<^sub>s\<^sub>k", unfolded subst_compose]
  from vl(1) vlt have aux:"\<forall>x\<in>vars_term l. \<sigma> x = ?\<tau>\<^sub>k\<^sub>s x \<cdot> \<sigma>\<^sub>s\<^sub>k" by auto
  hence l:"l \<cdot> \<tau> = l \<cdot> (?\<tau>\<^sub>k\<^sub>s \<circ>\<^sub>s \<sigma>\<^sub>s\<^sub>k)" unfolding conv[symmetric] subst_compose lst by auto
  with subst_compose have l:"l \<cdot> \<tau> = (l \<cdot> ?\<tau>\<^sub>k\<^sub>s) \<cdot> \<sigma>\<^sub>s\<^sub>k" by auto
  from aux vlt have aux:"\<forall>x\<in>vars_term l. \<tau> x = ?\<tau>\<^sub>k\<^sub>s x \<cdot> \<sigma>\<^sub>s\<^sub>k" by auto
  from orig_sig_subst_apply[OF l_sig v_sig] have "orig_sig (l \<cdot> (\<lambda>x. ks (\<tau> x)))" .
  with orig_sig_ctxt_apply D_sig have sig_Dl:"orig_sig (D\<langle>l \<cdot> (\<lambda>x. ks (\<tau> x))\<rangle>)" by auto
  from sl[unfolded C lst l subst_apply_term_ctxt_apply_distrib[symmetric]] inj_on_sigma_sk sig_Dl orig_sig_h have
    "sk.h s = D\<langle>l \<cdot> (\<lambda>x. ks (\<tau> x))\<rangle>" unfolding inj_on_def by blast
  { fix x
    assume "x \<in> vars_term r - vars_term l"
    with ext_subst have tx:"\<tau> x = Fun (FOrig c) []" unfolding \<tau>_def by fast
    hence sig:"orig_sig (\<tau> x)" using orig_sig_def by auto
    have ex:"(\<tau> x) \<cdot> \<sigma>\<^sub>s\<^sub>k = \<tau> x" unfolding tx by auto
    with ks_sk[OF sig] have ks:"ks (\<tau> x) = \<tau> x" by metis
    hence sig:"orig_sig (ks (\<tau> x))" using sig by auto
    from ex ks have "\<tau> x = ?\<tau>\<^sub>k\<^sub>s x \<cdot> \<sigma>\<^sub>s\<^sub>k" by auto
  } note vars_r_without_l = this
  with aux have "\<forall>x\<in>vars_term r. \<tau> x = ?\<tau>\<^sub>k\<^sub>s x \<cdot> \<sigma>\<^sub>s\<^sub>k" by auto
  with term_subst_eq_conv[of r \<tau> "?\<tau>\<^sub>k\<^sub>s \<circ>\<^sub>s \<sigma>\<^sub>s\<^sub>k"] have "r \<cdot> \<tau> = r \<cdot> (?\<tau>\<^sub>k\<^sub>s \<circ>\<^sub>s \<sigma>\<^sub>s\<^sub>k)" unfolding subst_compose by fast
  with subst_compose have r:"r \<cdot> \<tau> = (r \<cdot> ?\<tau>\<^sub>k\<^sub>s) \<cdot> \<sigma>\<^sub>s\<^sub>k" by simp
  from c_min_less_sk have leq:"r \<cdot> \<sigma> \<succeq>\<^sub>s\<^sub>k r \<cdot> \<tau>" using sk_run.okb_h.ext_subst_less[OF _ g_rsigma]
    unfolding \<tau>_def by auto
  let ?C = "D \<cdot>\<^sub>c \<sigma>\<^sub>s\<^sub>k"
  from leq sk_run.ctxt_h C(1) have "C\<langle>r \<cdot> \<sigma>\<rangle> \<succeq>\<^sub>s\<^sub>k ?C\<langle>r \<cdot> \<tau>\<rangle>" by auto
  with r tr have "t \<succeq>\<^sub>s\<^sub>k ?C\<langle>(r \<cdot> ?\<tau>\<^sub>k\<^sub>s) \<cdot> \<sigma>\<^sub>s\<^sub>k\<rangle>" by auto
  hence succ:"t \<succeq>\<^sub>s\<^sub>k D\<langle>r \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle> \<cdot> \<sigma>\<^sub>s\<^sub>k" by auto
  note s = sl[unfolded C lst l subst_apply_term_ctxt_apply_distrib[symmetric]]
  with inj_on_sigma_sk sig_Dl orig_sig_h have sk_s:"sk.h s = D\<langle>l \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle>"
    unfolding inj_on_def by blast
  from rstepI[OF lr] have step:"(sk.h s, D\<langle>r \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle>) \<in> rstep (sk.h\<^sub>R RR)" unfolding sk_s by auto
      
  with sk_s have "(D\<langle>l \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle>, D\<langle>r \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle>) \<in> rstep (sk.h\<^sub>R RR)" by auto
  from lr obtain l\<^sub>0 r\<^sub>0 where lr0:"l = sk.h l\<^sub>0" "r = sk.h r\<^sub>0"
    unfolding sk.h\<^sub>R_def map_funs_trs.simps map_funs_rule.simps by force
  define u where "u \<equiv> sk.h_inv (l \<cdot> ?\<tau>\<^sub>k\<^sub>s)"
  with sk_s[THEN sk.ctxt_image] have C:"D = sk.h\<^sub>C (sk.h\<^sub>C_inv D)" and lu:"sk.h u = l \<cdot> ?\<tau>\<^sub>k\<^sub>s"
    unfolding u_def by auto
  from sk.subst_image[OF lu[unfolded lr0]] lr0 have vl:"\<forall>x\<in>vars_term l. ?\<tau>\<^sub>k\<^sub>s x = sk.h (sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s x)" by auto
  with term_subst_eq_conv have l_subst:"l \<cdot> ?\<tau>\<^sub>k\<^sub>s = l \<cdot> (sk.h\<^sub>s (sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s))" by auto
      
  from C have "D\<langle>l \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle> = (sk.h\<^sub>C (sk.h\<^sub>C_inv D))\<langle>l \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle>" by auto
  hence l:"D\<langle>l \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle> = sk.h (sk.h\<^sub>C_inv D)\<langle>l\<^sub>0 \<cdot> (sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s)\<rangle>"
    unfolding lr0 l_subst[unfolded lr0] sk.subst_apply_h[symmetric] using sk.ctxt_apply_h by force
  
  have "\<forall>x\<in>vars_term r - vars_term l. ?\<tau>\<^sub>k\<^sub>s x = sk.h (sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s x)"
  proof
    fix x 
    assume "x \<in> vars_term r - vars_term l"
    with ext_subst have tx:"\<tau> x = Fun (FOrig c) []" unfolding \<tau>_def by fast
    hence sig:"orig_sig (\<tau> x)" unfolding orig_sig_def by auto
    hence ex:"\<tau> x \<in> {t |t. orig_sig t} \<and> \<tau> x \<cdot> \<sigma>\<^sub>s\<^sub>k = \<tau> x" unfolding tx by auto
    { fix t
      have "t \<cdot> \<sigma>\<^sub>s\<^sub>k = \<tau> x \<longrightarrow> t = \<tau> x" unfolding \<sigma>\<^sub>s\<^sub>k_def tx by (cases t, auto)
    }
    with ex have "\<exists>!t. t \<in> {t |t. orig_sig t} \<and> t \<cdot> \<sigma>\<^sub>s\<^sub>k = \<tau> x" by blast
    from the1_equality[OF this ex] have tx':"?\<tau>\<^sub>k\<^sub>s x = Fun (FOrig c) []"
      unfolding ks_def the_inv_into_def tx by auto
    have "sk.h (sk.h_inv (?\<tau>\<^sub>k\<^sub>s x)) = ?\<tau>\<^sub>k\<^sub>s x"
      unfolding tx' sk.h_def sk.h_inv_def term.map inv_f_f[OF sk.injF] by simp
    thus "?\<tau>\<^sub>k\<^sub>s x = sk.h (sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s x)"
      unfolding sk.h\<^sub>s_def tx' sk.h_def sk.h_inv_def map_funs_subst.simps using tx' by argo
  qed
  with vl have "\<forall>x\<in>vars_term r. ?\<tau>\<^sub>k\<^sub>s x = sk.h (sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s x)" by auto
  with term_subst_eq_conv have r_subst:"r \<cdot> ?\<tau>\<^sub>k\<^sub>s = r \<cdot> (sk.h\<^sub>s (sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s))" by auto
  from C have "D\<langle>r \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle> = (sk.h\<^sub>C (sk.h\<^sub>C_inv D))\<langle>r \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle>" by auto
  hence "D\<langle>r \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle> = (sk.h\<^sub>C (sk.h\<^sub>C_inv D))\<langle>sk.h r\<^sub>0 \<cdot> (sk.h\<^sub>s (sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s))\<rangle>" using r_subst lr0 by auto
  from this[unfolded sk.subst_apply_h[symmetric]] have r:"D\<langle>r \<cdot> ?\<tau>\<^sub>k\<^sub>s\<rangle> = sk.h (sk.h\<^sub>C_inv D)\<langle>r\<^sub>0 \<cdot> (sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s)\<rangle>"
    using sk.ctxt_apply_h by auto
  from step[unfolded sk_s l r] sk.rstep_h
    have step:"((sk.h\<^sub>C_inv D)\<langle>l\<^sub>0 \<cdot> sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s\<rangle>, (sk.h\<^sub>C_inv D)\<langle>r\<^sub>0 \<cdot> sk.h\<^sub>s_inv ?\<tau>\<^sub>k\<^sub>s\<rangle>) \<in> rstep RR" (is "(?s, ?t) \<in> _") by blast
  from sk_s have "sk ?s = sk s"
    unfolding sk_def sk.ctxt_apply_h' C[symmetric] sk.subst_apply_h lr0[symmetric] l_subst[symmetric] by auto
  with sk_inj have s:"?s = s" by auto
  from succ have "t \<succeq>\<^sub>s\<^sub>k sk (?t)"
    unfolding sk_def sk.ctxt_apply_h' C[symmetric] sk.subst_apply_h lr0[symmetric] r_subst[symmetric] by auto
  with step[unfolded s] obtain v where "(s, v) \<in> rstep RR" and "t \<succeq>\<^sub>s\<^sub>k sk v" by blast
  thus ?thesis by auto
qed
  
lemma rsteps_sk:"(s,t) \<in> (rstep RR)\<^sup>* \<Longrightarrow> ((sk s, sk t) \<in> (rstep (sk.h\<^sub>R RR))\<^sup>*)"
  using rstep_sk by (meson rtrancl_map)

context
  assumes fair:"\<forall>(s, t)\<in> PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>). (s, t) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>"
begin
  
abbreviation "Sh\<^sub>\<omega> \<equiv> sk_run.irun_h.Sw"
abbreviation "Eh\<^sub>\<omega> \<equiv> sk_run.irun_h.E\<^sub>\<omega>"
abbreviation "Rh\<^sub>\<omega> \<equiv> sk_run.irun_h.R\<^sub>\<omega>"
abbreviation Ehinf ("Eh\<^sub>\<infinity>") where "Ehinf \<equiv> (\<Union>i. sk_run.E\<^sub>h i)"

  
lemma sk_run_fair: "\<forall>(s, t)\<in> sk_run.irun_h.PCP_ext (sk_run.irun_h.R\<^sub>\<omega> \<union> sk_run.irun_h.E\<^sub>\<omega>\<^sup>\<leftrightarrow>). (s, t) \<in> (rstep Eh\<^sub>\<infinity>)\<^sup>\<leftrightarrow>"
proof
  fix s t
  assume pcp:"(s, t) \<in> sk_run.irun_h.PCP_ext (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
  from sk_run.PCP_h[unfolded sk_run.hRw sk_run.hEw, OF this]
    have "(sk.h_inv s, sk.h_inv t)\<in> PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" and id:"s = sk.h (sk.h_inv s) \<and> t = sk.h (sk.h_inv t)" by auto
  with fair have "(sk.h_inv s, sk.h_inv t) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>" by auto
  with sk.rstep_h[of _ _ "E\<^sub>\<infinity>"] have "(sk.h (sk.h_inv s), sk.h (sk.h_inv t)) \<in> (rstep (sk.h\<^sub>R E\<^sub>\<infinity>))\<^sup>\<leftrightarrow>"
    by blast
  with sk_run.hEinf have  "(sk.h (sk.h_inv s), sk.h (sk.h_inv t)) \<in> (rstep Eh\<^sub>\<infinity>)\<^sup>\<leftrightarrow>" by metis
  with id show "(s,t)  \<in> (rstep Eh\<^sub>\<infinity>)\<^sup>\<leftrightarrow>" by auto    
qed

lemma ground_RE_conv_S_conv:
  assumes "(s,t) \<in> (GROUND (rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>)))\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(s,t) \<in> (GROUND (rstep Sh\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
proof(cases "s=t")
  case True
  then show ?thesis by auto
next
  case False
  have "\<And>Rel. GROUND (Rel\<^sup>\<leftrightarrow>) = (GROUND Rel)\<^sup>\<leftrightarrow>" unfolding GROUND_def by blast
  with assms[unfolded conversion_def] have c:"(s, t) \<in> (GROUND ((rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>))\<^sup>\<leftrightarrow>))\<^sup>*" by metis
  from False GROUND_rtrancl[OF c]  GROUND_not_ground[OF c] have g:"ground s" "ground t"
    and conv:"(s,t) \<in> (rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_def by auto
  have E:"\<And>E. (rstep (Rh\<^sub>\<omega> \<union> E\<^sup>\<leftrightarrow>))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (Rh\<^sub>\<omega> \<union> E))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_def
    by (metis (no_types, lifting) rstep_simps(5) sup.right_idem sup_left_idem symcl_Un symcl_converse)
  note Sconv = sk_run.irun_h.gterms_ER_conv_implies_S_conv[OF sk_run.irun_h.Rw_less sk_run.irun_h.Rw_less g]
  with E conv show ?thesis unfolding sk_run.irun_h.Sw_def E_ord_def by blast
qed
  
context
  fixes \<R>::"('a,'b) trs"
  assumes R_less:"\<R> \<subseteq> {\<succ>}" and CR_R:"CR (rstep \<R>)"
  and conv_eq:"(rstep \<R>)\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>*"
begin

lemma SN_R:"SN (rstep \<R>)" 
  by (rule SN_subset [OF SN_less], insert compatible_rstep_imp_less[OF R_less], auto)
 
lemma R_NF_eq:
  assumes "(s,t) \<in> (rstep E\<^sub>\<omega>)\<^sup>\<leftrightarrow>"
  shows "\<exists>u. (s,u) \<in> (rstep \<R>)\<^sup>! \<and> (t,u) \<in> (rstep \<R>)\<^sup>!"
proof-
  from assms conv_eq oKBi_conversion_ERw have "(s, t) \<in> (rstep \<R>)\<^sup>\<leftrightarrow>\<^sup>*" by fastforce
  with CR_imp_conversionIff_join[OF CR_R] obtain w where
    w:"(s, w) \<in> (rstep \<R>)\<^sup>*" "(t, w) \<in> (rstep \<R>)\<^sup>*" by auto
  from SN_reaches_NF[of "rstep \<R>"] SN_R have "\<exists>w'. (w, w') \<in> (rstep \<R>)\<^sup>* \<and> w' \<in> NF_trs \<R>"
    unfolding normalizability_def by (simp add: SN_defs)
  then obtain w' where "(w, w') \<in> (rstep \<R>)\<^sup>*" and nf:"w' \<in> NF_trs \<R>" by auto
  with w have seqs:"(s, w') \<in> (rstep \<R>)\<^sup>*" "(t, w') \<in> (rstep \<R>)\<^sup>*" by auto
  show ?thesis by (rule exI[of _ w'], unfold normalizability_def, insert seqs nf, auto)
qed
  
lemma NF_R_NF_S:
  assumes "u \<in> NF (rstep \<R>)"
  shows "sk u \<in> NF (rstep Sh\<^sub>\<omega>)"
proof -
  { fix v
    assume S_step:"(sk u,v) \<in> rstep Sh\<^sub>\<omega>"
    from subsetD[OF sk_run.irun_h.Sw_implies_REw' this] rstep_simps(5) obtain C l\<^sub>s\<^sub>k r\<^sub>s\<^sub>k \<sigma> where
      lr:"sk u = C\<langle>l\<^sub>s\<^sub>k \<cdot> \<sigma>\<rangle>" "v = C\<langle>r\<^sub>s\<^sub>k \<cdot> \<sigma>\<rangle>" "(l\<^sub>s\<^sub>k,r\<^sub>s\<^sub>k) \<in> Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>\<^sup>\<leftrightarrow>" by blast
    let ?l = "sk.h_inv l\<^sub>s\<^sub>k" and ?r = "sk.h_inv r\<^sub>s\<^sub>k"
    from lr(3) sk_run.hRw sk_run.hEw sk.inv_hR_rule have
      rule:"(?l, ?r) \<in> (R\<^sub>\<omega> \<union> E\<^sub>\<omega>)\<^sup>\<leftrightarrow>" "sk.h ?l = l\<^sub>s\<^sub>k" "sk.h ?r = r\<^sub>s\<^sub>k" by blast+
    with conv_eq oKBi_conversion_ERw have "(?l, ?r) \<in> (rstep \<R>)\<^sup>\<leftrightarrow>\<^sup>*" by auto
    with CR_imp_conversionIff_join[OF CR_R] obtain w where
      w:"(?l, w) \<in> (rstep \<R>)\<^sup>*" "(?r, w) \<in> (rstep \<R>)\<^sup>*" by auto
    { assume eq:"?l = w"
      with w(2) compatible_rstep_trancl_imp_less[OF R_less] have "?r \<succeq> ?l"
        unfolding rtrancl_eq_or_trancl by auto
      with sk_run.compat_h[of ?r ?l] have "r\<^sub>s\<^sub>k \<succeq>\<^sub>s\<^sub>k l\<^sub>s\<^sub>k" using rule by auto
      with sk_run.subst_h sk_run.ctxt_h have leq:"C\<langle>r\<^sub>s\<^sub>k \<cdot> \<sigma>\<rangle> \<succeq>\<^sub>s\<^sub>k C\<langle>l\<^sub>s\<^sub>k \<cdot> \<sigma>\<rangle>" by auto
      from sk_run.irun_h.Sw_step_less S_step have "C\<langle>l\<^sub>s\<^sub>k \<cdot> \<sigma>\<rangle> \<succ>\<^sub>s\<^sub>k C\<langle>r\<^sub>s\<^sub>k \<cdot> \<sigma>\<rangle>" unfolding lr by auto
      with sk_run.trans_h[OF this] leq sk_run.okb_h.irrefl have False by auto
    }
    hence "?l \<noteq> w" by auto
    with w(1)[unfolded rtrancl_eq_or_trancl] tranclD obtain v where v:"(?l, v) \<in> rstep \<R>" by metis
    from lr rule have "sk u = C\<langle>sk.h ?l \<cdot> \<sigma>\<rangle>" by auto
    from sk_subst_ctxt_split2[OF this] v have "\<exists>v'. (u,v') \<in> rstep \<R>" by fast
  }
  with assms show ?thesis by blast
qed

lemma oriented_REw_step_Sw_step:
  assumes "(t,u) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>" and "t \<succ> u" and "u \<in> NF_trs \<R>"
  shows"(t,u) \<in> rstep S\<^sub>\<omega>"
proof-
  from assms obtain C l r \<sigma> where lr:"(l,r) \<in> (R\<^sub>\<omega> \<union> E\<^sub>\<omega>)\<^sup>\<leftrightarrow>" "t = C\<langle>l\<cdot>\<sigma>\<rangle>" "u = C\<langle>r\<cdot>\<sigma>\<rangle>" by blast
  from oKBi_conversion_ERw lr(1) have join:"(l,r) \<in> (rstep \<R>)\<^sup>\<down>"
    using CR_R[unfolded CR_iff_conversion_imp_join] unfolding conv_eq[symmetric] by force
  from assms(3)[unfolded lr] have "r \<in> NF_trs \<R>" unfolding NF_def by fast
  with NF_join_imp_reach[OF _ join] have lr_R_steps:"(l,r) \<in> (rstep \<R>)\<^sup>*" by auto
  from assms(2)[unfolded lr] irrefl have "l \<noteq> r" by (cases "l = r", auto)
  with lr_R_steps have lr_R_steps:"(l,r) \<in> (rstep \<R>)\<^sup>+" unfolding rtrancl_eq_or_trancl by auto
  with compatible_rstep_trancl_imp_less[OF R_less] subst have succ:"l\<cdot>\<sigma> \<succ> r\<cdot>\<sigma>" by auto
  from Rw_less subst ctxt have "(r,l) \<in> R\<^sub>\<omega> \<Longrightarrow> u \<succ> t" unfolding lr by auto
  with assms(2) trans irrefl have "(r,l) \<notin> R\<^sub>\<omega>" by fast
  with succ lr(1) show ?thesis unfolding Sw_def lr symcl_Un rstep_union by blast
qed
    
lemma ground_ctxt_less:
  assumes "ground s" and "ground t" and "C\<langle>s\<rangle> \<succ>\<^sub>s\<^sub>k C\<langle>t\<rangle>"
  shows "s \<succ>\<^sub>s\<^sub>k t"
proof-
  from assms sk_run.gtotal_h consider "s = t" | "s \<succ>\<^sub>s\<^sub>k t"  | "t \<succ>\<^sub>s\<^sub>k s" by auto
  thus ?thesis proof(cases)
    case 3
    with sk_run.okb_h.ctxt have "C\<langle>t\<rangle> \<succ>\<^sub>s\<^sub>k C\<langle>s\<rangle>" by auto
    from sk_run.okb_h.irrefl sk_run.okb_h.trans[OF assms(3) this] show ?thesis by auto
  qed (insert assms(3) sk_run.okb_h.irrefl, auto)
qed 

lemma ground_subt_less:
  assumes "ground s" and "s \<rhd> t"
  shows "s \<succ>\<^sub>s\<^sub>k t"
proof-
  from assms have gt:"ground t" "s \<noteq> t" by auto
  from assms obtain C where s:"s = C\<langle>t\<rangle>" by auto
  { assume "t \<succ>\<^sub>s\<^sub>k s"
    hence less:"t \<succ>\<^sub>s\<^sub>k C\<langle>t\<rangle>" unfolding s by auto
    { fix i
      have "(C ^ i)\<langle>t\<rangle> \<succ>\<^sub>s\<^sub>k (C ^ (Suc i))\<langle>t\<rangle>" using less sk_run.ctxt_h by (induct i, auto)
    }
    with less sk_run.SN_less_h[unfolded SN_on_def] have False by fast
  }
  with sk_run.gtotal_h assms(1) gt \<open>s \<noteq> t\<close> show ?thesis by auto
qed  

lemma Rstep_imp_S_step_sk:
  assumes "(s,t) \<in> rstep \<R>"
  shows "\<exists>u. (sk s, sk u) \<in> rstep Sh\<^sub>\<omega>"
proof-
  from assms compatible_rstep_imp_less[OF R_less] have st:"s \<succ> t" by auto
  from oKBi_conversion_ERw conv_eq have conv_eq':"(rstep \<R>)\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  with assms have tu_convERw:"(s,t) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  note sk_conv = this [unfolded conversion_def rstep_simps(5)[symmetric], THEN rsteps_sk]
  hence sk_tu:"(sk s, sk t) \<in> (rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding conversion_def rstep_simps(5) sk.hR_sym 
    unfolding sk.hR_union sk_run.hRw sk_run.hEw by auto
  from sk_run.irun_h.correctness_okb[OF sk_run_fair] have GCRh:"GCR (rstep Sh\<^sub>\<omega>)" and
    convh:"(rstep (sk_run.E\<^sub>h 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  note tu_E0_ground = gterm_conv_GROUND_conv[OF ground_sk ground_sk sk_tu]
  with ground_RE_conv_S_conv have "(sk s, sk t) \<in> (GROUND (rstep Sh\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by simp
  with GCRh[THEN CR_imp_conversionIff_join] 
  obtain v where v':"(sk s,v) \<in> (GROUND (rstep Sh\<^sub>\<omega>))\<^sup>*" "(sk t,v) \<in> (GROUND (rstep Sh\<^sub>\<omega>))\<^sup>*" by auto
  hence v:"(sk s,v) \<in> (rstep Sh\<^sub>\<omega>)\<^sup>*" "(sk t,v) \<in> (rstep Sh\<^sub>\<omega>)\<^sup>*" using rtrancl_mono[OF GROUND_subset]
    by auto
  note compat = sk_run.okb_h.compatible_rstep_trancl_imp_less[OF sk_run.irun_h.Sw_less]
  with v(2)[unfolded rtrancl_eq_or_trancl] have geq:"sk t \<succeq>\<^sub>s\<^sub>k v" by auto
  { assume eq:"v = sk s"
    from sk_run.subst_h[OF sk_run.compat_h[OF st]] have "sk s \<succ>\<^sub>s\<^sub>k sk t" unfolding sk_def by auto
    with sk_run.trans_h[of "sk t" "sk s" "sk t"] geq[unfolded eq]
      have False using sk_run.okb_h.irrefl by auto
  }
  with v'(1)[unfolded rtrancl_eq_or_trancl] obtain u where u:"(sk s, u) \<in> rstep Sh\<^sub>\<omega>" "ground u"
    unfolding trancl_unfold_left using GROUND_def by blast
      
  with sk_run.irun_h.Sw_implies_REw' have "(sk s, u) \<in> rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
  with sk.hR_union sk.hR_sym have "(sk s, u) \<in> rstep (sk.h\<^sub>R (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))"
    unfolding sk_run.hRw[symmetric] sk_run.hEw[symmetric] by metis
  from rstep_sk_exists[OF this u(2)] obtain u' where
    su:"(s, u') \<in> rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" "u \<succeq>\<^sub>s\<^sub>k (sk u')" by auto
  with u(1) sk_run.irun_h.Sw_step_less sk_run.trans_h have succ:"sk s \<succ>\<^sub>s\<^sub>k sk u'" by blast
  from su[unfolded rstep_sk] have "(sk s, sk u') \<in> rstep (sk.h\<^sub>R (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))" by auto
  hence "(sk s, sk u') \<in> rstep (sk.h\<^sub>R R\<^sub>\<omega> \<union> (sk.h\<^sub>R E\<^sub>\<omega>)\<^sup>\<leftrightarrow>)"
    unfolding sk.hR_union using sk.hR_union sk.hR_sym by metis
  with sk_run.hRw sk_run.hEw have tv_conv:"(sk s, sk u') \<in> rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
  then obtain l r C \<sigma> where lr:"(l,r) \<in> Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>\<^sup>\<leftrightarrow>" "sk s = C\<langle>l\<cdot>\<sigma>\<rangle>" "sk u' = C\<langle>r\<cdot>\<sigma>\<rangle>" by fast
  have Sstep_tv:"(sk s, sk u') \<in> rstep Sh\<^sub>\<omega>" proof(cases "(l,r) \<in> Rh\<^sub>\<omega>")
    case True
    with lr show ?thesis unfolding sk_run.irun_h.Sw_def by auto
  next
    case False
    with lr have Eh:"(l,r) \<in> Eh\<^sub>\<omega>\<^sup>\<leftrightarrow>" by auto
    from ground_sk[of s, unfolded lr] ground_sk[of u', unfolded lr]
    have g:"ground (l\<cdot>\<sigma>)" and "ground (r\<cdot>\<sigma>)" by auto
    from ground_ctxt_less[OF this succ[unfolded lr]] have "l \<cdot> \<sigma> \<succ>\<^sub>s\<^sub>k r \<cdot> \<sigma>" by auto
    with Eh show ?thesis unfolding lr sk_run.irun_h.Sw_def by blast
  qed
  thus ?thesis by auto
qed
    
lemma Sw_steps_lesseq:"(s,t) \<in> (rstep S\<^sub>\<omega>)\<^sup>* \<Longrightarrow> s \<succeq> t"
  unfolding rtrancl_eq_or_trancl using compatible_rstep_trancl_imp_less[OF Sw_less] by auto

abbreviation "lexless_sk \<equiv> lex_two {\<succ>\<^sub>s\<^sub>k} Id {\<succ>\<^sub>s\<^sub>k}"

lemma SN_lexless_sk: "SN lexless_sk"
  using sk_run.SN_less_h
  by (intro lex_two, auto)

lemma ground_REw_step_Sw_step:
  assumes "ground s" and "ground t" and "(s,t) \<in> rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
  shows "(s,t) \<in> ((rstep Sh\<^sub>\<omega>)\<^sup>\<leftrightarrow>)\<^sup>="
proof-
  from assms(3) obtain C \<sigma> l r where lr:"(l,r) \<in> Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>\<^sup>\<leftrightarrow>" "s = C\<langle>l\<cdot>\<sigma>\<rangle>" "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by fast
  from assms have "ground (l\<cdot>\<sigma>) \<and> ground (r\<cdot>\<sigma>)" unfolding lr by auto
  from sk_run.gtotal_h[rule_format, OF this] consider "l\<cdot>\<sigma> \<succ>\<^sub>s\<^sub>k r\<cdot>\<sigma>" | "r\<cdot>\<sigma> \<succ>\<^sub>s\<^sub>k l\<cdot>\<sigma>" | "l\<cdot>\<sigma> = r\<cdot>\<sigma>" by auto
  thus ?thesis proof(cases)
    case 1
    note rstep = rstepI[of "l\<cdot>\<sigma>" "r\<cdot>\<sigma>" _ s C Var t, simplified]
    have "(s,t) \<in> rstep Sh\<^sub>\<omega>" unfolding sk_run.irun_h.Sw_def rstep_union lr using lr(1) 1 by blast
    thus ?thesis by auto
  next
    case 2
    note rstep = rstepI[of "r\<cdot>\<sigma>" "l\<cdot>\<sigma>" _ t C Var s, simplified]
    have "(l,r) \<in> Rh\<^sub>\<omega> \<Longrightarrow> l\<cdot>\<sigma> \<succ>\<^sub>s\<^sub>k r\<cdot>\<sigma>" using sk_run.irun_h.Rw_less sk_run.subst_h by auto
    with sk_run.okb_h.irrefl sk_run.okb_h.trans[OF 2] have "(l,r) \<notin> Rh\<^sub>\<omega>" by auto
    hence "(t,s) \<in> rstep Sh\<^sub>\<omega>" unfolding sk_run.irun_h.Sw_def rstep_union lr using lr(1) 2 by blast
    thus ?thesis by auto
  qed (unfold lr, insert lr(1), auto)
qed
     
lemma ooverlap_exists:
  assumes "C\<langle>l\<^sub>1\<cdot>\<sigma>\<^sub>1\<rangle> = l\<^sub>2'\<cdot>\<sigma>\<^sub>2'" and "hole_pos C \<in> funposs l\<^sub>2'" and "(l\<^sub>1,r\<^sub>1) \<in> RR" and "(l\<^sub>2',r\<^sub>2') \<in> RR"
  and "\<not>(r\<^sub>1\<cdot>\<sigma>\<^sub>1 \<succ> l\<^sub>1\<cdot>\<sigma>\<^sub>1)" and "\<not>(r\<^sub>2'\<cdot>\<sigma>\<^sub>2' \<succ> l\<^sub>2'\<cdot>\<sigma>\<^sub>2')"
  shows "\<exists>r r' \<mu> \<tau> s t. ooverlap {\<succ>} RR r r' (hole_pos C) \<mu> s t \<and> C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle> = s \<cdot> \<tau> \<and>
                        r\<^sub>2' \<cdot> \<sigma>\<^sub>2' = t \<cdot> \<tau> \<and> l\<^sub>1\<cdot>\<sigma>\<^sub>1 = (fst r) \<cdot> \<mu> \<cdot> \<tau>"
proof-
  have rl1:"\<exists>p. p \<bullet> (l\<^sub>1, r\<^sub>1) \<in> RR" by (rule exI[of _ 0], insert assms(3), auto)
  from vars_rule_disjoint obtain \<pi>
    where \<pi>: "vars_rule (\<pi> \<bullet> (l\<^sub>2', r\<^sub>2')) \<inter> vars_rule (l\<^sub>1, r\<^sub>1) = {}" ..
  define l\<^sub>2 and r\<^sub>2 and \<sigma>\<^sub>2
    where "l\<^sub>2 = \<pi> \<bullet> l\<^sub>2'" and "r\<^sub>2 = \<pi> \<bullet> r\<^sub>2'" and "\<sigma>\<^sub>2 = (Var \<circ> Rep_perm (-\<pi>)) \<circ>\<^sub>s \<sigma>\<^sub>2'"
  note rename = l\<^sub>2_def r\<^sub>2_def \<sigma>\<^sub>2_def
  from assms(4) have "-\<pi> \<bullet> (l\<^sub>2,r\<^sub>2) \<in> RR" unfolding rename by auto
  hence rl2:"\<exists>p. p \<bullet> (l\<^sub>2,r\<^sub>2) \<in> RR" by (rule exI[of _ "-\<pi>"])
  from \<pi> have vars:"vars_rule (l\<^sub>1,r\<^sub>1) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}"
    unfolding rename rule_pt.permute_prod_eqvt by auto
  define p where "p \<equiv> hole_pos C" 
  from assms(2) have p:"p \<in> funposs l\<^sub>2" by (auto simp:rename p_def)
  note poss_p = funposs_imp_poss [OF p]
    
  have r_l1:"l\<^sub>2' = -\<pi> \<bullet> l\<^sub>2" by (auto simp:rename)
  from assms(1) have eq:"l\<^sub>1 \<cdot> \<sigma>\<^sub>1 = (l\<^sub>2' \<cdot> \<sigma>\<^sub>2' |_ p)" unfolding p_def by (metis subt_at_hole_pos)
  note subt_l1 = subt_at_subst[OF poss_p]
  from eq[unfolded r_l1 permute_term_subst_apply_term] have
    eq':"l\<^sub>1 \<cdot> \<sigma>\<^sub>1 = l\<^sub>2 |_ p \<cdot> \<sigma>\<^sub>2" 
    unfolding subt_at_subst[OF poss_p] unfolding rename
    by (metis \<open>l\<^sub>2 \<equiv> \<pi> \<bullet> l\<^sub>2'\<close> eq r_l1 subst.cop_add subt_l1 term_apply_subst_Var_Rep_perm)
      
  note coinc = coincidence_lemma' [of _ "vars_rule (l\<^sub>1, r\<^sub>1)"]
  define \<sigma> where "\<sigma> x = (if x \<in> vars_rule (l\<^sub>1, r\<^sub>1) then \<sigma>\<^sub>1 x else \<sigma>\<^sub>2 x)" for x
  have "l\<^sub>1 \<cdot> \<sigma> = l\<^sub>1 \<cdot> (\<sigma> |s vars_rule (l\<^sub>1, r\<^sub>1))"
    using coinc[of l\<^sub>1] by (simp add: vars_rule_def)
  also have "\<dots> = l\<^sub>1 \<cdot> (\<sigma>\<^sub>1 |s vars_rule (l\<^sub>1, r\<^sub>1))" by (simp add: \<sigma>_def [abs_def])
  finally have l1_coinc:"l\<^sub>1 \<cdot> \<sigma> = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using coinc[of l\<^sub>1] by (simp add: vars_rule_def)
  have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_term l\<^sub>2 = {}"
    using \<pi>[unfolded rename] by (auto simp: vars_rule_def rule_pt.permute_prod_eqvt l\<^sub>2_def)
  have unif: "l\<^sub>1 \<cdot> \<sigma> = (l\<^sub>2 |_ p) \<cdot> \<sigma>"
  proof -
    from disj have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_term (l\<^sub>2 |_ p) = {}" 
      using vars_term_subt_at [OF poss_p] by auto
    from l1_coinc have "l\<^sub>1 \<cdot> \<sigma> = (l\<^sub>2 |_ p) \<cdot> \<sigma>\<^sub>2" using eq' by simp
    also have "\<dots> = (l\<^sub>2 |_ p) \<cdot> (\<sigma>\<^sub>2 |s vars_term (l\<^sub>2 |_ p))"
      by (simp add: coincidence_lemma [symmetric])
    also have "\<dots> = (l\<^sub>2 |_ p) \<cdot> (\<sigma> |s vars_term (l\<^sub>2 |_ p))" using disj by (simp add: \<sigma>_def [abs_def])
    finally show ?thesis by (simp add: coincidence_lemma [symmetric])
  qed
    
  define \<mu> where "\<mu> = the_mgu l\<^sub>1 (l\<^sub>2 |_ p)"
  have is_mgu:"is_mgu \<mu> {(l\<^sub>1, l\<^sub>2 |_ p)}" by (rule is_mguI, insert unif the_mgu, auto simp: \<mu>_def)
  with unif obtain \<tau> where \<sigma>: "\<sigma> = \<mu> \<circ>\<^sub>s \<tau>" by (auto simp: is_mgu_def unifiers_def)
  from unif have mgu: "mgu l\<^sub>1 (l\<^sub>2 |_ p) = Some \<mu>"
    unfolding \<mu>_def the_mgu_def
    using unify_complete and unify_sound by (force split: option.splits simp: is_imgu_def unifiers_def)
      
  have "r\<^sub>1 \<cdot> \<sigma> = r\<^sub>1 \<cdot> (\<sigma> |s vars_rule (l\<^sub>1, r\<^sub>1))" using coinc[of r\<^sub>1] by (simp add: vars_rule_def)
  also have "\<dots> = r\<^sub>1 \<cdot> (\<sigma>\<^sub>1 |s vars_rule (l\<^sub>1, r\<^sub>1))" by (simp add: \<sigma>_def [abs_def])
  finally have r1_coinc:"r\<^sub>1 \<cdot> \<sigma> = r\<^sub>1 \<cdot> \<sigma>\<^sub>1" using coinc[of r\<^sub>1] by (simp add: vars_rule_def)
  from assms(5) subst have nonoriented1:"\<not> r\<^sub>1 \<cdot> \<mu> \<succ> l\<^sub>1 \<cdot> \<mu>"
    unfolding l1_coinc[symmetric] r1_coinc[symmetric] \<sigma> by auto
      
  note coinc = coincidence_lemma [symmetric]
  have "l\<^sub>2 \<cdot> \<sigma> = l\<^sub>2 \<cdot> (\<sigma> |s vars_term l\<^sub>2)" by (simp add: coinc)
  also have "\<dots> = l\<^sub>2 \<cdot> (\<sigma>\<^sub>2 |s vars_term l\<^sub>2)" using disj by (simp add: \<sigma>_def [abs_def])
  also have "\<dots> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" by (simp add: coinc)
  finally have l2_coinc:"l\<^sub>2 \<cdot> \<sigma> = l\<^sub>2' \<cdot> \<sigma>\<^sub>2'" unfolding rename by simp
  have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_term r\<^sub>2 = {}"
    using vars_term_subt_at [OF poss_p] and \<pi>[unfolded rename]
    by (auto simp: vars_rule_def rule_pt.permute_prod_eqvt r\<^sub>2_def)
  have "r\<^sub>2 \<cdot> \<sigma> = r\<^sub>2 \<cdot> (\<sigma> |s vars_term r\<^sub>2)" by (simp add: coinc)
  also have "\<dots> = r\<^sub>2 \<cdot> (\<sigma>\<^sub>2 |s vars_term r\<^sub>2)" using disj by (simp add: \<sigma>_def [abs_def])
  also have "\<dots> = r\<^sub>2 \<cdot> \<sigma>\<^sub>2" by (simp add: coinc)
  finally have r2_coinc:"r\<^sub>2 \<cdot> \<sigma> = r\<^sub>2' \<cdot> \<sigma>\<^sub>2'" unfolding rename by simp
  from assms(6) subst have nonoriented2:"\<not> r\<^sub>2 \<cdot> \<mu> \<succ> l\<^sub>2 \<cdot> \<mu>"
    unfolding l2_coinc[symmetric] r2_coinc[symmetric] \<sigma> by auto
  from assms(1) poss_p hole_pos_id_ctxt have "C = ctxt_of_pos_term p (l\<^sub>2 \<cdot> \<sigma>)"
    unfolding l1_coinc[symmetric] l2_coinc[symmetric] p_def by metis  
  hence cp_left:"C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle> = (ctxt_of_pos_term p (l\<^sub>2 \<cdot> \<mu>))\<langle>r\<^sub>1 \<cdot> \<mu>\<rangle> \<cdot> \<tau>"
    unfolding subst_apply_term_ctxt_apply_distrib r1_coinc[symmetric]
      ctxt_of_pos_term_subst[OF poss_imp_subst_poss[OF poss_p], symmetric]
      subst_subst_compose[symmetric] \<sigma>[symmetric] by simp
  from subst_subst_compose have cp_right:"r\<^sub>2' \<cdot> \<sigma>\<^sub>2' = r\<^sub>2 \<cdot> \<mu> \<cdot> \<tau>" unfolding r2_coinc[symmetric] \<sigma> by auto
  let ?s = "(ctxt_of_pos_term p (l\<^sub>2 \<cdot> \<mu>))\<langle>r\<^sub>1 \<cdot> \<mu>\<rangle>"
  let ?t = "r\<^sub>2 \<cdot> \<mu>"
  from l1_coinc \<sigma> have "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 = l\<^sub>1 \<cdot> \<mu> \<cdot> \<tau>" by auto
  note facts = rl1 rl2 vars p mgu nonoriented1 nonoriented2 cp_left cp_right this
  show ?thesis unfolding ooverlap_def
    by (rule exI[of _ "(l\<^sub>1,r\<^sub>1)"], rule exI[of _ "(l\<^sub>2,r\<^sub>2)"], rule exI[of _ \<mu>],
        rule exI[of _ \<tau>], rule exI[of _ ?s], rule exI[of _ ?t], insert facts,
        unfold fst_conv snd_conv p_def, auto)
qed
    
lemma Einf_sym_without_Ew_sym: assumes "(l,r) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" and "(l,r) \<notin>  E\<^sub>\<omega>\<^sup>\<leftrightarrow>"
  shows "\<exists>j. ((l, r) \<in> (E j) \<and> (l, r) \<notin> (E (Suc j))) \<or> ((r, l) \<in> (E j) \<and> (r, l) \<notin> (E (Suc j)))"
by (cases "(l,r) \<in> E\<^sub>\<infinity>", insert assms Einf_without_Ew, auto)

lemma Ew_reducible_aux:
  assumes "(\<exists>r'. (l,r') \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow> \<and> l\<cdot>\<sigma> \<succ> r'\<cdot>\<sigma>) \<or> l \<notin> NF (ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))"
  shows "l\<cdot>\<sigma> \<notin> NF (ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))"
proof-
  from assms consider "\<exists>r'. (l,r') \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow> \<and> l\<cdot>\<sigma> \<succ> r'\<cdot>\<sigma>" | "l \<notin> NF (ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))" by auto
  thus ?thesis proof(cases)
    case 1
    then obtain r' where r':"(l, r') \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "l \<cdot> \<sigma> \<succ> r' \<cdot> \<sigma>" by auto
    have "(l\<cdot>\<sigma>, r'\<cdot>\<sigma>) \<in> ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
      by (rule ordstep.intros[of _ _ _ _ \<box>], insert r', auto)
    then show ?thesis by auto
  next
    case 2
    then obtain t where "(l, t) \<in> ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
    from ordstep_subst_ctxt[OF subst_closed_less this, of \<box>]
      have "(l \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
    thus ?thesis by auto
  qed  
qed
  
lemma Ei_oriented_reducible_in_RiEw: 
  assumes "(l,r) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" and "l\<cdot>\<sigma> \<succ> r\<cdot>\<sigma>"
  shows "(\<exists>r'. (l,r') \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow> \<and> l\<cdot>\<sigma> \<succ> r'\<cdot>\<sigma>) \<or> l \<notin> NF (ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))"
  using assms
proof (induct "{#l, r#}" arbitrary: l r \<sigma> rule: SN_induct [OF lessenc_op.SN])
  case 1
  note IH = this
  have comm: "\<And>t u. {#t, u#} =  {#u, t#}" using add_mset_commute by auto
  show ?case
  proof (cases "(l, r) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>")
    case True
    with 1 show ?thesis by auto
  next
    case False
    with 1(2) Einf_sym_without_Ew_sym obtain j where lr_cases:
      "((l, r) \<in> (E j) \<and> (l, r) \<notin> (E (Suc j))) \<or> ((r, l) \<in> (E j) \<and> (r, l) \<notin> (E (Suc j)))" by auto
    define u' where "u' \<equiv> if (l, r) \<in> (E j) \<and> (l, r) \<notin> (E (Suc j)) then l else r"
    define v' where "v' \<equiv> if (l, r) \<in> (E j) \<and> (l, r) \<notin> (E (Suc j)) then r else l"
    from lr_cases have uv:"(u', v') \<in> (E j)" "(u', v') \<notin> E (Suc j)" by (auto simp: u'_def v'_def)
    have uv_cases:"(u' = l \<and> v' = r) \<or> (u' = r \<and> v' = l)" by (auto simp: u'_def v'_def)
    
    let ?ordstep = "ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
    from ordstep_rstep_conv[OF _ Rinf_less] subst.closedI[of "{\<succ>}"] subst have
      ordstep_rstep:"ordstep {\<succ>} R\<^sub>\<infinity> = rstep R\<^sub>\<infinity>" by auto
    note ordstep_mono = ordstep_mono[of _ "R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>", OF _ subset_refl, of _ "{\<succ>}"]
      
    from subst[of r l \<sigma>] trans[OF IH(3)] irrefl have not_r_succ_l:"\<not> (r \<succ> l)" by auto
    let ?enc = "encstep1 (E (Suc j)) (R (Suc j)) O (E (Suc j))\<^sup>\<leftrightarrow>"
    from uv oKBi_E_supset' [OF irun [of j]]
      consider "(u', v') \<in> R (Suc j)" | "(v', u') \<in> R (Suc j)" | "u' = v'" | "(u', v') \<in> ?enc\<^sup>\<leftrightarrow>"
      by auto blast
    then show ?thesis
    proof (cases)
      case 1
      from not_r_succ_l Ri_less[of "Suc j"] 1 have uv:"u' = l" "v' = r" using uv_cases by auto
      from 1 ordstep_rstep have "(l,r) \<in> ordstep {\<succ>} R\<^sub>\<infinity>" unfolding uv by auto
      with ordstep_mono show ?thesis by auto
    next
      case 2
      from not_r_succ_l Ri_less[of "Suc j"] 2 have uv:"u' = r" "v' = l" using uv_cases by auto
      from 2 ordstep_rstep have "(l,r) \<in> ordstep {\<succ>} R\<^sub>\<infinity>" unfolding uv by auto
      with ordstep_mono show ?thesis by auto
    next
      case 3
      from irrefl IH(3) uv_cases show ?thesis unfolding 3 by auto
    next
      case 4
      then obtain u v w where v: "(u, w) \<in> encstep1 (E (Suc j)) (R (Suc j))" "(w, v) \<in> (E (Suc j))\<^sup>\<leftrightarrow>"
        and uv_cases:"(u = l \<and> v = r) \<or> (u = r \<and> v = l)" using uv_cases by auto
          
      show ?thesis proof(cases "u = l")
        case True
        from v(1) have "u \<notin> NF (ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))"
        proof (cases)
          case (estep l\<^sub>0 r\<^sub>0 D \<tau>)
          from encompeq.intros [of u D "l\<^sub>0 \<cdot> \<tau>" Var] encompeq.intros [of "r\<^sub>0 \<cdot> \<tau>" \<box>] estep(2)
            have u_l0:"l\<^sub>0 \<cdot> \<tau> \<unlhd>\<cdot> u" "r\<^sub>0 \<unlhd>\<cdot> r\<^sub>0 \<cdot> \<tau>" by auto
          with estep have ur0: "u \<cdot>\<succ> r\<^sub>0" "u \<cdot>\<succ> l\<^sub>0" unfolding lessencp_def by blast+
          note mul = ns_s_mul_ext_union_multiset_l[OF ns_mul_ext_bottom, of "{#u#}" "{#l\<^sub>0, r\<^sub>0#}"]    
          with ur0 have "({#u, v#}, {#l\<^sub>0, r\<^sub>0#}) \<in> mulless" using multi_member_last by auto
          with uv_cases comm[of l r] have "({#l, r#}, {#l\<^sub>0, r\<^sub>0#}) \<in> mulless" by presburger
          from IH(1)[OF this _ estep(4)] estep(1) Ew_reducible_aux  
            have NF1:"l\<^sub>0 \<cdot> \<tau> \<notin> NF (ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))" by auto
          from u_l0 obtain C\<^sub>u \<sigma>\<^sub>u where u:"u = C\<^sub>u\<langle>l\<^sub>0 \<cdot> \<tau> \<cdot> \<sigma>\<^sub>u\<rangle>" unfolding encompeq.simps by auto
          from NF1 ordstep_subst_ctxt[OF subst_closed_less] show ?thesis unfolding u by blast
        next
          case rstep
          with ordstep_rstep have "(u, w) \<in> ordstep {\<succ>} R\<^sub>\<infinity>" by fast
          with ordstep_mono[of "R\<^sub>\<infinity>"] show ?thesis by auto
        qed
        with True show ?thesis by auto
      next
        case False
        with uv_cases have uv:"u = r" "v = l" by auto
        from v(1) encstep1_less have rw:"r \<succ> w" unfolding uv by auto
        from rw mset_two2[of r w l] have "({#l, r#}, {#l, w#}) \<in> mulless" using lessencp_def by fast
        from IH(1)[OF this _ trans[OF IH(3) subst, OF rw]] v(2) show ?thesis unfolding uv by fast
      qed
    qed
  qed
qed

lemma Sw_SS:"S\<^sub>\<omega> = \<S> R\<^sub>\<omega> E\<^sub>\<omega>" unfolding Sw_def E_ord_def by auto
  
lemma Ri_reducible_in_RwEw: 
  assumes "(l,r) \<in> R\<^sub>\<infinity>"
  shows "l \<notin> NF (rstep S\<^sub>\<omega>)"
  using assms
proof (induct "(l, r)" arbitrary: l r rule: SN_induct [OF SN_lexless])
  case 1
  note IH = this
  let ?ordstep = "ordstep {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
  show ?case 
  proof (cases "(l, r) \<in> R\<^sub>\<omega>")
    case True
    thus ?thesis unfolding Sw_def by auto
  next
    case False
    with Rinf_without_Rw[OF IH(2)] obtain i where
      not_i: "(l, r) \<notin> R (Suc i)" and pred_i: "(l, r) \<in> R i" by auto
    let ?Ri = "R (Suc i)" and ?Ei = "E (Suc i)"
    from pred_i oKBi_R_supset [OF irun [of i]] and not_i
      consider "(l, r) \<in> encstep2 ?Ei ?Ri O ?Ei" | "(l, r) \<in> ?Ri O (ostep ?Ei ?Ri)\<inverse>" by blast
    then show ?thesis
    proof (cases)
      case 1
      then obtain u where u: "(l, u) \<in> encstep2 ?Ei ?Ri" by auto
      thus ?thesis
      proof(cases)
        case (estep l\<^sub>0 r\<^sub>0 C \<sigma>)
        with Ei_oriented_reducible_in_RiEw[OF _ estep(4)] consider
          "\<exists>r'. (l\<^sub>0, r') \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow> \<and> l\<^sub>0 \<cdot> \<sigma> \<succ> r' \<cdot> \<sigma>" | "l\<^sub>0 \<notin> NF (ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))" by auto
        thus ?thesis proof(cases)
          case 1
          then obtain r' where r':"(l\<^sub>0, r') \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "l\<^sub>0 \<cdot> \<sigma> \<succ> r' \<cdot> \<sigma>" by auto
          hence "(l\<^sub>0 \<cdot> \<sigma>, r' \<cdot> \<sigma>) \<in> rstep S\<^sub>\<omega>" unfolding Sw_def by auto
          thus ?thesis unfolding estep by fast
        next
          case 2
          then obtain t where "(l\<^sub>0, t) \<in> ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
          with ordstep_Un consider "(l\<^sub>0, t) \<in> ordstep {\<succ>} (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" | "(l\<^sub>0, t) \<in> ordstep {\<succ>} R\<^sub>\<infinity>" by auto
          thus ?thesis proof(cases)
            case 1
            from ordstep_subst_ctxt[OF subst_closed_less 1] ordstep_mono
              have "(l, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> ordstep {\<succ>} (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" unfolding estep by auto
            with ordstep_mono[of "E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>"]  have "(l, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> ?ordstep" by auto
            with Sw_SS ordstep_subset_S[OF Rw_less]
              have "(l, C\<langle>t \<cdot> \<sigma>\<rangle>) \<in> rstep S\<^sub>\<omega>" unfolding ordstep_S_conv[OF Rw_less] by auto
            thus ?thesis by fast
          next
            case 2
            from ordstep_imp_rstep[OF this] obtain l\<^sub>1 r\<^sub>1 D \<tau>
              where lr1:"l\<^sub>0 = D\<langle>l\<^sub>1\<cdot>\<tau>\<rangle>" "t = D\<langle>r\<^sub>1\<cdot>\<tau>\<rangle>" "(l\<^sub>1,r\<^sub>1) \<in> R\<^sub>\<infinity>" by fast
            from encompeq.intros[OF lr1(1)] estep(5) have "l \<cdot>\<succ> l\<^sub>1" using lessencp_def by auto
            hence "((l, r), l\<^sub>1, r\<^sub>1) \<in> lexless" by auto
            from IH(1)[OF this lr1(3)] obtain t' where lt:"(l\<^sub>1, t') \<in> rstep S\<^sub>\<omega>" by auto
            thus ?thesis unfolding lr1 estep by fast
          qed
        qed
      next
        case (rstep l\<^sub>0 r\<^sub>0 C \<sigma>)
        from rstep(4) have "l \<cdot>\<succ> l\<^sub>0" using lessencp_def by auto
        hence "((l, r), l\<^sub>0, r\<^sub>0) \<in> lexless" using lessencp_def by auto
        from IH(1)[OF this] rstep(1) have "l\<^sub>0 \<notin> NF_trs S\<^sub>\<omega>" by auto
        then show ?thesis unfolding rstep(2) by blast
      qed
    next
      case 2
      then obtain r' where r':"(l,r') \<in> ?Ri" "(r,r') \<in> ostep ?Ei ?Ri" by fast
      from ostep_imp_less[OF Ri_less this(2)] have "r \<cdot>\<succ> r'" using lessencp_def by fast
      hence "((l, r), l, r') \<in> lexless" by auto
      from IH(1)[OF this] r'(1) show ?thesis by auto
    qed
  qed
qed

lemma Ei_oriented_reducible_in_RwEw: 
  assumes "(l,r) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" and "l\<cdot>\<sigma> \<succ> r\<cdot>\<sigma>"
  shows "l\<cdot>\<sigma> \<notin> NF_trs S\<^sub>\<omega>"
proof-
  from Ew_reducible_aux[OF Ei_oriented_reducible_in_RiEw[OF assms]]
    have "\<exists>t. (l\<cdot>\<sigma>, t) \<in> ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
  then obtain t where t:"(l\<cdot>\<sigma>, t) \<in> ordstep {\<succ>} (R\<^sub>\<infinity> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
  { assume "(l\<cdot>\<sigma>, t) \<in> ordstep {\<succ>} R\<^sub>\<infinity>"
    from ordstep_imp_rstep[OF this] obtain C \<tau> l\<^sub>0 r\<^sub>0 where
      lr0:"(l\<^sub>0, r\<^sub>0) \<in> R\<^sub>\<infinity>" and lt:"l \<cdot> \<sigma> = C\<langle>l\<^sub>0 \<cdot> \<tau>\<rangle>" by fast
    from Ri_reducible_in_RwEw[OF lr0] have "l \<cdot> \<sigma> \<notin> NF_trs S\<^sub>\<omega>" unfolding lt by fastforce
  } note Rinf = this
  { assume a:"(l\<cdot>\<sigma>, t) \<in> ordstep {\<succ>} (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
    with ordstep_mono[of "E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>"]  have "(l\<cdot>\<sigma>, t) \<in> ordstep {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by auto
            with Sw_SS ordstep_subset_S[OF Rw_less]
              have "(l\<cdot>\<sigma>, t) \<in> rstep S\<^sub>\<omega>" unfolding ordstep_S_conv[OF Rw_less] by auto
  }
  with t Rinf show ?thesis unfolding ordstep_Un by auto
qed
  
lemma Einf_Ew_cases:
  assumes "(s,t) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>"
shows "s \<notin> NF_trs S\<^sub>\<omega> \<or> (\<exists>w. (s,w) \<in> (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)\<^sup>= \<and> t \<succeq> w)"
proof-
  { fix s t
  assume st:"(s,t) \<in> E\<^sub>\<infinity>"
  define P_Ew where "P_Ew \<equiv> \<lambda>s t. \<exists>w. (s,w) \<in> (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)\<^sup>= \<and> t \<succeq> w"
  define P_ordstep where "P_ordstep \<equiv> \<lambda>t. t \<notin> NF_trs S\<^sub>\<omega>"
  from st have "(P_ordstep s \<or> P_Ew s t) \<and> (P_ordstep t \<or> P_Ew t s)" using st
  proof (induct "{#s, t#}" arbitrary: s t rule: SN_induct [OF lessenc_op.SN])
    case 1
    let ?R1 = "\<exists>u. (s,u) \<in> E\<^sub>\<omega>\<^sup>= \<and> (t,u) \<in> (mstep {#t#} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
    show ?case proof(cases "(s,t) \<in> E\<^sub>\<omega>")
      case True
      hence "P_Ew s t \<and> P_Ew t s" unfolding P_Ew_def by blast
      thus ?thesis by blast
    next
      case False
      let ?ordstep = "ordstep {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)"
      from False 1(2) Einf_without_Ew obtain i where
        not_i: "(s, t) \<notin> E (Suc i)" and pred_i: "(s, t) \<in> E i" by auto
      let ?Ei = "E (Suc i)" and ?Ri = "R (Suc i)"
      from oKBi_conversion_ERw oKBi_conversion_ERi
      have convi:"\<And>i.(rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E i \<union> R i))\<^sup>\<leftrightarrow>\<^sup>*" by auto  
      { fix s t
        assume a:"(s, t) \<in> ?Ri"
        with convi[of "Suc i"] have st:"(s,t) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
        from a Ri_reducible_in_RwEw 1 have "s \<notin> NF_trs S\<^sub>\<omega>" by auto
        hence 1:"P_ordstep s" unfolding P_ordstep_def by auto
        from a Ri_less have st:"s \<succ> t" by auto
        from st 1 have "P_ordstep s \<and> P_Ew t s" unfolding P_Ew_def by auto
      } note case_orient = this
        
      { fix s' t'
        assume a:"(s', t') \<in> encstep1 ?Ei ?Ri O ?Ei\<^sup>\<leftrightarrow>" "(s = s' \<and> t = t') \<or> (s = t' \<and> t = s')"
        then obtain u where u:"(s', u) \<in> encstep1 ?Ei ?Ri" "(u,t') \<in> ?Ei\<^sup>\<leftrightarrow>" by auto
        from pred_i a(2) have "(s', t') \<in> (E i)\<^sup>\<leftrightarrow>" by auto
        with convi[of i] have st:"(s', t') \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" unfolding conversion_def by fastforce
        from encstep1_ostep[OF u(1)] consider "(s', u) \<in> rstep ?Ri" | "(s', u) \<in> ordstep {\<succ>} (?Ei\<^sup>\<leftrightarrow>)"
          unfolding ostep_def by auto
        hence ordstep:"P_ordstep s'" proof(cases)
          case 1
          then obtain l r C \<sigma> where lr:"(l,r) \<in> R\<^sub>\<infinity>" "s' = C\<langle>l\<cdot>\<sigma>\<rangle>" "u = C\<langle>r\<cdot>\<sigma>\<rangle>" by fast
          from Ri_reducible_in_RwEw[OF lr(1)] obtain l' where "(l,l') \<in> rstep S\<^sub>\<omega>" by auto
          thus ?thesis unfolding lr P_ordstep_def by fast
        next
          case 2
          from this[unfolded ordstep.simps] obtain l r C \<sigma> where
            lr:"(l,r) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" "s' = C\<langle>l\<cdot>\<sigma>\<rangle>" "u = C\<langle>r\<cdot>\<sigma>\<rangle>" "l \<cdot> \<sigma> \<succ> r \<cdot> \<sigma>" by fast
          from Ei_oriented_reducible_in_RwEw[OF lr(1) lr(4)] obtain l' where
            "(l\<cdot>\<sigma>,l') \<in> rstep S\<^sub>\<omega>" by auto
          thus ?thesis unfolding lr P_ordstep_def by auto
        qed
          
        have comm: "\<And>t u. {#t, u#} =  {#u, t#}" using add_mset_commute by auto
        from u(2) have or:"(u,t') \<in> ?Ei \<or> (t',u) \<in> ?Ei" by auto
        from encstep1_less[OF u(1)] mset_two have "({#s', t'#}, {#u, t'#}) \<in> s_mul_ext {\<cdot>\<unrhd>} {\<cdot>\<succ>}"
          unfolding lessencp_def by fast
        with a(2) comm[of s' t'] have less1:"({#s, t#}, {#u, t'#}) \<in> s_mul_ext {\<cdot>\<unrhd>} {\<cdot>\<succ>}" by presburger
        with comm[of u t'] have less2:"({#s, t#}, {#t', u#}) \<in> s_mul_ext {\<cdot>\<unrhd>} {\<cdot>\<succ>}" by presburger
        from 1(1)[OF less1] 1(1)[OF less2] or u(2) 
          have "(P_Ew u t' \<or> P_ordstep u) \<and> (P_Ew t' u \<or> P_ordstep t')" by auto
        then consider "P_Ew t' u" | "P_ordstep t'" by auto
        hence "P_Ew t' s' \<or> P_ordstep t'"
        proof(cases)
          case 1
          then obtain v where v:"(t', v) \<in> (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)\<^sup>=" "u \<succeq> v"
            unfolding P_Ew_def by auto
          from encstep1_imp_rstep_sym[OF u(1)] rstep_UN have tu:"(s', u) \<in> rstep (E\<^sub>\<infinity>\<^sup>\<leftrightarrow> \<union> R\<^sub>\<infinity>)" by fast
          from encstep1_less[OF u(1)] v(2) trans have "s' \<succeq> v" by blast
          with v(1) show ?thesis unfolding P_Ew_def by blast
        next
          case 2
          thus ?thesis unfolding P_ordstep_def by auto
        qed
        with ordstep have "P_ordstep s' \<and> (P_Ew t' s' \<or> P_ordstep t')" by auto
      } note case_simplify = this
        
      let ?encl = "encstep1 ?Ei ?Ri O ?Ei"
      let ?encr = "?Ei O ((encstep1 ?Ei ?Ri)\<inverse>)"
      from pred_i oKBi_E_supset' [OF irun [of i]] and not_i
      consider "(s,t) \<in> ?Ri" | "(t,s) \<in> ?Ri" | "s = t" | "(s,t) \<in> ?encl" | "(s,t) \<in> ?encr" by blast
      then show ?thesis using case_orient case_simplify P_Ew_def
      proof (cases)
        case 4
        with case_simplify show ?thesis by auto
      next
        case 5
        hence "(t,s) \<in> encstep1 ?Ei ?Ri O ?Ei\<^sup>\<leftrightarrow>" by auto
        from case_simplify[OF this] show ?thesis by auto
      qed auto
    qed
  qed
  }
  with assms show ?thesis by auto
qed
  
lemma peak_cases:
  assumes step:"(s,t) \<in> rstep_r_p_s (E\<^sub>\<omega>\<^sup>\<leftrightarrow>) (l,r) p \<sigma>" and irstep:"(t,u) \<in> irstep False S\<^sub>\<omega>"
    and "\<not>(l\<cdot>\<sigma> \<succ> r\<cdot>\<sigma>)" and "\<not>(r\<cdot>\<sigma> \<succeq> l\<cdot>\<sigma>)"
  shows "s \<notin> NF_trs S\<^sub>\<omega> \<or> (\<exists>s'. s \<rhd> s' \<and> s' \<notin> NF_trs \<R>) \<or> (\<exists>w. (s,w) \<in> (rstep (E\<^sub>\<omega>\<^sup>\<leftrightarrow>))\<^sup>= \<and> u \<succeq> w)" (is "?A \<or> ?B \<or> ?C")
proof-
  from step obtain C where lr:"(l,r) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "s = C\<langle>l\<cdot>\<sigma>\<rangle>" "t = C\<langle>r\<cdot>\<sigma>\<rangle>" "C = ctxt_of_pos_term p s"
    unfolding rstep_r_p_s_def mem_Collect_eq split Let_def by simp
  show ?thesis proof(cases "C = \<box>")
    (* first consider the case where the step s \<rightarrow> t happens strictly below the root *)
    case False
    note gcomp = ground_subt_less[OF ground_sk]
    from False lr(2) have tt:"s \<rhd> l\<cdot>\<sigma>" by (cases C, auto)
    from lr(1) R_NF_eq[of "l\<cdot>\<sigma>" "r\<cdot>\<sigma>"] obtain w where
      w:"(l \<cdot> \<sigma>, w) \<in> (rstep \<R>)\<^sup>!" "(r \<cdot> \<sigma>, w) \<in> (rstep \<R>)\<^sup>!" by auto
    note w_nf = w[unfolded normalizability_def mem_Collect_eq split]
    from w_nf(2) compatible_rstep_trancl_imp_less[OF R_less] have rw:"r \<cdot> \<sigma> \<succeq> w"
      unfolding rtrancl_eq_or_trancl by fast
    with assms(4) have lw:"l \<cdot> \<sigma> \<noteq> w" by auto
    with w_nf(1) have "\<exists>l'. (l \<cdot> \<sigma>,l') \<in> rstep \<R>"
      unfolding lr(2) rtrancl_eq_or_trancl trancl_unfold_left by auto
    with nectxt_imp_supt_ctxt[OF False, of "l \<cdot> \<sigma>"] lr(2) show ?thesis by auto
  next
    (* here the step s \<rightarrow> t happens at the root *)
    case True
    note C_Hole = this
    from irstep[unfolded irstep_def, THEN qrstepE'] obtain D l\<^sub>2 r\<^sub>2 \<sigma>\<^sub>2 where
      lr2:"(l\<^sub>2,r\<^sub>2) \<in> S\<^sub>\<omega>" "t = D\<langle>l\<^sub>2\<cdot>\<sigma>\<^sub>2\<rangle>" "u = D\<langle>r\<^sub>2\<cdot>\<sigma>\<^sub>2\<rangle>" "\<forall>u\<lhd>l\<^sub>2 \<cdot> \<sigma>\<^sub>2. u \<in> NF_terms (lhss S\<^sub>\<omega>)" by blast 
    let ?p = "hole_pos D"
    (* do a CP Lemma-like analysis (standard CP lemma requires variable condition) *)
    show ?thesis proof(cases "?p \<in> funposs r")
      case False (* variable overlap *)
      note hole_D_not_in_funposs_r = this
      from lr(3) lr2(2) have p_r\<sigma>:"?p \<in> poss (r \<cdot> \<sigma>)" unfolding C_Hole by fastforce
      from poss_subst_apply_term[OF this hole_D_not_in_funposs_r] obtain q\<^sub>1 q\<^sub>2 x
        where p[simp]: "?p = q\<^sub>1 <#> q\<^sub>2" and q\<^sub>1: "q\<^sub>1 \<in> poss r"
          and rq\<^sub>1: "r |_ q\<^sub>1 = Var x" and q\<^sub>2: "q\<^sub>2 \<in> poss (\<sigma> x)" by auto
      note eq = lr(3)[unfolded lr2(2) True, simplified]
      hence [simp]: "r \<cdot> \<sigma> |_ ?p = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using subt_at_hole_pos by metis
      with rq\<^sub>1 q\<^sub>2 subt_at_append p_r\<sigma> q\<^sub>1 have [simp]: "\<sigma> x |_ q\<^sub>2 = l\<^sub>2 \<cdot>\<sigma>\<^sub>2" unfolding p by auto
      define \<sigma>' where "\<sigma>' y = (if y = x then replace_at (\<sigma> x) q\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2) else \<sigma> y)" for y
      have "(\<sigma> x, \<sigma>' x) \<in> rstep S\<^sub>\<omega>"
      proof -
        let ?C = "ctxt_of_pos_term q\<^sub>2 (\<sigma> x)"
        have "(?C\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>, ?C\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>) \<in> rstep S\<^sub>\<omega>" using lr2 by blast
        then show ?thesis using q\<^sub>2 by (simp add: \<sigma>'_def replace_at_ident)
      qed
      then have *: "\<And>x. (\<sigma> x, \<sigma>' x) \<in> (rstep S\<^sub>\<omega>)\<^sup>*" by (auto simp: \<sigma>'_def)
      then have left:"(l \<cdot> \<sigma>, l \<cdot> \<sigma>') \<in> (rstep S\<^sub>\<omega>)\<^sup>*" by (rule substs_rsteps)
      moreover
      have right:"(u, r \<cdot> \<sigma>') \<in> (rstep S\<^sub>\<omega>)\<^sup>*"
      proof -
        have "replace_at (r \<cdot> \<sigma>) ?p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2) = replace_at (r \<cdot> \<sigma>) q\<^sub>1 (\<sigma>' x)"
          using q\<^sub>1 and q\<^sub>2 by (simp add: \<sigma>'_def ctxt_of_pos_term_append rq\<^sub>1)
        moreover
        have "(replace_at (r \<cdot> \<sigma>) q\<^sub>1 (\<sigma>' x), r \<cdot> \<sigma>') \<in> (rstep S\<^sub>\<omega>)\<^sup>*"
          by (rule replace_at_subst_rsteps [OF * q\<^sub>1 rq\<^sub>1])
        ultimately
        have "(replace_at (r \<cdot> \<sigma>) ?p (r\<^sub>2 \<cdot> \<sigma>\<^sub>2), r \<cdot> \<sigma>') \<in> (rstep S\<^sub>\<omega>)\<^sup>*" by auto
        thus ?thesis unfolding lr2(3) unfolding eq[symmetric] ctxt_of_pos_term_hole_pos
          by blast 
      qed
      from lr(1) have E_step:"(l \<cdot> \<sigma>', r \<cdot> \<sigma>') \<in> (rstep E\<^sub>\<omega>)\<^sup>\<leftrightarrow>" by auto
      from left right Sw_steps_lesseq have leq:"u \<succeq> r \<cdot> \<sigma>'" unfolding lr(2) C_Hole by simp
      show ?thesis proof(cases "l \<cdot> \<sigma> = l \<cdot> \<sigma>'")
        case True
        have "?C" unfolding True lr(2) C_Hole by (rule exI[of _ "r \<cdot> \<sigma>'"], insert E_step leq, auto)
        thus ?thesis by auto
      next
        case False
        with left[unfolded rtrancl_eq_or_trancl trancl_unfold_left] show ?thesis
          unfolding lr(2) by auto
      qed
    next
      case True (* proper overlap *)
      from lr lr2 C_Hole have eq:"D\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle> = r \<cdot> \<sigma>" by auto
      note st = lr[unfolded C_Hole ctxt_apply_term.simps]
      note tu = lr2[unfolded C_Hole ctxt_apply_term.simps]
      note rl_cases = lr2(1)[unfolded Sw_def Un_iff[of _ "R\<^sub>\<omega>"]]
      have "\<exists>\<tau> l\<^sub>3 r\<^sub>3. l\<^sub>2 = l\<^sub>3\<cdot>\<tau> \<and> r\<^sub>2 = r\<^sub>3\<cdot>\<tau> \<and> (l\<^sub>3, r\<^sub>3) \<in> R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>"
        by (cases "(l\<^sub>2, r\<^sub>2) \<in> R\<^sub>\<omega>", rule exI[of _ Var], insert rl_cases, auto)
      then obtain \<tau> l\<^sub>3 r\<^sub>3 where lr3:"l\<^sub>2 = l\<^sub>3\<cdot>\<tau>" "r\<^sub>2 = r\<^sub>3\<cdot>\<tau>" "(l\<^sub>3, r\<^sub>3) \<in> R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" by auto
      from eq[unfolded this] subst_subst have eq':"D\<langle>l\<^sub>3 \<cdot> (\<tau> \<circ>\<^sub>s \<sigma>\<^sub>2)\<rangle> = r \<cdot> \<sigma>" by auto
      from lr(1) have rl0:"(r,l) \<in> R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" by auto
      from lr2(1) Sw_less have "l\<^sub>3 \<cdot> \<tau> \<succ> r\<^sub>3 \<cdot> \<tau>" unfolding lr3 by auto
      from subst[OF this] trans irrefl have unoriented':"\<not> r\<^sub>3 \<cdot> \<tau> \<circ>\<^sub>s \<sigma>\<^sub>2 \<succ> l\<^sub>3 \<cdot> \<tau> \<circ>\<^sub>s \<sigma>\<^sub>2"
        unfolding subst_subst[symmetric] by fast
          
      note ooverlap = ooverlap_exists[OF eq' True lr3(3) rl0 unoriented' assms(3)]
      hence ooverlap:"\<exists>r r' \<mu> \<tau> s t. ooverlap {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>) r r' (hole_pos D) \<mu> s t \<and>
               D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle> = s \<cdot> \<tau> \<and> l \<cdot> \<sigma> = t \<cdot> \<tau> \<and> l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = (fst r) \<cdot> \<mu> \<cdot> \<tau>" unfolding lr3 by auto
      then obtain rl\<^sub>1 rl\<^sub>2 \<mu> \<tau> s\<^sub>0 t\<^sub>0 where
        ooverlap:"ooverlap {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>) rl\<^sub>1 rl\<^sub>2 (hole_pos D) \<mu> s\<^sub>0 t\<^sub>0"
        and dec:"D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle> = s\<^sub>0 \<cdot> \<tau>" "l \<cdot> \<sigma> = t\<^sub>0 \<cdot> \<tau>" "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = (fst rl\<^sub>1) \<cdot> \<mu> \<cdot> \<tau>" by auto
      { fix u
        assume "u \<lhd> fst rl\<^sub>1 \<cdot> \<mu>"
        hence  "u \<cdot> \<tau> \<lhd> fst rl\<^sub>1 \<cdot> \<mu> \<cdot> \<tau>" by auto
        with lr2(4)[unfolded dec] have nf:"u \<in> NF_trs S\<^sub>\<omega>" unfolding NF_terms_lhss by fast
        have "\<And>s t. (s,t) \<in> ordstep {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>) \<Longrightarrow> (s,t) \<in> rstep S\<^sub>\<omega>"
          unfolding Sw_def rstep_union ordstep.simps by blast
        from nf NF_anti_mono[OF subrelI[OF this]] have "u \<in> NF (ordstep {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>))"
          by blast
      }
      hence prime:"(\<forall>u\<lhd>fst rl\<^sub>1 \<cdot> \<mu>. u \<in> NF (ordstep {\<succ>} (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)))" by auto
          
      with ooverlap have "(s\<^sub>0, t\<^sub>0) \<in> PCP_ext (R\<^sub>\<omega> \<union> E\<^sub>\<omega>\<^sup>\<leftrightarrow>)" unfolding PCP_ext_def mem_Collect_eq
        by blast
      with fair have "(s\<^sub>0, t\<^sub>0) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>" by auto
      with dec have "(D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>, l \<cdot> \<sigma>) \<in> (rstep E\<^sub>\<infinity>)\<^sup>\<leftrightarrow>" by auto
      with lr(2) lr2(3) have Einf_step:"(s, u) \<in> rstep (E\<^sub>\<infinity>\<^sup>\<leftrightarrow>)"
        unfolding C_Hole rstep_simps(5)[symmetric] by fastforce
      from rstep_imp_C_s_r[OF this] obtain l\<^sub>C\<^sub>P r\<^sub>C\<^sub>P C\<^sub>C\<^sub>P \<sigma>\<^sub>C\<^sub>P where
        decomp:"(l\<^sub>C\<^sub>P, r\<^sub>C\<^sub>P) \<in> E\<^sub>\<infinity>\<^sup>\<leftrightarrow>" "s = C\<^sub>C\<^sub>P\<langle>l\<^sub>C\<^sub>P \<cdot> \<sigma>\<^sub>C\<^sub>P\<rangle>" "u = C\<^sub>C\<^sub>P\<langle>r\<^sub>C\<^sub>P \<cdot> \<sigma>\<^sub>C\<^sub>P\<rangle>" by metis
      from Einf_Ew_cases[OF decomp(1)] consider
        "l\<^sub>C\<^sub>P \<notin> NF_trs S\<^sub>\<omega>" | "(\<exists>w. (l\<^sub>C\<^sub>P, w) \<in> (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)\<^sup>= \<and> r\<^sub>C\<^sub>P \<succeq> w)" by auto
      thus ?thesis proof(cases)
        case 2
        then obtain w' where w':"(l\<^sub>C\<^sub>P, w') \<in> (E\<^sub>\<omega>\<^sup>\<leftrightarrow>)\<^sup>=" "r\<^sub>C\<^sub>P \<succeq> w'" by auto
        define w where "w \<equiv> C\<^sub>C\<^sub>P\<langle>w' \<cdot> \<sigma>\<^sub>C\<^sub>P\<rangle>"
        with decomp w'(1) have estep:"(s, w) \<in> (rstep (E\<^sub>\<omega>\<^sup>\<leftrightarrow>))\<^sup>=" by blast
        from w' subst ctxt have vw:"u \<succeq> w" unfolding decomp w_def by auto
        with estep show ?thesis by auto
      next
        case 1
        then obtain w' where w':"(l\<^sub>C\<^sub>P, w') \<in> rstep S\<^sub>\<omega>" by auto
        from decomp w' have S_step:"(s,C\<^sub>C\<^sub>P\<langle>w' \<cdot> \<sigma>\<^sub>C\<^sub>P\<rangle>) \<in> rstep S\<^sub>\<omega>" by blast
        thus ?thesis by auto
      qed
    qed
  qed
qed
  
lemma R_norm_implies_Sw_steps:
  assumes tu:"(t,u) \<in> (rstep \<R>)\<^sup>!"
  shows "(t,u) \<in> (rstep S\<^sub>\<omega>)\<^sup>*"
proof -
  (* auxiliary *)
  from oKBi_conversion_ERw conv_eq have conv_eq':"(rstep \<R>)\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  from sk_run.compat_h[THEN sk_run.subst_h] have sk_compat_less:"\<And> s t. s \<succ> t \<Longrightarrow> sk s \<succ>\<^sub>s\<^sub>k sk t"
    unfolding sk_def by auto
  from sk_run.compat_h' sk_run.subst_h have sk_compat_leq:"\<And> s t. s \<succeq> t \<Longrightarrow> sk s \<succeq>\<^sub>s\<^sub>k sk t"
    unfolding sk_def by fastforce
  from sk_run.irun_h.Sw_step_less have Sstep_less:"\<And>s t. (s,t) \<in> rstep Sh\<^sub>\<omega> \<Longrightarrow> s \<succ>\<^sub>s\<^sub>k t" by auto
  have NF_R_less:"\<And>t u. (t,u) \<in> (rstep \<R>)\<^sup>! \<Longrightarrow> t \<noteq> u \<Longrightarrow> t \<succ> u"
    using compatible_rstep_trancl_imp_less[OF R_less] unfolding normalizability_def
    by (metis (no_types, lifting) case_prodD mem_Collect_eq rtranclD)
  (* obtain conversion in Sh\<^sub>\<omega> for (t,u) *)
  let ?steps = "\<lambda>ts n. \<forall>i<n. (sk (ts (Suc i)), sk (ts i)) \<in> rstep Sh\<^sub>\<omega>"
  (*from R_NF_Shw_conv[OF tu] obtain ts n where ts:"ts 0 = u" "ts n = t" "?steps ts n" by auto*)
  show ?thesis proof(cases "t = u", simp)
    case False
    with tu obtain t' where "(t,t') \<in> rstep \<R>"
      unfolding normalizability_def rtrancl_eq_or_trancl mem_Collect_eq split trancl_unfold_left by auto
  from Rstep_imp_S_step_sk[OF this] obtain t\<^sub>1 where "(sk t, sk t\<^sub>1) \<in> rstep Sh\<^sub>\<omega>" by auto
    from this tu show ?thesis
    proof(induct "(sk t, sk t\<^sub>1)" arbitrary:t t\<^sub>1 u rule: SN_induct [OF SN_lexless_sk])
      case 1
      note ind_step = this
      from 1(3) have u_NF:"u \<in> NF (rstep \<R>)" unfolding normalizability_def by auto
      (* auxiliary: a term convertible to t has NF u *)
      { fix t\<^sub>1
        assume "(t,t\<^sub>1) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
        with conv_eq' have "(t, t\<^sub>1) \<in> (rstep \<R>)\<^sup>\<leftrightarrow>\<^sup>*" by blast
        with 1(3) have "(u, t\<^sub>1) \<in> (rstep \<R>)\<^sup>\<leftrightarrow>\<^sup>*"
          unfolding normalizability_def mem_Collect_eq split conversion_def
          by (smt conversionE conversionI' conversion_inv relcomp.relcompI rtrancl_idemp_self_comp)
        with CR_R[unfolded CR_iff_conversion_imp_join] have "(t\<^sub>1, u) \<in> (rstep \<R>)\<^sup>\<down>" by auto
        with NF_join_imp_reach[OF u_NF] u_NF have "(t\<^sub>1, u) \<in> (rstep \<R>)\<^sup>!" unfolding normalizability_def
          by auto
      } note t_conv_R_NF = this
      from NF_R_NF_S 1(2) have "t \<notin> NF_trs \<R>" by auto
      with 1(3) have t_neq_u:"t \<noteq> u" by auto
      have nf_neq_step:"\<And>v u. (v, u) \<in> (rstep \<R>)\<^sup>! \<Longrightarrow> v \<noteq> u \<Longrightarrow> \<exists>v'.(v,v') \<in> rstep \<R>"
        unfolding normalizability_def rtrancl_eq_or_trancl mem_Collect_eq trancl_unfold_left by auto
      (* auxiliary: simpler induction hypothesis where already first conversion term is smaller *)
      { fix v u w
        assume nf:"(v, u) \<in> (rstep \<R>)\<^sup>!" and less:"sk t \<succ>\<^sub>s\<^sub>k sk v"
        have S_steps_vu:"(v, u) \<in> (rstep S\<^sub>\<omega>)\<^sup>*"
        proof(cases "v = u", simp)
          case False
          with nf_neq_step[OF nf] obtain v' where "(v,v') \<in> rstep \<R>" by auto
          from Rstep_imp_S_step_sk[OF this] obtain w where S_step:"(sk v, sk w) \<in> rstep Sh\<^sub>\<omega>" by auto
          from less have "((sk t, sk t\<^sub>1), (sk v, sk w)) \<in> lexless_sk" by auto
          from 1(1)[OF this S_step nf] show ?thesis by auto
        qed
      } note ih = this
        
      note ctxt_conv = ctxt.closed_conversion[OF ctxt_closed_rstep, THEN ctxt.closedD]
      from 1(2) sk_run.irun_h.Sw_implies_REw have "(sk t, sk t\<^sub>1) \<in> (rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>))\<^sup>\<leftrightarrow>" by auto
      with rstep_sk have step:"(t,t\<^sub>1) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>" unfolding rstep_union
        using sk.hR_union sk_run.hEw sk_run.hRw by auto
      show ?case proof(cases "t\<^sub>1 = u")
        case True
        from NF_R_less[OF ind_step(3) t_neq_u] have tu:"t \<succ> u" by auto
        from oriented_REw_step_Sw_step[OF step[unfolded True] tu] u_NF show ?thesis by auto
      next
        case False
        note t1_neq_u = this
        from step t_conv_R_NF have "(t\<^sub>1, u) \<in> (rstep \<R>)\<^sup>!" by blast
        with nf_neq_step[OF _ False] Rstep_imp_S_step_sk obtain t\<^sub>2 where "(sk t\<^sub>1, sk t\<^sub>2) \<in> rstep Sh\<^sub>\<omega>"
          by presburger
        with 1(2) sk_run.irun_h.Sw_step_less have sk_tt1:"sk t \<succ>\<^sub>s\<^sub>k sk t\<^sub>1" by auto
      
        (* apply induction hypothesis to t\<^sub>1 *)
        from 1(2) sk_run.irun_h.Sw_implies_REw have "(sk t, sk t\<^sub>1) \<in> (rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>))\<^sup>\<leftrightarrow>" by auto
        with rstep_sk have step:"(t,t\<^sub>1) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>" unfolding rstep_union
          using sk.hR_union sk_run.hEw sk_run.hRw by auto
        with t_conv_R_NF[of t\<^sub>1] have "(t\<^sub>1, u) \<in> (rstep \<R>)\<^sup>!" by blast
        from ih[OF this sk_tt1(1)] have t1_u:"(t\<^sub>1,u) \<in> (rstep S\<^sub>\<omega>)\<^sup>*" by auto

        (* if t \<rightarrow> t\<^sub>1 is an S\<^sub>\<omega>-step we are done*)
        show ?thesis proof(cases "(t,t\<^sub>1) \<in> rstep S\<^sub>\<omega>")
          case True
          with t1_u show ?thesis by auto
        next
          case False
          (* the first step must be an E\<^sub>\<omega> step *)
          { assume "(t\<^sub>1,t) \<in> rstep R\<^sub>\<omega>"
            with compatible_rstep_imp_less[OF Rw_less] sk_compat_less have "sk t\<^sub>1 \<succ>\<^sub>s\<^sub>k sk t" by auto
            with sk_tt1(1) have False using sk_run.okb_h.irrefl sk_run.trans_h by blast
          }
          with False[unfolded Sw_def rstep_union] step have t_t1:"(t,t\<^sub>1) \<in> (rstep E\<^sub>\<omega>)\<^sup>\<leftrightarrow>" by fastforce
          then obtain l and r C \<sigma> where lr:"(l,r) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow>" "t = C\<langle>l\<cdot>\<sigma>\<rangle>" "t\<^sub>1 = C\<langle>r\<cdot>\<sigma>\<rangle>" by fast
          have unoriented:"\<not>(l\<cdot>\<sigma> \<succ> r\<cdot>\<sigma>)" by (rule ccontr, insert lr False, unfold Sw_def, blast)
          
          (* we can choose the step from t\<^sub>1 to be innermost (required for PCP) *)
          from tranclD t1_neq_u t1_u[unfolded rtrancl_eq_or_trancl]
            obtain u' where "(t\<^sub>1, u') \<in> rstep S\<^sub>\<omega>" by metis
          from rstep_imp_irstep[OF this, of False] obtain v where irstep:"(t\<^sub>1, v) \<in> irstep False S\<^sub>\<omega>"
            by auto
          with qrstep_into_rstep have S_step_t1_v:"(t\<^sub>1, v) \<in> rstep S\<^sub>\<omega>" unfolding irstep_def by auto
          from this[unfolded rstep_sk] sk_run.hSw have "(sk t\<^sub>1, sk v) \<in> rstep Sh\<^sub>\<omega>" by auto
          with sk_run.irun_h.Sw_step_less have t1_gt_v:"sk t\<^sub>1 \<succ>\<^sub>s\<^sub>k sk v" by blast
          with sk_run.trans_h \<open>sk t  \<succ>\<^sub>s\<^sub>k sk t\<^sub>1\<close> have less_sk:"sk t \<succ>\<^sub>s\<^sub>k sk v" by blast
          (* ... apply IH again to get a sequence starting with an innermost step *)
          from S_step_t1_v Sw_implies_REw have "(t\<^sub>1, v) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by blast
          with converse_rtrancl_into_rtrancl[OF step] have "(t, v) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*"
            unfolding conversion_def by blast
          with t_conv_R_NF have nf:"(v, u) \<in> (rstep \<R>)\<^sup>!" by blast
          from ih[OF this less_sk] have "(v, u) \<in> (rstep S\<^sub>\<omega>)\<^sup>*" by auto
           
          (* consider peak analysis *)
          from lr have step:"(t, t\<^sub>1) \<in> rstep_r_p_s (E\<^sub>\<omega>\<^sup>\<leftrightarrow>) (l, r) (hole_pos C) \<sigma>"
            unfolding rstep_r_p_s_def by auto
          { assume "r\<cdot>\<sigma> \<succeq> l\<cdot>\<sigma>"
            with ctxt have "t\<^sub>1 \<succeq> t" unfolding lr by auto
            from sk_compat_leq[OF this] have "sk t\<^sub>1 \<succeq>\<^sub>s\<^sub>k sk t" by auto
            with sk_tt1 sk_run.trans_h[OF sk_tt1] sk_run.okb_h.irrefl have False by auto
          } note unoriented2 = this
          (* three cases can be distinguished with respect to the peak *)
          note peak_cases = unoriented2 peak_cases[OF step irstep unoriented]
          (* ... but the second case implies the first one, by the induction hypothesis: *)
          { assume "\<exists>s'\<lhd>t. s' \<notin> NF_trs \<R>"
            then obtain s' where s':"t \<rhd> s'" "s' \<notin> NF_trs \<R>" by auto
            with ground_subt_less[OF ground_sk] sk.subterm_strict_h supt_subst have lt:"sk t \<succ>\<^sub>s\<^sub>k sk s'"
              unfolding sk_def by metis
            from SN_reaches_NF[of "rstep \<R>"] SN_R have "\<exists>s. (s', s) \<in> (rstep \<R>)\<^sup>* \<and> s \<in> NF_trs \<R>"
              unfolding normalizability_def by (simp add: SN_defs)
            then obtain s'' where s:"(s', s'') \<in> (rstep \<R>)\<^sup>!" "s'' \<in> NF_trs \<R>"
              unfolding normalizability_def by auto
            with s' have "s'' \<noteq> s'" by auto
            with ih[OF s(1) lt, unfolded rtrancl_eq_or_trancl trancl_unfold_left] s'(1)
            have "t \<notin> NF_trs S\<^sub>\<omega>" by auto
          }
          with peak_cases consider "\<exists>w. (t, w) \<in> (rstep (E\<^sub>\<omega>\<^sup>\<leftrightarrow>))\<^sup>= \<and> v \<succeq> w" | "t \<notin> NF_trs S\<^sub>\<omega>" by auto
              
          thus ?thesis proof(cases)
            case 1
            then obtain w where w:"(t, w) \<in> (rstep (E\<^sub>\<omega>\<^sup>\<leftrightarrow>))\<^sup>=" "v \<succeq> w" by auto 
            note t1_ge = t1_gt_v sk_run.trans_h[OF t1_gt_v]
            with sk_compat_leq[OF w(2)] have lex:"((sk t, sk t\<^sub>1), sk t, sk w) \<in> lexless_sk" by auto
            from t1_ge sk_compat_leq[OF w(2)] have sk_t1_gt_w:"sk t\<^sub>1 \<succ>\<^sub>s\<^sub>k sk w" by auto 
            from sk_run.trans_h[OF \<open>sk t \<succ>\<^sub>s\<^sub>k sk t\<^sub>1\<close> sk_t1_gt_w] have tw:"sk t \<succ>\<^sub>s\<^sub>k sk w" by auto
            with sk_run.okb_h.irrefl sk_compat_leq have t_neq_w:"t \<noteq> w" by auto
            with w(1) have "(t,w) \<in> rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>) \<or> (w,t) \<in> rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>)" by auto
            hence sk_step:"(sk t, sk w) \<in> (rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>))\<^sup>\<leftrightarrow>"
              unfolding rstep_sk[of _ _ "R\<^sub>\<omega> \<union> E\<^sub>\<omega>"] sk.hR_union sk_run.hRw sk_run.hEw by blast
            from sk_run.trans_h[OF tw] sk_run.okb_h.irrefl have nwt:"\<not>(sk w \<succ>\<^sub>s\<^sub>k sk t)" by auto
            note sk_less = sk_run.okb_h.compatible_rstep_imp_less[OF sk_run.irun_h.Rw_less]
            from this[of "sk w" "sk t"] nwt have "(sk w, sk t) \<notin> rstep Rh\<^sub>\<omega>" by fast
            with sk_step have "(sk t, sk w) \<in> rstep (Rh\<^sub>\<omega> \<union> Eh\<^sub>\<omega>\<^sup>\<leftrightarrow>)" by blast
            from ground_REw_step_Sw_step[OF ground_sk ground_sk this] t_neq_w sk_inj
            have "(sk t, sk w) \<in> (rstep Sh\<^sub>\<omega>)\<^sup>\<leftrightarrow>" by fast
            with nwt Sstep_less have S_step:"(sk t, sk w) \<in> rstep Sh\<^sub>\<omega>" by blast
            from ind_step(1)[OF lex S_step t_conv_R_NF[of t]] show ?thesis by auto
          next
            case 2
            then obtain w where w:"(t, w) \<in> rstep S\<^sub>\<omega>" by auto
            with Sw_step_less sk_compat_less have tw_less_sk:"sk t \<succ>\<^sub>s\<^sub>k sk w" by auto
            from w Sw_implies_REw have tw_conv:"(t, w) \<in> (rstep (R\<^sub>\<omega> \<union> E\<^sub>\<omega>))\<^sup>\<leftrightarrow>\<^sup>*" by blast
            from ih t_conv_R_NF[OF this] tw_less_sk have "(w, u) \<in> (rstep S\<^sub>\<omega>)\<^sup>*" by auto
            with w show ?thesis by auto
          qed
        qed
      qed
    qed
  qed
qed

lemma NF_Sw_subset_NF_R:"NF_trs S\<^sub>\<omega> \<subseteq> NF_trs \<R>"
proof-
  { fix s
    assume a:"s \<notin> NF_trs \<R>"
    from SN_R SN_reaches_NF[of "rstep \<R>" s] obtain u where nf:"(s, u) \<in> (rstep \<R>)\<^sup>!"
      unfolding SN_defs by blast
    with a have su:"s \<noteq> u" unfolding normalizability_def NF_def rtrancl_eq_or_trancl by auto
    from nf R_norm_implies_Sw_steps have "(s, u) \<in> (rstep S\<^sub>\<omega>)\<^sup>*" by auto
    with su have "(s, u) \<in> (rstep S\<^sub>\<omega>)\<^sup>+" unfolding rtrancl_eq_or_trancl by auto
    from tranclD[OF this] have "s \<notin> NF_trs S\<^sub>\<omega>" by auto
  }
  thus ?thesis by auto
qed

lemma completeness:
  assumes Ew_irreducible:"\<forall>s t. (s,t) \<in> E\<^sub>\<omega>\<^sup>\<leftrightarrow> \<longrightarrow> s \<in> NF_trs S\<^sub>\<omega>"
  and Ew_nontrivial:"\<forall>t. (t,t) \<notin> E\<^sub>\<omega>"
  and "reduced \<R>" and "reduced R\<^sub>\<omega>"
  shows "E\<^sub>\<omega> = {} \<and> \<R> \<doteq> R\<^sub>\<omega>"
proof-
  have "E\<^sub>\<omega> = {}"
  proof(rule ccontr)
    assume "E\<^sub>\<omega> \<noteq> {}"
    then obtain s t where st:"(s,t) \<in> E\<^sub>\<omega>" by fastforce
    with R_NF_eq[of s t] obtain u where "(s, u) \<in> (rstep \<R>)\<^sup>!" "(t, u) \<in> (rstep \<R>)\<^sup>!" by auto
    with R_norm_implies_Sw_steps have rsteps_u:"(s, u) \<in> (rstep S\<^sub>\<omega>)\<^sup>*" "(t, u) \<in> (rstep S\<^sub>\<omega>)\<^sup>*" by auto
    from st Ew_irreducible[rule_format] have nfs:"s \<in> NF_trs S\<^sub>\<omega>" "t \<in> NF_trs S\<^sub>\<omega>" by auto
    note nfs = nfs[unfolded NF_def] rsteps_u[unfolded rtrancl_eq_or_trancl] tranclD
    hence "s = u \<and> t = u" by fast
    with st Ew_nontrivial show False by auto
  qed
  hence Sw_Rw:"S\<^sub>\<omega> = R\<^sub>\<omega>" unfolding Sw_def by auto
  interpret complete_ars "(rstep \<R>)" by (unfold_locales, insert SN_R CR_R, auto)
  from Sw_implies_REw have rstep_Sw_R:"rstep S\<^sub>\<omega> \<subseteq> (rstep \<R>)\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding conv_eq[unfolded oKBi_conversion_ERw] unfolding conversion_def by blast
  note complete_eq = complete_NE_intro1[of "rstep S\<^sub>\<omega>", OF rstep_Sw_R SN_Sw_step NF_Sw_subset_NF_R]
  hence rstep_R_Sw:"(rstep \<R>)\<^sup>! = (rstep S\<^sub>\<omega>)\<^sup>!" by auto
  note vars = SN_imp_variable_condition[OF SN_R] SN_imp_variable_condition[OF SN_Sw_step]
  from reduced_NE_imp_unique[OF vars \<open>reduced \<R>\<close> _ rstep_R_Sw] \<open>reduced R\<^sub>\<omega>\<close> show ?thesis
    using \<open>E\<^sub>\<omega> = {}\<close> unfolding Sw_Rw by auto
qed
end
end
end
end